/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The system module encapsulates the most basic features used by other modules.
 * @module system
 * @requires require
 * @requires jquery
 */
define('durandal/system',['require', 'jquery'], function(require, $) {
    var isDebugging = false,
        nativeKeys = Object.keys,
        hasOwnProperty = Object.prototype.hasOwnProperty,
        toString = Object.prototype.toString,
        system,
        treatAsIE8 = false,
        nativeIsArray = Array.isArray,
        slice = Array.prototype.slice;

    //see http://patik.com/blog/complete-cross-browser-console-log/
    // Tell IE9 to use its built-in console
    if (Function.prototype.bind && (typeof console === 'object' || typeof console === 'function') && typeof console.log == 'object') {
        try {
            ['log', 'info', 'warn', 'error', 'assert', 'dir', 'clear', 'profile', 'profileEnd']
                .forEach(function(method) {
                    console[method] = this.call(console[method], console);
                }, Function.prototype.bind);
        } catch (ex) {
            treatAsIE8 = true;
        }
    }

    // callback for dojo's loader 
    // note: if you wish to use Durandal with dojo's AMD loader,
    // currently you must fork the dojo source with the following
    // dojo/dojo.js, line 1187, the last line of the finishExec() function: 
    //  (add) signal("moduleLoaded", [module.result, module.mid]);
    // an enhancement request has been submitted to dojo to make this
    // a permanent change. To view the status of this request, visit:
    // http://bugs.dojotoolkit.org/ticket/16727

    if (require.on) {
        require.on("moduleLoaded", function(module, mid) {
            system.setModuleId(module, mid);
        });
    }

    // callback for require.js loader
    if (typeof requirejs !== 'undefined') {
        requirejs.onResourceLoad = function(context, map, depArray) {
            system.setModuleId(context.defined[map.id], map.id);
        };
    }

    var noop = function() { };

    var log = function() {
        try {
            // Modern browsers
            if (typeof console != 'undefined' && typeof console.log == 'function') {
                // Opera 11
                if (window.opera) {
                    var i = 0;
                    while (i < arguments.length) {
                        console.log('Item ' + (i + 1) + ': ' + arguments[i]);
                        i++;
                    }
                }
                // All other modern browsers
                else if ((slice.call(arguments)).length == 1 && typeof slice.call(arguments)[0] == 'string') {
                    console.log((slice.call(arguments)).toString());
                } else {
                    console.log.apply(console, slice.call(arguments));
                }
            }
            // IE8
            else if ((!Function.prototype.bind || treatAsIE8) && typeof console != 'undefined' && typeof console.log == 'object') {
                Function.prototype.call.call(console.log, console, slice.call(arguments));
            }

            // IE7 and lower, and other old browsers
        } catch (ignore) { }
    };

    var logError = function(error) {
        if(error instanceof Error){
            throw error;
        }

        throw new Error(error);
    };

    /**
     * @class SystemModule
     * @static
     */
    system = {
        /**
         * Durandal's version.
         * @property {string} version
         */
        version: "2.0.1",
        /**
         * A noop function.
         * @method noop
         */
        noop: noop,
        /**
         * Gets the module id for the specified object.
         * @method getModuleId
         * @param {object} obj The object whose module id you wish to determine.
         * @return {string} The module id.
         */
        getModuleId: function(obj) {
            if (!obj) {
                return null;
            }

            if (typeof obj == 'function') {
                return obj.prototype.__moduleId__;
            }

            if (typeof obj == 'string') {
                return null;
            }

            return obj.__moduleId__;
        },
        /**
         * Sets the module id for the specified object.
         * @method setModuleId
         * @param {object} obj The object whose module id you wish to set.
         * @param {string} id The id to set for the specified object.
         */
        setModuleId: function(obj, id) {
            if (!obj) {
                return;
            }

            if (typeof obj == 'function') {
                obj.prototype.__moduleId__ = id;
                return;
            }

            if (typeof obj == 'string') {
                return;
            }

            obj.__moduleId__ = id;
        },
        /**
         * Resolves the default object instance for a module. If the module is an object, the module is returned. If the module is a function, that function is called with `new` and it's result is returned.
         * @method resolveObject
         * @param {object} module The module to use to get/create the default object for.
         * @return {object} The default object for the module.
         */
        resolveObject: function(module) {
            if (system.isFunction(module)) {
                return new module();
            } else {
                return module;
            }
        },
        /**
         * Gets/Sets whether or not Durandal is in debug mode.
         * @method debug
         * @param {boolean} [enable] Turns on/off debugging.
         * @return {boolean} Whether or not Durandal is current debugging.
         */
        debug: function(enable) {
            if (arguments.length == 1) {
                isDebugging = enable;
                if (isDebugging) {
                    this.log = log;
                    this.error = logError;
                    this.log('Debug:Enabled');
                } else {
                    this.log('Debug:Disabled');
                    this.log = noop;
                    this.error = noop;
                }
            }

            return isDebugging;
        },
        /**
         * Logs data to the console. Pass any number of parameters to be logged. Log output is not processed if the framework is not running in debug mode.
         * @method log
         * @param {object} info* The objects to log.
         */
        log: noop,
        /**
         * Logs an error.
         * @method error
         * @param {string|Error} obj The error to report.
         */
        error: noop,
        /**
         * Asserts a condition by throwing an error if the condition fails.
         * @method assert
         * @param {boolean} condition The condition to check.
         * @param {string} message The message to report in the error if the condition check fails.
         */
        assert: function (condition, message) {
            if (!condition) {
                system.error(new Error(message || 'Assert:Failed'));
            }
        },
        /**
         * Creates a deferred object which can be used to create a promise. Optionally pass a function action to perform which will be passed an object used in resolving the promise.
         * @method defer
         * @param {function} [action] The action to defer. You will be passed the deferred object as a paramter.
         * @return {Deferred} The deferred object.
         */
        defer: function(action) {
            return $.Deferred(action);
        },
        /**
         * Creates a simple V4 UUID. This should not be used as a PK in your database. It can be used to generate internal, unique ids. For a more robust solution see [node-uuid](https://github.com/broofa/node-uuid).
         * @method guid
         * @return {string} The guid.
         */
        guid: function() {
            return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
                var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
                return v.toString(16);
            });
        },
        /**
         * Uses require.js to obtain a module. This function returns a promise which resolves with the module instance. You can pass more than one module id to this function or an array of ids. If more than one or an array is passed, then the promise will resolve with an array of module instances.
         * @method acquire
         * @param {string|string[]} moduleId The id(s) of the modules to load.
         * @return {Promise} A promise for the loaded module(s).
         */
        acquire: function() {
            var modules,
                first = arguments[0],
                arrayRequest = false;

            if(system.isArray(first)){
                modules = first;
                arrayRequest = true;
            }else{
                modules = slice.call(arguments, 0);
            }

            return this.defer(function(dfd) {
                require(modules, function() {
                    var args = arguments;
                    setTimeout(function() {
                        if(args.length > 1 || arrayRequest){
                            dfd.resolve(slice.call(args, 0));
                        }else{
                            dfd.resolve(args[0]);
                        }
                    }, 1);
                }, function(err){
                    dfd.reject(err);
                });
            }).promise();
        },
        /**
         * Extends the first object with the properties of the following objects.
         * @method extend
         * @param {object} obj The target object to extend.
         * @param {object} extension* Uses to extend the target object.
         */
        extend: function(obj) {
            var rest = slice.call(arguments, 1);

            for (var i = 0; i < rest.length; i++) {
                var source = rest[i];

                if (source) {
                    for (var prop in source) {
                        obj[prop] = source[prop];
                    }
                }
            }

            return obj;
        },
        /**
         * Uses a setTimeout to wait the specified milliseconds.
         * @method wait
         * @param {number} milliseconds The number of milliseconds to wait.
         * @return {Promise}
         */
        wait: function(milliseconds) {
            return system.defer(function(dfd) {
                setTimeout(dfd.resolve, milliseconds);
            }).promise();
        }
    };

    /**
     * Gets all the owned keys of the specified object.
     * @method keys
     * @param {object} object The object whose owned keys should be returned.
     * @return {string[]} The keys.
     */
    system.keys = nativeKeys || function(obj) {
        if (obj !== Object(obj)) {
            throw new TypeError('Invalid object');
        }

        var keys = [];

        for (var key in obj) {
            if (hasOwnProperty.call(obj, key)) {
                keys[keys.length] = key;
            }
        }

        return keys;
    };

    /**
     * Determines if the specified object is an html element.
     * @method isElement
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */
    system.isElement = function(obj) {
        return !!(obj && obj.nodeType === 1);
    };

    /**
     * Determines if the specified object is an array.
     * @method isArray
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */
    system.isArray = nativeIsArray || function(obj) {
        return toString.call(obj) == '[object Array]';
    };

    /**
     * Determines if the specified object is...an object. ie. Not an array, string, etc.
     * @method isObject
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */
    system.isObject = function(obj) {
        return obj === Object(obj);
    };

    /**
     * Determines if the specified object is a boolean.
     * @method isBoolean
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */
    system.isBoolean = function(obj) {
        return typeof(obj) === "boolean";
    };

    /**
     * Determines if the specified object is a promise.
     * @method isPromise
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */
    system.isPromise = function(obj) {
        return obj && system.isFunction(obj.then);
    };

    /**
     * Determines if the specified object is a function arguments object.
     * @method isArguments
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    /**
     * Determines if the specified object is a function.
     * @method isFunction
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    /**
     * Determines if the specified object is a string.
     * @method isString
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    /**
     * Determines if the specified object is a number.
     * @method isNumber
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    /**
     * Determines if the specified object is a date.
     * @method isDate
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    /**
     * Determines if the specified object is a boolean.
     * @method isBoolean
     * @param {object} object The object to check.
     * @return {boolean} True if matches the type, false otherwise.
     */

    //isArguments, isFunction, isString, isNumber, isDate, isRegExp.
    var isChecks = ['Arguments', 'Function', 'String', 'Number', 'Date', 'RegExp'];

    function makeIsFunction(name) {
        var value = '[object ' + name + ']';
        system['is' + name] = function(obj) {
            return toString.call(obj) == value;
        };
    }

    for (var i = 0; i < isChecks.length; i++) {
        makeIsFunction(isChecks[i]);
    }

    return system;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The viewEngine module provides information to the viewLocator module which is used to locate the view's source file. The viewEngine also transforms a view id into a view instance.
 * @module viewEngine
 * @requires system
 * @requires jquery
 */
define('durandal/viewEngine',['durandal/system', 'jquery'], function (system, $) {
    var parseMarkup;

    if ($.parseHTML) {
        parseMarkup = function (html) {
            return $.parseHTML(html);
        };
    } else {
        parseMarkup = function (html) {
            return $(html).get();
        };
    }

    /**
     * @class ViewEngineModule
     * @static
     */
    return {
        /**
         * The file extension that view source files are expected to have.
         * @property {string} viewExtension
         * @default .html
         */
        viewExtension: '.html',
        /**
         * The name of the RequireJS loader plugin used by the viewLocator to obtain the view source. (Use requirejs to map the plugin's full path).
         * @property {string} viewPlugin
         * @default text
         */
        viewPlugin: 'text',
        /**
         * Determines if the url is a url for a view, according to the view engine.
         * @method isViewUrl
         * @param {string} url The potential view url.
         * @return {boolean} True if the url is a view url, false otherwise.
         */
        isViewUrl: function (url) {
            return url.indexOf(this.viewExtension, url.length - this.viewExtension.length) !== -1;
        },
        /**
         * Converts a view url into a view id.
         * @method convertViewUrlToViewId
         * @param {string} url The url to convert.
         * @return {string} The view id.
         */
        convertViewUrlToViewId: function (url) {
            return url.substring(0, url.length - this.viewExtension.length);
        },
        /**
         * Converts a view id into a full RequireJS path.
         * @method convertViewIdToRequirePath
         * @param {string} viewId The view id to convert.
         * @return {string} The require path.
         */
        convertViewIdToRequirePath: function (viewId) {
            return this.viewPlugin + '!' + viewId + this.viewExtension;
        },
        /**
         * Parses the view engine recognized markup and returns DOM elements.
         * @method parseMarkup
         * @param {string} markup The markup to parse.
         * @return {DOMElement[]} The elements.
         */
        parseMarkup: parseMarkup,
        /**
         * Calls `parseMarkup` and then pipes the results through `ensureSingleElement`.
         * @method processMarkup
         * @param {string} markup The markup to process.
         * @return {DOMElement} The view.
         */
        processMarkup: function (markup) {
            var allElements = this.parseMarkup(markup);
            return this.ensureSingleElement(allElements);
        },
        /**
         * Converts an array of elements into a single element. White space and comments are removed. If a single element does not remain, then the elements are wrapped.
         * @method ensureSingleElement
         * @param {DOMElement[]} allElements The elements.
         * @return {DOMElement} A single element.
         */
        ensureSingleElement:function(allElements){
            if (allElements.length == 1) {
                return allElements[0];
            }

            var withoutCommentsOrEmptyText = [];

            for (var i = 0; i < allElements.length; i++) {
                var current = allElements[i];
                if (current.nodeType != 8) {
                    if (current.nodeType == 3) {
                        var result = /\S/.test(current.nodeValue);
                        if (!result) {
                            continue;
                        }
                    }

                    withoutCommentsOrEmptyText.push(current);
                }
            }

            if (withoutCommentsOrEmptyText.length > 1) {
                return $(withoutCommentsOrEmptyText).wrapAll('<div class="durandal-wrapper"></div>').parent().get(0);
            }

            return withoutCommentsOrEmptyText[0];
        },
        /**
         * Creates the view associated with the view id.
         * @method createView
         * @param {string} viewId The view id whose view should be created.
         * @return {Promise} A promise of the view.
         */
        createView: function(viewId) {
            var that = this;
            var requirePath = this.convertViewIdToRequirePath(viewId);

            return system.defer(function(dfd) {
                system.acquire(requirePath).then(function(markup) {
                    var element = that.processMarkup(markup);
                    element.setAttribute('data-view', viewId);
                    dfd.resolve(element);
                }).fail(function(err){
                        that.createFallbackView(viewId, requirePath, err).then(function(element){
                            element.setAttribute('data-view', viewId);
                            dfd.resolve(element);
                        });
                    });
            }).promise();
        },
        /**
         * Called when a view cannot be found to provide the opportunity to locate or generate a fallback view. Mainly used to ease development.
         * @method createFallbackView
         * @param {string} viewId The view id whose view should be created.
         * @param {string} requirePath The require path that was attempted.
         * @param {Error} requirePath The error that was returned from the attempt to locate the default view.
         * @return {Promise} A promise for the fallback view.
         */
        createFallbackView: function (viewId, requirePath, err) {
            var that = this,
                message = 'View Not Found. Searched for "' + viewId + '" via path "' + requirePath + '".';

            return system.defer(function(dfd) {
                dfd.resolve(that.processMarkup('<div class="durandal-view-404">' + message + '</div>'));
            }).promise();
        }
    };
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The viewLocator module collaborates with the viewEngine module to provide views (literally dom sub-trees) to other parts of the framework as needed. The primary consumer of the viewLocator is the composition module.
 * @module viewLocator
 * @requires system
 * @requires viewEngine
 */
define('durandal/viewLocator',['durandal/system', 'durandal/viewEngine'], function (system, viewEngine) {
    function findInElements(nodes, url) {
        for (var i = 0; i < nodes.length; i++) {
            var current = nodes[i];
            var existingUrl = current.getAttribute('data-view');
            if (existingUrl == url) {
                return current;
            }
        }
    }
    
    function escape(str) {
        return (str + '').replace(/([\\\.\+\*\?\[\^\]\$\(\)\{\}\=\!\<\>\|\:])/g, "\\$1");
    }

    /**
     * @class ViewLocatorModule
     * @static
     */
    return {
        /**
         * Allows you to set up a convention for mapping module folders to view folders. It is a convenience method that customizes `convertModuleIdToViewId` and `translateViewIdToArea` under the covers.
         * @method useConvention
         * @param {string} [modulesPath] A string to match in the path and replace with the viewsPath. If not specified, the match is 'viewmodels'.
         * @param {string} [viewsPath] The replacement for the modulesPath. If not specified, the replacement is 'views'.
         * @param {string} [areasPath] Partial views are mapped to the "views" folder if not specified. Use this parameter to change their location.
         */
        useConvention: function(modulesPath, viewsPath, areasPath) {
            modulesPath = modulesPath || 'viewmodels';
            viewsPath = viewsPath || 'views';
            areasPath = areasPath || viewsPath;

            var reg = new RegExp(escape(modulesPath), 'gi');

            this.convertModuleIdToViewId = function (moduleId) {
                return moduleId.replace(reg, viewsPath);
            };

            this.translateViewIdToArea = function (viewId, area) {
                if (!area || area == 'partial') {
                    return areasPath + '/' + viewId;
                }
                
                return areasPath + '/' + area + '/' + viewId;
            };
        },
        /**
         * Maps an object instance to a view instance.
         * @method locateViewForObject
         * @param {object} obj The object to locate the view for.
         * @param {string} [area] The area to translate the view to.
         * @param {DOMElement[]} [elementsToSearch] An existing set of elements to search first.
         * @return {Promise} A promise of the view.
         */
        locateViewForObject: function(obj, area, elementsToSearch) {
            var view;

            if (obj.getView) {
                view = obj.getView();
                if (view) {
                    return this.locateView(view, area, elementsToSearch);
                }
            }

            if (obj.viewUrl) {
                return this.locateView(obj.viewUrl, area, elementsToSearch);
            }

            var id = system.getModuleId(obj);
            if (id) {
                return this.locateView(this.convertModuleIdToViewId(id), area, elementsToSearch);
            }

            return this.locateView(this.determineFallbackViewId(obj), area, elementsToSearch);
        },
        /**
         * Converts a module id into a view id. By default the ids are the same.
         * @method convertModuleIdToViewId
         * @param {string} moduleId The module id.
         * @return {string} The view id.
         */
        convertModuleIdToViewId: function(moduleId) {
            return moduleId;
        },
        /**
         * If no view id can be determined, this function is called to genreate one. By default it attempts to determine the object's type and use that.
         * @method determineFallbackViewId
         * @param {object} obj The object to determine the fallback id for.
         * @return {string} The view id.
         */
        determineFallbackViewId: function (obj) {
            var funcNameRegex = /function (.{1,})\(/;
            var results = (funcNameRegex).exec((obj).constructor.toString());
            var typeName = (results && results.length > 1) ? results[1] : "";

            return 'views/' + typeName;
        },
        /**
         * Takes a view id and translates it into a particular area. By default, no translation occurs.
         * @method translateViewIdToArea
         * @param {string} viewId The view id.
         * @param {string} area The area to translate the view to.
         * @return {string} The translated view id.
         */
        translateViewIdToArea: function (viewId, area) {
            return viewId;
        },
        /**
         * Locates the specified view.
         * @method locateView
         * @param {string|DOMElement} viewOrUrlOrId A view, view url or view id to locate.
         * @param {string} [area] The area to translate the view to.
         * @param {DOMElement[]} [elementsToSearch] An existing set of elements to search first.
         * @return {Promise} A promise of the view.
         */
        locateView: function(viewOrUrlOrId, area, elementsToSearch) {
            if (typeof viewOrUrlOrId === 'string') {
                var viewId;

                if (viewEngine.isViewUrl(viewOrUrlOrId)) {
                    viewId = viewEngine.convertViewUrlToViewId(viewOrUrlOrId);
                } else {
                    viewId = viewOrUrlOrId;
                }

                if (area) {
                    viewId = this.translateViewIdToArea(viewId, area);
                }

                if (elementsToSearch) {
                    var existing = findInElements(elementsToSearch, viewId);
                    if (existing) {
                        return system.defer(function(dfd) {
                            dfd.resolve(existing);
                        }).promise();
                    }
                }

                return viewEngine.createView(viewId);
            }

            return system.defer(function(dfd) {
                dfd.resolve(viewOrUrlOrId);
            }).promise();
        }
    };
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The binder joins an object instance and a DOM element tree by applying databinding and/or invoking binding lifecycle callbacks (binding and bindingComplete).
 * @module binder
 * @requires system
 * @requires knockout
 */
define('durandal/binder',['durandal/system', 'knockout'], function (system, ko) {
    var binder,
        insufficientInfoMessage = 'Insufficient Information to Bind',
        unexpectedViewMessage = 'Unexpected View Type',
        bindingInstructionKey = 'durandal-binding-instruction',
        koBindingContextKey = '__ko_bindingContext__';

    function normalizeBindingInstruction(result){
        if(result === undefined){
            return { applyBindings: true };
        }

        if(system.isBoolean(result)){
            return { applyBindings:result };
        }

        if(result.applyBindings === undefined){
            result.applyBindings = true;
        }

        return result;
    }

    function doBind(obj, view, bindingTarget, data){
        if (!view || !bindingTarget) {
            if (binder.throwOnErrors) {
                system.error(insufficientInfoMessage);
            } else {
                system.log(insufficientInfoMessage, view, data);
            }
            return;
        }

        if (!view.getAttribute) {
            if (binder.throwOnErrors) {
                system.error(unexpectedViewMessage);
            } else {
                system.log(unexpectedViewMessage, view, data);
            }
            return;
        }

        var viewName = view.getAttribute('data-view');

        try {
            var instruction;

            if (obj && obj.binding) {
                instruction = obj.binding(view);
            }

            instruction = normalizeBindingInstruction(instruction);
            binder.binding(data, view, instruction);

            if(instruction.applyBindings){
                system.log('Binding', viewName, data);
                ko.applyBindings(bindingTarget, view);
            }else if(obj){
                ko.utils.domData.set(view, koBindingContextKey, { $data:obj });
            }

            binder.bindingComplete(data, view, instruction);

            if (obj && obj.bindingComplete) {
                obj.bindingComplete(view);
            }

            ko.utils.domData.set(view, bindingInstructionKey, instruction);
            return instruction;
        } catch (e) {
            e.message = e.message + ';\nView: ' + viewName + ";\nModuleId: " + system.getModuleId(data);
            if (binder.throwOnErrors) {
                system.error(e);
            } else {
                system.log(e.message);
            }
        }
    }

    /**
     * @class BinderModule
     * @static
     */
    return binder = {
        /**
         * Called before every binding operation. Does nothing by default.
         * @method binding
         * @param {object} data The data that is about to be bound.
         * @param {DOMElement} view The view that is about to be bound.
         * @param {object} instruction The object that carries the binding instructions.
         */
        binding: system.noop,
        /**
         * Called after every binding operation. Does nothing by default.
         * @method bindingComplete
         * @param {object} data The data that has just been bound.
         * @param {DOMElement} view The view that has just been bound.
         * @param {object} instruction The object that carries the binding instructions.
         */
        bindingComplete: system.noop,
        /**
         * Indicates whether or not the binding system should throw errors or not.
         * @property {boolean} throwOnErrors
         * @default false The binding system will not throw errors by default. Instead it will log them.
         */
        throwOnErrors: false,
        /**
         * Gets the binding instruction that was associated with a view when it was bound.
         * @method getBindingInstruction
         * @param {DOMElement} view The view that was previously bound.
         * @return {object} The object that carries the binding instructions.
         */
        getBindingInstruction:function(view){
            return ko.utils.domData.get(view, bindingInstructionKey);
        },
        /**
         * Binds the view, preserving the existing binding context. Optionally, a new context can be created, parented to the previous context.
         * @method bindContext
         * @param {KnockoutBindingContext} bindingContext The current binding context.
         * @param {DOMElement} view The view to bind.
         * @param {object} [obj] The data to bind to, causing the creation of a child binding context if present.
         */
        bindContext: function(bindingContext, view, obj) {
            if (obj && bindingContext) {
                bindingContext = bindingContext.createChildContext(obj);
            }

            return doBind(obj, view, bindingContext, obj || (bindingContext ? bindingContext.$data : null));
        },
        /**
         * Binds the view, preserving the existing binding context. Optionally, a new context can be created, parented to the previous context.
         * @method bind
         * @param {object} obj The data to bind to.
         * @param {DOMElement} view The view to bind.
         */
        bind: function(obj, view) {
            return doBind(obj, view, obj, obj);
        }
    };
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The activator module encapsulates all logic related to screen/component activation.
 * An activator is essentially an asynchronous state machine that understands a particular state transition protocol.
 * The protocol ensures that the following series of events always occur: `canDeactivate` (previous state), `canActivate` (new state), `deactivate` (previous state), `activate` (new state).
 * Each of the _can_ callbacks may return a boolean, affirmative value or promise for one of those. If either of the _can_ functions yields a false result, then activation halts.
 * @module activator
 * @requires system
 * @requires knockout
 */
define('durandal/activator',['durandal/system', 'knockout'], function (system, ko) {
    var activator;

    function ensureSettings(settings) {
        if (settings == undefined) {
            settings = {};
        }

        if (!settings.closeOnDeactivate) {
            settings.closeOnDeactivate = activator.defaults.closeOnDeactivate;
        }

        if (!settings.beforeActivate) {
            settings.beforeActivate = activator.defaults.beforeActivate;
        }

        if (!settings.afterDeactivate) {
            settings.afterDeactivate = activator.defaults.afterDeactivate;
        }

        if(!settings.affirmations){
            settings.affirmations = activator.defaults.affirmations;
        }

        if (!settings.interpretResponse) {
            settings.interpretResponse = activator.defaults.interpretResponse;
        }

        if (!settings.areSameItem) {
            settings.areSameItem = activator.defaults.areSameItem;
        }

        return settings;
    }

    function invoke(target, method, data) {
        if (system.isArray(data)) {
            return target[method].apply(target, data);
        }

        return target[method](data);
    }

    function deactivate(item, close, settings, dfd, setter) {
        if (item && item.deactivate) {
            system.log('Deactivating', item);

            var result;
            try {
                result = item.deactivate(close);
            } catch(error) {
                system.error(error);
                dfd.resolve(false);
                return;
            }

            if (result && result.then) {
                result.then(function() {
                    settings.afterDeactivate(item, close, setter);
                    dfd.resolve(true);
                }, function(reason) {
                    system.log(reason);
                    dfd.resolve(false);
                });
            } else {
                settings.afterDeactivate(item, close, setter);
                dfd.resolve(true);
            }
        } else {
            if (item) {
                settings.afterDeactivate(item, close, setter);
            }

            dfd.resolve(true);
        }
    }

    function activate(newItem, activeItem, callback, activationData) {
        if (newItem) {
            if (newItem.activate) {
                system.log('Activating', newItem);

                var result;
                try {
                    result = invoke(newItem, 'activate', activationData);
                } catch (error) {
                    system.error(error);
                    callback(false);
                    return;
                }

                if (result && result.then) {
                    result.then(function() {
                        activeItem(newItem);
                        callback(true);
                    }, function(reason) {
                        system.log(reason);
                        callback(false);
                    });
                } else {
                    activeItem(newItem);
                    callback(true);
                }
            } else {
                activeItem(newItem);
                callback(true);
            }
        } else {
            callback(true);
        }
    }

    function canDeactivateItem(item, close, settings) {
        settings.lifecycleData = null;

        return system.defer(function (dfd) {
            if (item && item.canDeactivate) {
                var resultOrPromise;
                try {
                    resultOrPromise = item.canDeactivate(close);
                } catch(error) {
                    system.error(error);
                    dfd.resolve(false);
                    return;
                }

                if (resultOrPromise.then) {
                    resultOrPromise.then(function(result) {
                        settings.lifecycleData = result;
                        dfd.resolve(settings.interpretResponse(result));
                    }, function(reason) {
                        system.error(reason);
                        dfd.resolve(false);
                    });
                } else {
                    settings.lifecycleData = resultOrPromise;
                    dfd.resolve(settings.interpretResponse(resultOrPromise));
                }
            } else {
                dfd.resolve(true);
            }
        }).promise();
    };

    function canActivateItem(newItem, activeItem, settings, activationData) {
        settings.lifecycleData = null;

        return system.defer(function (dfd) {
            if (newItem == activeItem()) {
                dfd.resolve(true);
                return;
            }

            if (newItem && newItem.canActivate) {
                var resultOrPromise;
                try {
                    resultOrPromise = invoke(newItem, 'canActivate', activationData);
                } catch (error) {
                    system.error(error);
                    dfd.resolve(false);
                    return;
                }

                if (resultOrPromise.then) {
                    resultOrPromise.then(function(result) {
                        settings.lifecycleData = result;
                        dfd.resolve(settings.interpretResponse(result));
                    }, function(reason) {
                        system.error(reason);
                        dfd.resolve(false);
                    });
                } else {
                    settings.lifecycleData = resultOrPromise;
                    dfd.resolve(settings.interpretResponse(resultOrPromise));
                }
            } else {
                dfd.resolve(true);
            }
        }).promise();
    };

    /**
     * An activator is a read/write computed observable that enforces the activation lifecycle whenever changing values.
     * @class Activator
     */
    function createActivator(initialActiveItem, settings) {
        var activeItem = ko.observable(null);
        var activeData;

        settings = ensureSettings(settings);

        var computed = ko.computed({
            read: function () {
                return activeItem();
            },
            write: function (newValue) {
                computed.viaSetter = true;
                computed.activateItem(newValue);
            }
        });

        computed.__activator__ = true;

        /**
         * The settings for this activator.
         * @property {ActivatorSettings} settings
         */
        computed.settings = settings;
        settings.activator = computed;

        /**
         * An observable which indicates whether or not the activator is currently in the process of activating an instance.
         * @method isActivating
         * @return {boolean}
         */
        computed.isActivating = ko.observable(false);

        /**
         * Determines whether or not the specified item can be deactivated.
         * @method canDeactivateItem
         * @param {object} item The item to check.
         * @param {boolean} close Whether or not to check if close is possible.
         * @return {promise}
         */
        computed.canDeactivateItem = function (item, close) {
            return canDeactivateItem(item, close, settings);
        };

        /**
         * Deactivates the specified item.
         * @method deactivateItem
         * @param {object} item The item to deactivate.
         * @param {boolean} close Whether or not to close the item.
         * @return {promise}
         */
        computed.deactivateItem = function (item, close) {
            return system.defer(function(dfd) {
                computed.canDeactivateItem(item, close).then(function(canDeactivate) {
                    if (canDeactivate) {
                        deactivate(item, close, settings, dfd, activeItem);
                    } else {
                        computed.notifySubscribers();
                        dfd.resolve(false);
                    }
                });
            }).promise();
        };

        /**
         * Determines whether or not the specified item can be activated.
         * @method canActivateItem
         * @param {object} item The item to check.
         * @param {object} activationData Data associated with the activation.
         * @return {promise}
         */
        computed.canActivateItem = function (newItem, activationData) {
            return canActivateItem(newItem, activeItem, settings, activationData);
        };

        /**
         * Activates the specified item.
         * @method activateItem
         * @param {object} newItem The item to activate.
         * @param {object} newActivationData Data associated with the activation.
         * @return {promise}
         */
        computed.activateItem = function (newItem, newActivationData) {
            var viaSetter = computed.viaSetter;
            computed.viaSetter = false;

            return system.defer(function (dfd) {
                if (computed.isActivating()) {
                    dfd.resolve(false);
                    return;
                }

                computed.isActivating(true);

                var currentItem = activeItem();
                if (settings.areSameItem(currentItem, newItem, activeData, newActivationData)) {
                    computed.isActivating(false);
                    dfd.resolve(true);
                    return;
                }

                computed.canDeactivateItem(currentItem, settings.closeOnDeactivate).then(function (canDeactivate) {
                    if (canDeactivate) {
                        computed.canActivateItem(newItem, newActivationData).then(function (canActivate) {
                            if (canActivate) {
                                system.defer(function (dfd2) {
                                    deactivate(currentItem, settings.closeOnDeactivate, settings, dfd2);
                                }).promise().then(function () {
                                    newItem = settings.beforeActivate(newItem, newActivationData);
                                    activate(newItem, activeItem, function (result) {
                                        activeData = newActivationData;
                                        computed.isActivating(false);
                                        dfd.resolve(result);
                                    }, newActivationData);
                                });
                            } else {
                                if (viaSetter) {
                                    computed.notifySubscribers();
                                }

                                computed.isActivating(false);
                                dfd.resolve(false);
                            }
                        });
                    } else {
                        if (viaSetter) {
                            computed.notifySubscribers();
                        }

                        computed.isActivating(false);
                        dfd.resolve(false);
                    }
                });
            }).promise();
        };

        /**
         * Determines whether or not the activator, in its current state, can be activated.
         * @method canActivate
         * @return {promise}
         */
        computed.canActivate = function () {
            var toCheck;

            if (initialActiveItem) {
                toCheck = initialActiveItem;
                initialActiveItem = false;
            } else {
                toCheck = computed();
            }

            return computed.canActivateItem(toCheck);
        };

        /**
         * Activates the activator, in its current state.
         * @method activate
         * @return {promise}
         */
        computed.activate = function () {
            var toActivate;

            if (initialActiveItem) {
                toActivate = initialActiveItem;
                initialActiveItem = false;
            } else {
                toActivate = computed();
            }

            return computed.activateItem(toActivate);
        };

        /**
         * Determines whether or not the activator, in its current state, can be deactivated.
         * @method canDeactivate
         * @return {promise}
         */
        computed.canDeactivate = function (close) {
            return computed.canDeactivateItem(computed(), close);
        };

        /**
         * Deactivates the activator, in its current state.
         * @method deactivate
         * @return {promise}
         */
        computed.deactivate = function (close) {
            return computed.deactivateItem(computed(), close);
        };

        computed.includeIn = function (includeIn) {
            includeIn.canActivate = function () {
                return computed.canActivate();
            };

            includeIn.activate = function () {
                return computed.activate();
            };

            includeIn.canDeactivate = function (close) {
                return computed.canDeactivate(close);
            };

            includeIn.deactivate = function (close) {
                return computed.deactivate(close);
            };
        };

        if (settings.includeIn) {
            computed.includeIn(settings.includeIn);
        } else if (initialActiveItem) {
            computed.activate();
        }

        computed.forItems = function (items) {
            settings.closeOnDeactivate = false;

            settings.determineNextItemToActivate = function (list, lastIndex) {
                var toRemoveAt = lastIndex - 1;

                if (toRemoveAt == -1 && list.length > 1) {
                    return list[1];
                }

                if (toRemoveAt > -1 && toRemoveAt < list.length - 1) {
                    return list[toRemoveAt];
                }

                return null;
            };

            settings.beforeActivate = function (newItem) {
                var currentItem = computed();

                if (!newItem) {
                    newItem = settings.determineNextItemToActivate(items, currentItem ? items.indexOf(currentItem) : 0);
                } else {
                    var index = items.indexOf(newItem);

                    if (index == -1) {
                        items.push(newItem);
                    } else {
                        newItem = items()[index];
                    }
                }

                return newItem;
            };

            settings.afterDeactivate = function (oldItem, close) {
                if (close) {
                    items.remove(oldItem);
                }
            };

            var originalCanDeactivate = computed.canDeactivate;
            computed.canDeactivate = function (close) {
                if (close) {
                    return system.defer(function (dfd) {
                        var list = items();
                        var results = [];

                        function finish() {
                            for (var j = 0; j < results.length; j++) {
                                if (!results[j]) {
                                    dfd.resolve(false);
                                    return;
                                }
                            }

                            dfd.resolve(true);
                        }

                        for (var i = 0; i < list.length; i++) {
                            computed.canDeactivateItem(list[i], close).then(function (result) {
                                results.push(result);
                                if (results.length == list.length) {
                                    finish();
                                }
                            });
                        }
                    }).promise();
                } else {
                    return originalCanDeactivate();
                }
            };

            var originalDeactivate = computed.deactivate;
            computed.deactivate = function (close) {
                if (close) {
                    return system.defer(function (dfd) {
                        var list = items();
                        var results = 0;
                        var listLength = list.length;

                        function doDeactivate(item) {
                            computed.deactivateItem(item, close).then(function () {
                                results++;
                                items.remove(item);
                                if (results == listLength) {
                                    dfd.resolve();
                                }
                            });
                        }

                        for (var i = 0; i < listLength; i++) {
                            doDeactivate(list[i]);
                        }
                    }).promise();
                } else {
                    return originalDeactivate();
                }
            };

            return computed;
        };

        return computed;
    }

    /**
     * @class ActivatorSettings
     * @static
     */
    var activatorSettings = {
        /**
         * The default value passed to an object's deactivate function as its close parameter.
         * @property {boolean} closeOnDeactivate
         * @default true
         */
        closeOnDeactivate: true,
        /**
         * Lower-cased words which represent a truthy value.
         * @property {string[]} affirmations
         * @default ['yes', 'ok', 'true']
         */
        affirmations: ['yes', 'ok', 'true'],
        /**
         * Interprets the response of a `canActivate` or `canDeactivate` call using the known affirmative values in the `affirmations` array.
         * @method interpretResponse
         * @param {object} value
         * @return {boolean}
         */
        interpretResponse: function(value) {
            if(system.isObject(value)) {
                value = value.can || false;
            }

            if(system.isString(value)) {
                return ko.utils.arrayIndexOf(this.affirmations, value.toLowerCase()) !== -1;
            }

            return value;
        },
        /**
         * Determines whether or not the current item and the new item are the same.
         * @method areSameItem
         * @param {object} currentItem
         * @param {object} newItem
         * @param {object} currentActivationData
         * @param {object} newActivationData
         * @return {boolean}
         */
        areSameItem: function(currentItem, newItem, currentActivationData, newActivationData) {
            return currentItem == newItem;
        },
        /**
         * Called immediately before the new item is activated.
         * @method beforeActivate
         * @param {object} newItem
         */
        beforeActivate: function(newItem) {
            return newItem;
        },
        /**
         * Called immediately after the old item is deactivated.
         * @method afterDeactivate
         * @param {object} oldItem The previous item.
         * @param {boolean} close Whether or not the previous item was closed.
         * @param {function} setter The activate item setter function.
         */
        afterDeactivate: function(oldItem, close, setter) {
            if(close && setter) {
                setter(null);
            }
        }
    };

    /**
     * @class ActivatorModule
     * @static
     */
    activator = {
        /**
         * The default settings used by activators.
         * @property {ActivatorSettings} defaults
         */
        defaults: activatorSettings,
        /**
          * Creates a new activator.
          * @method create
          * @param {object} [initialActiveItem] The item which should be immediately activated upon creation of the ativator.
          * @param {ActivatorSettings} [settings] Per activator overrides of the default activator settings.
          * @return {Activator} The created activator.
          */
        create: createActivator,
        /**
         * Determines whether or not the provided object is an activator or not.
         * @method isActivator
         * @param {object} object Any object you wish to verify as an activator or not.
         * @return {boolean} True if the object is an activator; false otherwise.
         */
        isActivator:function(object){
            return object && object.__activator__;
        }
    };

    return activator;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The composition module encapsulates all functionality related to visual composition.
 * @module composition
 * @requires system
 * @requires viewLocator
 * @requires binder
 * @requires viewEngine
 * @requires activator
 * @requires jquery
 * @requires knockout
 */
define('durandal/composition',['durandal/system', 'durandal/viewLocator', 'durandal/binder', 'durandal/viewEngine', 'durandal/activator', 'jquery', 'knockout'], function (system, viewLocator, binder, viewEngine, activator, $, ko) {
    var dummyModel = {},
        activeViewAttributeName = 'data-active-view',
        composition,
        compositionCompleteCallbacks = [],
        compositionCount = 0,
        compositionDataKey = 'durandal-composition-data',
        partAttributeName = 'data-part',
        bindableSettings = ['model', 'view', 'transition', 'area', 'strategy', 'activationData'],
        visibilityKey = "durandal-visibility-data",
        composeBindings = ['compose:'];

    function getHostState(parent) {
        var elements = [];
        var state = {
            childElements: elements,
            activeView: null
        };

        var child = ko.virtualElements.firstChild(parent);

        while (child) {
            if (child.nodeType == 1) {
                elements.push(child);
                if (child.getAttribute(activeViewAttributeName)) {
                    state.activeView = child;
                }
            }

            child = ko.virtualElements.nextSibling(child);
        }

        if(!state.activeView){
            state.activeView = elements[0];
        }

        return state;
    }

    function endComposition() {
        compositionCount--;

        if (compositionCount === 0) {
            setTimeout(function(){
                var i = compositionCompleteCallbacks.length;

                while(i--) {
                    try{
                        compositionCompleteCallbacks[i]();
                    }catch(e){
                        system.error(e);
                    }
                }

                compositionCompleteCallbacks = [];
            }, 1);
        }
    }

    function cleanUp(context){
        delete context.activeView;
        delete context.viewElements;
    }

    function tryActivate(context, successCallback, skipActivation) {
        if(skipActivation){
            successCallback();
        } else if (context.activate && context.model && context.model.activate) {
            var result;

            try{
                if(system.isArray(context.activationData)) {
                    result = context.model.activate.apply(context.model, context.activationData);
                } else {
                    result = context.model.activate(context.activationData);
                }

                if(result && result.then) {
                    result.then(successCallback, function(reason) {
                        system.error(reason);
                        successCallback();
                    });
                } else if(result || result === undefined) {
                    successCallback();
                } else {
                    endComposition();
                    cleanUp(context);
                }
            }
            catch(e){
                system.error(e);
            }
        } else {
            successCallback();
        }
    }

    function triggerAttach() {
        var context = this;

        if (context.activeView) {
            context.activeView.removeAttribute(activeViewAttributeName);
        }

        if (context.child) {
            try{
                if (context.model && context.model.attached) {
                    if (context.composingNewView || context.alwaysTriggerAttach) {
                        context.model.attached(context.child, context.parent, context);
                    }
                }

                if (context.attached) {
                    context.attached(context.child, context.parent, context);
                }

                context.child.setAttribute(activeViewAttributeName, true);

                if (context.composingNewView && context.model && context.model.detached) {
                    ko.utils.domNodeDisposal.addDisposeCallback(context.child, function () {
                        try{
                            context.model.detached(context.child, context.parent, context);
                        }catch(e2){
                            system.error(e2);
                        }
                    });
                }
            }catch(e){
                system.error(e);
            }
        }

        context.triggerAttach = system.noop;
    }

    function shouldTransition(context) {
        if (system.isString(context.transition)) {
            if (context.activeView) {
                if (context.activeView == context.child) {
                    return false;
                }

                if (!context.child) {
                    return true;
                }

                if (context.skipTransitionOnSameViewId) {
                    var currentViewId = context.activeView.getAttribute('data-view');
                    var newViewId = context.child.getAttribute('data-view');
                    return currentViewId != newViewId;
                }
            }

            return true;
        }

        return false;
    }

    function cloneNodes(nodesArray) {
        for (var i = 0, j = nodesArray.length, newNodesArray = []; i < j; i++) {
            var clonedNode = nodesArray[i].cloneNode(true);
            newNodesArray.push(clonedNode);
        }
        return newNodesArray;
    }

    function replaceParts(context){
        var parts = cloneNodes(context.parts);
        var replacementParts = composition.getParts(parts, null, true);
        var standardParts = composition.getParts(context.child);

        for (var partId in replacementParts) {
            $(standardParts[partId]).replaceWith(replacementParts[partId]);
        }
    }

    function removePreviousView(context){
        var children = ko.virtualElements.childNodes(context.parent), i, len;

        if(!system.isArray(children)){
            var arrayChildren = [];
            for(i = 0, len = children.length; i < len; i++){
                arrayChildren[i] = children[i];
            }
            children = arrayChildren;
        }

        for(i = 1,len = children.length; i < len; i++){
            ko.removeNode(children[i]);
        }
    }

    function hide(view) {
        ko.utils.domData.set(view, visibilityKey, view.style.display);
        view.style.display = "none";
    }

    function show(view) {
        view.style.display = ko.utils.domData.get(view, visibilityKey);
    }

    function hasComposition(element){
        var dataBind = element.getAttribute('data-bind');
        if(!dataBind){
            return false;
        }

        for(var i = 0, length = composeBindings.length; i < length; i++){
            if(dataBind.indexOf(composeBindings[i]) > -1){
                return true;
            }
        }

        return false;
    }

    /**
     * @class CompositionTransaction
     * @static
     */
    var compositionTransaction = {
        /**
         * Registers a callback which will be invoked when the current composition transaction has completed. The transaction includes all parent and children compositions.
         * @method complete
         * @param {function} callback The callback to be invoked when composition is complete.
         */
        complete: function (callback) {
            compositionCompleteCallbacks.push(callback);
        }
    };

    /**
     * @class CompositionModule
     * @static
     */
    composition = {
        /**
         * An array of all the binding handler names (includeing :) that trigger a composition.
         * @property {string} composeBindings
         * @default ['compose:']
         */
        composeBindings:composeBindings,
        /**
         * Converts a transition name to its moduleId.
         * @method convertTransitionToModuleId
         * @param {string} name The name of the transtion.
         * @return {string} The moduleId.
         */
        convertTransitionToModuleId: function (name) {
            return 'transitions/' + name;
        },
        /**
         * The name of the transition to use in all compositions.
         * @property {string} defaultTransitionName
         * @default null
         */
        defaultTransitionName: null,
        /**
         * Represents the currently executing composition transaction.
         * @property {CompositionTransaction} current
         */
        current: compositionTransaction,
        /**
         * Registers a binding handler that will be invoked when the current composition transaction is complete.
         * @method addBindingHandler
         * @param {string} name The name of the binding handler.
         * @param {object} [config] The binding handler instance. If none is provided, the name will be used to look up an existing handler which will then be converted to a composition handler.
         * @param {function} [initOptionsFactory] If the registered binding needs to return options from its init call back to knockout, this function will server as a factory for those options. It will receive the same parameters that the init function does.
         */
        addBindingHandler:function(name, config, initOptionsFactory){
            var key,
                dataKey = 'composition-handler-' + name,
                handler;

            config = config || ko.bindingHandlers[name];
            initOptionsFactory = initOptionsFactory || function(){ return undefined;  };

            handler = ko.bindingHandlers[name] = {
                init: function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
                    if(compositionCount > 0){
                        var data = {
                            trigger:ko.observable(null)
                        };

                        composition.current.complete(function(){
                            if(config.init){
                                config.init(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext);
                            }

                            if(config.update){
                                ko.utils.domData.set(element, dataKey, config);
                                data.trigger('trigger');
                            }
                        });

                        ko.utils.domData.set(element, dataKey, data);
                    }else{
                        ko.utils.domData.set(element, dataKey, config);

                        if(config.init){
                            config.init(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext);
                        }
                    }

                    return initOptionsFactory(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext);
                },
                update: function (element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
                    var data = ko.utils.domData.get(element, dataKey);

                    if(data.update){
                        return data.update(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext);
                    }

                    if(data.trigger){
                        data.trigger();
                    }
                }
            };

            for (key in config) {
                if (key !== "init" && key !== "update") {
                    handler[key] = config[key];
                }
            }
        },
        /**
         * Gets an object keyed with all the elements that are replacable parts, found within the supplied elements. The key will be the part name and the value will be the element itself.
         * @method getParts
         * @param {DOMElement\DOMElement[]} elements The element(s) to search for parts.
         * @return {object} An object keyed by part.
         */
        getParts: function(elements, parts, isReplacementSearch) {
            parts = parts || {};

            if (!elements) {
                return parts;
            }

            if (elements.length === undefined) {
                elements = [elements];
            }

            for (var i = 0, length = elements.length; i < length; i++) {
                var element = elements[i];

                if (element.getAttribute) {
                    if(!isReplacementSearch && hasComposition(element)){
                        continue;
                    }

                    var id = element.getAttribute(partAttributeName);
                    if (id) {
                        parts[id] = element;
                    }

                    if(!isReplacementSearch && element.hasChildNodes()){
                        composition.getParts(element.childNodes, parts);
                    }
                }
            }

            return parts;
        },
        cloneNodes:cloneNodes,
        finalize: function (context) {
            if(context.transition === undefined) {
                context.transition = this.defaultTransitionName;
            }

            if(!context.child && !context.activeView){
                if (!context.cacheViews) {
                    ko.virtualElements.emptyNode(context.parent);
                }

                context.triggerAttach();
                endComposition();
                cleanUp(context);
            }else if (shouldTransition(context)) {
                var transitionModuleId = this.convertTransitionToModuleId(context.transition);

                system.acquire(transitionModuleId).then(function (transition) {
                    context.transition = transition;

                    transition(context).then(function () {
                        if (!context.cacheViews) {
                            if(!context.child){
                                ko.virtualElements.emptyNode(context.parent);
                            }else{
                                removePreviousView(context);
                            }
                        }else if(context.activeView){
                            var instruction = binder.getBindingInstruction(context.activeView);
                            if(instruction && instruction.cacheViews != undefined && !instruction.cacheViews){
                                ko.removeNode(context.activeView);
                            }
                        }

                        context.triggerAttach();
                        endComposition();
                        cleanUp(context);
                    });
                }).fail(function(err){
                    system.error('Failed to load transition (' + transitionModuleId + '). Details: ' + err.message);
                });
            } else {
                if (context.child != context.activeView) {
                    if (context.cacheViews && context.activeView) {
                        var instruction = binder.getBindingInstruction(context.activeView);
                        if(!instruction || (instruction.cacheViews != undefined && !instruction.cacheViews)){
                            ko.removeNode(context.activeView);
                        }else{
                            hide(context.activeView);
                        }
                    }

                    if (!context.child) {
                        if (!context.cacheViews) {
                            ko.virtualElements.emptyNode(context.parent);
                        }
                    } else {
                        if (!context.cacheViews) {
                            removePreviousView(context);
                        }

                        show(context.child);
                    }
                }

                context.triggerAttach();
                endComposition();
                cleanUp(context);
            }
        },
        bindAndShow: function (child, context, skipActivation) {
            context.child = child;

            if (context.cacheViews) {
                context.composingNewView = (ko.utils.arrayIndexOf(context.viewElements, child) == -1);
            } else {
                context.composingNewView = true;
            }

            tryActivate(context, function () {
                if (context.binding) {
                    context.binding(context.child, context.parent, context);
                }

                if (context.preserveContext && context.bindingContext) {
                    if (context.composingNewView) {
                        if(context.parts){
                            replaceParts(context);
                        }

                        hide(child);
                        ko.virtualElements.prepend(context.parent, child);

                        binder.bindContext(context.bindingContext, child, context.model);
                    }
                } else if (child) {
                    var modelToBind = context.model || dummyModel;
                    var currentModel = ko.dataFor(child);

                    if (currentModel != modelToBind) {
                        if (!context.composingNewView) {
                            ko.removeNode(child);
                            viewEngine.createView(child.getAttribute('data-view')).then(function(recreatedView) {
                                composition.bindAndShow(recreatedView, context, true);
                            });
                            return;
                        }

                        if(context.parts){
                            replaceParts(context);
                        }

                        hide(child);
                        ko.virtualElements.prepend(context.parent, child);

                        binder.bind(modelToBind, child);
                    }
                }

                composition.finalize(context);
            }, skipActivation);
        },
        /**
         * Eecutes the default view location strategy.
         * @method defaultStrategy
         * @param {object} context The composition context containing the model and possibly existing viewElements.
         * @return {promise} A promise for the view.
         */
        defaultStrategy: function (context) {
            return viewLocator.locateViewForObject(context.model, context.area, context.viewElements);
        },
        getSettings: function (valueAccessor, element) {
            var value = valueAccessor(),
                settings = ko.utils.unwrapObservable(value) || {},
                activatorPresent = activator.isActivator(value),
                moduleId;

            if (system.isString(settings)) {
                if (viewEngine.isViewUrl(settings)) {
                    settings = {
                        view: settings
                    };
                } else {
                    settings = {
                        model: settings,
                        activate: true
                    };
                }

                return settings;
            }

            moduleId = system.getModuleId(settings);
            if (moduleId) {
                settings = {
                    model: settings,
                    activate: true
                };

                return settings;
            }

            if(!activatorPresent && settings.model) {
                activatorPresent = activator.isActivator(settings.model);
            }

            for (var attrName in settings) {
                if (ko.utils.arrayIndexOf(bindableSettings, attrName) != -1) {
                    settings[attrName] = ko.utils.unwrapObservable(settings[attrName]);
                } else {
                    settings[attrName] = settings[attrName];
                }
            }

            if (activatorPresent) {
                settings.activate = false;
            } else if (settings.activate === undefined) {
                settings.activate = true;
            }

            return settings;
        },
        executeStrategy: function (context) {
            context.strategy(context).then(function (child) {
                composition.bindAndShow(child, context);
            });
        },
        inject: function (context) {
            if (!context.model) {
                this.bindAndShow(null, context);
                return;
            }

            if (context.view) {
                viewLocator.locateView(context.view, context.area, context.viewElements).then(function (child) {
                    composition.bindAndShow(child, context);
                });
                return;
            }

            if (!context.strategy) {
                context.strategy = this.defaultStrategy;
            }

            if (system.isString(context.strategy)) {
                system.acquire(context.strategy).then(function (strategy) {
                    context.strategy = strategy;
                    composition.executeStrategy(context);
                }).fail(function(err){
                    system.error('Failed to load view strategy (' + context.strategy + '). Details: ' + err.message);
                });
            } else {
                this.executeStrategy(context);
            }
        },
        /**
         * Initiates a composition.
         * @method compose
         * @param {DOMElement} element The DOMElement or knockout virtual element that serves as the parent for the composition.
         * @param {object} settings The composition settings.
         * @param {object} [bindingContext] The current binding context.
         */
        compose: function (element, settings, bindingContext, fromBinding) {
            compositionCount++;

            if(!fromBinding){
                settings = composition.getSettings(function() { return settings; }, element);
            }

            if (settings.compositionComplete) {
                compositionCompleteCallbacks.push(function () {
                    settings.compositionComplete(settings.child, settings.parent, settings);
                });
            }

            compositionCompleteCallbacks.push(function () {
                if(settings.composingNewView && settings.model && settings.model.compositionComplete){
                    settings.model.compositionComplete(settings.child, settings.parent, settings);
                }
            });

            var hostState = getHostState(element);

            settings.activeView = hostState.activeView;
            settings.parent = element;
            settings.triggerAttach = triggerAttach;
            settings.bindingContext = bindingContext;

            if (settings.cacheViews && !settings.viewElements) {
                settings.viewElements = hostState.childElements;
            }

            if (!settings.model) {
                if (!settings.view) {
                    this.bindAndShow(null, settings);
                } else {
                    settings.area = settings.area || 'partial';
                    settings.preserveContext = true;

                    viewLocator.locateView(settings.view, settings.area, settings.viewElements).then(function (child) {
                        composition.bindAndShow(child, settings);
                    });
                }
            } else if (system.isString(settings.model)) {
                system.acquire(settings.model).then(function (module) {
                    settings.model = system.resolveObject(module);
                    composition.inject(settings);
                }).fail(function(err){
                    system.error('Failed to load composed module (' + settings.model + '). Details: ' + err.message);
                });
            } else {
                composition.inject(settings);
            }
        }
    };

    ko.bindingHandlers.compose = {
        init: function() {
            return { controlsDescendantBindings: true };
        },
        update: function (element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
            var settings = composition.getSettings(valueAccessor, element);
            if(settings.mode){
                var data = ko.utils.domData.get(element, compositionDataKey);
                if(!data){
                    var childNodes = ko.virtualElements.childNodes(element);
                    data = {};

                    if(settings.mode === 'inline'){
                        data.view = viewEngine.ensureSingleElement(childNodes);
                    }else if(settings.mode === 'templated'){
                        data.parts = cloneNodes(childNodes);
                    }

                    ko.virtualElements.emptyNode(element);
                    ko.utils.domData.set(element, compositionDataKey, data);
                }

                if(settings.mode === 'inline'){
                    settings.view = data.view.cloneNode(true);
                }else if(settings.mode === 'templated'){
                    settings.parts = data.parts;
                }

                settings.preserveContext = true;
            }

            composition.compose(element, settings, bindingContext, true);
        }
    };

    ko.virtualElements.allowedBindings.compose = true;

    return composition;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Durandal events originate from backbone.js but also combine some ideas from signals.js as well as some additional improvements.
 * Events can be installed into any object and are installed into the `app` module by default for convenient app-wide eventing.
 * @module events
 * @requires system
 */
define('durandal/events',['durandal/system'], function (system) {
    var eventSplitter = /\s+/;
    var Events = function() { };

    /**
     * Represents an event subscription.
     * @class Subscription
     */
    var Subscription = function(owner, events) {
        this.owner = owner;
        this.events = events;
    };

    /**
     * Attaches a callback to the event subscription.
     * @method then
     * @param {function} callback The callback function to invoke when the event is triggered.
     * @param {object} [context] An object to use as `this` when invoking the `callback`.
     * @chainable
     */
    Subscription.prototype.then = function (callback, context) {
        this.callback = callback || this.callback;
        this.context = context || this.context;
        
        if (!this.callback) {
            return this;
        }

        this.owner.on(this.events, this.callback, this.context);
        return this;
    };

    /**
     * Attaches a callback to the event subscription.
     * @method on
     * @param {function} [callback] The callback function to invoke when the event is triggered. If `callback` is not provided, the previous callback will be re-activated.
     * @param {object} [context] An object to use as `this` when invoking the `callback`.
     * @chainable
     */
    Subscription.prototype.on = Subscription.prototype.then;

    /**
     * Cancels the subscription.
     * @method off
     * @chainable
     */
    Subscription.prototype.off = function () {
        this.owner.off(this.events, this.callback, this.context);
        return this;
    };

    /**
     * Creates an object with eventing capabilities.
     * @class Events
     */

    /**
     * Creates a subscription or registers a callback for the specified event.
     * @method on
     * @param {string} events One or more events, separated by white space.
     * @param {function} [callback] The callback function to invoke when the event is triggered. If `callback` is not provided, a subscription instance is returned.
     * @param {object} [context] An object to use as `this` when invoking the `callback`.
     * @return {Subscription|Events} A subscription is returned if no callback is supplied, otherwise the events object is returned for chaining.
     */
    Events.prototype.on = function(events, callback, context) {
        var calls, event, list;

        if (!callback) {
            return new Subscription(this, events);
        } else {
            calls = this.callbacks || (this.callbacks = {});
            events = events.split(eventSplitter);

            while (event = events.shift()) {
                list = calls[event] || (calls[event] = []);
                list.push(callback, context);
            }

            return this;
        }
    };

    /**
     * Removes the callbacks for the specified events.
     * @method off
     * @param {string} [events] One or more events, separated by white space to turn off. If no events are specified, then the callbacks will be removed.
     * @param {function} [callback] The callback function to remove. If `callback` is not provided, all callbacks for the specified events will be removed.
     * @param {object} [context] The object that was used as `this`. Callbacks with this context will be removed.
     * @chainable
     */
    Events.prototype.off = function(events, callback, context) {
        var event, calls, list, i;

        // No events
        if (!(calls = this.callbacks)) {
            return this;
        }

        //removing all
        if (!(events || callback || context)) {
            delete this.callbacks;
            return this;
        }

        events = events ? events.split(eventSplitter) : system.keys(calls);

        // Loop through the callback list, splicing where appropriate.
        while (event = events.shift()) {
            if (!(list = calls[event]) || !(callback || context)) {
                delete calls[event];
                continue;
            }

            for (i = list.length - 2; i >= 0; i -= 2) {
                if (!(callback && list[i] !== callback || context && list[i + 1] !== context)) {
                    list.splice(i, 2);
                }
            }
        }

        return this;
    };

    /**
     * Triggers the specified events.
     * @method trigger
     * @param {string} [events] One or more events, separated by white space to trigger.
     * @chainable
     */
    Events.prototype.trigger = function(events) {
        var event, calls, list, i, length, args, all, rest;
        if (!(calls = this.callbacks)) {
            return this;
        }

        rest = [];
        events = events.split(eventSplitter);
        for (i = 1, length = arguments.length; i < length; i++) {
            rest[i - 1] = arguments[i];
        }

        // For each event, walk through the list of callbacks twice, first to
        // trigger the event, then to trigger any `"all"` callbacks.
        while (event = events.shift()) {
            // Copy callback lists to prevent modification.
            if (all = calls.all) {
                all = all.slice();
            }

            if (list = calls[event]) {
                list = list.slice();
            }

            // Execute event callbacks.
            if (list) {
                for (i = 0, length = list.length; i < length; i += 2) {
                    list[i].apply(list[i + 1] || this, rest);
                }
            }

            // Execute "all" callbacks.
            if (all) {
                args = [event].concat(rest);
                for (i = 0, length = all.length; i < length; i += 2) {
                    all[i].apply(all[i + 1] || this, args);
                }
            }
        }

        return this;
    };

    /**
     * Creates a function that will trigger the specified events when called. Simplifies proxying jQuery (or other) events through to the events object.
     * @method proxy
     * @param {string} events One or more events, separated by white space to trigger by invoking the returned function.
     * @return {function} Calling the function will invoke the previously specified events on the events object.
     */
    Events.prototype.proxy = function(events) {
        var that = this;
        return (function(arg) {
            that.trigger(events, arg);
        });
    };

    /**
     * Creates an object with eventing capabilities.
     * @class EventsModule
     * @static
     */

    /**
     * Adds eventing capabilities to the specified object.
     * @method includeIn
     * @param {object} targetObject The object to add eventing capabilities to.
     */
    Events.includeIn = function(targetObject) {
        targetObject.on = Events.prototype.on;
        targetObject.off = Events.prototype.off;
        targetObject.trigger = Events.prototype.trigger;
        targetObject.proxy = Events.prototype.proxy;
    };

    return Events;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The app module controls app startup, plugin loading/configuration and root visual display.
 * @module app
 * @requires system
 * @requires viewEngine
 * @requires composition
 * @requires events
 * @requires jquery
 */
define('durandal/app',['durandal/system', 'durandal/viewEngine', 'durandal/composition', 'durandal/events', 'jquery'], function(system, viewEngine, composition, Events, $) {
    var app,
        allPluginIds = [],
        allPluginConfigs = [];

    function loadPlugins(){
        return system.defer(function(dfd){
            if(allPluginIds.length == 0){
                dfd.resolve();
                return;
            }

            system.acquire(allPluginIds).then(function(loaded){
                for(var i = 0; i < loaded.length; i++){
                    var currentModule = loaded[i];

                    if(currentModule.install){
                        var config = allPluginConfigs[i];
                        if(!system.isObject(config)){
                            config = {};
                        }

                        currentModule.install(config);
                        system.log('Plugin:Installed ' + allPluginIds[i]);
                    }else{
                        system.log('Plugin:Loaded ' + allPluginIds[i]);
                    }
                }

                dfd.resolve();
            }).fail(function(err){
                system.error('Failed to load plugin(s). Details: ' + err.message);
            });
        }).promise();
    }

    /**
     * @class AppModule
     * @static
     * @uses Events
     */
    app = {
        /**
         * The title of your application.
         * @property {string} title
         */
        title: 'Application',
        /**
         * Configures one or more plugins to be loaded and installed into the application.
         * @method configurePlugins
         * @param {object} config Keys are plugin names. Values can be truthy, to simply install the plugin, or a configuration object to pass to the plugin.
         * @param {string} [baseUrl] The base url to load the plugins from.
         */
        configurePlugins:function(config, baseUrl){
            var pluginIds = system.keys(config);
            baseUrl = baseUrl || 'plugins/';

            if(baseUrl.indexOf('/', baseUrl.length - 1) === -1){
                baseUrl += '/';
            }

            for(var i = 0; i < pluginIds.length; i++){
                var key = pluginIds[i];
                allPluginIds.push(baseUrl + key);
                allPluginConfigs.push(config[key]);
            }
        },
        /**
         * Starts the application.
         * @method start
         * @return {promise}
         */
        start: function() {
            system.log('Application:Starting');

            if (this.title) {
                document.title = this.title;
            }

            return system.defer(function (dfd) {
                $(function() {
                    loadPlugins().then(function(){
                        dfd.resolve();
                        system.log('Application:Started');
                    });
                });
            }).promise();
        },
        /**
         * Sets the root module/view for the application.
         * @method setRoot
         * @param {string} root The root view or module.
         * @param {string} [transition] The transition to use from the previous root (or splash screen) into the new root.
         * @param {string} [applicationHost] The application host element or id. By default the id 'applicationHost' will be used.
         */
        setRoot: function(root, transition, applicationHost) {
            var hostElement, settings = { activate:true, transition: transition };

            if (!applicationHost || system.isString(applicationHost)) {
                hostElement = document.getElementById(applicationHost || 'applicationHost');
            } else {
                hostElement = applicationHost;
            }

            if (system.isString(root)) {
                if (viewEngine.isViewUrl(root)) {
                    settings.view = root;
                } else {
                    settings.model = root;
                }
            } else {
                settings.model = root;
            }

            composition.compose(hostElement, settings);
        }
    };

    Events.includeIn(app);

    return app;
});

define('browserSupport',[], function () {
    var ua = navigator.userAgent.toLowerCase();

    var browserSupport = {
        isSupportedMobile: null,
        isSupportedBrowser: null,
        isMobileDevice: null,
        isFirefox: null,
        isChromeWithPageCoordsBug: null,
        isSafari: false,

        initialize: initialize
    };

    return browserSupport;

    function initialize() {
        var browserInfo = getBrowserInfo();
        browserSupport.isFirefox = isFirefox(browserInfo);

        browserSupport.isMobileDevice = isMobileDevice();
        browserSupport.isSupportedMobile = isSupportedMobile();
        browserSupport.isSupportedBrowser = isSupportedBrowser(browserInfo);
        browserSupport.isChromeWithPageCoordsBug = isChromeWithPageCoordsBug();
        browserSupport.isSafari = /^((?!chrome|android).)*safari/i.test(navigator.userAgent);
    };

    function isChromeWithPageCoordsBug() {
        if (ua.match(/(chrome)\/?\s*([\d\.]+)/i)) {
            return window.navigator.appVersion.match(/Chrome\/(.*?) /)[1] == "38.0.2125.102";
        }
        return false;
    }



    function getBrowserInfo() {
        var N = navigator.appName,
            tem,
            M = ua.match(/(chrome|safari|firefox|msie)\/?\s*([\d\.]+)/i) || [];

        M = M[2] ? [M[1], M[2]] : [N, navigator.appVersion, '-?'];
        if (M && (tem = ua.match(/version\/([\.\d]+)/i)) != null) M[2] = tem[1];

        var browser = M[0].toLowerCase();
        var version = parseInt(M[1], 10);

        return { browser: browser, version: version };
    };

    function isFirefox(browserInfo) {
        return browserInfo.browser == "firefox";
    };

    function isMobileDevice() {
        if (ua.indexOf("ipod") != -1 ||
            ua.indexOf("iphone") != -1 ||
            ua.indexOf("ipad") != -1 ||
            ua.indexOf("android") != -1)
            return true;
        return false;
    };

    function isSupportedMobile() {
        if (/iP(hone|od|ad)/.test(navigator.platform)) {
            var v = (navigator.appVersion).match(/OS (\d+)_(\d+)_?(\d+)?/);
            if (parseInt(v[1], 10) >= 6) {
                return true;
            }
        }

        //Android is supported but not chrome, opera or firefox
        if (ua.indexOf("android") != -1 &&
            ua.indexOf("opera") == -1 &&
            ua.indexOf("firefox") == -1) {
            return true;
        }

        return false;
    };

    function isSupportedBrowser(browserInfo) {
        // Opera is not supported
        if (navigator.appName.toLowerCase() == "opera" || navigator.userAgent.indexOf("OPR") != -1)
            return false;

        //Chrome 28+, Firefox 22+, Safari 5+ are supported
        var browser = browserInfo.browser;
        var version = browserInfo.version;

        if (browser == "chrome" && version >= 28 ||
                browser == "firefox" && version >= 22 ||
                browser == "safari" && version >= 533)
            return true;

        // IE 11 is supported
        if (ua.indexOf('rv:11.0') != -1) {
            return true;
        }

        return false;
    };
});
define('bootstrapping/browserSupportTask',['browserSupport'], function (browserSupport) {
    return {
        execute: function () {
            browserSupport.initialize();
        }
    };
});
define('bootstrapping/viewLocatorTask',['durandal/viewLocator'], function (viewLocator) {

    return {
        execute: function () {
            viewLocator.useConvention();
        }
    };

});
define('bootstrapping/knockoutBindings/areaBinding',[],function () {
    ko.bindingHandlers.area = {
        init: function (element, valueAccessor) {
            var
                value = valueAccessor(),
                click = value ? value.click : undefined,
                offset, x, y
            ;

            $(element).on('click', handler);
            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $(element).off('click', handler);
            });

            function handler(e) {
                offset = $(element).offset();
                x = e.pageX - offset.left;
                y = e.pageY - offset.top;

                if (typeof (click) == "function") {
                    click({ x: x, y: y });
                }
            }

        }
    };
});
define('bootstrapping/knockoutBindings/autofocusBinding',['knockout', 'jquery', 'durandal/composition'], function (ko, $, composition) {

    ko.bindingHandlers.autofocus = {
        init: function (element) {
            var $element = $(element);
            $element.find('.autofocus').first().focus();
        }
    };

    composition.addBindingHandler('autofocus');

});
define('bootstrapping/knockoutBindings/autosizeBinding',['jquery', 'knockout', 'durandal/composition'], function ($, ko, composition) {
    ko.bindingHandlers.autosize = {
        init: function (element, valueAccessor, allBindingsAccessor, data, context) {
            autosize($(element), { setOverflowX: false, setOverflowY: false });
        },
        update: function (element, valueAccessor, allBindingsAccessor, data, context) {
            autosize.update($(element));
        }
    }

    composition.addBindingHandler('autosize');
});
define('bootstrapping/knockoutBindings/backgroundBinding',[],function () {
    ko.bindingHandlers.background = {
        init: function (element) {
            $(element)
                .css('background-position', '0 0')
                .css('background-repeat', 'no-repeat');
        },
        update: function (element, valueAccessor) {
            var value = valueAccessor();
            if (value) {
                var src = ko.utils.unwrapObservable(value);
                var image = new Image();
                image.src = src;

                image.onload = function () {
                    $(element)
                        .css('background-image', 'url(' + src + ')')
                        .css('height', image.height)
                        .css('width', image.width);
                }
            }
        }
    };
});
define('templateSettings',[],function() {
    'use strict';

    return {
        init: function(settings) {
            for (var key in settings) {
                this[key] = settings[key];
            }
        },
        getColorValue: function(color) {
            var settingColor = _.find(this.colors, function(c) {
                return c.key === color;
            });

            if (!settingColor) {
                throw 'Variable "' + color + '" does not exist in template settings';
            }

            return settingColor.value;
        }
    };
});
define('bootstrapping/knockoutBindings/circleProgressBinding',['durandal/composition', 'templateSettings', 'less'], function (composition, templateSettings, less) {
    
        ko.bindingHandlers.circleProgress = {
            update: function(element, valueAccessor) {
    
                var $element = $(element),
                    score = valueAccessor().progress || 0,
                    lineWidth = parseInt($element.css('column-width')) || valueAccessor().lineWidth || 4,
    
                    centerX = element.width / 2,
                    centerY = element.height / 2,
                    radius = valueAccessor().radius || (centerX < centerY ? centerX : centerY - lineWidth / 2 - 1),
                    progress = score / 100,
                    cnxt = element.getContext('2d'),
                    masteryScore = valueAccessor().masteryScore,
                    isMastered = valueAccessor().isMastered;
    
                if (isMastered) {
                    $element.addClass('mastered')
                }

                var colors = {
                    basic: $element.css('color') || valueAccessor().basicColor || getDefaultBasicColor(),
                    progress: $element.css('border-top-color') || $element.css('border-color') || valueAccessor().progressColor || templateSettings.getColorValue('@secondary-color'),
                    mastered: $element.css('outline-color') || valueAccessor().masteredColor || templateSettings.getColorValue('@correct-color')
                };
    
                if (progress > 0) {
                    cnxt.beginPath();
                    cnxt.strokeStyle = isMastered ? colors.mastered : colors.progress;
                    cnxt.lineWidth = lineWidth;
    
                    if (progress == 1) {
                        cnxt.arc(centerX, centerY, radius, 0, 2 * Math.PI);
                        cnxt.closePath();
                        cnxt.stroke();
                    } else {
                        cnxt.arc(centerX, centerY, radius, 1.5 * Math.PI, (progress * 2 - .5) * Math.PI);
                        cnxt.stroke();       
    
                        cnxt.beginPath();
                        cnxt.arc(centerX, centerY, radius, (progress * 2 - .5) * Math.PI, 1.5 * Math.PI);
                        cnxt.strokeStyle = colors.basic;
                        cnxt.lineWidth = lineWidth;
                        cnxt.stroke();
                    }
                } else {                    
                    cnxt.beginPath();
                    cnxt.arc(centerX, centerY, radius, 0, 2 * Math.PI);
                    cnxt.strokeStyle = colors.basic;
                    cnxt.lineWidth = lineWidth;
                    cnxt.closePath();
                    cnxt.stroke();
                }
    
            }
        };

        function getDefaultBasicColor() {
            var c = new less.tree.Color(templateSettings.getColorValue('@text-color').slice(1));
            var dimension = new less.tree.Dimension('95');
            var func = less.functions.functionRegistry.get('tint');
            return func(c, dimension).toRGB();
        }
    });
define('bootstrapping/knockoutBindings/contextBinding',[], function () {

    ko.bindingHandlers.context = {
        init: function (element, valueAccessor, allBindingsAccessor, viewModel) {
            viewModel.__context__ = element.getContext('2d');
        },
        update: function (element, valueAccessor, allBindingsAccessor, viewModel) {
            var callback = ko.utils.unwrapObservable(allBindingsAccessor().contextCallback);
            callback.call(viewModel, viewModel.__context__);
        }
    };
});
define('helpers/url',[],function() {
    return function Url(url) {
        url = url || '';

        var that = this;
        that.toString = function() {
          return url;
        };
        that.addQueryStringParam = function (key, value) {
            value = encodeURIComponent(value);

            var regex = new RegExp('([?&])' + key + '=.*?(&|$)', 'i');
            var separator = url.indexOf('?') !== -1 ? '&' : '?';
            if (url.match(regex)) {
                url = url.replace(regex, '$1' + key + '=' + value + '$2');
            } else {
                url = url + separator + key + '=' + value;
            }

            return that;
        }
    };
});

define('bootstrapping/knockoutBindings/elementsWrapBinding',['knockout', 'jquery', 'durandal/composition', 'helpers/url'], function (ko, $, composition, Url) {

    ko.bindingHandlers.elementsWrap = {
        init: function (element) {
            wrapElement(element);
        },

        update: function (element) {
            wrapElement(element);
        }
    };

    function wrapElement(element) {
        var $element = $(element),
                imageWrapper = '<figure class="image-wrapper"></figure>',
                tableWrapper = '<figure class="table-wrapper"></figure>';

        $('img', $element).each(function (index, image) {
            var $image = $(image),
                $wrapper = $(imageWrapper).css('float', $image[0].style.cssFloat),
                $parent = $image.parent();

            if ($image.closest('.cropped-image').length > 0 || $parent.hasClass('image-wrapper')) {
                return;
            }

            if ($parent.prop('tagName') == "TD" && $parent[0].style.width == "") {
                $wrapper.css('width', $image[0].style.width);
                $wrapper.css('height', $image[0].style.height);
            }

            // moved to css, because of IE11 crash on Windows 10 => .image-wrapper img {height:auto!important;float:none!important;}
            //$image.height('auto');
            //$image.css('float', 'none');
            //
            $image.wrap($wrapper);
        });

        $('table', $element).each(function (index, table) {
            var $table = $(table),
                $wrapper = $(tableWrapper).css('float', $table.attr('align'));

            $table.attr('align', 'center');
            $table.wrap($wrapper);
        });

        $(
          ".audio-editor iframe, .video-editor iframe," +
            'iframe[data-media-type="old-editor-video"], iframe[data-media-type="old-editor-audio"]',
          $element
        ).each(function(index, iframe) {
          var $iframe = $(iframe);
          var src = $iframe.attr("src");

          $iframe.attr(
            "src",
            new Url(src)
              .addQueryStringParam("style_variables", getStyles())
              .toString()
          );
        });
    }

    composition.addBindingHandler('elementsWrap');

    function getStyles() {
        return window.LessProcessor && window.LessProcessor.vars ? JSON.stringify({ '@main-color': window.LessProcessor.vars['@main-color'], '@content-body-color': window.LessProcessor.vars['@content-body-color'], '@text-color': window.LessProcessor.vars['@text-color'] }) : undefined;
    }

});
define('bootstrapping/knockoutBindings/focusHandlerBinding',[],function() {

    ko.bindingHandlers.focusHandler = {
        update: function() {

            var elem = '.xapi-form-input',
                wrapper = elem + '-wrapper';

            $(elem).focusin(function () { })
                .blur(function () { })
                .focusin();

            $(wrapper).click(function() {
                if ($(this).parent().find(elem).is(':focus')) {
                    $(this).bind('click mousedown mouseup', function() {
                        return false;
                    });
                } else {
                    $(this).bind('click mousedown mouseup', function() {
                        return false;
                    }).parent().find(elem).focus();
                }
            });

            $(elem).hover(function() {
                $(this).parent().find(wrapper).toggleClass('hover');
            });

            $(elem).focusin(function() {
                $(this).parent().find(wrapper).addClass('focus');
            });
            $(elem).blur(function() {
                $(this).parent().find(wrapper).removeClass('focus');
            });
        }
    };

});
define('bootstrapping/knockoutBindings/toggleVisibleBindingHandler',[],function () {

    ko.bindingHandlers.toggleVisible = {
        update: function (element, valueAccessor) {
            var $element = $(element),
                isExpanded = valueAccessor().isExpanded,
                $animationContainer = $element.children('[data-animate]'),
                speed = 300;

            if (isExpanded) {
                $animationContainer.css('height', '').hide().slideDown(speed, function(){$animationContainer.css('overflow', '')});
            } else {
                $animationContainer.css('overflow','hidden').animate({height: 0}, speed);
            }
        }
    }
});
define('bootstrapping/knockoutBindings/dropDownBinding',[],function () {

    ko.bindingHandlers.dropDown = {
        init: function (element) {
            var $element = $(element),
                expandedClass = 'expanded',
                disabledClass = 'disabled',
                $html = $('html');

            $element.on('click', function () {
                if ($element.hasClass(expandedClass) || $element.hasClass(disabledClass)) {
                    return;
                } else {
                    $element.addClass(expandedClass);

                    _.defer(function () {
                        $html.on('click', hide);
                    });
                }
            });

            function hide() {
                $element.removeClass(expandedClass);
                $html.off('click', hide);
            };

        }
    }
});
define('bootstrapping/knockoutBindings/scrollTopBinding',['knockout'], function (ko) {
    'use strict';

    ko.bindingHandlers.scrollTop = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var time = ko.unwrap(valueAccessor());
            var $element = $(element);
            var $document = $(document);
            var isScrolling = false;
            var isEnabled = false;

            $element.hide().css('opacity', '0');
            
            $document.scroll(function() {
                var y = $(this).scrollTop();

                if (y > 0) {
                    if(!isEnabled) {
                        $element
                            .show()       
                            .animate({bottom: "+=10", opacity: '1'}, 300);

                        isEnabled = true;
                    }
                } else {
                    $element                   
                        .animate({bottom: "-=10", opacity: '0'}, 300, function(){
                                if(!isEnabled)
                                    $element.hide();
                            });

                    isEnabled = false;
                }
            }); 

            $element.bind('click', function(){
                if(!isScrolling && isEnabled){
                    isScrolling = true;

                    $('html, body').animate({scrollTop: 0}, time, function(){
                        isScrolling = false;
                    });
                }
            });
        }
    };
});
define('bootstrapping/knockoutBindings/scrollIntoViewBinding',['durandal/composition'], function (composition) {

    ko.bindingHandlers.scrollIntoView = {
        update: function (element, valueAccessor) {
            var $element = $(element),
                isVisible = ko.utils.unwrapObservable(valueAccessor());

            if (!isVisible) {
                return;
            }

            scrollToElement($element);
        }
    };

    function scrollToElement($element) {
        _.defer(function () {
            if (!$element.is(':hidden')){
                $('html:not(:animated),body:not(:animated)').animate({
                    scrollTop: $element.offset().top
                }, function () { });
            }
        });
    }

    composition.addBindingHandler('scrollIntoView');

});
define('localizationManager',[], function () {
    return {
        getLocalizedText: function(value) {
            if (_.isString(value)) {
                return TranslationPlugin.getTextByKey(value);
            } else if (_.isObject(value)) {
                var text = TranslationPlugin.getTextByKey(value.key);

                for (var replacement in value.replace) {
                    text = text.replace('{' + replacement + '}', value.replace[replacement]);
                }

                return text;
            }
        }
    }
});
define('constants',[],function () {
    'use strict';

    return {
        questionTypes: {
            multipleSelect: 'multipleSelect',
            fillInTheBlank: 'fillInTheBlank',
            dragAndDrop: 'dragAndDropText',
            singleSelectText: 'singleSelectText',
            singleSelectImage: 'singleSelectImage',
            textMatching: 'textMatching',
            hotspot: 'hotspot',
            statement: 'statement',
            informationContent: 'informationContent',
            openQuestion: 'openQuestion',
            scenario: 'scenario',
            rankingText: 'rankingText'
        },
        defaultImageUrl: 'css/img/single-select-image-anwer.png',
        events: {
            appClosed: 'appClosed',
            onError: 'onError',
            appInitialized: 'appInitialized',
            imagePreviewClosed: 'image:preview:closed'
        },
        localStorageProgressKey: 'course_progress',
        localStorageResultKey: 'course_result',
        course: {
            statuses: {
                completed: 'completed',
                failed: 'failed',
                inProgress: 'inProgress'
            }
        },
        progressContext: {
            statuses: {
                undefined: 'undefined',
                saved: 'saved',
                error: 'error',
                ignored: 'ignored'
            }
        },
        documents: {
            types: {
                pdf: 'pdf',
                word: 'word',
                exel: 'exel',
                powerpoint: 'powerpoint',
                zip: 'zip'
            },
            downloadLocalizationKey: '[download]',
            containerSelector: '.document-container',
            downloadBtnSelector: '.download-document-btn',
            documentTitleWrapperSelector: '.document-title-wrapper',
            iframeSelector: '.document-block-iframe-wrapper iframe',

            sizeAttrName: 'data-document-size',
            typeAttrName: 'data-document-type'
        },
        dialogs: {
            resendResults: {
                webhooks: {
                    resultsSendErrorTitleKey: '[results send error title]',
                    endpointNameKey: '[webhooks endpoint name key]'
                }    
            }
        },
        loginTypes: {
            byToken: 'byToken',
            byEmailAndName: 'byEmailAndName',
            byEmailAndNameWithoutCrossDeviceSaving: 'byEmailAndNameWithoutCrossDeviceSaving',
            default: 'default'
        }
    };

});

define('helpers/documentBlock',['localizationManager', 'constants'], function (localizationManager, constants) {
    return {
        getDocumentBlockContent: function (html) {
            var $output = $('<output>').html(html),
                $container = $output.find(constants.documents.containerSelector);

            var downloadText = localizationManager.getLocalizedText(constants.documents.downloadLocalizationKey);
            var documentData = {
                type: $container.attr(constants.documents.typeAttrName),
                size: +$container.attr(constants.documents.sizeAttrName)
            };

            var documentSizeString = getSize(documentData.size);
            var downloadBtnText = downloadText + ' (' + documentSizeString + ')';

            if(documentData.type === constants.documents.types.zip) {
                $container.addClass(documentData.type);
            }

            $output.find(constants.documents.downloadBtnSelector)
                .text(downloadBtnText);

            var iconClass = documentData.type === constants.documents.types.zip ? 'icon-zip' : 'icon-file';
            var $typeIcon = $('<div class="icon-container">' +
                '<span class="document-type-text">' + documentData.type + '</span>' +
                '</div>')
                .addClass(iconClass);
            var $typeIconWrapper = $('<div class="document-icon"></div>')
                .append($typeIcon);

            var $documentInfo = $(constants.documents.documentTitleWrapperSelector, $container)
                .prepend($typeIconWrapper);

            var content = $output.children()[0];
            return content;
        }
    };
    /**
     * Pluralize document size to user friendly string
     *
     * @param {Number} size - size of document in bytes
     * */
    function getSize(size) {
        var sizeStr = '';
        if (!size || size < 1024) {
            return '0 Kb';
        }
        sizeStr = (size / (1024 * 1024)).toFixed(2);

        return sizeStr + ' MB';
    }
});

define('bootstrapping/knockoutBindings/contentBlockBinding',['durandal/composition', 'helpers/documentBlock'], function (composition, documentBlockHelper) {
    ko.bindingHandlers.contentBlock = {
        init: function (element, valueAccessor) {
            var $element = $(element),
                html = valueAccessor();
            
            var dataType = getContentBlockType(html);
            
            switch(dataType){
                case 'hotspot': {
                    var hotspotOnImage = HotspotStorage.create($(html)[0]);
                        
                    $element.addClass('hotspot-on-image-container');
                    $element.html(hotspotOnImage.element);

                    ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                        HotspotStorage.remove(hotspotOnImage);
                    });

                    break;
                }
                case 'document': {
                    var content = documentBlockHelper.getDocumentBlockContent(html);
                    $element.append(content);
                    break;
                }
                default: {
                    $element.html(html);
                    return ko.bindingHandlers.elementsWrap.init(element, valueAccessor);
                }
            }
        }
    };
    
    composition.addBindingHandler('contentBlock');
    
    function getContentBlockType(data) {
        var $output = $('<output>').html(data),
            dataType = $('[data-type]', $output).attr('data-type');
        
        return dataType;
    }
});
define('bootstrapping/knockoutBindings/toolTipBindingHandler',['durandal/composition'], function (composition) {
    ko.bindingHandlers.toolTip = {
        init: function (element, valueAccessor) {
            var $element = $(element),
               lineWidth = valueAccessor().lineWidth || 4,
               score = valueAccessor().progress || 0,
               masteryScore = valueAccessor().masteryScore,
               centerX = $element.width() / 2,
               centerY = $element.height() / 2,
               radius = valueAccessor().radius || (centerX < centerY ? centerX : centerY - lineWidth / 2 - 1),
               $canvasParent = $element.parent();
            if (masteryScore) {
                var $toolTip = $('<div />')
                    .addClass('mastered-score-tooltip').appendTo($canvasParent);


                var $icon = $('<span> /')
                    .addClass('material-icons')
                    .text('done')
                    .appendTo($toolTip)

               

                var $tootlTipText = $('<span />')
                    .addClass('mastered-score-tooltip-text')
                    .text(getScoreTooltipText(masteryScore, score))
                    .appendTo($toolTip);


                $canvasParent.hover(showTooltip, hideTooltip);
                $toolTip.hover(showTooltip, hideTooltip);

                $canvasParent.focusin(showTooltip);
                $canvasParent.focusout(hideTooltip);

                $toolTip.focusin(showTooltip);
                $toolTip.focusout(hideTooltip);

                updateTooltipPosition();
                
                ko.utils.domData.set(element, 'ko_tooltip', $toolTip);
                ko.utils.domData.set(element, 'ko_tooltip_text', $tootlTipText);

                $(window).on('resize', updateTooltipPosition);

                ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                    $toolTip.remove();
                    $(window).off('resize', updateTooltipPosition);
                });
            }

            function showTooltip() {
                $tootlTipText.stop().animate({
                    'opacity': 1,
                    'margin-bottom': '10px'
                }, 200);
            }
            function hideTooltip(event) {
                var e = event.toElement || event.relatedTarget;
                if (e && (e == $canvasParent[0] || e.parentNode == $canvasParent[0] || e == $toolTip[0] || e.parentNode == $toolTip[0])) {
                    return;
                }
                $tootlTipText.stop().animate({
                    'opacity': 0,
                    'margin-bottom': '15px'
                }, 200);
            }

            function updateTooltipPosition() {
                
                var masteryScoreAngle = 2 * Math.PI * (masteryScore / 100) - 0.5 * Math.PI,
                    masteryScoreX = centerX + (Math.cos(masteryScoreAngle) * (radius)),
                    masteryScoreY = centerY - (Math.sin(masteryScoreAngle) * (radius)),
                    tooltipPosX = masteryScoreX - 6,
                    toolTipPosY = masteryScoreY - 6;

                $toolTip.css({
                    'left': tooltipPosX,
                    'bottom': toolTipPosY
                });
            }
        },
        update: function (element, valueAccessor) {
            var score = valueAccessor().progress || 0,
              masteryScore = valueAccessor().masteryScore,
              $toolTip = ko.utils.domData.get(element, 'ko_tooltip'),
              $tootlTipText = ko.utils.domData.get(element, 'ko_tooltip_text');

            if ($tootlTipText) {
                $tootlTipText.text(getScoreTooltipText(masteryScore, score));
            }

            if (score >= masteryScore && $toolTip) {
                $toolTip.addClass('mastered');
            }
        }
    }

    function getScoreTooltipText(mastery, score) {
        var scoreToComplete = mastery - score;
        return scoreToComplete > 0 ? scoreToComplete + '% ' + TranslationPlugin.getTextByKey('[to complete]') : TranslationPlugin.getTextByKey('[completed]');
    }

    composition.addBindingHandler('toolTip');
});
define('helpers/fileDownloader',["download", "jquery"], function($) {
  "use strict";

  function downloadFile(url, filename) {
    return fetch(url)
      .then(function(response) {
        if (response.status !== 200) {
          return response.json().then(function(error) {
            if (error && error.message) {
              throw new Error(error.message);
            }
            throw new Error();
          });
        }

        return response.blob();
      })
      .then(function(blob) {
        var isChrome = navigator.userAgent.match("CriOS");
        var isFirefox = navigator.userAgent.match("FxiOS");
        if (!isChrome && !isFirefox) {
          download(blob, filename, blob.type);
        } else {
          //for chrome in apple devices
          var reader = new FileReader();
          var out = new Blob([blob], { type: blob.type });
          reader.onload = function(e) {
            var win = window.open();
            win.document.write(
              '<iframe src="' +
                reader.result +
                '" frameborder="0" style="display: block; width: 100vw; height: 100vh; max-width: 100%;" allowfullscreen />'
            );
          };
          reader.readAsDataURL(out);
        }
      });
  }

  return {
    downloadFile: downloadFile
  };
});

define('bootstrapping/knockoutBindings/downloadAsPdfBinding',['helpers/fileDownloader', 'helpers/url'], function (fileDownloader, Url) {
    var buttonStatuses = {
        default: 'default',
        proggress: 'proggress',
        error: 'error',
        serverError: 'server-error'
    };

    ko.bindingHandlers.downloadAsPdf = {
        init: function (element, valueAccessor) {

            var
                $element = $(element),
                filename = getFullPdfFilename((valueAccessor().title || $element.attr('title'))),
                serviceUrl = valueAccessor().serviceUrl,
                version = valueAccessor().version;

            if (location.href.indexOf('/preview/') !== -1) {
                disableLink($element);
                return;
            }

            var convertionUrl = new Url(serviceUrl + '/convert/')
                .addQueryStringParam('url', getBaseUrl() + 'pdf/index.html')
                .addQueryStringParam('version', version).toString();

            var timeoutId;

            $element.click(function () {
                if (timeoutId) {
                    clearTimeout(timeoutId);
                    timeoutId = null;
                }

                setStatus($element, buttonStatuses.proggress);

                fileDownloader.downloadFile(convertionUrl, filename)
                    .then(function(){
                        setStatus($element, buttonStatuses.default);
                    })
                    .catch(function(error) {
                        if(error.message) {
                            var el = $element.find('.server-error').text(error.message);
                            setStatus($element, buttonStatuses.serverError);
                        } else {
                            setStatus($element, buttonStatuses.error);
                        }
                        timeoutId = setTimeout(function () {
                            setStatus($element, buttonStatuses.default);
                        }, 5000);
                    });
                return false;
            });
        }
    };

    function setStatus($element, status) {
        $element
            .toggleClass(buttonStatuses.default, status === buttonStatuses.default)
            .toggleClass(buttonStatuses.proggress, status === buttonStatuses.proggress)
            .toggleClass(buttonStatuses.error, status === buttonStatuses.error)
            .toggleClass(buttonStatuses.serverError, status === buttonStatuses.serverError);
    }

    function getDateTimeString() {
        var now = new Date();
        return now.toLocaleDateString() + ' ' + now.toLocaleTimeString();
    }

    function disableLink($link) {
        $link.css({
            'opacity': '0.5'
        });
        $link.addClass('disabled').removeClass('default');
    }

    function getBaseUrl() {
        var baseUrl = location.href.replace(location.hash, '');
        baseUrl = baseUrl.replace('#', '');

        if (location.pathname && location.pathname.length > 1) {
            var pageName = location.pathname.split("/").pop();
            baseUrl = baseUrl.replace(pageName, '');
        }

        if (location.search && location.search.length > 1) {
            baseUrl = baseUrl.replace(location.search, '');
        }

        return baseUrl;
    }

    /* Fix for IE11 and Edge (files can`t saved without extension) */
    function getFullPdfFilename(title) {
        return title + '.pdf';
    }

    function hasHostname(hostname) {
        return location.host.indexOf(hostname) !== -1;
    }
})
;
define('bootstrapping/knockoutBindings/mediaPlayerBinding',['helpers/url'], function (Url) {
    ko.bindingHandlers.mediaPlayer = {
        init: function (element, valueAccessor) {

            var
                $element = $(element),
                args = valueAccessor(),
                embedCode = ko.utils.unwrapObservable(args.embedCode);

            $element.html(getMediaEmbedCode());

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $iframe = $element.find('iframe');
                if ($iframe.length) {
                    $iframe.attr('src', 'about:blank');
                }
                $element.empty();
            });

            function getMediaEmbedCode() {

                if (!embedCode)
                    return embedCode;

                var srcAttrName = 'src',
                    $container = $('<div/>').html(embedCode),
                    $iframe = $container.find('iframe'),
                    src = $iframe.attr(srcAttrName),
                    variablesList = getVariablesList();

                src = new Url(src).addQueryStringParam('style_variables', variablesList).toString();

                $iframe.attr(srcAttrName, src);
                return $container.html();
            }
        }
    };

    function getVariablesList() {
        return window.LessProcessor && window.LessProcessor.vars ? JSON.stringify({ '@main-color': window.LessProcessor.vars['@main-color'], '@content-body-color': window.LessProcessor.vars['@content-body-color'], '@text-color': window.LessProcessor.vars['@text-color'] }) : undefined;
    }
});
define('bootstrapping/knockoutBindings/customFocusBinding',[],function () {
    ko.bindingHandlers.customFocus = {
        init: function (element, valueAccessor) {
            var $element = $(element),
                value = valueAccessor(),
                focusClass = value.class || 'focused';

            $('html').bind('touchstart', removeFocusClass);
            $('html').bind('click', removeFocusClass);
            $(window).bind('blur', removeFocusClass);

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $('html').unbind('touchstart', removeFocusClass);
                $('html').unbind('click', removeFocusClass);
                $(window).unbind('blur', removeFocusClass);
            });

            $element.on('touchstart', function (e) {
                e.stopPropagation();
            });

            $element.on('click', function (e) {
                $element.addClass(focusClass);
                e.stopPropagation();
            });

            function removeFocusClass() {
                $element.removeClass(focusClass);
            }

        }
    }
});
/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * This module is based on Backbone's core history support. It abstracts away the low level details of working with browser history and url changes in order to provide a solid foundation for a router.
 * @module history
 * @requires system
 * @requires jquery
 */
define('plugins/history',['durandal/system', 'jquery'], function (system, $) {
    // Cached regex for stripping a leading hash/slash and trailing space.
    var routeStripper = /^[#\/]|\s+$/g;

    // Cached regex for stripping leading and trailing slashes.
    var rootStripper = /^\/+|\/+$/g;

    // Cached regex for detecting MSIE.
    var isExplorer = /msie [\w.]+/;

    // Cached regex for removing a trailing slash.
    var trailingSlash = /\/$/;

    // Update the hash location, either replacing the current entry, or adding
    // a new one to the browser history.
    function updateHash(location, fragment, replace) {
        if (replace) {
            var href = location.href.replace(/(javascript:|#).*$/, '');
            location.replace(href + '#' + fragment);
        } else {
            // Some browsers require that `hash` contains a leading #.
            location.hash = '#' + fragment;
        }
    };

    /**
     * @class HistoryModule
     * @static
     */
    var history = {
        /**
         * The setTimeout interval used when the browser does not support hash change events.
         * @property {string} interval
         * @default 50
         */
        interval: 50,
        /**
         * Indicates whether or not the history module is actively tracking history.
         * @property {string} active
         */
        active: false
    };
    
    // Ensure that `History` can be used outside of the browser.
    if (typeof window !== 'undefined') {
        history.location = window.location;
        history.history = window.history;
    }

    /**
     * Gets the true hash value. Cannot use location.hash directly due to a bug in Firefox where location.hash will always be decoded.
     * @method getHash
     * @param {string} [window] The optional window instance
     * @return {string} The hash.
     */
    history.getHash = function(window) {
        var match = (window || history).location.href.match(/#(.*)$/);
        return match ? match[1] : '';
    };
    
    /**
     * Get the cross-browser normalized URL fragment, either from the URL, the hash, or the override.
     * @method getFragment
     * @param {string} fragment The fragment.
     * @param {boolean} forcePushState Should we force push state?
     * @return {string} he fragment.
     */
    history.getFragment = function(fragment, forcePushState) {
        if (fragment == null) {
            if (history._hasPushState || !history._wantsHashChange || forcePushState) {
                fragment = history.location.pathname + history.location.search;
                var root = history.root.replace(trailingSlash, '');
                if (!fragment.indexOf(root)) {
                    fragment = fragment.substr(root.length);
                }
            } else {
                fragment = history.getHash();
            }
        }
        
        return fragment.replace(routeStripper, '');
    };

    /**
     * Activate the hash change handling, returning `true` if the current URL matches an existing route, and `false` otherwise.
     * @method activate
     * @param {HistoryOptions} options.
     * @return {boolean|undefined} Returns true/false from loading the url unless the silent option was selected.
     */
    history.activate = function(options) {
        if (history.active) {
            system.error("History has already been activated.");
        }

        history.active = true;

        // Figure out the initial configuration. Do we need an iframe?
        // Is pushState desired ... is it available?
        history.options = system.extend({}, { root: '/' }, history.options, options);
        history.root = history.options.root;
        history._wantsHashChange = history.options.hashChange !== false;
        history._wantsPushState = !!history.options.pushState;
        history._hasPushState = !!(history.options.pushState && history.history && history.history.pushState);

        var fragment = history.getFragment();
        var docMode = document.documentMode;
        var oldIE = (isExplorer.exec(navigator.userAgent.toLowerCase()) && (!docMode || docMode <= 7));

        // Normalize root to always include a leading and trailing slash.
        history.root = ('/' + history.root + '/').replace(rootStripper, '/');

        if (oldIE && history._wantsHashChange) {
            history.iframe = $('<iframe src="javascript:0" tabindex="-1" />').hide().appendTo('body')[0].contentWindow;
            history.navigate(fragment, false);
        }

        // Depending on whether we're using pushState or hashes, and whether
        // 'onhashchange' is supported, determine how we check the URL state.
        if (history._hasPushState) {
            $(window).on('popstate', history.checkUrl);
        } else if (history._wantsHashChange && ('onhashchange' in window) && !oldIE) {
            $(window).on('hashchange', history.checkUrl);
        } else if (history._wantsHashChange) {
            history._checkUrlInterval = setInterval(history.checkUrl, history.interval);
        }

        // Determine if we need to change the base url, for a pushState link
        // opened by a non-pushState browser.
        history.fragment = fragment;
        var loc = history.location;
        var atRoot = loc.pathname.replace(/[^\/]$/, '$&/') === history.root;

        // Transition from hashChange to pushState or vice versa if both are requested.
        if (history._wantsHashChange && history._wantsPushState) {
            // If we've started off with a route from a `pushState`-enabled
            // browser, but we're currently in a browser that doesn't support it...
            if (!history._hasPushState && !atRoot) {
                history.fragment = history.getFragment(null, true);
                history.location.replace(history.root + history.location.search + '#' + history.fragment);
                // Return immediately as browser will do redirect to new url
                return true;

            // Or if we've started out with a hash-based route, but we're currently
            // in a browser where it could be `pushState`-based instead...
            } else if (history._hasPushState && atRoot && loc.hash) {
                this.fragment = history.getHash().replace(routeStripper, '');
                this.history.replaceState({}, document.title, history.root + history.fragment + loc.search);
            }
        }

        if (!history.options.silent) {
            return history.loadUrl();
        }
    };

    /**
     * Disable history, perhaps temporarily. Not useful in a real app, but possibly useful for unit testing Routers.
     * @method deactivate
     */
    history.deactivate = function() {
        $(window).off('popstate', history.checkUrl).off('hashchange', history.checkUrl);
        clearInterval(history._checkUrlInterval);
        history.active = false;
    };

    /**
     * Checks the current URL to see if it has changed, and if it has, calls `loadUrl`, normalizing across the hidden iframe.
     * @method checkUrl
     * @return {boolean} Returns true/false from loading the url.
     */
    history.checkUrl = function() {
        var current = history.getFragment();
        if (current === history.fragment && history.iframe) {
            current = history.getFragment(history.getHash(history.iframe));
        }

        if (current === history.fragment) {
            return false;
        }

        if (history.iframe) {
            history.navigate(current, false);
        }
        
        history.loadUrl();
    };
    
    /**
     * Attempts to load the current URL fragment. A pass-through to options.routeHandler.
     * @method loadUrl
     * @return {boolean} Returns true/false from the route handler.
     */
    history.loadUrl = function(fragmentOverride) {
        var fragment = history.fragment = history.getFragment(fragmentOverride);

        return history.options.routeHandler ?
            history.options.routeHandler(fragment) :
            false;
    };

    /**
     * Save a fragment into the hash history, or replace the URL state if the
     * 'replace' option is passed. You are responsible for properly URL-encoding
     * the fragment in advance.
     * The options object can contain `trigger: false` if you wish to not have the
     * route callback be fired, or `replace: true`, if
     * you wish to modify the current URL without adding an entry to the history.
     * @method navigate
     * @param {string} fragment The url fragment to navigate to.
     * @param {object|boolean} options An options object with optional trigger and replace flags. You can also pass a boolean directly to set the trigger option. Trigger is `true` by default.
     * @return {boolean} Returns true/false from loading the url.
     */
    history.navigate = function(fragment, options) {
        if (!history.active) {
            return false;
        }

        if(options === undefined) {
            options = {
                trigger: true
            };
        }else if(system.isBoolean(options)) {
            options = {
                trigger: options
            };
        }

        fragment = history.getFragment(fragment || '');

        if (history.fragment === fragment) {
            return;
        }

        history.fragment = fragment;

        var url = history.root + fragment;

        // Don't include a trailing slash on the root.
        if(fragment === '' && url !== '/') {
            url = url.slice(0, -1);
        }

        // If pushState is available, we use it to set the fragment as a real URL.
        if (history._hasPushState) {
            history.history[options.replace ? 'replaceState' : 'pushState']({}, document.title, url);

            // If hash changes haven't been explicitly disabled, update the hash
            // fragment to store history.
        } else if (history._wantsHashChange) {
            updateHash(history.location, fragment, options.replace);
            
            if (history.iframe && (fragment !== history.getFragment(history.getHash(history.iframe)))) {
                // Opening and closing the iframe tricks IE7 and earlier to push a
                // history entry on hash-tag change.  When replace is true, we don't
                // want history.
                if (!options.replace) {
                    history.iframe.document.open().close();
                }
                
                updateHash(history.iframe.location, fragment, options.replace);
            }

            // If you've told us that you explicitly don't want fallback hashchange-
            // based history, then `navigate` becomes a page refresh.
        } else {
            return history.location.assign(url);
        }

        if (options.trigger) {
            return history.loadUrl(fragment);
        }
    };

    /**
     * Navigates back in the browser history.
     * @method navigateBack
     */
    history.navigateBack = function() {
        history.history.back();
    };

    /**
     * @class HistoryOptions
     * @static
     */

    /**
     * The function that will be called back when the fragment changes.
     * @property {function} routeHandler
     */

    /**
     * The url root used to extract the fragment when using push state.
     * @property {string} root
     */

    /**
     * Use hash change when present.
     * @property {boolean} hashChange
     * @default true
     */

    /**
     * Use push state when present.
     * @property {boolean} pushState
     * @default false
     */

    /**
     * Prevents loading of the current url when activating history.
     * @property {boolean} silent
     * @default false
     */

    return history;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Connects the history module's url and history tracking support to Durandal's activation and composition engine allowing you to easily build navigation-style applications.
 * @module router
 * @requires system
 * @requires app
 * @requires activator
 * @requires events
 * @requires composition
 * @requires history
 * @requires knockout
 * @requires jquery
 */
define('plugins/router',['durandal/system', 'durandal/app', 'durandal/activator', 'durandal/events', 'durandal/composition', 'plugins/history', 'knockout', 'jquery'], function(system, app, activator, events, composition, history, ko, $) {
    var optionalParam = /\((.*?)\)/g;
    var namedParam = /(\(\?)?:\w+/g;
    var splatParam = /\*\w+/g;
    var escapeRegExp = /[\-{}\[\]+?.,\\\^$|#\s]/g;
    var startDeferred, rootRouter;
    var trailingSlash = /\/$/;

    function routeStringToRegExp(routeString) {
        routeString = routeString.replace(escapeRegExp, '\\$&')
            .replace(optionalParam, '(?:$1)?')
            .replace(namedParam, function(match, optional) {
                return optional ? match : '([^\/]+)';
            })
            .replace(splatParam, '(.*?)');

        return new RegExp('^' + routeString + '$');
    }

    function stripParametersFromRoute(route) {
        var colonIndex = route.indexOf(':');
        var length = colonIndex > 0 ? colonIndex - 1 : route.length;
        return route.substring(0, length);
    }

    function endsWith(str, suffix) {
        return str.indexOf(suffix, str.length - suffix.length) !== -1;
    }

    function compareArrays(first, second) {
        if (!first || !second){
            return false;
        }

        if (first.length != second.length) {
            return false;
        }

        for (var i = 0, len = first.length; i < len; i++) {
            if (first[i] != second[i]) {
                return false;
            }
        }

        return true;
    }

    /**
     * @class Router
     * @uses Events
     */

    /**
     * Triggered when the navigation logic has completed.
     * @event router:navigation:complete
     * @param {object} instance The activated instance.
     * @param {object} instruction The routing instruction.
     * @param {Router} router The router.
     */

    /**
     * Triggered when the navigation has been cancelled.
     * @event router:navigation:cancelled
     * @param {object} instance The activated instance.
     * @param {object} instruction The routing instruction.
     * @param {Router} router The router.
     */

    /**
     * Triggered right before a route is activated.
     * @event router:route:activating
     * @param {object} instance The activated instance.
     * @param {object} instruction The routing instruction.
     * @param {Router} router The router.
     */

    /**
     * Triggered right before a route is configured.
     * @event router:route:before-config
     * @param {object} config The route config.
     * @param {Router} router The router.
     */

    /**
     * Triggered just after a route is configured.
     * @event router:route:after-config
     * @param {object} config The route config.
     * @param {Router} router The router.
     */

    /**
     * Triggered when the view for the activated instance is attached.
     * @event router:navigation:attached
     * @param {object} instance The activated instance.
     * @param {object} instruction The routing instruction.
     * @param {Router} router The router.
     */

    /**
     * Triggered when the composition that the activated instance participates in is complete.
     * @event router:navigation:composition-complete
     * @param {object} instance The activated instance.
     * @param {object} instruction The routing instruction.
     * @param {Router} router The router.
     */

    /**
     * Triggered when the router does not find a matching route.
     * @event router:route:not-found
     * @param {string} fragment The url fragment.
     * @param {Router} router The router.
     */

    var createRouter = function() {
        var queue = [],
            isProcessing = ko.observable(false),
            currentActivation,
            currentInstruction,
            activeItem = activator.create();

        var router = {
            /**
             * The route handlers that are registered. Each handler consists of a `routePattern` and a `callback`.
             * @property {object[]} handlers
             */
            handlers: [],
            /**
             * The route configs that are registered.
             * @property {object[]} routes
             */
            routes: [],
            /**
             * The route configurations that have been designated as displayable in a nav ui (nav:true).
             * @property {KnockoutObservableArray} navigationModel
             */
            navigationModel: ko.observableArray([]),
            /**
             * The active item/screen based on the current navigation state.
             * @property {Activator} activeItem
             */
            activeItem: activeItem,
            /**
             * Indicates that the router (or a child router) is currently in the process of navigating.
             * @property {KnockoutComputed} isNavigating
             */
            isNavigating: ko.computed(function() {
                var current = activeItem();
                var processing = isProcessing();
                var currentRouterIsProcesing = current
                    && current.router
                    && current.router != router
                    && current.router.isNavigating() ? true : false;
                return  processing || currentRouterIsProcesing;
            }),
            /**
             * An observable surfacing the active routing instruction that is currently being processed or has recently finished processing.
             * The instruction object has `config`, `fragment`, `queryString`, `params` and `queryParams` properties.
             * @property {KnockoutObservable} activeInstruction
             */
            activeInstruction:ko.observable(null),
            __router__:true
        };

        events.includeIn(router);

        activeItem.settings.areSameItem = function (currentItem, newItem, currentActivationData, newActivationData) {
            if (currentItem == newItem) {
                return compareArrays(currentActivationData, newActivationData);
            }

            return false;
        };

        function hasChildRouter(instance) {
            return instance.router && instance.router.parent == router;
        }

        function setCurrentInstructionRouteIsActive(flag) {
            if (currentInstruction && currentInstruction.config.isActive) {
                currentInstruction.config.isActive(flag)
            }
        }

        function completeNavigation(instance, instruction) {
            system.log('Navigation Complete', instance, instruction);

            var fromModuleId = system.getModuleId(currentActivation);
            if (fromModuleId) {
                router.trigger('router:navigation:from:' + fromModuleId);
            }

            currentActivation = instance;

            setCurrentInstructionRouteIsActive(false);
            currentInstruction = instruction;
            setCurrentInstructionRouteIsActive(true);

            var toModuleId = system.getModuleId(currentActivation);
            if (toModuleId) {
                router.trigger('router:navigation:to:' + toModuleId);
            }

            if (!hasChildRouter(instance)) {
                router.updateDocumentTitle(instance, instruction);
            }

            rootRouter.explicitNavigation = false;
            rootRouter.navigatingBack = false;
            router.trigger('router:navigation:complete', instance, instruction, router);
        }

        function cancelNavigation(instance, instruction) {
            system.log('Navigation Cancelled');

            router.activeInstruction(currentInstruction);

            if (currentInstruction) {
                router.navigate(currentInstruction.fragment, false);
            }

            isProcessing(false);
            rootRouter.explicitNavigation = false;
            rootRouter.navigatingBack = false;
            router.trigger('router:navigation:cancelled', instance, instruction, router);
        }

        function redirect(url) {
            system.log('Navigation Redirecting');

            isProcessing(false);
            rootRouter.explicitNavigation = false;
            rootRouter.navigatingBack = false;
            router.navigate(url, { trigger: true, replace: true });
        }

        function activateRoute(activator, instance, instruction) {
            rootRouter.navigatingBack = !rootRouter.explicitNavigation && currentActivation != instruction.fragment;
            router.trigger('router:route:activating', instance, instruction, router);

            activator.activateItem(instance, instruction.params).then(function(succeeded) {
                if (succeeded) {
                    var previousActivation = currentActivation;
                    completeNavigation(instance, instruction);

                    if (hasChildRouter(instance)) {
                        var fullFragment = instruction.fragment;
                        if (instruction.queryString) {
                            fullFragment += "?" + instruction.queryString;
                        }

                        instance.router.loadUrl(fullFragment);
                    }

                    if (previousActivation == instance) {
                        router.attached();
                        router.compositionComplete();
                    }
                } else if(activator.settings.lifecycleData && activator.settings.lifecycleData.redirect){
                    redirect(activator.settings.lifecycleData.redirect);
                }else{
                    cancelNavigation(instance, instruction);
                }

                if (startDeferred) {
                    startDeferred.resolve();
                    startDeferred = null;
                }
            }).fail(function(err){
                system.error(err);
            });;
        }

        /**
         * Inspects routes and modules before activation. Can be used to protect access by cancelling navigation or redirecting.
         * @method guardRoute
         * @param {object} instance The module instance that is about to be activated by the router.
         * @param {object} instruction The route instruction. The instruction object has config, fragment, queryString, params and queryParams properties.
         * @return {Promise|Boolean|String} If a boolean, determines whether or not the route should activate or be cancelled. If a string, causes a redirect to the specified route. Can also be a promise for either of these value types.
         */
        function handleGuardedRoute(activator, instance, instruction) {
            var resultOrPromise = router.guardRoute(instance, instruction);
            if (resultOrPromise) {
                if (resultOrPromise.then) {
                    resultOrPromise.then(function(result) {
                        if (result) {
                            if (system.isString(result)) {
                                redirect(result);
                            } else {
                                activateRoute(activator, instance, instruction);
                            }
                        } else {
                            cancelNavigation(instance, instruction);
                        }
                    });
                } else {
                    if (system.isString(resultOrPromise)) {
                        redirect(resultOrPromise);
                    } else {
                        activateRoute(activator, instance, instruction);
                    }
                }
            } else {
                cancelNavigation(instance, instruction);
            }
        }

        function ensureActivation(activator, instance, instruction) {
            if (router.guardRoute) {
                handleGuardedRoute(activator, instance, instruction);
            } else {
                activateRoute(activator, instance, instruction);
            }
        }

        function canReuseCurrentActivation(instruction) {
            return currentInstruction
                && currentInstruction.config.moduleId == instruction.config.moduleId
                && currentActivation
                && ((currentActivation.canReuseForRoute && currentActivation.canReuseForRoute.apply(currentActivation, instruction.params))
                || (!currentActivation.canReuseForRoute && currentActivation.router && currentActivation.router.loadUrl));
        }

        function dequeueInstruction() {
            if (isProcessing()) {
                return;
            }

            var instruction = queue.shift();
            queue = [];

            if (!instruction) {
                return;
            }

            isProcessing(true);
            router.activeInstruction(instruction);

            if (canReuseCurrentActivation(instruction)) {
                ensureActivation(activator.create(), currentActivation, instruction);
            } else {
                system.acquire(instruction.config.moduleId).then(function(module) {
                    var instance = system.resolveObject(module);
                    ensureActivation(activeItem, instance, instruction);
                }).fail(function(err){
                        system.error('Failed to load routed module (' + instruction.config.moduleId + '). Details: ' + err.message);
                    });
            }
        }

        function queueInstruction(instruction) {
            queue.unshift(instruction);
            dequeueInstruction();
        }

        // Given a route, and a URL fragment that it matches, return the array of
        // extracted decoded parameters. Empty or unmatched parameters will be
        // treated as `null` to normalize cross-browser behavior.
        function createParams(routePattern, fragment, queryString) {
            var params = routePattern.exec(fragment).slice(1);

            for (var i = 0; i < params.length; i++) {
                var current = params[i];
                params[i] = current ? decodeURIComponent(current) : null;
            }

            var queryParams = router.parseQueryString(queryString);
            if (queryParams) {
                params.push(queryParams);
            }

            return {
                params:params,
                queryParams:queryParams
            };
        }

        function configureRoute(config){
            router.trigger('router:route:before-config', config, router);

            if (!system.isRegExp(config)) {
                config.title = config.title || router.convertRouteToTitle(config.route);
                config.moduleId = config.moduleId || router.convertRouteToModuleId(config.route);
                config.hash = config.hash || router.convertRouteToHash(config.route);
                config.routePattern = routeStringToRegExp(config.route);
            }else{
                config.routePattern = config.route;
            }

            config.isActive = config.isActive || ko.observable(false);
            router.trigger('router:route:after-config', config, router);
            router.routes.push(config);

            router.route(config.routePattern, function(fragment, queryString) {
                var paramInfo = createParams(config.routePattern, fragment, queryString);
                queueInstruction({
                    fragment: fragment,
                    queryString:queryString,
                    config: config,
                    params: paramInfo.params,
                    queryParams:paramInfo.queryParams
                });
            });
        };

        function mapRoute(config) {
            if(system.isArray(config.route)){
                var isActive = config.isActive || ko.observable(false);

                for(var i = 0, length = config.route.length; i < length; i++){
                    var current = system.extend({}, config);

                    current.route = config.route[i];
                    current.isActive = isActive;

                    if(i > 0){
                        delete current.nav;
                    }

                    configureRoute(current);
                }
            }else{
                configureRoute(config);
            }

            return router;
        }

        /**
         * Parses a query string into an object.
         * @method parseQueryString
         * @param {string} queryString The query string to parse.
         * @return {object} An object keyed according to the query string parameters.
         */
        router.parseQueryString = function (queryString) {
            var queryObject, pairs;

            if (!queryString) {
                return null;
            }

            pairs = queryString.split('&');

            if (pairs.length == 0) {
                return null;
            }

            queryObject = {};

            for (var i = 0; i < pairs.length; i++) {
                var pair = pairs[i];
                if (pair === '') {
                    continue;
                }

                var parts = pair.split('=');
                queryObject[parts[0]] = parts[1] && decodeURIComponent(parts[1].replace(/\+/g, ' '));
            }

            return queryObject;
        };

        /**
         * Add a route to be tested when the url fragment changes.
         * @method route
         * @param {RegEx} routePattern The route pattern to test against.
         * @param {function} callback The callback to execute when the route pattern is matched.
         */
        router.route = function(routePattern, callback) {
            router.handlers.push({ routePattern: routePattern, callback: callback });
        };

        /**
         * Attempt to load the specified URL fragment. If a route succeeds with a match, returns `true`. If no defined routes matches the fragment, returns `false`.
         * @method loadUrl
         * @param {string} fragment The URL fragment to find a match for.
         * @return {boolean} True if a match was found, false otherwise.
         */
        router.loadUrl = function(fragment) {
            var handlers = router.handlers,
                queryString = null,
                coreFragment = fragment,
                queryIndex = fragment.indexOf('?');

            if (queryIndex != -1) {
                coreFragment = fragment.substring(0, queryIndex);
                queryString = fragment.substr(queryIndex + 1);
            }

            if(router.relativeToParentRouter){
                var instruction = this.parent.activeInstruction();
                coreFragment = instruction.params.join('/');

                if(coreFragment && coreFragment.charAt(0) == '/'){
                    coreFragment = coreFragment.substr(1);
                }

                if(!coreFragment){
                    coreFragment = '';
                }

                coreFragment = coreFragment.replace('//', '/').replace('//', '/');
            }

            coreFragment = coreFragment.replace(trailingSlash, '');

            for (var i = 0; i < handlers.length; i++) {
                var current = handlers[i];
                if (current.routePattern.test(coreFragment)) {
                    current.callback(coreFragment, queryString);
                    return true;
                }
            }

            system.log('Route Not Found');
            router.trigger('router:route:not-found', fragment, router);

            if (currentInstruction) {
                history.navigate(currentInstruction.fragment, { trigger:false, replace:true });
            }

            rootRouter.explicitNavigation = false;
            rootRouter.navigatingBack = false;

            return false;
        };

        /**
         * Updates the document title based on the activated module instance, the routing instruction and the app.title.
         * @method updateDocumentTitle
         * @param {object} instance The activated module.
         * @param {object} instruction The routing instruction associated with the action. It has a `config` property that references the original route mapping config.
         */
        router.updateDocumentTitle = function(instance, instruction) {
            if (instruction.config.title) {
                if (app.title) {
                    document.title = instruction.config.title + " | " + app.title;
                } else {
                    document.title = instruction.config.title;
                }
            } else if (app.title) {
                document.title = app.title;
            }
        };

        /**
         * Save a fragment into the hash history, or replace the URL state if the
         * 'replace' option is passed. You are responsible for properly URL-encoding
         * the fragment in advance.
         * The options object can contain `trigger: false` if you wish to not have the
         * route callback be fired, or `replace: true`, if
         * you wish to modify the current URL without adding an entry to the history.
         * @method navigate
         * @param {string} fragment The url fragment to navigate to.
         * @param {object|boolean} options An options object with optional trigger and replace flags. You can also pass a boolean directly to set the trigger option. Trigger is `true` by default.
         * @return {boolean} Returns true/false from loading the url.
         */
        router.navigate = function(fragment, options) {
            if(fragment && fragment.indexOf('://') != -1){
                window.location.href = fragment;
                return true;
            }

            rootRouter.explicitNavigation = true;
            return history.navigate(fragment, options);
        };

        /**
         * Navigates back in the browser history.
         * @method navigateBack
         */
        router.navigateBack = function() {
            history.navigateBack();
        };

        router.attached = function() {
            router.trigger('router:navigation:attached', currentActivation, currentInstruction, router);
        };

        router.compositionComplete = function(){
            isProcessing(false);
            router.trigger('router:navigation:composition-complete', currentActivation, currentInstruction, router);
            dequeueInstruction();
        };

        /**
         * Converts a route to a hash suitable for binding to a link's href.
         * @method convertRouteToHash
         * @param {string} route
         * @return {string} The hash.
         */
        router.convertRouteToHash = function(route) {
            if(router.relativeToParentRouter){
                var instruction = router.parent.activeInstruction(),
                    hash = instruction.config.hash + '/' + route;

                if(history._hasPushState){
                    hash = '/' + hash;
                }

                hash = hash.replace('//', '/').replace('//', '/');
                return hash;
            }

            if(history._hasPushState){
                return route;
            }

            return "#" + route;
        };

        /**
         * Converts a route to a module id. This is only called if no module id is supplied as part of the route mapping.
         * @method convertRouteToModuleId
         * @param {string} route
         * @return {string} The module id.
         */
        router.convertRouteToModuleId = function(route) {
            return stripParametersFromRoute(route);
        };

        /**
         * Converts a route to a displayable title. This is only called if no title is specified as part of the route mapping.
         * @method convertRouteToTitle
         * @param {string} route
         * @return {string} The title.
         */
        router.convertRouteToTitle = function(route) {
            var value = stripParametersFromRoute(route);
            return value.substring(0, 1).toUpperCase() + value.substring(1);
        };

        /**
         * Maps route patterns to modules.
         * @method map
         * @param {string|object|object[]} route A route, config or array of configs.
         * @param {object} [config] The config for the specified route.
         * @chainable
         * @example
 router.map([
    { route: '', title:'Home', moduleId: 'homeScreen', nav: true },
    { route: 'customer/:id', moduleId: 'customerDetails'}
 ]);
         */
        router.map = function(route, config) {
            if (system.isArray(route)) {
                for (var i = 0; i < route.length; i++) {
                    router.map(route[i]);
                }

                return router;
            }

            if (system.isString(route) || system.isRegExp(route)) {
                if (!config) {
                    config = {};
                } else if (system.isString(config)) {
                    config = { moduleId: config };
                }

                config.route = route;
            } else {
                config = route;
            }

            return mapRoute(config);
        };

        /**
         * Builds an observable array designed to bind a navigation UI to. The model will exist in the `navigationModel` property.
         * @method buildNavigationModel
         * @param {number} defaultOrder The default order to use for navigation visible routes that don't specify an order. The default is 100 and each successive route will be one more than that.
         * @chainable
         */
        router.buildNavigationModel = function(defaultOrder) {
            var nav = [], routes = router.routes;
            var fallbackOrder = defaultOrder || 100;

            for (var i = 0; i < routes.length; i++) {
                var current = routes[i];

                if (current.nav) {
                    if (!system.isNumber(current.nav)) {
                        current.nav = ++fallbackOrder;
                    }

                    nav.push(current);
                }
            }

            nav.sort(function(a, b) { return a.nav - b.nav; });
            router.navigationModel(nav);

            return router;
        };

        /**
         * Configures how the router will handle unknown routes.
         * @method mapUnknownRoutes
         * @param {string|function} [config] If not supplied, then the router will map routes to modules with the same name.
         * If a string is supplied, it represents the module id to route all unknown routes to.
         * Finally, if config is a function, it will be called back with the route instruction containing the route info. The function can then modify the instruction by adding a moduleId and the router will take over from there.
         * @param {string} [replaceRoute] If config is a module id, then you can optionally provide a route to replace the url with.
         * @chainable
         */
        router.mapUnknownRoutes = function(config, replaceRoute) {
            var catchAllRoute = "*catchall";
            var catchAllPattern = routeStringToRegExp(catchAllRoute);

            router.route(catchAllPattern, function (fragment, queryString) {
                var paramInfo = createParams(catchAllPattern, fragment, queryString);
                var instruction = {
                    fragment: fragment,
                    queryString: queryString,
                    config: {
                        route: catchAllRoute,
                        routePattern: catchAllPattern
                    },
                    params: paramInfo.params,
                    queryParams: paramInfo.queryParams
                };

                if (!config) {
                    instruction.config.moduleId = fragment;
                } else if (system.isString(config)) {
                    instruction.config.moduleId = config;
                    if(replaceRoute){
                        history.navigate(replaceRoute, { trigger:false, replace:true });
                    }
                } else if (system.isFunction(config)) {
                    var result = config(instruction);
                    if (result && result.then) {
                        result.then(function() {
                            router.trigger('router:route:before-config', instruction.config, router);
                            router.trigger('router:route:after-config', instruction.config, router);
                            queueInstruction(instruction);
                        });
                        return;
                    }
                } else {
                    instruction.config = config;
                    instruction.config.route = catchAllRoute;
                    instruction.config.routePattern = catchAllPattern;
                }

                router.trigger('router:route:before-config', instruction.config, router);
                router.trigger('router:route:after-config', instruction.config, router);
                queueInstruction(instruction);
            });

            return router;
        };

        /**
         * Resets the router by removing handlers, routes, event handlers and previously configured options.
         * @method reset
         * @chainable
         */
        router.reset = function() {
            currentInstruction = currentActivation = undefined;
            router.handlers = [];
            router.routes = [];
            router.off();
            delete router.options;
            return router;
        };

        /**
         * Makes all configured routes and/or module ids relative to a certain base url.
         * @method makeRelative
         * @param {string|object} settings If string, the value is used as the base for routes and module ids. If an object, you can specify `route` and `moduleId` separately. In place of specifying route, you can set `fromParent:true` to make routes automatically relative to the parent router's active route.
         * @chainable
         */
        router.makeRelative = function(settings){
            if(system.isString(settings)){
                settings = {
                    moduleId:settings,
                    route:settings
                };
            }

            if(settings.moduleId && !endsWith(settings.moduleId, '/')){
                settings.moduleId += '/';
            }

            if(settings.route && !endsWith(settings.route, '/')){
                settings.route += '/';
            }

            if(settings.fromParent){
                router.relativeToParentRouter = true;
            }

            router.on('router:route:before-config').then(function(config){
                if(settings.moduleId){
                    config.moduleId = settings.moduleId + config.moduleId;
                }

                if(settings.route){
                    if(config.route === ''){
                        config.route = settings.route.substring(0, settings.route.length - 1);
                    }else{
                        config.route = settings.route + config.route;
                    }
                }
            });

            return router;
        };

        /**
         * Creates a child router.
         * @method createChildRouter
         * @return {Router} The child router.
         */
        router.createChildRouter = function() {
            var childRouter = createRouter();
            childRouter.parent = router;
            return childRouter;
        };

        return router;
    };

    /**
     * @class RouterModule
     * @extends Router
     * @static
     */
    rootRouter = createRouter();
    rootRouter.explicitNavigation = false;
    rootRouter.navigatingBack = false;

    /**
     * Verify that the target is the current window
     * @method targetIsThisWindow
     * @return {boolean} True if the event's target is the current window, false otherwise.
     */
    rootRouter.targetIsThisWindow = function(event) {
        var targetWindow = $(event.target).attr('target');
        
        if (!targetWindow ||
            targetWindow === window.name ||
            targetWindow === '_self' ||
            (targetWindow === 'top' && window === window.top)) { return true; }
        
        return false;
    };

    /**
     * Activates the router and the underlying history tracking mechanism.
     * @method activate
     * @return {Promise} A promise that resolves when the router is ready.
     */
    rootRouter.activate = function(options) {
        return system.defer(function(dfd) {
            startDeferred = dfd;
            rootRouter.options = system.extend({ routeHandler: rootRouter.loadUrl }, rootRouter.options, options);

            history.activate(rootRouter.options);

            if(history._hasPushState){
                var routes = rootRouter.routes,
                    i = routes.length;

                while(i--){
                    var current = routes[i];
                    current.hash = current.hash.replace('#', '');
                }
            }

            $(document).delegate("a", 'click', function(evt){
                if(history._hasPushState){
                    if(!evt.altKey && !evt.ctrlKey && !evt.metaKey && !evt.shiftKey && rootRouter.targetIsThisWindow(evt)){
                        var href = $(this).attr("href");

                        // Ensure the protocol is not part of URL, meaning its relative.
                        // Stop the event bubbling to ensure the link will not cause a page refresh.
                        if (href != null && !(href.charAt(0) === "#" || /^[a-z]+:/i.test(href))) {
                            rootRouter.explicitNavigation = true;
                            evt.preventDefault();
                            history.navigate(href);
                        }
                    }
                }else{
                    rootRouter.explicitNavigation = true;
                }
            });

            if(history.options.silent && startDeferred){
                startDeferred.resolve();
                startDeferred = null;
            }
        }).promise();
    };

    /**
     * Disable history, perhaps temporarily. Not useful in a real app, but possibly useful for unit testing Routers.
     * @method deactivate
     */
    rootRouter.deactivate = function() {
        history.deactivate();
    };

    /**
     * Installs the router's custom ko binding handler.
     * @method install
     */
    rootRouter.install = function(){
        ko.bindingHandlers.router = {
            init: function() {
                return { controlsDescendantBindings: true };
            },
            update: function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
                var settings = ko.utils.unwrapObservable(valueAccessor()) || {};

                if (settings.__router__) {
                    settings = {
                        model:settings.activeItem(),
                        attached:settings.attached,
                        compositionComplete:settings.compositionComplete,
                        activate: false
                    };
                } else {
                    var theRouter = ko.utils.unwrapObservable(settings.router || viewModel.router) || rootRouter;
                    settings.model = theRouter.activeItem();
                    settings.attached = theRouter.attached;
                    settings.compositionComplete = theRouter.compositionComplete;
                    settings.activate = false;
                }

                composition.compose(element, settings, bindingContext);
            }
        };

        ko.virtualElements.allowedBindings.router = true;
    };

    return rootRouter;
});

define('bootstrapping/knockoutBindings/reviewableBinding',['durandal/composition', 'plugins/router'], function (composition, router) {

    ko.bindingHandlers.reviewable = {
        init: function (element, valueAccessor) {
            if (!router.getQueryStringValue('reviewApiUrl'))
                return;

            var $element = $(element),
                context = valueAccessor();

            $element.addClass('reviewable');
            $element.data('review-context', context);
        }
    };

    composition.addBindingHandler('reviewable');

});
ko.bindingHandlers.limitedLengthText = {
    init: function () {
    },
    update: function (element, valueAccessor) {
        var text = ko.unwrap(valueAccessor().text),
            maxLength = ko.unwrap(valueAccessor().maxLength),
            $element = $(element);

        if (text.length > maxLength) {
            text = text.substr(0, maxLength - 3) + "...";
        }

        if (text !== $element.text()) {
            $element.text(text);
        }
    }
};
define("bootstrapping/knockoutBindings/limitedLengthTextBinding", function(){});

define('bootstrapping/knockoutBindings/scrollbarBinding',['durandal/composition', 'perfectScrollbar', 'browserSupport'], function (composition, perfectScrollbar, browserSupport) {

    ko.bindingHandlers.scrollbar = {
        init: function (element, valueAccessors) {
            if (browserSupport.isMobileDevice) {
                $(element).css({
                    'overflow-y': 'auto'
                });
                return;
            }

            var byClass = valueAccessors().byClass;
            var checkDOMChanges = valueAccessors().checkDOMChanges;
            var scrollToEndAfterDOMChanged = valueAccessors().scrollToEndAfterDOMChanged;
            var cssClasses = {
                scrollStarted: 'cs-scroll-started',
                scrollFinished: 'cs-scroll-finished',
                scrollEnabled: 'cs-scroll-enabled'
            };

            var customScrollbarContainer = byClass ? element.getElementsByClassName(byClass)[0] : element;

            perfectScrollbar.initialize(customScrollbarContainer, {
                suppressScrollX: true
            });

            if (checkDOMChanges) {
                refreshScrollAfterDomChanged();
            }

            document.addEventListener('ps-y-reach-start', function () {
                customScrollbarContainer.classList.remove(cssClasses.scrollStarted);
            });

            document.addEventListener('ps-y-reach-end', function () {
                customScrollbarContainer.classList.add(cssClasses.scrollFinished);
            });

            document.addEventListener('ps-scroll-up', function () {
                customScrollbarContainer.classList.remove(cssClasses.scrollFinished);
            });

            document.addEventListener('ps-scroll-down', function () {
                customScrollbarContainer.classList.add(cssClasses.scrollStarted);
            });

            scrollEnabled();

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                perfectScrollbar.destroy(customScrollbarContainer);
            });

            function refreshScrollAfterDomChanged() {
                var scrollHeight = customScrollbarContainer.clientHeight;
                var timeout = setTimeout(function () {
                    perfectScrollbar.update(customScrollbarContainer);
                    if (scrollToEndAfterDOMChanged && scrollHeight !== customScrollbarContainer.clientHeight) {
                        scrollHeight = customScrollbarContainer.clientHeight;
                        customScrollbarContainer.scrollTop = customScrollbarContainer.scrollHeight;
                    }
                    scrollEnabled();

                    refreshScrollAfterDomChanged();
                }, 500);

                ko.utils.domNodeDisposal.addDisposeCallback(element, function () { clearTimeout(timeout); });
            }

            function scrollEnabled() {
                customScrollbarContainer.classList.toggle(cssClasses.scrollEnabled, customScrollbarContainer.scrollHeight > customScrollbarContainer.clientHeight);
            }
        }
    };

    composition.addBindingHandler('scrollbar');

});
define('bootstrapping/knockoutBindings/localizeBinding',['knockout', 'localizationManager'], function (ko, localizationManager) {
    ko.bindingHandlers.localize = {
        update: function (element, valueAccessor) {
            localizeValue(element, valueAccessor);
        }
    };

    function localizeValue(element, valueAccessor) {
        var value = valueAccessor();

        if (_.isEmpty(value)) {
            return;
        }

        if (!_.isUndefined(value['text'])) {
            $(element).text(localizationManager.getLocalizedText(value['text']));
        }
        if (!_.isUndefined(value['placeholder'])) {
            $(element).attr('placeholder', localizationManager.getLocalizedText(value['placeholder']));
        }
        if (!_.isUndefined(value['value'])) {
            $(element).prop('value', localizationManager.getLocalizedText(value['value']));
        }
        if (!_.isUndefined(value['title'])) {
            $(element).prop('title', localizationManager.getLocalizedText(value['title']));
        }
        if (!_.isUndefined(value['html'])) {
            $(element).html(localizationManager.getLocalizedText(value['html']));
        }
        if (!_.isUndefined(value['data-text'])) {
            $(element).attr('data-text', localizationManager.getLocalizedText(value['data-text']));
        }
    }
});
define('bootstrapping/knockoutBindings/fontScaleBinding',[],function () {

    ko.bindingHandlers.fontScale = {
        init: function (element, valueAccessor) {
            var $element = $(element);
            var text = ko.unwrap(valueAccessor());

            $element.text(text);

            if(text.length > 100) {
                /*
                    Formula description:

                    Max text length = 255 symbols;
                    Min text length = 100 symbols;
                    maxLen - minLen = 155; 

                    coeficient = 1 - (length - minLen) / 155;

                    Max font size = 1.385;
                    Min font size = 0.65;
                    maxFont - minFont = 0.825;

                    Max line height = 38;
                    Min line height = 25;
                    maxLHeight - minLHeight = 13;

                    Result font size = coeficient * 0.825 + minFont;
                    Result line height = coeficient * 13 + minLHeight;
                */

                var coef = 1 - (text.length - 100) / 155;

                var fSize = coef * 0.825 + 0.8;
                var lHeight = coef * 13 + 25;

                $element.css('fontSize', fSize + 'em');
                $element.css('lineHeight', lHeight + 'px');
            } else {
                $element.css('fontSize', '1.385em');
                $element.css('lineHeight', '38px');
            }
        }
    };
});
define('treeOfContent/utils/screenResolutionChecker',[], function () {

    var checker = {
        isLowResolution: ko.observable()
    };

    checkResolution();

    var lazyWindowResizeHandler = _.debounce(checkResolution, 300);
    $(window).on('resize', lazyWindowResizeHandler);

    function checkResolution() {
        checker.isLowResolution($(window).width() < 820);
    }

    return checker;
});
define('treeOfContent/bindingHandlers/treeOfContentAnimationsBinding',['treeOfContent/utils/screenResolutionChecker', 'browserSupport'], function (screenResolutionChecker, browserSupport) {

    ko.bindingHandlers.treeOfContentAnimations = {
        init: function (element, valueAccessor) {
            var $html = $('html'),
                $body = $('body'),
                $element = $(element),
                resolutionChangeSubscription = screenResolutionChecker.isLowResolution.subscribe(onResolutionChanged),
                expandedChangeSubscription = valueAccessor().expanded.subscribe(onIsExpandedChanged);

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                resolutionChangeSubscription.dispose();
                expandedChangeSubscription.dispose();
            });

            onResolutionChanged(screenResolutionChecker.isLowResolution());
            onIsExpandedChanged(ko.utils.unwrapObservable(valueAccessor().expanded));

            function onResolutionChanged(isLowResolution) {
                var isExpanded = ko.utils.unwrapObservable(valueAccessor().expanded);
                if (isLowResolution) {
                    setVisibilityWithoutAnimations(isExpanded ? 'visible' : 'hidden');
                } else {
                    $element.css({
                        visibility: 'visible',
                        webkitTtransition: 'left 0.2s ease-out',
                        transition: 'left 0.2s ease-out'
                    });
                }

                $body.css({
                    'max-height': isExpanded && isLowResolution ? '100vh' : 'none',
                    'overflow': isExpanded && isLowResolution ? 'hidden' : 'visible'
                });

                $html.css('overflow-y', isExpanded && isLowResolution ? 'hidden' : 'scroll');

                function setVisibilityWithoutAnimations(visibility) {
                    $element.css({
                        transition: 'transform 0s, opacity 0s, visibility 0s',
                        webkitTtransition: 'transform 0s, opacity 0s, visibility 0s'
                    });

                    $element.css({
                        visibility: visibility
                    });

                    $element.css({
                        webkitTtransition: 'transform 0.2s, opacity 0.2s, visibility 0.2s',
                        transition: 'transform 0.2s, opacity 0.2s, visibility 0.2s'
                    });
                }
            }

            function onIsExpandedChanged(isExpanded) {
                if (screenResolutionChecker.isLowResolution()) {
                    $html.css('overflow-y', isExpanded ? 'hidden' : 'scroll');

                    $body.css({
                        'max-height': isExpanded ? '100vh' : 'none',
                        'overflow': isExpanded ? 'hidden' : 'visible'
                    });

                    $element.css({
                        visibility: isExpanded ? 'visible' : 'hidden'
                    });
                }
            }
        }
    };
});
define('bootstrapping/knockoutBindings/selectTextBinding',[],function () {

    ko.bindingHandlers.selectText = {
        init: function (element, valueAccessor) {
            var $element = $(element);
            
            $element.on('click', function(){
                this.select();
            });
        }
    };
});
define('bootstrapping/knockoutBindings/submitOnEnterBinding',[],function () {

    ko.bindingHandlers.submitOnEnter = {
        init: function (element, valueAccessor, allBindings, viewModel) {
            var $element = $(element),
                callback = valueAccessor();
            
            $element.keydown(function (e) {
                if (e.which == 13) {
                    e.preventDefault();
                    e.stopPropagation();

                    if (typeof callback === 'function') {
                        callback.call(viewModel);
                    } else {
                        $element.click();
                    }
                }
            });
        }
    };
});
define('bootstrapping/knockoutBindings/keyDownBinding',[],function () {

    ko.bindingHandlers.keyDown = {
        init: function (element, valueAccessor, allBindings, viewModel) {
            var bindingData = valueAccessor();
            var keyDownEventHandler = function (event) {
                if(bindingData.hasOwnProperty(event.keyCode)){
                    if (_.isFunction(bindingData[event.keyCode].handler)) {
                        bindingData[event.keyCode].handler(viewModel);
                    }

                    if(!bindingData[event.keyCode].hasOwnProperty('preventDefault') || bindingData[event.keyCode].preventDefault) {
                        event.preventDefault();
                    }

                    if(!bindingData[event.keyCode].hasOwnProperty('stopPropagation') || bindingData[event.keyCode].stopPropagation) {                        
                        event.stopPropagation();
                    }                    
                }
            };

            $(element).on('keydown', keyDownEventHandler);

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $(element).off('keydown', keyDownEventHandler);
            });
        }
    };
});
define('bootstrapping/knockoutBindings/tabFocusBinding',['knockout', 'jquery', 'durandal/composition'], function (ko, $) {

    ko.bindingHandlers.tabFocus = {
        init: function (element, valueAccessor) {
            var $element = $(element),
                keyboardClass = valueAccessor() || 'keyboard-navigation';

            $element.keydown(function (e) {
                if (e.which == 9 && !$element.hasClass(keyboardClass)) {
                    $element.addClass(keyboardClass);
                }
            });

            $element.click(function() {
                if ($element.hasClass(keyboardClass)) {
                    $element.removeClass(keyboardClass);
                }
            });
        }
    };
});
define('bootstrapping/knockoutBindings/copyToClipboardBinding',['knockout'], function(ko) {
    ko.bindingHandlers.copyToClipboard = {
        init: function(element, valueAccessor) {
            var $element = $(element),
                targets = valueAccessor().targets,
                texts = valueAccessor().texts,
                copied = valueAccessor().copied;

            var cplipboard = new Clipboard(element, {
                text: function() {
                    var targetTexts = [];
                    var text = '';
                    targets.forEach(function(item) {
                        var itemText = document.getElementById(item);
                        targetTexts.push(itemText.value);
                    });
                    copied(true);
                    setTimeout(function(){
                        copied(false);
                    }, 3000);
                    if (texts) {
                        texts.forEach(function(item, index) {
                            if (index == texts.length - 1) {
                                text += item + ': ' + targetTexts[index];
                            } else {
                                text += item + ': ' + targetTexts[index] + ', ';
                            }
                        });
                        return text;
                    }
                    return targetTexts[0];
                }
            });
        }
    };
});
define('bootstrapping/knockoutBindings/changeCssClassByDelayBinding',['knockout', 'underscore'], function (ko, _) {
    ko.bindingHandlers.changeClass = {
        update: function (element, valueAccessor) {
            var delay = valueAccessor().delay || 400,
                fromClass = valueAccessor().fromClass || 'hide',
                toClass = valueAccessor().toClass || 'show',
                value = ko.utils.unwrapObservable(valueAccessor().value) || false; 

            _.delay(function(){
                element.classList.toggle(toClass, value);
                element.classList.toggle(fromClass, !value);
            }, delay);
        }
    };
});
define('bootstrapping/knockoutBindings/containImageBinding',[],function () {
    var imageResizerQueryPattern = /\?width=\d+\&height=\d+&scaleBySmallerSide=\w+/;

    ko.bindingHandlers.containImage = {
        init: function (element, valueAccessor) {
            var $element = $(element);
            var imageProps = valueAccessor();
            var imageUrl = getOriginalImageUrl(imageProps.imageUrl);
            var imageWidth = imageProps.imageWidth;
            var imageHeight = imageProps.imageHeight;

            return getImageSize(imageUrl).then(function(size){
                var resizedImageUrl = imageUrl;
                var backgroundSize = 'initial';
                if(size.width >= imageWidth || size.height >= imageHeight) {
                    resizedImageUrl = getResizedSectionThumbnailUrl(imageUrl, imageWidth, imageHeight);
                    backgroundSize = 'contain';
                }
                if(size.width >= imageWidth && size.height >= imageHeight) {
                    backgroundSize = 'cover';
                }
                $element.css('backgroundSize', backgroundSize);
                $element.css('backgroundImage', 'url(' + resizedImageUrl + ')');
            });
        }
    };

    function getOriginalImageUrl(imageUrl) {
        var originalImage = imageUrl;
        var coincidences = imageResizerQueryPattern.exec(imageUrl);
        if (coincidences && coincidences.length) {
            originalImage = imageUrl.substring(0, coincidences.index);
        }
        return originalImage;
    }

    function getImageSize(imageUrl) {
        var dfd = $.Deferred();
        var img = new Image();
        img.onload = function() {
            dfd.resolve({
                width: this.width,
                height: this.height
            });
        }
        img.onerror = img.onabort = function() {
            dfd.reject('can not load image, url: ' + imageUrl);
        }
        img.src = imageUrl;
        return dfd.promise();
    }

    function getResizedSectionThumbnailUrl(imageUrl, imageWidth, imageHeight) {
        return imageUrl + '?width=' + imageWidth + '&height=' + imageHeight;
    }
});
define('bootstrapping/knockoutBindings/lessColorBinding',['underscore', 'less'], function (_, less) {

    ko.bindingHandlers.lessColor = {
        init: function (element, valueAccessor) {
            var conditions = valueAccessor();

            _.each(conditions, function (condition) {
                var outputColor = condition.color;
                if (!condition.attr) return;
                if (condition.func && condition.dimension) {
                    var c = new less.tree.Color(condition.color.slice(1)),
                        dimension = new less.tree.Dimension(condition.dimension),
                        func = less.functions.functionRegistry.get(condition.func);
                    outputColor = func(c, dimension).toRGB();
                }
                
                element.setAttribute(condition.attr, outputColor);
            });
        }
    };
});
ko.bindingHandlers.editableText = {
    init: function(element, valueAccessor) {
        var $element = $(element);
        var text = valueAccessor().text;
        var onEnterKeyPress = valueAccessor().onEnterKeyPress;
        var updateValue = function() {
            if (ko.utils.unwrapObservable(text) !== $element.text()) {
                text($element.text());
            }
        };

        $element.text(ko.utils.unwrapObservable(text)).on('focus', function() {
            $element.on('DOMSubtreeModified', updateValue);
        }).on('blur', function() {
            $element.scrollLeft(0);
            $element.off('DOMSubtreeModified', updateValue);
        }).on('keypress', function(event) {
            if (event.keyCode !== 13) {
                return;
            }

            $element.blur();
            if (onEnterKeyPress) {
                onEnterKeyPress();
            }

            event.preventDefault();
            event.stopPropagation();
        }).on('drop dragover', function(event) {
            event.preventDefault();
            event.stopPropagation();
        });
    },
    update: function(element, valueAccessor) {
        var text = ko.utils.unwrapObservable(valueAccessor().text);
        var $element = $(element);

        if (text !== $element.text()) {
            $element.text(text);
        }

        if ($element.text().length === 0) {
            $element.text('');
        }
    }
};
define("bootstrapping/knockoutBindings/editableTextBindingHandler", function(){});

define('eventManager',['durandal/app'],
    function (app) {

        var
            events = {
                courseStarted: "courseStarted",
                courseFinished: "courseFinished",
                courseFinalized: "courseFinalized",
                courseEvaluated: 'courseEvaluated',
                answersSubmitted: "answersSubmitted",
                learningContentExperienced: "learningContentExperienced"
            },

            turnAllEventsOff = function () {
                _.each(events, function (event) {
                    app.off(event);
                });
            },

            subscribeForEvent = function (event) {
                if (!events.hasOwnProperty(event)) {
                    throw 'Unsupported event';
                }

                return app.on(event);
            },

            unsubscribeForEvent = function(event, callback) {
                if (!events.hasOwnProperty(event)) {
                    throw 'Unsupported event';
                }

                return app.off(event, callback);
            },

            courseStarted = function (data) {
                app.trigger(events.courseStarted, data);
            },

            courseEvaluated = function (data) {
                return executeAfterSubscribersDone(events.courseEvaluated, data);
            },

            courseFinished = function (data) {
                return executeAfterSubscribersDone(events.courseFinished, data);
            },

            courseFinalized = function () {
                return executeAfterSubscribersDone(events.courseFinalized, {});
            },

            answersSubmitted = function (data, sendParentProgress) {
                app.trigger(events.answersSubmitted, data, sendParentProgress);
            },
            
            learningContentExperienced = function (data, spentTime) {
                app.trigger(events.learningContentExperienced, data, spentTime);
            },
            
            executeAfterSubscribersDone = function (event, eventData) {
                if (_.isNullOrUndefined(app.callbacks) || _.isNullOrUndefined(app.callbacks[event])) {
                    return Q();
                }

                var promises = [];
                _.each(app.callbacks[event], function (handler) {
                    if (_.isFunction(handler)) {
                        var promise = handler(eventData);

                        if (Q.isPromise(promise)) {
                            promises.push(promise);
                        }
                    }
                });

                return Q.all(promises);
            };

        return {
            events: events,
            turnAllEventsOff: turnAllEventsOff,
            subscribeForEvent: subscribeForEvent,
            unsubscribeForEvent: unsubscribeForEvent,

            courseStarted: courseStarted,
            courseFinished: courseFinished,
            courseFinalized: courseFinalized,
            courseEvaluated: courseEvaluated,
            answersSubmitted: answersSubmitted,
            learningContentExperienced: learningContentExperienced
        };
    }
);
define('models/course',["eventManager", "constants", "templateSettings"], function(
  eventManager,
  constants,
  templateSettings
) {
  var ctor = function(spec) {
    var course = {
      id: spec.id,
      templateId: spec.templateId,
      title: spec.title,

      createdBy: spec.createdBy,
      createdOn: spec.createdOn,

      authorContactEmail: spec.authorContactEmail,
      authorPersonalPhone: spec.authorPersonalPhone,
      authorShortBio: spec.authorShortBio,
      authorAvatarUrl: spec.authorAvatarUrl,

      hasIntroductionContent: spec.hasIntroductionContent,
      introductions: spec.introductions,
      sections: spec.sections,
      isFinished: false
    };

    var affectProgressSections = _.filter(course.sections, function(section) {
      return section.affectProgress;
    });

    course.score = function() {
      var sum = 0;
      var itemCount = 0;
      if (templateSettings.masteryScore.isOverall) {
        itemCount = _.reduce(
          affectProgressSections,
          function(memo, section) {
            var questions = _.filter(section.questions, function(question) {
              return question.affectProgress;
            });
            questions.forEach(function(question) {
              sum += question.score();
            });
            return (memo += questions.length);
          },
          0
        );
      } else {
        itemCount = affectProgressSections.length;
        sum = _.reduce(
          affectProgressSections,
          function(memo, section) {
            return memo + section.score();
          },
          0
        );
      }

      return itemCount == 0 ? 0 : Math.floor(sum / itemCount);
    };

    course.result = function() {
      return course.score() / 100;
    };

    course.isCompleted = function() {
      if (templateSettings.masteryScore.isOverall) {
        if (affectProgressSections.length === 0) {
          return true;
        }

        return course.score() >= templateSettings.masteryScore.score;
      }

      return !_.some(affectProgressSections, function(section) {
        return !section.isCompleted();
      });
    };

    course.getStatus = function() {
      if (!course.isFinished) {
        return constants.course.statuses.inProgress;
      }

      return course.isCompleted()
        ? constants.course.statuses.completed
        : constants.course.statuses.failed;
    };

    course.setFinishedStatus = function() {
      course.isFinished = true;
    };

    course.finish = function(callback, withoutEvent) {
      course.setFinishedStatus();
      course.sections.forEach(function(section) {
        section.questions.forEach(function(question) {
          question.markAsAnswered();
        });
      });

      if (withoutEvent) {
        callback && callback();
      } else {
        eventManager
          .courseFinished(course)
          .fail(function(error) {
            alert(error);
          })
          .fin(function() {
            callback && callback();
          });
      }
    };

    course.finalize = function(callback) {
      eventManager.courseFinalized().then(function() {
        eventManager.turnAllEventsOff();
        if (callback) {
          callback();
        }
      });
    };

    course.evaluate = function(score, callback) {
      eventManager.courseEvaluated(score).fin(function() {
        callback && callback();
      });
    };

    return course;
  };

  return ctor;
});

define('models/section',['templateSettings'],
    function (templateSettings) {

        var ctor = function (spec) {

            var section = {
                id: spec.id,
                title: spec.title,
                imageUrl: spec.imageUrl,
                questions: spec.questions
            };

            var questions = _.filter(section.questions, function (question) {
                return question.affectProgress;
            });

            section.affectProgress = questions.length > 0;

            section.hasSurveyQuestions = !!_.filter(section.questions, function (question) {
                return question.isSurvey;
            }).length;

            section.score = ko.computed(function () {
                var result = _.reduce(questions, function (memo, question) {
                    return memo + question.score();
                }, 0);
                var questionsLength = questions.length;
                return questionsLength === 0 ? templateSettings.masteryScore.score : Math.floor(result / questionsLength);
            });

            section.isCompleted = ko.computed(function () {
                return templateSettings && templateSettings.masteryScore ? section.score() >= templateSettings.masteryScore.score : false;
            });

            return section;
        };

        return ctor;
    }
);
define('publishSettings',[],function () {
    'use strict';

    return {
        init: function(settings){
            for (var key in settings){
                this[key] = settings[key]; 
            }
        }
    };
});

define('guard',[], function () {

    return {
        throwIfNotAnObject: function (obj, message) {
            if (!_.isObject(obj)) {
                throw message;
            }
        },

        throwIfNotString: function (str, message) {
            if (!_.isString(str)) {
                throw message;
            }
        },

        throwIfNotBoolean: function (bool, message) {
            if (!_.isBoolean(bool)) {
                throw message;
            }
        },

        throwIfNotArray: function (array, message) {
            if (!_.isArray(array)) {
                throw message;
            }
        },

        throwIfNotDate: function (date, message) {
            if (!_.isDate(date)) {
                throw message;
            }
        },
        
        throwIfNotNumber: function (number, message) {
            if (!_.isNumber(number)) {
                throw message;
            }
        }
    };

});
define('models/contentBlock',[],
    function () {

        function ContentBlock(id, contentUrl, children) {
            this.id = id;
            this.contentUrl = contentUrl;
            this.children = children;

            this.content;
        }

        return ContentBlock;
    }
);
define('requester/httpWrapper',['jquery'], function ($) {
  'use strict';

  return {
    get: get,
    post: post,
    put: put
  };

  function get(options) {
    options.query = options.query || {};
    options.headers = options.headers || {};
    options.cacheBuster = options.cacheBuster || '';
    options.useCaching = options.useCaching || false;

    if (options.withCredentials) {
      return $.ajax(options.url, {
        data: Object.assign({}, options.query, { _: options.cacheBuster }),
        headers: _buildHeaders(options.headers, options.useCaching),
        cache: options.useCaching,
        xhrFields: {
          withCredentials: true
        },
        crossDomain: true,
        dataType: options.dataType,
        contentType: options.contentType
      });
    } else {
      return $.ajax(options.url, {
        data: Object.assign({}, options.query, { _: options.cacheBuster }),
        headers: _buildHeaders(options.headers, options.useCaching),
        cache: options.useCaching,
        dataType: options.dataType,
        contentType: options.contentType
      });
    }
  }

  function post(options) {
    options.data = options.data || {};
    options.headers = options.headers || {};
    if (options.withCredentials) {
      return $.ajax(options.url, {
        method: 'POST',
        data: JSON.stringify(options.data),
        headers: options.headers,
        cache: false,
        xhrFields: {
          withCredentials: true
        },
        crossDomain: true
      });
    } else {
      return $.ajax(options.url, {
        method: 'POST',
        data: JSON.stringify(options.data),
        headers: options.headers,
        cache: false
      });
    }
  }

  function put(options) {
    options.data = options.data || {};
    options.headers = options.headers || {};
    if (options.withCredentials) {
      return $.ajax(options.url, {
        method: 'PUT',
        data: JSON.stringify(options.data),
        headers: options.headers,
        cache: false,
        xhrFields: {
          withCredentials: true
        },
        crossDomain: true
      });
    } else {
      return $.ajax(options.url, {
        method: 'PUT',
        data: JSON.stringify(options.data),
        headers: options.headers,
        cache: false
      });
    }
  }

  function _buildHeaders(headers, cacheBuster) {
    headers = headers || {};
    cacheBuster = cacheBuster || false;
    const _headers = Object.assign({}, headers);
    const CACHE_EXPIRATION_DATE_IN_SECONDS = 2678400;
    if (cacheBuster) {
      _headers['Cache-Control'] = 'public';
      _headers.Pragma = 'public';
      _headers.Expires = CACHE_EXPIRATION_DATE_IN_SECONDS;
    } else {
      _headers['Cache-Control'] = 'no-cache';
      _headers.Pragma = 'no-cache';
      _headers.Expires = 0;
    }

    return _headers;
  }
});

define('requester/resourceLoader',['requester/httpWrapper'], function (requester) {
  'use strict';

  function ResourceLoader() {
    this.expires = +new Date();

    this.setCacheBuster = function(cacheBuster) {
      if (!cacheBuster) {
        return;
      }
      this.expires = +Date.parse(cacheBuster);
    }

    this.getCacheBuster = function() {
      return this.expires;
    }

    this.getLocalResource = function(data) {
      const req = Object.assign({ url: data.url, useCaching: true, cacheBuster: this.expires }, data.requestOptions);
      return requester.get(req);
    }
  }
  return new ResourceLoader();
});

﻿define('models/questions/question',['eventManager', 'guard', 'requester/resourceLoader', 'constants'], function (
  eventManager,
  guard,
  resourceLoader,
  constants
) {
  'use strict';

  function Question(spec, _protected) {
    if (typeof spec == typeof undefined) {
      throw 'You should provide a specification to create an Question';
    }

    this.id = spec.id;
    this.shortId = spec.shortId;
    this.sectionId = spec.sectionId;
    this.title = spec.title;
    this.type = spec.type;
    this.score = ko.observable(spec.score);
    this.learningContents = spec.learningContents;
    this.questionInstructions = spec.questionInstructions;
    this.isAnswered = false;
    this.isCorrectAnswered = false;
    this.affectProgress = true;
    this.isInformationContent = spec.type === constants.questionTypes.informationContent;
    this.correctFeedback = spec.correctFeedback;
    this.incorrectFeedback = spec.incorrectFeedback;

    if (spec.hasOwnProperty('isSurvey')) {
      this.isSurvey = spec.isSurvey;
      this.affectProgress = !this.isSurvey;
    }

    this.feedback = {
      hasCorrect: spec.hasCorrectFeedback,
      correct: null,

      hasIncorrect: spec.hasIncorrectFeedback,
      incorrect: null
    };
    (this.loadFeedback = loadFeedback), (this.loadLearningContent = loadLearningContent);
    this.loadQuestionInstructions = loadQuestionInstructions;
    this.load = load;

    this.voiceOver = spec.voiceOver;

    this.learningContentExperienced = learningContentExperienced;

    this.loadContent = loadContent;

    this.progress = function (data) {
      if (!this.affectProgress && !this.hasOwnProperty('isSurvey')) return 0;

      if (!_.isNullOrUndefined(data)) {
        _protected.restoreProgress.call(this, data);

        this.markAsAnswered();
        this.isCorrectAnswered = this.score() === 100;
      } else {
        return _protected.getProgress.call(this);
      }
    };

    this.submitAnswer = function () {
      this.markAsAnswered();

      if (!this.affectProgress && !this.hasOwnProperty('isSurvey')) return;

      this.score(_protected.submit.apply(this, arguments));

      this.isCorrectAnswered = this.score() === 100;

      eventManager.answersSubmitted(this, true);
    };

    this.markAsAnswered = function () {
      this.isAnswered = true;
    };
  }

  return Question;

  function learningContentExperienced(spentTime) {
    eventManager.learningContentExperienced(this, spentTime);
  }

  function loadQuestionInstructions() {
    return this.loadContent(this.questionInstructions);
  }

  function loadLearningContent() {
    return this.loadContent(this.learningContents);
  }

  function load() {
    var that = this;
    return that.loadQuestionInstructions().then(function () {
      return that.loadLearningContent().then(function () {
        return that.loadFeedback();
      });
    });
  }

  function loadFeedback() {
    if (this.feedback.hasIncorrect) {
      this.loadContent(this.incorrectFeedback);
    }

    if (this.feedback.hasCorrect) {
      this.loadContent(this.correctFeedback);
    }
  }

  function loadContent(items) {
    var that = this;
    var promises = [];
    _.each(items, function (item) {
      if (typeof item.content === typeof undefined) {
        promises.push(
          loadPage(item.contentUrl).then(function (content) {
            item.content = content;
          })
        );
      }

      return that.loadContent(item.children);
    });

    return Q.allSettled(promises);
  }

  function loadPage(contentUrl) {
    return resourceLoader.getLocalResource({
      url: contentUrl
    });
  }
});

define('models/answers/answer',[],
    function () {

        function Answer(spec) {
            this.id = spec.id;
            this.shortId = spec.shortId;
            this.text = spec.text;
            this.isCorrect = spec.isCorrect;
            this.matchCase = spec.matchCase;
        }

        return Answer;
    }
);
define('models/answers/checkableAnswer',['models/answers/answer'], function (Answer) {
    "use strict";

    function CheckableAnswer(spec) {
        Answer.call(this, spec);

        this.isChecked = false;
    }

    return CheckableAnswer;

});
define('models/questions/multipleSelectQuestion',['guard', 'models/questions/question', 'models/answers/checkableAnswer'],
    function (guard, Question, CheckableAnswer) {
        "use strict";

        function MultipleSelectQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.answers = (function () {
                var index = 0;
                return _.map(spec.answers, function (answer) {
                    return new CheckableAnswer({
                        id: answer.id,
                        shortId: index++,
                        isCorrect: answer.isCorrect,
                        text: answer.text
                    });
                });
            })();
        }

        return MultipleSelectQuestion;

        function submitAnswer(checkedAnswerIds) {
            guard.throwIfNotArray(checkedAnswerIds, 'Checked answer ids is not an array');

            _.each(this.answers, function (answer) {
                answer.isChecked = _.contains(checkedAnswerIds, answer.id);
            });

            return calculateScore(this.answers);
        }

        function calculateScore(answers) {
            var hasIncorrectCheckedAnswer = _.some(answers, function (answer) {
                return answer.isChecked != answer.isCorrect;
            });

            return hasIncorrectCheckedAnswer ? 0 : 100;
        }

        function getProgress() {
            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.chain(this.answers)
                    .filter(function (answer) {
                        return answer.isChecked;
                    })
                    .map(function (answer) {
                        return answer.shortId;
                    }).value();
            }
        }

        function restoreProgress(progress) {
            _.each(this.answers, function (answer) {
                answer.isChecked = progress === 100 ? answer.isCorrect : progress && _.indexOf(progress, answer.shortId) > -1;
            });
            this.score(calculateScore(this.answers));
        }

    });
define('models/answers/answerGroup',[],function () {
    "use strict";

    function AnswerGroup(spec) {
        this.id = spec.id;
        this.shortId = spec.shortId;
        this.answeredText = '';
        this.answers = spec.answers;

        this.getCorrectText = getCorrectText;
        this.checkIsCorrect = checkIsCorrect;
    }

    return AnswerGroup;

    function getCorrectText() {
        var correctAnswers = _.where(this.answers, {
            isCorrect: true
        });
        return _.map(correctAnswers, function (answer) {
            return answer.text;
        });
    }

    function checkIsCorrect(answerGroupValue) {
        if (!_.isString(answerGroupValue.value)) {
            return false;
        }

        return _.some(this.answers, function (answer) {
            return answer.isCorrect &&
                (answer.matchCase ? answer.text === answerGroupValue.value : answer.text.toLowerCase() === answerGroupValue.value.toLowerCase());
        });
    }
});
define('models/questions/fillInTheBlankQuestion',['guard', 'models/questions/question', 'models/answers/answerGroup', 'models/answers/answer'],
    function (guard, Question, AnswerGroup, Answer) {
        "use strict";

        function FillInTheBlankQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.hasContent = spec.hasContent;
            this.answerGroupsValues = null;

            this.answerGroups = (function () {
                var index = 0;
                return _.map(spec.answerGroups, function (answerGroup) {
                    return new AnswerGroup({
                        id: answerGroup.id,
                        shortId: index++,
                        type: answerGroup.type,
                        answers: _.map(answerGroup.answers, function (answer) {
                            return new Answer({
                                id: answer.id,
                                isCorrect: answer.isCorrect,
                                text: answer.text,
                                matchCase: answer.matchCase
                            });
                        })
                    });
                });
            })();
        }

        return FillInTheBlankQuestion;

        function submitAnswer(inputValues) {
            guard.throwIfNotArray(inputValues, 'Input values is not array.');

            saveAnsweredTexts(inputValues, this.answerGroups);
            this.answerGroupsValues = inputValues;

            return calculateScore(inputValues, this.answerGroups);
        }

        function calculateScore(answerGroupValues, pointer) {
            var hasIncorrectAnswer = _.some(answerGroupValues, function (answerGroupValue) {
                return _.some(pointer, function (answerGroup) {
                    if (_.isFunction(answerGroup.checkIsCorrect) && answerGroupValue.id == answerGroup.id) {
                        return !answerGroup.checkIsCorrect(answerGroupValue);
                    }
                    return false;
                });
            });

            return hasIncorrectAnswer ? 0 : 100;
        }

        function saveAnsweredTexts(answerGroupValues, answerGroups) {
            _.each(answerGroupValues, function (answerGroupValue) {
                var answerGroup = _.find(answerGroups, function (answerGroup) {
                    return answerGroup.id === answerGroupValue.id;
                });

                answerGroup.answeredText = answerGroupValue.value;
            });
        }

        function getProgress() {
            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.chain(this.answerGroups)
                    .filter(function (group) {
                        return !!group.answeredText;
                    })
                    .reduce(function (obj, ctx) {
                        obj[ctx.shortId] = ctx.answeredText;
                        return obj;
                    }, {})
                    .value();
            }
        }

        function restoreProgress(progress) {
            var inputValues = _.chain(this.answerGroups)
                .map(function (answerGroup) {
                    var correct = _.chain(answerGroup.answers)
                        .find(function (answer) {
                            return answer.isCorrect;
                        }).value();

                    return {
                        id: answerGroup.id,
                        value: progress === 100 ? correct.text : progress[answerGroup.shortId],
                        answers: answerGroup.answers
                    };
                }).value();

            saveAnsweredTexts(inputValues, this.answerGroups);
            this.score(calculateScore(inputValues, this.answerGroups));
        }

    });
define('models/answers/draggableAnswer',['models/answers/answer'], function (Answer) {
    "use strict";

    function DraggableAnswer(spec) {
        Answer.call(this, spec);

        this.correctPosition = { x: spec.x, y: spec.y };

        this.currentPosition = { x: -1, y: -1 };
    }

    return DraggableAnswer;

});
define('models/questions/dragAndDropQuestion',['guard', 'models/questions/question', 'models/answers/draggableAnswer'],
    function (guard, Question, DraggableAnswer) {
        "use strict";

        function DragAndDropQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.background = spec.background;

            this.answers = (function () {
                var index = 0;
                return _.map(spec.dropspots, function (answer) {
                    return new DraggableAnswer({
                        id: answer.id,
                        shortId: index++,
                        isCorrect: true,
                        text: answer.text,
                        x: answer.x,
                        y: answer.y
                    });
                });
            })();
        }

        return DragAndDropQuestion;

        function submitAnswer(dragableTexts) {
            guard.throwIfNotArray(dragableTexts, 'Dragable texts is not array.');

            var that = this;
            var answersArray = [];
            _.each(that.answers, function (answer) {
                answer.currentPosition = { x: -1, y: -1 };
            })
            _.each(dragableTexts, function (dragableText) {
                answersArray = _.where(that.answers, { text: dragableText.text });
                var answer = _.find(answersArray, function (answer) {
                    return answer.text === dragableText.text && answer.correctPosition.x === dragableText.x && answer.correctPosition.y === dragableText.y
                })
                if (answer && answer.currentPosition) {
                    answer.currentPosition.x = answer.correctPosition.x,
                    answer.currentPosition.y = answer.correctPosition.y
                }
                else {
                    var answer = _.find(that.answers, function (answer) {
                        return answer.text == dragableText.text
                    })
                    answer.currentPosition.x = dragableText.x,
                    answer.currentPosition.y = dragableText.y
                }
            });

            return calculateScore(this.answers);

        }

        function calculateScore(answers) {

            var hasIncorrectAnswer = _.some(answers, function (answer) {
                return answer.currentPosition.x != answer.correctPosition.x || answer.currentPosition.y != answer.correctPosition.y;
            });

            return hasIncorrectAnswer ? 0 : 100;
        }

        function getProgress() {

            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.chain(this.answers)
                    .filter(function (answer) {
                        return answer.currentPosition
                            && answer.currentPosition.x
                            && answer.currentPosition.x > -1
                            && answer.currentPosition.y
                            && answer.currentPosition.y > -1;
                    })
                    .reduce(function (obj, ctx) {
                        obj[ctx.shortId] = ctx.currentPosition;
                        return obj;
                    }, {})
                    .value();
            }
        }

        function restoreProgress(progress) {
            if (progress === 100) {
                _.each(this.answers, function (answer) {
                    answer.currentPosition = { x: answer.correctPosition.x, y: answer.correctPosition.y }
                });
            } else {
                _.each(this.answers, function (answer) {
                    if (progress[answer.shortId]) {
                        answer.currentPosition = progress[answer.shortId];
                    }
                });

            }
            this.score(calculateScore(this.answers));
        }

    });
define('models/answers/checkableImageAnswer',['models/answers/checkableAnswer'], function (CheckableAnswer) {
    "use strict";

    function CheckableImageAnswer(spec) {
        this.id = spec.id;
        this.shortId = spec.shortId;
        this.image = spec.image;
    }

    return CheckableImageAnswer;

});
define('models/questions/singleSelectImageQuestion',['guard', 'constants', 'models/questions/question', 'models/answers/checkableImageAnswer'],
    function (guard, constants, Question, CheckableImageAnswer) {
        "use strict";

        function SingleSelectImageQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.correctAnswerId = spec.correctAnswerId;
            this.checkedAnswerId = null;
            this.answers = (function () {
                var index = 0;
                return _.map(spec.answers, function (answer) {
                    return new CheckableImageAnswer({
                        id: answer.id,
                        shortId: index++,
                        image: answer.image || constants.defaultImageUrl
                    });
                });
            })();
        }

        return SingleSelectImageQuestion;

        function submitAnswer(checkedAnswerId) {

            this.checkedAnswerId = checkedAnswerId;

            return calculateScore(checkedAnswerId, this.correctAnswerId);
        }

        function calculateScore(answerId, correctAnswerId) {
            return answerId != correctAnswerId ? 0 : 100;
        }

        function getProgress() {
            var that = this;
            if (that.isCorrectAnswered) {
                return 100;
            } else {
                var checked = _.find(that.answers, function (answer) {
                    return answer.id == that.checkedAnswerId;
                });

                return checked ? checked.shortId : undefined;
            }
        }

        function restoreProgress(progress) {
            var that = this;
            if (progress === 100) {
                that.checkedAnswerId = that.correctAnswerId;
            } else {
                var checked = _.find(that.answers, function (answer) {
                    return answer.shortId === progress;
                });

                if (checked) {
                    that.checkedAnswerId = checked.id;
                }

            }
            that.score(calculateScore(that.checkedAnswerId, that.correctAnswerId));
        }

    });
define('models/questions/textMatchingQuestion',['models/questions/question'],
    function (Question) {
        "use strict";

        function TextMatchingQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.answers = (function () {
                var index = 0;
                return _.map(spec.answers, function (answer) {
                    return {
                        id: answer.id,
                        shortId: index++,
                        key: answer.key,
                        value: answer.value,
                        attemptedValue: null
                    };
                });
            })();
        }

        return TextMatchingQuestion;

        function submitAnswer(pairs) {
            _.each(this.answers, function (answer) {
                var attempted = _.find(pairs, function (pair) {
                    return pair.id == answer.id;
                });
                if (attempted) {
                    answer.attemptedValue = attempted.value;
                } else {
                    answer.attemptedValue = null;
                }
            });

            return getScore(this.answers);
        }

        function getScore(answers) {
            return _.every(answers, function (answer) {
                return answer.value == answer.attemptedValue;
            }) ? 100 : 0;
        }

        function getProgress() {
            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.chain(this.answers)
                    .filter(function (answer) {
                        return !!answer.attemptedValue;
                    })
                    .map(function (answer) {
                        return {
                            shortId: answer.shortId,
                            attemptedValue: answer.attemptedValue
                        }
                    })
                    .reduce(function (obj, ctx) {
                        obj[ctx.shortId] = ctx.attemptedValue;
                        return obj;
                    }, {})
                    .value();
            }
        }

        function restoreProgress(progress) {
            if (progress === 100) {
                _.each(this.answers, function (answer) {
                    answer.attemptedValue = answer.value;
                });
            } else {
                _.each(this.answers, function (answer) {
                    if (progress[answer.shortId]) {
                        answer.attemptedValue = progress[answer.shortId];
                    }
                });
            }
            this.score(getScore(this.answers));
        }
    });
define('models/questions/informationContent',['models/questions/question', 'templateSettings'],
    function (Question, templateSettings) {
        "use strict";

        function InformationContent(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submit
            });

            this.affectProgress = templateSettings.allowContentPagesScoring;

            var superSubmitAnswer = this.submitAnswer;
            this.submitAnswer = function () {
                if (this.isAnswered)
                    return;

                superSubmitAnswer.call(this);
            }
        }

        return InformationContent;

        function getProgress() {
            return this.score();
        }

        function restoreProgress(progress) {
            if (!_.isNaN(progress)) {
                this.score(progress === 100 ? progress : 0);
            }
        }

        function submit() {
            return 100;
        }
});
define('models/answers/statementAnswer',['models/answers/answer'], function (Answer) {
    "use strict";

    function SatementAnswer(spec) {
        Answer.call(this, spec);
        this.userAnswer = null;
    }

    return SatementAnswer;
});
define('models/questions/statementQuestion',['models/questions/question', 'models/answers/statementAnswer', 'guard'],
    function (Question, StatementAnswer, guard) {
        "use strict";

        function StatementQuestion(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.statements = (function () {
                var index = 0;
                return _.map(spec.statements, function (statement) {
                    return new StatementAnswer({
                        id: statement.id,
                        shortId: index++,
                        text: statement.text,
                        isCorrect: statement.isCorrect
                    });
                });
            })();
        }

        return StatementQuestion;

        function submitAnswer(userAnswers) {
            guard.throwIfNotArray(userAnswers, 'userAnswers is not an array');

            _.each(this.statements, function (statement) {
                var userAnswer = _.find(userAnswers, function (answer) { return answer.id == statement.id; });
                statement.userAnswer = !_.isNullOrUndefined(userAnswer) ? userAnswer.answer : null;
            });

            return calculateScore.call(this);
        }

        function calculateScore() {
            return _.every(this.statements, function (statement) {
                return statement.userAnswer == statement.isCorrect;
            }) ? 100 : 0;
        }

        function getProgress() {
            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.chain(this.statements)
                    .filter(function (statement) {
                        return _.isBoolean(statement.userAnswer);
                    })
                    .reduce(function (obj, ctx) {
                        obj[ctx.shortId] = ctx.userAnswer ? 1 : 0;
                        return obj;
                    }, {})
                    .value();

            }
        }

        function restoreProgress(progress) {
            _.chain(this.statements)
                .each(function (statement) {
                    if (progress == 100) {
                        statement.userAnswer = statement.isCorrect;
                    } else {
                        statement.userAnswer = !!progress[statement.shortId];
                    }
                });
            this.score(calculateScore.call(this));
        }

    });
define('models/questions/hotspot',['models/questions/question', 'guard'],
    function (Question, guard) {
        "use strict";

        function Hotspot(spec) {

            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.background = spec.background;
            this.spots = spec.spots;
            this.isMultiple = spec.isMultiple;
            this.placedMarks = [];
        };

        return Hotspot;

        function submitAnswer(marks) {
            guard.throwIfNotArray(marks, 'Marks is not array.');

            this.placedMarks = _.map(marks, function (mark) { return { x: mark.x, y: mark.y }; });

            return calculateScore(this.isMultiple, this.spots, this.placedMarks);
        };

        function calculateScore(isMultiple, spots, placedMarks) {
            if (!_.isArray(spots) || spots.length == 0) {
                return placedMarks.length ? 0 : 100;
            }

            var answerCorrect;
            if (!isMultiple) {
                answerCorrect = _.some(spots, function (spot) {
                    return _.some(placedMarks, function (mark) {
                        return markIsInSpot(mark, spot);
                    });
                });
            } else {
                var spotsWithMarks = [];
                var marksOnSpots = [];

                _.each(placedMarks, function (mark) {
                    _.each(spots, function (spot) {
                        if (markIsInSpot(mark, spot)) {
                            spotsWithMarks.push(spot);
                            marksOnSpots.push(mark);
                        }
                    });
                });

                answerCorrect = _.uniq(spotsWithMarks).length === spots.length && _.uniq(marksOnSpots).length === placedMarks.length;
            }

            return answerCorrect ? 100 : 0;
        }

        function markIsInSpot(mark, spot) {
            var x = mark.x, y = mark.y;

            var inside = false;
            for (var i = 0, j = spot.length - 1; i < spot.length; j = i++) {
                var xi = spot[i].x, yi = spot[i].y;
                var xj = spot[j].x, yj = spot[j].y;

                var intersect = ((yi > y) != (yj > y))
                    && (x < (xj - xi) * (y - yi) / (yj - yi) + xi);
                if (intersect) inside = !inside;
            }

            return inside;
        };

        function getProgress() {
            return this.placedMarks;
        }

        function restoreProgress(progress) {
            var that = this;
            _.each(progress, function (mark) {
                that.placedMarks.push(mark);
            });

            this.score(calculateScore(that.isMultiple, that.spots, that.placedMarks));
        }
    });
define('models/questions/openQuestion',['models/questions/question', 'guard'], function (Question, guard) {
    "use strict";

    function OpenQuestion(spec) {
        Question.call(this, spec, {
            getProgress: getProgress,
            restoreProgress: restoreProgress,

            submit: submitAnswer
        });

        this.answeredText = '';
    }

    return OpenQuestion;

    function getProgress() {
        return this.isAnswered ? this.answeredText : 0;
    }

    function restoreProgress(progress) {
        if (progress === 0) {
            return;
        }

        this.answeredText = progress;
        this.score(progress ? 100 : 0);
    }

    function submitAnswer(answeredText) {
        guard.throwIfNotString(answeredText);
        this.answeredText = answeredText;
        return answeredText ? 100 : 0;
    }
});
define('models/questions/scenarioQuestion',['guard', 'models/questions/question'],
    function (guard, Question) {
        "use strict";

        function ScenarioQuestion(spec) {
            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submitAnswer
            });

            this.embedCode = spec.embedCode;

            this.embedUrl = spec.embedUrl;

            this.projectId = spec.projectId;

            this.masteryScore = spec.masteryScore;
        }

        return ScenarioQuestion;

        function getProgress() {
            return this.score();
        }

        function restoreProgress(progress) {
            if (!_.isNaN(progress)) {
                this.score(progress === 100 ? progress : 0);
            }
        }

        function submitAnswer(score) {
            guard.throwIfNotNumber(score, 'Score is not a number');

            var correctScore = score >= this.masteryScore ? 100 : 0;

            return correctScore;
        }
    });
define('models/questions/rankingTextQuestion',['models/questions/question'],
    function (Question) {
        "use strict";

        function RankingTextQuestion(spec) {
            Question.call(this, spec, {
                getProgress: getProgress,
                restoreProgress: restoreProgress,

                submit: submit
            });

            this.correctOrder = _.map(spec.rankingItems, function (item) {
                return {
                    text: item.text
                };
            });

            this.rankingItems = _.shuffle(this.correctOrder);
        }

        return RankingTextQuestion;

        function submit(items) {
            var pattern = _.pluck(this.correctOrder, 'text'),
                answered = _.pluck(items, 'text');

            return _.isEqual(pattern, answered) ? 100 : 0;
        }

        function getProgress() {
            if (this.isCorrectAnswered) {
                return 100;
            } else {
                return _.pluck(this.rankingItems, 'text');
            }
        }

        function restoreProgress(progress) {
            if (progress === 100) {
                this.rankingItems = this.correctOrder;
                this.score(100);
            } else {
                var pattern = _.pluck(this.correctOrder, 'text');
                if (pattern.length === progress.length && _.intersection(pattern, progress).length === pattern.length) {
                    this.rankingItems = _.map(progress, function(item) {
                        return {
                            text: item
                        };
                    });
                }
                this.score(0);
            }
        }
    });
define('models/questions/questionsFactory',['guard', 'constants', 'models/contentBlock', 'models/questions/multipleSelectQuestion',
        'models/questions/fillInTheBlankQuestion', 'models/questions/dragAndDropQuestion',
        'models/questions/singleSelectImageQuestion', 'models/questions/textMatchingQuestion',
        'models/questions/informationContent', 'models/questions/statementQuestion', 'models/questions/hotspot',
        'models/questions/openQuestion', 'models/questions/scenarioQuestion', 'models/questions/rankingTextQuestion'
    ],
    function (guard, constants, ContentBlock, MultipleSelectQuestion, FillInTheBlankQuestion, DragAndDropQuestion,
        SingleSelectImageQuestion, TextMatchingQuestion, InformationContent,
        StatementQuestion, Hotspot, OpenQuestion, ScenarioQuestion, RankingText) {
        "use strict";

        return {
            createQuestion: createQuestion,
        };

        function createQuestion(sectionId, question, shortId) {
            guard.throwIfNotString(sectionId, 'sectionId is invalid');
            guard.throwIfNotAnObject(question, 'Question data is invalid');
            guard.throwIfNotString(question.type, 'Question type is invalid');
            
            var questionData = {
                id: question.id,
                shortId: shortId,
                sectionId: sectionId,
                title: question.title,
                type: question.type,
                learningContents: _.map(question.learningContents, function (learningContent) {
                    return mapContentBlock(learningContent, sectionId, question.id);
                }),
                questionInstructions: _.map(question.questionInstructions, function (instruction) {
                    return mapContentBlock(instruction, sectionId, question.id);
                }),
                score: 0,
                voiceOver: question.voiceOver,
                hasCorrectFeedback: question.hasCorrectFeedback,
                hasIncorrectFeedback: question.hasIncorrectFeedback,
                correctFeedback: _.map(question.questionCorrectFeedbacks, function(correctFeedback) {
                    return mapContentBlock(correctFeedback, sectionId, question.id);
                }),
                incorrectFeedback: _.map(question.questionIncorrectFeedbacks, function(incorrectFeedback) {
                    return mapContentBlock(incorrectFeedback, sectionId, question.id);
                })
            };

            if (question.hasOwnProperty('isSurvey')) {
                questionData.isSurvey = question.isSurvey;
            }

            switch (question.type) {
                case constants.questionTypes.multipleSelect:
                case constants.questionTypes.singleSelectText:
                    questionData.answers = question.answers;
                    return new MultipleSelectQuestion(questionData);
                case constants.questionTypes.dragAndDrop:
                    questionData.background = question.background;
                    questionData.dropspots = question.dropspots;
                    return new DragAndDropQuestion(questionData);
                case constants.questionTypes.fillInTheBlank:
                    questionData.answerGroups = question.answerGroups;
                    questionData.hasContent = question.hasContent;
                    return new FillInTheBlankQuestion(questionData);
                case constants.questionTypes.singleSelectImage:
                    questionData.correctAnswerId = question.correctAnswerId;
                    questionData.answers = question.answers;
                    return new SingleSelectImageQuestion(questionData);
                case constants.questionTypes.textMatching:
                    questionData.answers = question.answers;
                    return new TextMatchingQuestion(questionData);
                case constants.questionTypes.informationContent:
                    return new InformationContent(questionData);
                case constants.questionTypes.statement:
                    questionData.statements = question.answers;
                    return new StatementQuestion(questionData);
                case constants.questionTypes.hotspot:
                    questionData.spots = question.spots;
                    questionData.isMultiple = question.isMultiple;
                    questionData.background = question.background;
                    return new Hotspot(questionData);
                case constants.questionTypes.openQuestion:
                    return new OpenQuestion(questionData);
                case constants.questionTypes.scenario:
                    questionData.embedCode = question.embedCode;
                    questionData.masteryScore = question.masteryScore;
                    questionData.projectId = question.projectId;
                    questionData.embedUrl = question.embedUrl;
                    return new ScenarioQuestion(questionData);
                case constants.questionTypes.rankingText:
                    questionData.rankingItems = question.answers;
                    return new RankingText(questionData);
                default:
                    return null;
            }
        }

        function mapContentBlock(item, sectionId, questionId) {
            var contentUrl = 'content/' + sectionId + '/' + questionId + '/' + item.id + '.html';
            var children = _.map(item.children, function(childItem) {
                return mapContentBlock(childItem, sectionId, questionId);
            })
            
            return new ContentBlock(item.id, contentUrl, children)
        }
    });
define('context',[
  'models/course',
  'models/section',
  'publishSettings',
  'models/questions/questionsFactory',
  'requester/resourceLoader'
], function (Course, Section, publishSettings, questionsFactory, resourceLoader) {
  let _response;

  function mapCourse(response) {
    var questionShortIds = publishSettings.questionShortIds || {};

    this.course = new Course({
      id: response.id,
      templateId: response.templateId,
      title: response.title,
      hasIntroductionContent: response.hasIntroductionContent,
      introductions: _.chain(response.introductions)
        .filter(function (item) {
          return !_.isNullOrUndefined(item.id);
        })
        .map(function (item) {
          return {
            id: item.id,
            children: item.children
          };
        })
        .value(),
      sections: _.chain(response.sections)
        .filter(function (item) {
          return !_.isNullOrUndefined(item.questions) && item.questions.length > 0;
        })
        .map(function (section) {
          return new Section({
            id: section.id,
            title: section.title,
            imageUrl: section.imageUrl,
            questions: _.chain(section.questions)
              .map(function (question) {
                return questionsFactory.createQuestion(section.id, question, questionShortIds[question.id]);
              })
              .filter(function (question) {
                return question != null;
              })
              .value()
          });
        })
        .value(),
      createdOn: new Date(response.createdOn),
      createdBy: response.createdBy,
      authorContactEmail: response.authorContactEmail,
      authorPersonalPhone: response.authorPersonalPhone,
      authorShortBio: response.authorShortBio,
      authorAvatarUrl: response.authorAvatarUrl
    });
  }

  var course = {},
    initialize = function () {
      var that = this;

      var dfd = Q.defer();

      if (_response) {
        mapCourse.call(that, _response);
        return;
      }

      resourceLoader
        .getLocalResource({
          url: 'content/data.js',
          requestOptions: {
            contentType: 'application/json',
            dataType: 'json'
          }
        })
        .done(function (response) {
          _response = response;
          mapCourse.call(that, _response);
          dfd.resolve({
            course: that.course
          });
        })
        .fail(function () {
          dfd.reject('Unable to load data.js');
        });

      return dfd.promise;
    };

  return {
    initialize: initialize,
    course: course,
    isInReviewAttemptMode: function () {
      return !!this.course.isFinished;
    }
  };
});

define('displaySettings',[],function() {
  "use strict";

  return {
    backgroundSettings: {
      repeat: {
        repeat: "repeat"
      },
      fullscreen: {
        size: "cover",
        position: "center"
      },
      top: {
        position: "top"
      },
      bottom: {
        position: "bottom"
      },
      left: {
        position: "left"
      },
      right: {
        position: "right"
      },
      leftTop: {
        position: "left top"
      },
      rightTop: {
        position: "right top"
      },
      leftBottom: {
        position: "left bottom"
      },
      rightBottom: {
        position: "right bottom"
      },
      cover: {
        size: "cover"
      }
    }
  };
});

define('header/header',[
    'knockout', 'underscore', 'plugins/router', 'context', 'templateSettings', 'publishSettings', 'displaySettings'
], function (ko, _, router, context, templateSettings, publishSettings, displaySettings) {

        'use strict';
        var viewmodel = {
            version: '',
            title: '',
            logo: {
                url: '',
                maxWidth: '',
                maxHeight: '',
            },
            viewSettings: null,
            backgroundProps: null,
            pdfServiceUrl: null,
            pdfExportEnabled: ko.observable(false),
            height: ko.observable(null),

            cssName: ko.computed(getCssName),
            activate: activate
        };

        return viewmodel;

        //public methods
        function activate(viewSettings, pdfExportEnabled) {
            viewmodel.viewSettings = viewSettings;
            viewmodel.pdfExportEnabled = pdfExportEnabled;
            viewmodel.pdfServiceUrl = '//' + publishSettings.pdfConverterUrl;
            viewmodel.version = context.course.id + (+new Date(context.course.createdOn));
            viewmodel.title = context.course.title;
            viewmodel.logo.url = templateSettings.logo.url;
            viewmodel.logo.maxWidth = templateSettings.logo.maxWidth || '300px';
            viewmodel.logo.maxHeight = templateSettings.logo.maxHeight || '100px';

            var background = templateSettings.background,
                backgroundType = background.header.image && background.header.image.option || 'fullscreen',
                bgSettings = displaySettings.backgroundSettings[backgroundType];

            viewmodel.backgroundProps = {
                brightness: background.header.brightness || 0,
                color: background.header.color || 'transparent',
                isBodyEnabled: background.body.enabled,
                image: {
                    url: background.header.image && background.header.image.url,
                    backgroundPosition: bgSettings && bgSettings.position || '0 0',
                    backgroundSize: bgSettings && bgSettings.size || 'auto',
                    backgroundRepeat: bgSettings && bgSettings.repeat || 'no-repeat'
                }
            };
        }

        function getCssName() {
            var activeItem = router.activeItem();

            if (_.isObject(activeItem)) {
                var moduleId = activeItem.__moduleId__;

                return moduleId.slice(moduleId.lastIndexOf('/') + 1);
            }

            return '';
        }
    });
define('bootstrapping/knockoutBindings/headerHeightDepender',['underscore', 'header/header'], function(_, header) {
    ko.bindingHandlers.headerHeightDepender = {
        init: function(element, valueAccessor, allBindings, viewModel, bindingContext) {
            var deactivate = viewModel.deactivate;

            viewModel.deactivate = function() {
                if(_.isFunction(deactivate)) {
                    deactivate();
                }

                header.height(null);
            }
        },
        update: function (element) {
            setTimeout(function(){
                var innerHeight = $(element).actual('innerHeight');
    
                header.height(innerHeight);
            }, 0)
        }
    }
});
define('viewmodels/questions/dragAndDrop/bindingHandlers/draggableTextBindingHandler',['knockout'], function (ko) {
    'use strict';

    ko.bindingHandlers.draggableText = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var allBindings = allBindingsAccessor();
            var scope = ko.utils.unwrapObservable(allBindings.scope) || 'question';
            var $element = $(element);
			
			$element.css('touch-action','none');
            $element.draggable({
                appendTo: '.application-wrapper',
                scope: scope,
                revert: 'invalid',
                helper: 'clone',
                start: function () {
                    $element.css({ visibility: 'hidden' });
                },
                stop: function () {
                    $element.css({ visibility: 'visible' });
                }
            });
        }
    };
});
define('viewmodels/questions/dragAndDrop/bindingHandlers/draggableTextContainerBindingHandler',['knockout'], function (ko) {
    'use strict';

    ko.bindingHandlers.draggableTextContainer = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var allBindings = allBindingsAccessor();
            var scope = ko.utils.unwrapObservable(allBindings.scope) || 'question';

            $(element).parent()
                .on('dragstart', '.drag-and-drop-text-draggable', function () {
                    $(element).addClass('active');
                })
                .on('dragstop', '.drag-and-drop-text-draggable', function () {
                    $(element).removeClass('active');
                    if ($(element).children('.drag-and-drop-text-draggable').length) {
                        $(element).children('.drag-and-drop-text-draggable-container-message').hide();
                    } else {
                        $(element).children('.drag-and-drop-text-draggable-container-message').show();
                    }
                });

            $(element).droppable({
                accept: '.drag-and-drop-text-draggable',
                scope: scope,
                drop: function (e, ui) {
                    ui.draggable.css('left', '').css('top', '').appendTo(this);
                    var text = ko.dataFor(ui.draggable.get(0));
                    if (text.dropSpot && text.dropSpot.text) {
                        text.dropSpot.text(undefined);
                    }
                }
            });
        }
    };
});
define('viewmodels/questions/dragAndDrop/bindingHandlers/dropspotBindingHandler',['knockout', 'durandal/composition'], function (ko, composition) {
    'use strict';

    ko.bindingHandlers.dropspot = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var value = valueAccessor();
            var allBindings = allBindingsAccessor();

            var left = ko.utils.unwrapObservable(value.x);
            var top = ko.utils.unwrapObservable(value.y);
            var scope = ko.utils.unwrapObservable(allBindings.scope) || 'question';

            $('.ui-draggable')
                .on('dragstart', function (event, ui) {
                    $(element).addClass('active');
                    $(this).closest('.drag-and-drop-text-dropspot').removeClass('dropped');

                    if ($(element).children('.drag-and-drop-text-draggable').length) {
                        return;
                    }

                    $(element).css('min-width', ui.helper.outerWidth());
                    $(element).height(ui.helper.outerHeight());

                });
                $('html').on('mouseup', function (event, ui) {
                    $(element).removeClass('active');
                    $(element).css('min-width', '');
                    $(element).css('height', '');
                    $(this).closest('.drag-and-drop-text-dropspot').addClass('dropped');
                });

            $(element).droppable({
                accept: function (arg) {
                    if ($(element).find(arg).length) {
                        return true;
                    }

                    return $(element).find('.drag-and-drop-text-draggable').length == 0;
                },
                tolerance: 'intersect',
                scope: scope,
                drop: function (e, ui) {
                    var text = ko.dataFor(ui.draggable.get(0));

                    ui.draggable.css('left', '').css('top', '').appendTo(this);

                    if (ko.isWriteableObservable(value.text)) {
                        text.placed(true);
                        value.text(text);
                        text.dropSpot = value;
                        $(element).addClass('dropped');
                    }
                }
            });

            $(element).css('left', left + 'px').css('top', top + 'px');
        },
        update: function (element, valueAccessor) {
            var value = valueAccessor(),
                text = ko.utils.unwrapObservable(value.text),
                $draggableTextContainer = $('.drag-and-drop-text-draggable-container'),
                draggableTextClass = '.drag-and-drop-text-draggable',
                draggableContainerMessageClass = '.drag-and-drop-text-draggable-container-message';

            if (text) {
                
                var $textChildren = $draggableTextContainer.children(draggableTextClass);
                $.each($textChildren, function (index, item) {
                    var data = ko.dataFor(item);
                    if (data.text === text.text && data.placed() && $(element).children('.drag-and-drop-text-draggable').length ==0) {
                        data.dropSpot = value;
                        $(item).css('left', '').css('top', '').appendTo($(element));
                    }
                });
                if ($draggableTextContainer.children(draggableTextClass).length) {
                    $draggableTextContainer.children(draggableContainerMessageClass).hide();
                } else {
                    $draggableTextContainer.addClass('empty');
                    $draggableTextContainer.children(draggableContainerMessageClass).show();
                }
            } else {
                $(element).closest('.drag-and-drop-text-dropspot').removeClass('dropped');
                $(element).children(draggableTextClass).css('left', '').css('top', '').appendTo($draggableTextContainer);
                $draggableTextContainer.removeClass('empty');
                $draggableTextContainer.children(draggableContainerMessageClass).hide();
            }
        }
    };
    composition.addBindingHandler('dropspot');
});
define('viewmodels/questions/rankingText/bindingHandlers/preventDefaultTouchEventBinding',['knockout'], function (ko) {
    'use strict';

    ko.bindingHandlers.preventDefaultTouchEvent = {
        update: function (element, valueAccessor) {
            if(valueAccessor()) {
                element.addEventListener("touchmove", preventBehavior, false);
            }

            function preventBehavior(e) {
                e.preventDefault(); 
            };
        }
    };
});
define('viewmodels/questions/fillInTheBlank/bindingHandlers/fillInTheBlankAnswersBinding',['knockout', 'durandal/composition'], function (ko, composition) {
    ko.bindingHandlers.fillInTheBlankAnswers = {
        init: function (element, valueAccessor) {
            var $element = $(element),
                value = valueAccessor().inputValues(),
                content = valueAccessor().content;

            $element.html(content);

            $(".blankSelect").select({
                defaultText: TranslationPlugin.getTextByKey('[fill in the blank choose answer]')
            });

            _.each(value, function (blank) {
                var source = $('[data-group-id=' + blank.id + ']', $element),
                    handler = function () {
                        blank.value = source.val().trim();
                    },
                    stopPropagationHandler = function (event) {
                        event.stopPropagation();
                    };

                source.val(undefined)
                    .on('blur change', handler)
                    .on('keydown', stopPropagationHandler);

                ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                    source.off('blur change', handler);
                    source.off('keydown', stopPropagationHandler);
                });
            });
        },
        update: function (element, valueAccessor) {
            var $element = $(element),
                isAnswered = ko.utils.unwrapObservable(valueAccessor().isAnswered),
                inputValues = ko.utils.unwrapObservable(valueAccessor().inputValues);

            if (!isAnswered) {
                $('.blankSelect')
                    .select('refresh')
                    .select('enabled', true);

                $('.blankInput').each(function () {
                    $(this).val('');
                    $(this).removeAttr('disabled');
                });
            } else {
                _.each(inputValues, function (blank) {
                    var $source = $('[data-group-id=' + blank.id + ']', $element);
                    var defaultText = '...';

                    if ($source.is('input')) {
                        $source.val(blank.value || defaultText);
                        $source.attr('disabled', 'disabled');
                    } else if ($source.is('select')) {
                        $source.select('updateValue', blank.value || defaultText);
                        $source.select('enabled', false);
                    }
                });
            }
        }

    };

    composition.addBindingHandler('fillInTheBlankAnswers');

});
define('viewmodels/questions/textMatching/bindingHandlers/draggableBindingHandler',['knockout'], function (ko) {

    ko.bindingHandlers.draggable = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var allBindings = allBindingsAccessor();
			
			$(element).css('touch-action','none');
            $(element).draggable({
                appendTo: '.application-wrapper',
                helper: function () {
                    return $(element)
                        .clone()
                        .addClass('handle')
                        .css({
                            width: $(this).outerWidth(),
                            height: $(this).outerHeight()
                        });
                },
                scope: ko.utils.unwrapObservable(allBindings.scope) || 'default',
                tolerance: 'pointer',
                revert: true,
                revertDuration: 0,
                start: function () {
                    $(element).hide();
                },
                stop: function () {
                    $(element).show();
                }
            });
        }
    }

});
define('viewmodels/questions/textMatching/bindingHandlers/droppableBindingHandler',['knockout'], function (ko) {

    ko.bindingHandlers.droppable = {
        init: function (element, valueAccessor, allBindingsAccessor) {
            var
                value = valueAccessor(),
                allBindings = allBindingsAccessor()
            ;

            $(element).droppable({
                accept: function (arg) {
                    if ($(element).find(arg).length) {
                        return true;
                    }

                    if (allBindings.accept) {
                        return allBindings.accept > $(element).find('.ui-draggable').length;
                    }

                    return $(arg).hasClass('ui-draggable');
                },
                activeClass: 'active',
                hoverClass: 'hover',
                tolerance: 'pointer',
                drop: function (event, ui) {
                    ui.draggable.trigger('dragstop');
                    $(ui.helper).remove();
                   
                    var draggable = ko.dataFor(ui.draggable.get(0));
                    var droppable = ko.dataFor(ui.draggable.closest('.ui-droppable').get(0));

                    if (droppable != value) {
                        if (_.isFunction(value.acceptValue)) {
                            value.acceptValue(draggable);
                        }

                        if (_.isFunction(droppable.rejectValue)) {
                            droppable.rejectValue(draggable);
                        }
                    }
                }
            });

        }
    }

});
define('viewmodels/questions/textMatching/bindingHandlers/overlayBindingHandler',['knockout'], function (ko) {

    ko.bindingHandlers.overlay = {
        update: function (element, valueAccessor) {
            var value = valueAccessor();
            if (ko.utils.unwrapObservable(value)) {
                $(element).addClass('overlay-binding-wrapper');
                $('<div />').addClass('overlay-binding').appendTo(element);
            } else {
                $(element).removeClass('overlay-binding-wrapper');
                $(element).children('.overlay-binding').remove();
            }
        }
    };

});
define('viewmodels/questions/textMatching/bindingHandlers/syncHeightBindingHandler',['knockout', 'durandal/composition'], function (ko, composition) {

    ko.bindingHandlers.syncHeight = {
        init: function (element) {
            var $element = $(element),
                handler = function () {
                    var height = 0;
                    $element.find('.text-matching-row').each(function () {
                        $(this).find('.text-matching-column').height('auto');
                        $(this).height('auto');

                        if ($(this).height() > height) {
                            height = $(this).height();
                        }
                    });
                    $element.find('.text-matching-row, .text-matching-column').each(function () {
                        $(this).height(height);
                    });

                },
                debounced = _.debounce(handler, 10);

            $(window).on('resize', debounced);

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $(window).off('resize', debounced);
            });

            handler();
        }
    };

    composition.addBindingHandler('syncHeight');

});
define('viewmodels/questions/openQuestion/bindingHandlers/textAreaBindingHandler',['jquery', 'knockout'], function ($, ko) {
    ko.bindingHandlers.textArea = {
        init: function (element, valueAccessor, allBindingsAccessor, data, context) {
            ko.bindingHandlers.autosize.init(element, valueAccessor, allBindingsAccessor, data, context);
            ko.bindingHandlers.value.init(element, valueAccessor, allBindingsAccessor, data, context);
        },
        update: function (element, valueAccessor, allBindingsAccessor, data, context) {
            ko.bindingHandlers.value.update(element, valueAccessor, allBindingsAccessor, data, context);
            ko.bindingHandlers.autosize.update(element, valueAccessor, allBindingsAccessor, data, context);
        }
    };
});
define('widgets/imagePreview/viewmodel',['durandal/system', 'durandal/composition', 'durandal/app', 'constants'], function (system, composition, app, constants) {
    "use strict";

    var imagePreviewModel = {
        imageId: ko.observable(null),
        imageUrl: ko.observable(''),
        isVisible: ko.observable(false),
        isLoaded: ko.observable(false),

        openPreviewImage: openPreviewImage,
        closePreview: closePreview,
        activate: activate
    };

    return imagePreviewModel;

    function activate() {
        
    }

    function openPreviewImage(data) {
        showOverlay();
        setPreviewImage(data);
    }

    function setPreviewImage(data) {
        imagePreviewModel.imageId(data.id);
        imagePreviewModel.imageUrl(data.image);
    }

    function showOverlay() {
        imagePreviewModel.isVisible(true);
    }

    function closePreview() {
        hideOverlay();
    }

    function hideOverlay() {
        imagePreviewModel.isVisible(false);
        app.trigger(constants.events.imagePreviewClosed, imagePreviewModel.imageId());
    }
});
define('viewmodels/questions/singleSelectImage/bindingHandlers/previewImageBinding',['durandal/app', 'constants', 'knockout', 'durandal/composition', 'widgets/imagePreview/viewmodel'],
    function (app, constants, ko, composition, imagePreview) {
        app.on(constants.events.imagePreviewClosed, function (id) {
            $('#' + id).focus();
        });

        ko.bindingHandlers.previewImage = {
            init: function (element, valueAccessor) {
                var $element = $(element),
                    $data = valueAccessor();

                if ($data) {
                    $element.on('click', function () {
                        imagePreview.openPreviewImage($data);
                    }).on('keydown', function (event) {
                        if (event.which == 13) {
                            event.preventDefault();
                            event.stopPropagation();
                            $element.click();
                        }
                    });
                }
            }
        };

        composition.addBindingHandler('previewImage');
    });
define('widgets/imagePreview/bindingHandlers/fadeBindingHandler',['durandal/composition'], function (composition) {
    ko.bindingHandlers.fade = {
        init: function (element, valueAccessor) {
            var $element = $(element),
                data = valueAccessor() || {},
                value = ko.utils.unwrapObservable(data);

            if (value) {
                $element.stop().show();
            } else {
                $element.stop().hide();
            }
        },
        update: function (element, valueAccessor) {
            var $element = $(element),
                data = valueAccessor() || {},
                value = ko.utils.unwrapObservable(data);

            if (value) {
                $element.stop().fadeIn();
            } else {
                $element.stop().fadeOut();
            }
        }
    }

    composition.addBindingHandler('fade');
})
;
define('widgets/imagePreview/bindingHandlers/imageLoaderBindingHandler',['durandal/composition'], function (composition) {
    var image = new Image();
    image.className = 'image';
    image.style.display = 'none';
    image.style.width = 'auto'; //Fix for IE
    image.style.height = 'auto';

    ko.bindingHandlers.imageLoader = {
        init: function (element, valueAccessor) {

            var $element = $(element),
                data = valueAccessor() || {},
                isLoaded = data.isLoaded;

            var resizeHandler = $(window).on('resize', function () {
                if (isLoaded()) {
                    updatePreviewImageSize($element);
                }
            });

            var orientationChangeHandler = $(window).on("orientationchange", function () {
                if (isLoaded()) {
                    updatePreviewImageSize($element);
                }
            });

            ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
                $(window).unbind('resize', resizeHandler);
                $(window).unbind('orientationchange', orientationChangeHandler);
            });

            $element.append(image);
        },
        update: function (element, valueAccessor) {

            var $element = $(element),
                data = valueAccessor() || {},
                isLoaded = data.isLoaded,
                imageUrl = data.imageUrl();
            $(image).hide();
            isLoaded(false);

            var browserWidth = Math.max(document.documentElement.clientWidth, window.innerWidth || 0);
            var browserHeight = Math.max(document.documentElement.clientHeight, window.innerHeight || 0);
            var maxSize = browserWidth > browserHeight ? browserWidth : browserHeight;

            var resizedImageUrl = imageUrl + '?height=' + maxSize + '&width=' + maxSize;
            image.onload = function () {
                if (data.imageUrl() === imageUrl) {
                    updatePreviewImageSize($element);
                    isLoaded(true);
                    $(image).fadeIn();
                }
            }
            image.src = resizedImageUrl;
        }
    }

    function updatePreviewImageSize($element) {
        var browserWidth = Math.max(document.documentElement.clientWidth, window.innerWidth || 0) - 46; // 46 - padding for close button
        var browserHeight = Math.max(document.documentElement.clientHeight, window.innerHeight || 0) - 46;

        $('.image', $element).css('maxWidth', browserWidth - 46 + 'px').css('maxHeight', browserHeight - 46 + 'px');
    }

    composition.addBindingHandler('imageLoader');
});

define('dialogs/bindingHandlers/blockBodyScrollBinding',['knockout'], function (ko) {
    'use strict';
    
    ko.bindingHandlers.blockBodyScroll = {
        update: function (element, valueAccessor) {
            var isBlocked = ko.utils.unwrapObservable(valueAccessor()),
                $body = $('body');
            
            $body.css({
                'max-height': isBlocked ? '100vh' : 'none',
                'overflow': isBlocked ? 'hidden' : 'visible'
            });
        }
    };
});
var css = {
        ratingContainer: 'rating-container',
        ratingItem: 'rating-item',
        ratingItemBtn: 'rating-item-btn',
        ratingTipContainer: 'rating-tip-container',
        checked: 'checked',
        highlighted: 'highlighted'
    },
    keys = {
        rating: 'rating'
    };

ko.bindingHandlers.rating = {
    init: function(element, valueAccessors) {
        var $element = $(element),
            value = ko.utils.unwrapObservable(valueAccessors().value) || 0,
            max = ko.utils.unwrapObservable(valueAccessors().max) || 5,
            onValueUpdated = valueAccessors().onValueUpdated;

        var $container = $('<ul>');
        $container.addClass(css.ratingContainer)
            .appendTo($element)
            .mouseleave(function() {
                $container.children().each(function(index, item) {
                    $(item).removeClass(css.highlighted);
                });
            });

        for (var i = 0; i < max; i++) {
            var $li = $('<li>');
            $li.appendTo($container);
            $li.addClass(css.ratingItem)

            var $item = $('<button>');
            $item.addClass(css.ratingItemBtn)
                .data(keys.rating, i + 1)
                .text(i + 1)
                .click(function() {
                    var rating = $(this).data(keys.rating);
                    valueAccessors().value(rating);
                    applyRatingStyles(rating, css.checked);
                    if (_.isFunction(onValueUpdated)) {
                        onValueUpdated();
                    }
                })
                .hover(function() {
                    var rating = $(this).data(keys.rating);
                    applyRatingStyles(rating, css.highlighted);
                })
                .appendTo($li);
        }

        applyRatingStyles(value, css.checked);

        function applyRatingStyles(rating, className) {
            $container.children().each(function(index, item) {
                $(item).find('.' + css.ratingItemBtn).toggleClass(className, index < rating);
            });
        }
    }
};
define("dialogs/bindingHandlers/ratingBindingHandler", function(){});

define('bootstrapping/knockoutBindingTask',[
    'bootstrapping/knockoutBindings/areaBinding',
    'bootstrapping/knockoutBindings/autofocusBinding',
    'bootstrapping/knockoutBindings/autosizeBinding',
    'bootstrapping/knockoutBindings/backgroundBinding',
    'bootstrapping/knockoutBindings/circleProgressBinding',
    'bootstrapping/knockoutBindings/contextBinding',
    'bootstrapping/knockoutBindings/elementsWrapBinding',
    'bootstrapping/knockoutBindings/focusHandlerBinding',
    'bootstrapping/knockoutBindings/toggleVisibleBindingHandler',
    'bootstrapping/knockoutBindings/dropDownBinding',
    'bootstrapping/knockoutBindings/scrollTopBinding',
    'bootstrapping/knockoutBindings/scrollIntoViewBinding',
    'bootstrapping/knockoutBindings/contentBlockBinding',
    'bootstrapping/knockoutBindings/toolTipBindingHandler',
    'bootstrapping/knockoutBindings/downloadAsPdfBinding',
    'bootstrapping/knockoutBindings/mediaPlayerBinding',
    'bootstrapping/knockoutBindings/customFocusBinding',
    'bootstrapping/knockoutBindings/reviewableBinding',
    'bootstrapping/knockoutBindings/limitedLengthTextBinding',
    'bootstrapping/knockoutBindings/scrollbarBinding',
    'bootstrapping/knockoutBindings/localizeBinding',
    'bootstrapping/knockoutBindings/fontScaleBinding',
    'treeOfContent/bindingHandlers/treeOfContentAnimationsBinding',
    'bootstrapping/knockoutBindings/selectTextBinding',
    'bootstrapping/knockoutBindings/submitOnEnterBinding',
    'bootstrapping/knockoutBindings/keyDownBinding',
    'bootstrapping/knockoutBindings/tabFocusBinding',
    'bootstrapping/knockoutBindings/copyToClipboardBinding',
    'bootstrapping/knockoutBindings/changeCssClassByDelayBinding',
    'bootstrapping/knockoutBindings/containImageBinding',
    'bootstrapping/knockoutBindings/lessColorBinding',
    'bootstrapping/knockoutBindings/editableTextBindingHandler',
    'bootstrapping/knockoutBindings/headerHeightDepender',
    'viewmodels/questions/dragAndDrop/bindingHandlers/draggableTextBindingHandler',
    'viewmodels/questions/dragAndDrop/bindingHandlers/draggableTextContainerBindingHandler',
    'viewmodels/questions/dragAndDrop/bindingHandlers/dropspotBindingHandler',
    'viewmodels/questions/rankingText/bindingHandlers/preventDefaultTouchEventBinding',
    'viewmodels/questions/fillInTheBlank/bindingHandlers/fillInTheBlankAnswersBinding',
    'viewmodels/questions/textMatching/bindingHandlers/draggableBindingHandler',
    'viewmodels/questions/textMatching/bindingHandlers/droppableBindingHandler',
    'viewmodels/questions/textMatching/bindingHandlers/overlayBindingHandler',
    'viewmodels/questions/textMatching/bindingHandlers/syncHeightBindingHandler',
    'viewmodels/questions/openQuestion/bindingHandlers/textAreaBindingHandler',
    'viewmodels/questions/singleSelectImage/bindingHandlers/previewImageBinding',
    'widgets/imagePreview/bindingHandlers/fadeBindingHandler',
    'widgets/imagePreview/bindingHandlers/imageLoaderBindingHandler',
    'dialogs/bindingHandlers/blockBodyScrollBinding',
    'dialogs/bindingHandlers/ratingBindingHandler'
], function () {

    return {
        execute: function () { }
    };

});
define('bootstrapping/binders/background',['templateSettings', 'displaySettings'], function (templateSettings, displaySettings) {
    'use strict';

    ko.bindingHandlers.secondaryBackground = {
        init: function (element) {
            var body = templateSettings && templateSettings.background && templateSettings.background.body;

            if (body) {
                if (body.texture && body.texture.length) {
                    applyImage(element, body.texture, "repeat");
                }

                if (body.color && body.color.length) {
                    applyColor(element, body.color);
                }

                if (body.brightness) {
                    applyBrightness($('<div />').width('100%').height('100%').appendTo(element), body.brightness);
                }
            }
        }
    };

    return {
        apply: applyOnce()
    };

    function applyOnce() {
        var executed = false;

        return function () {
            if (!executed) {
                executed = true;
                apply(templateSettings.background);
            }
        }
    }

    function apply(background) {

        var elementsClasses = {
            body: {
                element: "body",
                brightness: ".body-layout-wrapper"
            }
        };

        if (background.body.texture && background.body.texture.length) {
            applyImage(elementsClasses.body.element, background.body.texture);
        };

        if (background.body.color && background.body.color.length) {
            applyColor(elementsClasses.body.element, background.body.color);
        };

        if (background.body.brightness) {
            applyBrightness(elementsClasses.body.brightness, background.body.brightness);
        };
    }

    function applyBrightness(element, brightness) {
        var $element = $(element);

        if (brightness === 0) {
            return;
        }

        $element.css({
            "background-color": brightness > 0 ? 'white' : 'black',
            "opacity": brightness > 0 ? brightness : -brightness
        });
    }

    function applyImage(element, url, type) {
        type = type || "repeat";

        var $element = $(element),
            image = new Image(),
            position = displaySettings.backgroundSettings[type].position || '0 0',
            repeat = displaySettings.backgroundSettings[type].repeat || 'no-repeat',
            size = displaySettings.backgroundSettings[type].size || 'auto';

        image.onload = function () {
            $element.css({
                'top': '0',
                'bottom': '0',
                'background-image': 'url(' + url + ')',
                'background-position': position,
                '-webkit-background-size': size,
                'background-size': size,
                'background-repeat': repeat
            })
        };
        image.src = url;
    }

    function applyColor(element, color) {
        $(element).css({
            'background-color': color
        });
    }

});
define('bootstrapping/binderTask',['durandal/binder', './binders/background'], function (binder, background) {

    return {
        execute: function () {
            binder.binding = function (obj, view) {
                TranslationPlugin.localize(view);
                background.apply();
            }
        }
    };

});
define('routing/routerExtender',['plugins/router'], function (router) {
    "use strict";

    function execute() {

        router.isNavigationLocked = ko.observable(false);

        router.activeInstruction.subscribe(function (instruction) {
            if (_.isObject(instruction)
                && !_.isNullOrUndefined(instruction.queryParams)
                && !_.isNullOrUndefined(instruction.queryParams.lock)) {

                var lock = instruction.queryParams.lock.toLowerCase() == "true";
                router.isNavigationLocked(lock);

                var activationData = instruction.params;
                _.each(activationData, function (item) {
                    if (_.isObject(item) && !_.isNullOrUndefined(item.lock)) {
                        delete item['lock'];
                    }

                    if (_.isEmpty(item)) {
                        var i = _.indexOf(activationData, item);
                        activationData.splice(i, 1);
                    }
                });

                instruction.params = activationData;
            }
        });

        router.getQueryStringValue = function (key) {
            var urlParams = window.location.href;
            var regex = new RegExp("(?:\\|\?|&|&amp;)" + key + "=([^&#]*)");
            var results = regex.exec(urlParams);
            return results === null ? null : decodeURIComponent(results[1].replace(/\+/g, " "));
        };

    }

    return {
        execute: execute
    };
})
;
define('bootstrapping/routingTask',['routing/routerExtender'], function (extender) {
    "use strict";

    var task = {
        execute: execute
    };

    return task;

    function execute() {
        extender.execute();
    }

});
define('bootstrapping/reviewTask',['plugins/router', 'durandal/app', 'context', 'constants', 'templateSettings'], function (router, app, context, constants, templateSettings) {
    "use strict";

    var task = {
        execute: execute
    };

    return task;

    function execute() {
        if (!supportedBrowser.isBrowserSupported())
            return;

        var reviewApiUrl = router.getQueryStringValue('reviewApiUrl');
        var authoringToolDomain = router.getQueryStringValue('authoringToolDomain');
        if (!reviewApiUrl)
            return;

        postMessage({ supportsNativeReview: true });

        app.on(constants.events.appInitialized, function () {
            pluginsLocalizationService.init(templateSettings.languages.selected);

            var reviewPlugin = new ReviewPlugin();
            reviewPlugin.init({
                reviewApiUrl: decodeURIComponent(reviewApiUrl),
                authoringToolDomain: decodeURIComponent(authoringToolDomain),
                courseId: context.course.id
            });

            router.on('router:navigation:composition-complete').then(function () {
                reviewPlugin.renderSpots();
            });
        });

        function postMessage(data) {
            var editorWindow = window.parent;
            editorWindow.postMessage(data, '*');
        }
    }
});

define('bootstrapping/cssCustomBrowserFixTask',[],function () {
    "use strict";

    var task = {
        execute: execute
    };

    var mainFolder = 'css/';

    var browsers = {
        ie: {
            pattern: /MSIE|Trident\//i,
            css: 'ie.css'
        },
        edge: {
            pattern: /Edge/i,
            css: 'edge.css'
        },
    };

    return task;

    function execute() {
        for(var key in browsers) {
            var browser = browsers[key];

            if(browser.pattern.test(navigator.userAgent)) {
                addStyle(browser.css);
            }
        }
    }

    function addStyle(name) {
        var head  = document.getElementsByTagName('head')[0];
        var link  = document.createElement('link');

        link.rel  = 'stylesheet';
        link.type = 'text/css';
        link.href = mainFolder + name;
        link.media = 'all';

        head.appendChild(link);
    }

});
define('bootstrapper.tasks',['require','bootstrapping/browserSupportTask','bootstrapping/viewLocatorTask','bootstrapping/knockoutBindingTask','bootstrapping/binderTask','bootstrapping/routingTask','bootstrapping/reviewTask','bootstrapping/cssCustomBrowserFixTask'],function (require) {
    return {
        getCollection: function () {
            return [
                require('bootstrapping/browserSupportTask'),
                require('bootstrapping/viewLocatorTask'),
                require('bootstrapping/knockoutBindingTask'),
                require('bootstrapping/binderTask'),
                require('bootstrapping/routingTask'),
                require('bootstrapping/reviewTask'),
                require('bootstrapping/cssCustomBrowserFixTask')
            ];
        }
    };
});
define('bootstrapper',['durandal/system', 'bootstrapper.tasks'], function (system, tasks) {

    return {
        run: run
    };


    function run() {
        _.each(tasks.getCollection(), function (task) {
            if (_.isFunction(task.execute)) {
                task.execute();
            } else {
                system.log('Bootstrapper task ' + system.getModuleId(task) + 'is not executable');
            }
        });
     }
});
define('datajsInitializer',['requester/resourceLoader', 'q'], function (resourceLoader, Q) {
  const initialize = function () {
    const dfd = Q.defer();

    resourceLoader
      .getLocalResource({
        url: 'content/data.js',
        requestOptions: {
          contentType: 'application/json',
          dataType: 'json'
        }
      })
      .done(function (response) {
        resourceLoader.setCacheBuster(response.createdOn);
        dfd.resolve();
      })
      .fail(function () {
        dfd.reject('Unable to load data.js');
      });

    return dfd.promise;
  };

  return {
    initialize: initialize
  };
});

define('includedModules/moduleLoader',[],function () {

    return {
        loadModule: loadModule
    };

    function loadModule(modulePath) {
        var defer = Q.defer();
        require([modulePath],
            function (module) {
                defer.resolve(module);
            },
            function (error) {
                defer.reject({ modulePath: error.requireModules && error.requireModules[0], messsage: error.message });
            });

        return defer.promise;
    }
});

define('modules/progress/progressStorage/urlProvider',["q", "publishSettings"], function(Q, publishSettings) {
  "use strict";

  function UrlProvider() {
    this.authServiceUrl = "";
    this.learnServiceUrl = "";
    this.courseLink = "";
  }

  UrlProvider.prototype.initialize = initialize;

  var instance = new UrlProvider();

  return instance;

  function initialize() {
    var self = this;
    return Q.fcall(function() {
      self.authServiceUrl = publishSettings.authServiceUrl
        ? publishSettings.authServiceUrl
        : "https://auth-staging.easygenerator.com";
      self.learnServiceUrl = publishSettings.learnServiceUrl
        ? publishSettings.learnServiceUrl
        : "https://learn-staging.easygenerator.com";

      self.courseLink =
        window.location.protocol +
        "//" +
        window.location.host +
        window.location.pathname +
        "#/";
    });
  }
});

define('modules/progress/constants',[],function() {
  "use strict";

  return {
    progressKey: "course_progress",
    resultKey: "course_result",
    contentTypeJson: "application/json",
    learnerCoursesApi:  "/api/learner/me/courses/"
  };
});

define('helpers/idToUuid',[],function() {
  function isUUID(id) {
    var regex = new RegExp(
      /^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$/i
    );
    return regex.test(id);
  }

  return function IdToUuid(id) {
    if (isUUID(id)) {
      return id;
    }
    return (
      id.slice(0, 8) +
      "-" +
      id.slice(8, 12) +
      "-" +
      id.slice(12, 16) +
      "-" +
      id.slice(16, 20) +
      "-" +
      id.slice(20, 32)
    );
  };
});

define('modules/progress/progressStorage/auth',['requester/httpWrapper', './urlProvider', '../constants', 'helpers/idToUuid'], function (
  httpWrapper,
  urlProvider,
  constants,
  idToUuid
) {
  'use strict';

  var _private = {
    tokenKey: 'token.learn',
    shortTermAccessKey: 'shortTermAccess.auth.template',
    getToken: function () {
      return sessionStorage.getItem(_private.tokenKey);
    },
    removeToken: function () {
      sessionStorage.removeItem(_private.tokenKey);
    }
  };

  var auth = {};

  Object.defineProperty(auth, 'headers', {
    get: function () {
      return {};
    }
  });

  Object.defineProperty(auth, 'authenticated', {
    get: function () {
      return _private.getToken() !== null;
    }
  });

  Object.defineProperty(auth, 'shortTermAccess', {
    get: function () {
      return sessionStorage.getItem(_private.shortTermAccessKey) === 'true';
    },
    set: function (value) {
      if (typeof value !== 'undefined' && value !== '') {
        sessionStorage.setItem(_private.shortTermAccessKey, value);
      } else if (value === '') {
        sessionStorage.removeItem(_private.shortTermAccessKey);
      }
    }
  });

  auth.setToken = function (value) {
    if (typeof value !== 'undefined' && value !== null) {
      sessionStorage.setItem(_private.tokenKey, value);
    }
  };

  auth.getToken = function () {
    return _private.getToken();
  };

  auth.exists = function (email) {
    return httpWrapper
      .get({
        url: urlProvider.authServiceUrl + '/api/account/exists',
        query: { email: email },
        headers: {
          'Content-Type': constants.contentTypeJson
        },
        withCredentials: true
      })
      .then(function (response) {
        return response.exists;
      });
  };

  auth.authorize = function () {
    return httpWrapper
      .get({
        url: urlProvider.authServiceUrl + '/api/account/token',
        withCredentials: true
      })
      .then(function (response) {
        auth.setToken(response.token);
      })
      .fail(function (fail) {
        auth.setToken('');
        console.error(fail);
      });
  };

  auth.identify = function () {
    return httpWrapper
      .get({
        url: urlProvider.learnServiceUrl + '/api/learner/me',
        headers: {
          Authorization: 'Bearer ' + auth.getToken()
        },
        withCredentials: false
      })
      .fail(function (fail) {
        if (fail.status === 401) {
          return auth.signout();
        }
        console.error('User info request fails with error: ', fail);
      });
  };

  auth.signin = function (email, password, rememberMe) {
    return httpWrapper
      .post({
        url: urlProvider.authServiceUrl + '/api/auth/credentials',
        data: {
          email: email,
          password: password,
          shortTermAccess: !rememberMe
        },
        headers: {
          'Content-Type': constants.contentTypeJson
        },
        withCredentials: true
      })
      .then(function (response) {
        auth.setToken(response.token);
        auth.shortTermAccess = !rememberMe;
      });
  };

  auth.register = function (email, password, name, rememberMe) {
    return httpWrapper
      .post({
        url: urlProvider.authServiceUrl + '/api/credentials/register',
        data: {
          email: email,
          password: password,
          name: name,
          shortTermAccess: !rememberMe
        },
        headers: {
          'Content-Type': constants.contentTypeJson
        },
        withCredentials: true
      })
      .then(function (response) {
        auth.setToken(response.token);
        auth.shortTermAccess = !rememberMe;
      });
  };

  auth.getValueFromUrl = function (key) {
    key = key || 'token';
    var url = window.location.href,
      regex = new RegExp('[?&]' + key + '(=([^&#]*)|&|#|$)', 'i'),
      result = regex.exec(url);

    if (!result || !result[2]) {
      return null;
    }

    return decodeURIComponent(result[2].replace(/\+/g, ' '));
  };

  auth.signout = function () {
    return httpWrapper
      .post({
        url: urlProvider.authServiceUrl + '/api/account/logout',
        withCredentials: true
      })
      .then(function (response) {
        _private.removeToken();
        auth.shortTermAccess = '';
      })
      .fail(function (fail) {
        console.error('Logout failed with error: ', fail);
      });
  };

  auth.sendSecretLink = function (courseId) {
    return httpWrapper.post({
      url: urlProvider.learnServiceUrl + '/api/learner/me/courses/' + idToUuid(courseId) + '/send-secret-link',
      headers: {
        'Content-Type': constants.contentTypeJson,
        Authorization: 'Bearer ' + _private.getToken()
      },
      withCredentials: false
    });
  };

  auth.forgotpassword = function (email) {
    return httpWrapper.post({
      url: urlProvider.authServiceUrl + '/api/credentials/forgot-password',
      data: {
        email: email,
        returnUrl: urlProvider.courseLink
      },
      headers: {
        'Content-Type': constants.contentTypeJson
      }
    });
  };

  return auth;
});

define('helpers/loginHelper',['plugins/router', 'q', 'constants', 'modules/progress/progressStorage/auth', 'templateSettings'], function (router, Q, constants, auth, templateSettings) {

    function LoginHelper() {
        this.token = '';
        this.email = '';
        this.username = '';
    }

    var instance = new LoginHelper();

    LoginHelper.prototype.initialize = initialize.bind(instance);

    LoginHelper.prototype.getLoginMethod = getLoginMethod.bind(instance);

    return instance;

    function initialize() {
        var self = this;
        return Q.fcall(function() {
            self.token = auth.getValueFromUrl('token');
            self.email = router.getQueryStringValue('email');
            self.username = router.getQueryStringValue('name');
        });
    }

    function getLoginMethod() {
        if(this.token) {
            return constants.loginTypes.byToken;
        } else if (this.email && this.username && templateSettings.allowCrossDeviceSaving) {
            return constants.loginTypes.byEmailAndName;
        } else if (this.email && this.username && !templateSettings.allowCrossDeviceSaving) {
            return constants.loginTypes.byEmailAndNameWithoutCrossDeviceSaving;
        }

        return constants.loginTypes.default;
    }
});
define('userContext',['plugins/router', 'durandal/app', 'helpers/loginHelper'], function(router, app, loginHelper) {

    var _initialized = false;

    var context = {
        use: use,
        initialize: initialize,
        getCurrentUser: getCurrentUser,
        user: new UserContext(),
        clear: clear
    };
    
    return context;

    function UserContext() {
        this.account = null;
        this.email = '';
        this.username = '';
        this.keepMeLoggedIn = false;
    }

    function getCurrentUser() {
        return (context.user.email || context.user.account) && context.user.username ? context.user : null;
    }

    function use(userInfoProvider) {
        if(!userInfoProvider) {
            return;
        }
        var accountId = userInfoProvider.getAccountId(),
            accountHomePage = userInfoProvider.getAccountHomePage(),
            username = userInfoProvider.getUsername();
        if(!accountId || !accountHomePage || !username) {
            return;
        }
        context.user.email = accountId;
        context.user.username = username;
        context.user.account = {
            homePage: accountHomePage,
            name: accountId
        };
        _initialized = true;
    }

    function clear() {
        context.user = new UserContext();
    }

    function initialize() {
        return Q.fcall(function () {
            if(_initialized){
                return;
            }
            
            app.on('user:authenticated').then(authenticated);

            var username = loginHelper.username,
                email = loginHelper.email;

            if (username || email) {
                context.user.email = email ? email : '';
                context.user.username = username ? username : '';
            }

            _initialized = true;
        });
    }

    function authenticated(authenticatedUser) {
        context.user.email = authenticatedUser.email;
        context.user.username = authenticatedUser.username;
    }
});
define('progressContext',['durandal/system', 'durandal/app', 'plugins/router', 'eventManager', 'context', 'userContext', 'constants'],
    function (system, app, router, eventManager, dataContext, userContext, constants) {

        var
            statuses = constants.progressContext.statuses,
            self = {
                storage: null,
                progress: {
                    url: null,
                    answers: {},
                    user: null,
                    attemptId: system.guid(),
                    finished: false
                }
            },
            context = {
                get: get,
                remove: remove,
                finish: finish,

                use: use,
                ready: ready,
                restoreProgress: restoreProgress,

                status: ko.observable(statuses.undefined)
            };

        return context;

        function save() {
            if (!self.storage) {
                return;
            }

            var savedProgressResults = self.storage.saveProgress(self.progress);
            if (_.isObject(savedProgressResults) && _.isFunction(savedProgressResults.then)
                && _.isFunction(savedProgressResults.fail)) {
                savedProgressResults.then(function () {
                    context.status(statuses.saved);
                }).fail(function () {
                    context.status(statuses.error);
                });
            } else {
                context.status(savedProgressResults ? statuses.saved : statuses.error);
            }

            saveResults();
            return savedProgressResults;
        }

        function navigated(obj, instruction) {
            if (self.progress.finished) {
                return;
            }

            if (instruction.config.moduleId === 'xApi/viewmodels/xAPIError' ||
                instruction.config.moduleId === 'viewmodels/404' ||
                instruction.config.moduleId === 'account/account' ||
                instruction.config.moduleId === 'account/signin/index' ||
                instruction.config.moduleId === 'account/register/index' ||
                instruction.config.moduleId === 'account/limitAccess/index') {
                return;
            }
            if (_.isNull(self.progress.url)) {
                self.progress.url = instruction.fragment;
                save();
            } else if (self.progress.url != instruction.fragment) {
                self.progress.url = instruction.fragment;
                save();
            }
        }

      function authenticated(user) {
            delete user.password;
            self.progress.user = user;
        }

        function authenticationSkipped() {
            self.progress.user = 0;
        }

        function questionAnswered(question) {
            try {
                self.progress.answers[question.shortId] = question.progress();
                save();
            } catch (e) {
                console.error(e);
            }
        }

        function finish(callback) {
            callback = callback || function () { };

            if (self.progress) {
                self.progress.finished = true;
                var promise = save();
                if (_.isObject(promise) && _.isFunction(promise.then)) {
                    promise.then(callback);
                    return;
                }
            }

            callback();
        }

        function get() {
            return self.progress;
        }

        function remove(callback) {
            callback = callback || function () { };
            if (!self.storage) {
                callback();
                return;
            }
            saveResults();
            if (_.isFunction(self.storage.removeProgress)) {
                var promise = self.storage.removeProgress();
                if (_.isObject(promise) && _.isFunction(promise.then)) {
                    promise.then(callback);
                } else {
                    callback();
                }
            } else {
                callback();
            }
        }

        function saveResults() {
            if (_.isFunction(self.storage.saveResults)) {
                self.storage.saveResults();
            }
        }

        function use(storage) {
            if (_.isFunction(storage.getProgress) && _.isFunction(storage.saveProgress)) {
                clear();

                self.storage = storage;

                restore(userContext.getCurrentUser());

                eventManager.subscribeForEvent(eventManager.events.answersSubmitted).then(questionAnswered);

                app.on('user:authenticated').then(authenticated);
                app.on('user:authentication-skipped').then(authenticationSkipped);

                router.on('router:route:activating', navigated);

                window.onbeforeunload = function () {
                    if (context.status() === statuses.error) {
                        return TranslationPlugin.getTextByKey('[progress cannot be saved]');
                    }
                }
            } else {
                throw 'Cannot use this storage';
            }
        }

        function ready() {
            return !!self.storage;
        }

        function clear() {
            self = {
                storage: null,
                progress: {
                    url: null,
                    answers: {},
                    user: null,
                    attemptId: system.guid(),
                    finished: false
                }
            }
        }

        function restore(user) {
            var progress = self.storage.getProgress();
            var isProgressNotEmpty = !_.isEmpty(progress);
            var isProgressString = isProgressNotEmpty && _.isString(progress.attemptId);
            var isSuitableProgress = isProgressString && (!user || !progress.user || user.email === progress.user.email);

            if (isSuitableProgress) {
                self.progress = progress;
            }
        }

        function restoreProgress() {
            dataContext.initialize();
            var progress = null;

            if (context.ready()) {
                progress = context.get();
                if (_.isObject(progress)) {
                    if (progress.finished) {
                        dataContext.course.finish(null, true);
                    }
                    if (_.isObject(progress.answers)) {
                        _.each(dataContext.course.sections, function (section) {
                            _.each(section.questions, function (question) {
                                if (!_.isNullOrUndefined(progress.answers[question.shortId])) {
                                    question.progress(progress.answers[question.shortId]);
                                }
                                if (progress.finished) {
                                    question.markAsAnswered();
                                }
                            });
                        });
                    }
                }
                window.location.hash = !_.isString(progress.url) ? '' : progress.url.replace('objective', 'section'); //fix for old links
            }
        }
    });
define('errorTracking/ravenWrapper',[],function() {
  function RavenWrapper() {
    var self = this;

    self.raven = window.Raven;

    self.initialize = function(
      errorTrackingServiceUrl,
      email
    ) {
      if (!self.raven) {
        return;
      }

      self.raven.config(errorTrackingServiceUrl, {
        logger: 'javascript',
        ignoreErrors: [
          /LMSSetValue/,
          /LMSInitialize/,
          /LMSCommit/
        ],
        ignoreUrls: [
          /vendor\.min\.js/
        ],
        autoBreadcrumbs: {
          xhr: true,
          console: false,
          dom: true,
          location: true
        }
      }).install();

      self.raven.setUserContext({ email: email });
    };

    self.captureError = function(error) {
      if (!self.raven || !error) {
        return;
      }

      self.raven.captureMessage(error, { level: 'error' });
    };
  }
  return new RavenWrapper();
});

define('errorTracking/errorTracker',["underscore", "errorTracking/ravenWrapper", "durandal/app"], function(
  _,
  ravenWrapper,
  app
) {
  function ErrorTracker() {
    return {
      init: init,
      trackError: trackError
    };

    function init(errorTrackingServiceUrl) {
      if (!errorTrackingServiceUrl) {
        return;
      }

      app
        .on("user:authenticated")
        .then(function (userInfo) {
          ravenWrapper.initialize(errorTrackingServiceUrl, userInfo.email)
        });
    }

    function trackError(error) {
      if(!error) {
        return;
      }

      ravenWrapper.captureError(error);
    }
  }
  return new ErrorTracker();
});

define('includedModules/appExtender',['eventManager', 'progressContext', 'userContext', 'errorTracking/errorTracker'],
    function (eventManager, progressContext, userContext, errorTracker) {
        'use strict';

        return {
            extend: extend
        };

        function extend(module) {
            if (_.isFunction(module.courseFinished)) {
                eventManager.subscribeForEvent(eventManager.events.courseFinished).then(module.courseFinished);
            }
            if (_.isFunction(module.courseFinalized)) {
                eventManager.subscribeForEvent(eventManager.events.courseFinalized).then(module.courseFinalized);
            }
            if (_.isFunction(module.addErrorHandler)) {
                module.addErrorHandler(errorTracker.trackError);
            }
            if (_.isObject(module.userInfoProvider)) {
                userContext.use(module.userInfoProvider);
            }
            if (_.isObject(module.progressProvider)) {
                progressContext.use(module.progressProvider);
            }
        }
    });
define('includedModules/modulesInitializer',['underscore', 'q', './moduleLoader', './appExtender'],
    function (_, Q, moduleLoader, appExtender) {
        'use strict';

        var _loadedModules = [];

        return {
            hasModule: hasModule,
            load: load
        };

        function hasModule(moduleName) {
            moduleName = moduleName || '';
            return _.contains(_loadedModules, moduleName);
        }

        function load(modules) {
            var modulesIds = _.map(modules, function(module){
                return module.name;
            });

            var defer = Q.defer();


            loadModulesSequentially();

            return defer.promise;

            function loadModulesSequentially() {
                if (modulesIds.length == 0) {
                    defer.resolve();
                    return;
                }


                var module = modulesIds.shift();
                moduleLoader.loadModule('../includedModules/' + module)
                    .then(initializeModule)
                    .then(extendAppFromModule)
                    .then(function () {
                        _loadedModules.push(module);
                        loadModulesSequentially();
                    })
                    .fail(onModuleLoadingFailed);
            }
        }

        function initializeModule(module) {
            return Q.fcall(function () {
                if (_.isFunction(module.initialize)) {
                    module.initialize();
                }
                return module;
            });
        }

        function extendAppFromModule(module) {
            appExtender.extend(module);
        }

        function onModuleLoadingFailed(error) {
            throw 'Cannot load module"' + error.modulePath + '". because of next error "' + error.message + '".';
        }
    }
);
define('modules/index',[],function(){
    'use strict';

    return {
        init: init
    };

    function init(templateSettings, manifest, publishSettings, customisations){
        var promises = [];

        window.WebFontLoader && promises.push(window.WebFontLoader.load(templateSettings.fonts, manifest, publishSettings));
        window.LessProcessor && promises.push(window.LessProcessor.load(templateSettings.colors, templateSettings.fonts));
        window.LessProcessor && customisations && customisations.styles && promises.push(window.LessProcessor.loadCustomStyles(customisations.styles));
        
        return Q.all(promises);
    }
});

define('modules/publishModeProvider',['includedModules/modulesInitializer', 'templateSettings'], function(modulesInitializer, templateSettings) {

    "use strict";

    var viewmodel = {
        init: init,
        isPreview: false,
        isScormEnabled: false,
        isReview: false
    };

    function init(publishMode) {
        if (publishMode === 'Lms' || publishMode === 'Lms2004') {
            viewmodel.isScormEnabled = modulesInitializer.hasModule('lms');
            templateSettings.allowCrossDeviceSaving = false;
        }
        if (publishMode === 'Review') {
            viewmodel.isReview = true;
        }
        if (publishMode === 'Preview') {
            viewmodel.isPreview = true;
        }

        if (viewmodel.isPreview || viewmodel.isReview) {
            templateSettings.allowCrossDeviceSaving = false;
            templateSettings.xApi.enabled = false;
            templateSettings.nps.enabled = false;
        }
    }

    return viewmodel;
});
requirejs.config({
  paths: {
    text: '../js/text',
    durandal: '../js/durandal',
    plugins: '../js/durandal/plugins',
    transitions: '../js/durandal/transitions',
    requester: './requester'
  }
});

define('jquery', function () {
  return jQuery;
});
define('knockout', function () {
  return ko;
});
define('WebFont', function () {
  return WebFont;
});
define('q', function () {
  return Q;
});
define('underscore', function () {
  return _;
});
define('perfectScrollbar', function () {
  return Ps;
});

define('less', function () {
  return less;
});

define('download', function () {
  return download;
});

define('main',[
  'durandal/app',
  'durandal/system',
  'underscore',
  'bootstrapper',
  'datajsInitializer',
  'templateSettings',
  'publishSettings',
  'includedModules/modulesInitializer',
  'modules/index',
  'modules/publishModeProvider',
  'errorTracking/errorTracker',
  'requester/resourceLoader'
], function (
  app,
  system,
  _,
  bootstrapper,
  dataJsInitializer,
  templateSettings,
  publishSettings,
  modulesInitializer,
  modulesLoader,
  publishModeProvider,
  errorTracker,
  resourceLoader
) {
  app.title = 'easygenerator';

  system.debug(false);

  app.configurePlugins({
    router: true,
    dialog: true,
    http: true,
    widget: true
  });

  app.start().then(function () {
    bootstrapper.run();
    dataJsInitializer.initialize().then(function () {
      return ConfigurationReader.read('', resourceLoader.getCacheBuster())
        .then(function (configsFiles) {
          var configs = ConfigurationReader.init(configsFiles);
          templateSettings.init(configs.templateSettings);
          TranslationPlugin.init(configs.translations);
          publishSettings.init(configsFiles.publishSettings);

          return modulesLoader
            .init(templateSettings, configsFiles.manifest, publishSettings, configsFiles.customisations)
            .then(function () {
              if (publishSettings.modules) {
                return modulesInitializer.load(publishSettings.modules).then(function () {
                  initializeApp(
                    publishSettings.publishMode,
                    publishSettings.errorTrackingServiceUrl,
                    templateSettings.errorTracking
                  );
                });
              } else {
                initializeApp(
                  publishSettings.publishMode,
                  publishSettings.errorTrackingServiceUrl,
                  templateSettings.errorTracking
                );
              }
            });
        })
        .catch(function (e) {
          console.error(e);
        });
    });
  });

  function initializeApp(publishMode, errorTrackingServiceUrl, errorTracking) {
    if (errorTrackingServiceUrl && errorTracking.enabled) {
      errorTracker.init(errorTrackingServiceUrl);
    }
    publishModeProvider.init(publishMode);
    app.setRoot('viewmodels/shell');
  }
});

define('account/header/index',[],function(){
	'use strict';

	var _instance = null;

	function Header(courseTitle){
		if (_instance) {
			return _instance;
		}
		this.courseTitle = courseTitle;
		this.isCourseTitleTooLong = this.courseTitle > 100;
	}

	return Header;
});

define('account/socialNetworks/index',[],function () {
    'use strict';

    var _instance = null;

    var _getAuthLink = function (courseTitle, authServiceUrl, socialNetwork) {
        const addTokenQueryParam =
          window.location.host.indexOf('easygenerator.com') === -1 ? '&addTokenQuery=true' : '';
        return authServiceUrl + '/api/auth/' + socialNetwork +
            '?callbackUrl=' + encodeURIComponent(window.location.href) + addTokenQueryParam;
    };

    function SocialNetworks(courseTitle, authServiceUrl) {
        if (_instance) {
			return _instance;
		}
        var getAuthLink = _getAuthLink.bind(this, courseTitle, authServiceUrl);
        this.facebookAuthLink = getAuthLink('facebook');
        this.linkedinAuthLink = getAuthLink('linkedin');
        this.googleAuthLink = getAuthLink('google');
    }

    return SocialNetworks;
});

define('account/helpers/validatedValue',['knockout'], function(ko) {
    'use strict';

    return function (validateCallback){
        var value = ko.observable('');
        value.trim = function() {
            value(ko.utils.unwrapObservable(value).trim());
        };
        value.isValid = ko.computed(validateCallback.bind(null, value));
        value.isModified = ko.observable(false);
        value.hasFocus = ko.observable(false);
        value.markAsModified = function() {
            value.isModified(true);
            return value;
        };
        return value; 
    }
});
define('xApi/constants',[],
    function () {

        var verbs = {
            started: {
                id: "http://adlnet.gov/expapi/verbs/launched",
                display: {
                    "en-US": "started"
                }
            },
            stopped: {
                id: "http://adlnet.gov/expapi/verbs/exited",
                display: {
                    "en-US": "stopped"
                }
            },
            passed: {
                id: "http://adlnet.gov/expapi/verbs/passed",
                display: {
                    "en-US": "passed"
                }
            },
            failed: {
                id: "http://adlnet.gov/expapi/verbs/failed",
                display: {
                    "en-US": "failed"
                }
            },
            experienced: {
                id: "http://adlnet.gov/expapi/verbs/experienced",
                display: {
                    "en-US": "experienced"
                }
            },
            answered: {
                id: "http://adlnet.gov/expapi/verbs/answered",
                display: {
                    "en-US": "answered"
                }
            },
            mastered: {
                id: "http://adlnet.gov/expapi/verbs/mastered",
                display: {
                    "en-US": "mastered"
                }
            },
            progressed: {
                id: "http://adlnet.gov/expapi/verbs/progressed",
                display: {
                    "en-US": "progressed"
                }
            },
            evaluated: {
                id: "http://www.tincanapi.co.uk/verbs/evaluated",
                display: {
                    "en-US": "evaluated"
                }
            }
        };

        var interactionTypes = {
            choice: "choice",
            fillIn: "fill-in",
            matching: "matching",
            dragAndDrop: "dragAndDrop",
            hotspot: "hotspot",
            sequencing: 'sequencing',
            other: "other"
        };

        var activityTypes = {
            course: "http://adlnet.gov/expapi/activities/course",
            objective: "http://adlnet.gov/expapi/activities/objective"
        }

        var extenstionKeys = {
            courseId: "http://easygenerator/expapi/course/id",
            surveyMode: "http://easygenerator/expapi/question/survey",
            questionType: "http://easygenerator/expapi/question/type",
            imageUrl: "http://easygenerator/expapi/question/imageUrl",
            content: "http://easygenerator/expapi/question/content",
            answerTexts: "http://easygenerator/expapi/question/answers",
        };

        var patterns = {
            email: /^[^@\s]+@[^@\s]+$/,
            isoDuration: /^PT[0-9]{1,2}H[0-9]{1,2}M[0-9]{1,2}S$/
        };

        return {
            verbs: verbs,
            interactionTypes: interactionTypes,
            activityTypes: activityTypes,
            extenstionKeys: extenstionKeys,
            patterns: patterns
        };
    }
);
define('account/limitAccess/accessLimiter',['underscore'], function (_) {

    "use strict";

    var accessLimitation = { enabled: false };

    function accessLimitationEnabled() {
        return accessLimitation.enabled;
    }
    function userHasAccess(user) {
        if (!accessLimitationEnabled())
            return true;

        return _.some(accessLimitation.allowedUsers, function(item) {
            return _.isString(item.email) && (item.email.trim().toLowerCase() === user.email.trim().toLowerCase());
        });
    }
    function initialize(accessLimitationSettings) {
        if (!accessLimitationSettings)
            return;

        accessLimitation = accessLimitationSettings;
    }

    return {
        userHasAccess: userHasAccess,
        accessLimitationEnabled: accessLimitationEnabled,
        initialize: initialize
    };
});
define('xApi/configuration/routes',[], function () {

    return [
        {
            route: 'xapierror(/:backUrl)',
            moduleId: 'xApi/viewmodels/xAPIError',
            title: 'xAPI Error'
        }
    ];

});
define('xApi/routingManager',['plugins/router', './configuration/routes'], function(router, routes) {
    'use strict';

    return {
        mapRoutes: function() {
            router.map(routes);
        }
    };
});
define('xApi/configuration/xApiSettings',['../constants', 'publishSettings'],
    function (constants, publishSettings) {

        var settings = {
            scoresDistribution: {
                positiveVerb: constants.verbs.passed
            },

            anonymousCredentials: {
                username: "",
                password: ""
            },

            xApi: {
                allowedVerbs: []
            },

            timeout: 120000,//2 minutes

            defaultLanguage: "en-US",
            xApiVersion: "1.0.0",

            initxApi: initxApi,
            initNps: initNps
        };


        var lrsHost = publishSettings.defaultLRSUrl || 'https://reports.easygenerator.com';

        var defaultXapi = {
            lrs: {
                uri: lrsHost + '/xApi/statements',
                authenticationRequired: false,
                credentials: {
                    username: '',
                    password: ''
                }
            },
            allowedVerbs: ['started', 'stopped', 'experienced', 'mastered', 'answered', 'passed', 'failed', 'progressed']
        };

        var npsHost = publishSettings.defaultNpsUrl || 'https://nps.easygenerator.com';

        var defaultNps = {
            nps: {
                uri: npsHost + '/xApi/statements',
                authenticationRequired: false,
                credentials: {
                    username: '',
                    password: ''
                }
            }
        };

        return settings;

        function initxApi(templateSettings) {
            $.extend(settings.xApi, templateSettings);

            if (templateSettings.selectedLrs == 'default') {
                $.extend(settings.xApi, defaultXapi);
            }
        }

        function initNps() {
            $.extend(settings.xApi, defaultNps);
        }

    }
);

define(
    'xApi/base64',[],function() {

        var base64 = {

            // private property
            _keyStr: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",

            // public method for encoding
            encode: function (input) {
                var output = "";
                var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
                var i = 0;

                input = base64._utf8_encode(input);

                while (i < input.length) {

                    chr1 = input.charCodeAt(i++);
                    chr2 = input.charCodeAt(i++);
                    chr3 = input.charCodeAt(i++);

                    enc1 = chr1 >> 2;
                    enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
                    enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
                    enc4 = chr3 & 63;

                    if (isNaN(chr2)) {
                        enc3 = enc4 = 64;
                    } else if (isNaN(chr3)) {
                        enc4 = 64;
                    }

                    output = output +
                    this._keyStr.charAt(enc1) + this._keyStr.charAt(enc2) +
                    this._keyStr.charAt(enc3) + this._keyStr.charAt(enc4);

                }

                return output;
            },

            // public method for decoding
            decode: function (input) {
                var output = "";
                var chr1, chr2, chr3;
                var enc1, enc2, enc3, enc4;
                var i = 0;

                input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");

                while (i < input.length) {

                    enc1 = this._keyStr.indexOf(input.charAt(i++));
                    enc2 = this._keyStr.indexOf(input.charAt(i++));
                    enc3 = this._keyStr.indexOf(input.charAt(i++));
                    enc4 = this._keyStr.indexOf(input.charAt(i++));

                    chr1 = (enc1 << 2) | (enc2 >> 4);
                    chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
                    chr3 = ((enc3 & 3) << 6) | enc4;

                    output = output + String.fromCharCode(chr1);

                    if (enc3 != 64) {
                        output = output + String.fromCharCode(chr2);
                    }
                    if (enc4 != 64) {
                        output = output + String.fromCharCode(chr3);
                    }

                }

                output = base64._utf8_decode(output);

                return output;

            },

            // private method for UTF-8 encoding
            _utf8_encode: function (string) {
                string = string.replace(/\r\n/g, "\n");
                var utftext = "";

                for (var n = 0; n < string.length; n++) {

                    var c = string.charCodeAt(n);

                    if (c < 128) {
                        utftext += String.fromCharCode(c);
                    }
                    else if ((c > 127) && (c < 2048)) {
                        utftext += String.fromCharCode((c >> 6) | 192);
                        utftext += String.fromCharCode((c & 63) | 128);
                    }
                    else {
                        utftext += String.fromCharCode((c >> 12) | 224);
                        utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                        utftext += String.fromCharCode((c & 63) | 128);
                    }

                }

                return utftext;
            },

            // private method for UTF-8 decoding
            _utf8_decode: function (utftext) {
                var string = "";
                var i = 0;
                var c = c1 = c2 = 0;

                while (i < utftext.length) {

                    c = utftext.charCodeAt(i);

                    if (c < 128) {
                        string += String.fromCharCode(c);
                        i++;
                    }
                    else if ((c > 191) && (c < 224)) {
                        c2 = utftext.charCodeAt(i + 1);
                        string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                        i += 2;
                    }
                    else {
                        c2 = utftext.charCodeAt(i + 1);
                        c3 = utftext.charCodeAt(i + 2);
                        string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                        i += 3;
                    }

                }

                return string;
            }

        };

        return base64;

    }
);
define('xApi/errorsHandler',['durandal/app', 'constants'],
    function (app, constants) {
        "use strict";

        var
            errorMessages = {
                invalidEndpoint: "Invalid endpoint",
                notFoundEndpoint: "Not found endpoint",
                invalidCredentials: "Invalid credentials",
                invalidEmail: "Invalid e-mail",
                invalidProtocol: "Invalid protocol",
                xDomainRequestError: "XDomainRequest error",
                timeoutError: "Timeout error",

                badRequest: "Bad request: ",
                unhandledMessage: "Unhandled error: ",

                verbIsIncorrect: "Vebr object is not well formed",
                actorDataIsIncorrect: "Actor data is incorrect",

                notEnoughDataInSettings: "Request failed: Not enough data in the settings"
            },

            handleError = function (message) {
                if (window.location.hash.indexOf('xapierror') !== -1) {
                    return;
                }

                var hash = window.location.hash.slice(1, window.location.hash.length);
                var navigateUrl = 'xapierror/' + encodeURIComponent(_.isEmpty(hash) ? '' : hash);

                app.trigger(constants.events.onError, { navigateUrl: navigateUrl });
            };

        return {
            errors: errorMessages,
            handleError: handleError
        };
    }
);
define('helpers/requestResender',["q", "jquery"], function(Q, $) {
  var constants = {
    sendResultAttemptsTimeout: 10000,
    sendResultAttemptsCount: 10,
    serverProblemStatus: 500,
    noInternetConnectionStatus: 0
  };

  function send(requestOptions, failedCallback, defer, attemptNumber) {
    if (typeof defer === "undefined") {
      defer = Q.defer();
    }

    if (typeof attemptNumber === "undefined") {
      attemptNumber = 0;
    }

    $.ajax(requestOptions)
      .done(function() {
        defer.resolve();
      })
      .fail(function(response, textStatus, error) {
        if (
          attemptNumber >= constants.sendResultAttemptsCount ||
          (response.status < constants.serverProblemStatus &&
            response.status !== constants.noInternetConnectionStatus)
        ) {
          return defer.reject(
            failedCallback && failedCallback(response, textStatus, error)
          );
        }

        setTimeout(function() {
          send(requestOptions, failedCallback, defer, ++attemptNumber);
        }, constants.sendResultAttemptsTimeout);
      });

    return defer.promise;
  }

  return {
    send: send
  };
});

define('xApi/requestManager',['./configuration/xApiSettings', './base64', './errorsHandler', 'helpers/requestResender'],
    function (xApiSettings, base64, errorsHandler, requestResender) {

        var states = {
            notInitialized: 0,
            initialized: 1,
            initializationFailed: 2
        };

        var state = states.notInitialized;

        var eventManager = {
            init: init,
            sendStatement: sendStatement
        };
        return eventManager;

        function init() {
            var dfd = Q.defer();
            switch (state) {
                case states.notInitialized:
                    var initPromise = Q.fcall(function () {
                        try {
                            initXDomainRequestTransport();
                            state = states.initialized;
                        } catch (e) {
                            state = states.initializationFailed;
                            throw e;
                        };
                    });

                    dfd.resolve(initPromise);
                    break;
                case states.initialized:
                    dfd.resolve();
                    break;
                case states.initializationFailed:
                    dfd.reject();
                    break;
            }

            return dfd.promise;
        }

        function sendStatement(statement, uri, username, password) {
            var requestOptions = createRequestOptions(statement, uri, username, password);

            if (!$.support.cors) {
                requestOptions = getOptionsForIEMode(requestOptions);
            }

            return requestResender.send(requestOptions, onError);
        }

        function onError(request, textStatus, error) {
            return getErrorMessage(request, textStatus, error);
        }

        function initXDomainRequestTransport() {
            if (window.XDomainRequest) {
                jQuery.ajaxTransport(function (s) {

                    var xdr;

                    return {

                        send: function (headers, complete) {

                            function callback(status, statusText, responses, responseHeaders) {
                                xdr.onload = xdr.onerror = xdr.ontimeout = jQuery.noop;
                                xdr = undefined;
                                complete(status, statusText, responses, responseHeaders);
                            }

                            xdr = new window.XDomainRequest();

                            xdr.onload = function () {
                                callback(200, "OK", { text: xdr.responseText });
                            };

                            xdr.onerror = function () {
                                callback(-1, errorsHandler.errors.xDomainRequestError);
                            };

                            if (s.timeout) {
                                xdr.timeout = s.timeout;
                                xdr.ontimeout = function () {
                                    callback(-1, errorsHandler.errors.timeoutError);
                                };
                            }

                            try {
                                xdr.open(s.type, s.url, true);
                            }
                            catch (e) {
                                var errorMessage;

                                if (location.protocol != s.url.split("/")[0]) {
                                    errorMessage = errorsHandler.errors.invalidProtocol;
                                } else {
                                    errorMessage = e.message;
                                }

                                callback(-1, errorMessage);
                                return;
                            }

                            xdr.send((s.hasContent && s.data) || null);
                        },

                        abort: function () {
                            if (xdr) {
                                xdr.onerror = jQuery.noop();
                                xdr.abort();
                            }
                        }
                    };

                });
            }
        }

        function getOptionsForIEMode(options) {
            var newUrl = options.url;

            //Everything that was on query string goes into form vars
            var formData = new Array();
            var qsIndex = newUrl.indexOf('?');
            if (qsIndex > 0) {
                formData.push(newUrl.substr(qsIndex + 1));
                newUrl = newUrl.substr(0, qsIndex);
            }

            //Method has to go on querystring, and nothing else
            options.url = newUrl + '?method=' + options.type;

            //Headers
            if (!_.isUndefined(options.headers) && _.isArray(options.headers)) {
                for (var headerName in options.headers) {
                    formData.push(headerName + "=" + encodeURIComponent(options.headers[headerName]));
                }
            }

            options.headers = {};

            //The original data is repackaged as "content" form var
            if (!_.isUndefined(options.data)) {
                formData.push('content=' + encodeURIComponent(options.data));
            }

            options.data = formData.join("&");
            options.type = "POST";

            return options;
        }

        function createRequestOptions(statement, url, userName, password) {
            var headers = [];
            headers["X-Experience-API-Version"] = xApiSettings.xApiVersion;
            headers["Content-Type"] = "application/json";
            var auth = "Basic " + base64.encode(userName + ':' + password);
            headers["Authorization"] = auth;

            var options = {};

            options.url = url;
            options.data = JSON.stringify(statement);
            options.type = 'POST';
            options.headers = headers;
            options.timeout = xApiSettings.timeout;
            options.contentType = 'application/json';
            options.async = true;

            options.beforeSend = function (xmlHttpRequest) {
                if (!_.isUndefined(options.headers) && _.isArray(options.headers)) {
                    for (var headerName in options.headers)
                        xmlHttpRequest.setRequestHeader(headerName, options.headers[headerName]);
                }
            };

            return options;
        }

        function getErrorMessage(request, textStatus, error) {
            switch (request.status) {
                case 0:
                    error = errorsHandler.errors.invalidEndpoint;
                    break;
                case 400:
                    if (request.responseText.indexOf("Mbox") !== -1) {
                        error = errorsHandler.errors.invalidEmail;
                    } else if (request.responseText.indexOf("URL") !== -1 || request.responseText.indexOf("endpoint") !== -1) {
                        error = errorsHandler.errors.invalidEndpoint;
                    } else {
                        error = errorsHandler.errors.badRequest + request.responseText;
                    }

                    break;
                case 401:
                    error = errorsHandler.errors.invalidCredentials;
                    break;
                case 404:
                    error = errorsHandler.errors.notFoundEndpoint;
                    break;
                default:
                    error = errorsHandler.errors.unhandledMessage + request.statusText;
                    break;
            }

            return error;
        }
    }
);

define('xApi/guard',['./constants'],
    function (constants) {

        var

            throwIfNotISODuration = function (duration, message) {
                if (!_.isString(duration)) {
                    throw message;
                }

                if (!constants.patterns.isoDuration.test(duration)) {
                    throw message;
                }
            },

            throwIfNotMbox = function (mbox, message) {
                if (!_.isString(mbox)) {
                    throw message;
                }

                if (mbox.indexOf('mailto:') == -1) {
                    throw message;
                }

                try {
                    throwIfNotEmail(mbox.split('mailto:')[1]);
                } catch (e) {
                    throw message;
                }
            },

            throwIfNotEmail = function (email, message) {
                if (!_.isString(email)) {
                    throw message;
                }

                if (!constants.patterns.email.test(email)) {
                    throw message;
                }
            },

            throwIfNotLanguageMap = function (display, message) {
                if (!_.isObject(display)) {
                    throw message;
                }

                for (var locale in display) {
                    if (!_.isString(display[locale])) {
                        throw message;
                    }
                }
            },

            throwIfNotVerbId = function (id, message) {
                if (!_.isString(id)) {
                    throw message;
                }

                var isVerIdValid = false;
                for (var verb in constants.verbs) {
                    if (constants.verbs[verb].id === id) {
                        isVerIdValid = true;
                        break;
                    }
                }

                if (!isVerIdValid) {
                    throw message;
                }
            },

            throwIfNotString = function (text, message) {
                if (!_.isString(text)) {
                    throw message;
                }
            },

            throwIfNotAnObject = function (item, message) {
                if (!_.isObject(item)) {
                    throw message;
                }
            },

            throwIfNotNumber = function (item, message) {
                if (!_.isNumber(item)) {
                    throw message;
                }
            };

        return {
            throwIfNotMbox: throwIfNotMbox,
            throwIfNotEmail: throwIfNotEmail,
            throwIfNotLanguageMap: throwIfNotLanguageMap,
            throwIfNotVerbId: throwIfNotVerbId,
            throwIfNotString: throwIfNotString,
            throwIfNotAnObject: throwIfNotAnObject,
            throwIfNotNumber: throwIfNotNumber,
            throwIfNotISODuration: throwIfNotISODuration
        };
    }
);
define('xApi/models/statement',['../guard', 'durandal/system'],
    function (guard, system) {

        var statement = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create Statement';
            }

            guard.throwIfNotAnObject(spec.actor, 'You should provide actor for Statement');
            guard.throwIfNotAnObject(spec.verb, 'You should provide verb for Statement');
            guard.throwIfNotAnObject(spec.object, 'You should provide object for Statement');

            var obj = {};
            obj.id = system.guid();
            obj.actor = spec.actor;
            obj.verb = spec.verb;
            obj.object = spec.object;
            obj.timestamp = (new Date()).toISOString();

            if (_.isObject(spec.context)) {
                obj.context = spec.context;
            }

            if (_.isObject(spec.result)) {
                obj.result = spec.result;
            }

            return obj;
        };

        return statement;
    }
);
define('xApi/models/object',['../guard'],
    function (guard) {
        
        var object = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create Object';
            }

            guard.throwIfNotString(spec.objectType, 'You should provide object type');

            return {
                objectType: spec.objectType
            };
        };

        return object;

    }
);
define('xApi/models/activityDefinition',['../guard'],
    function (guard) {

        var activityDefinition = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create an ActivityDefinition';
            }

            var obj = {};

            guard.throwIfNotLanguageMap(spec.name, 'You should provide name for ActivityDefinition');

            if (typeof spec.name != typeof undefined) {
                obj.name = spec.name;
            }

            if (typeof spec.type != typeof undefined) {
                obj.type = spec.type;
            }

            if (typeof spec.interactionType != typeof undefined) {
                obj.interactionType = spec.interactionType;
            }

            if (typeof spec.correctResponsesPattern != typeof undefined) {
                obj.correctResponsesPattern = spec.correctResponsesPattern;
            }

            if (typeof spec.choices != typeof undefined) {
                obj.choices = spec.choices;
            }

            if (typeof spec.source != typeof undefined) {
                obj.source = spec.source;
            }

            if (typeof spec.target != typeof undefined) {
                obj.target = spec.target;
            }

            return obj;
        };

        return activityDefinition;
    }
);
define('xApi/models/activity',['../guard', './object', './activityDefinition'],
    function (guard, object, activityDefinition) {

        var activity = function (spec) {
            
            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create an Activity';
            }

            guard.throwIfNotString(spec.id, 'You should provide identity for Activity');
            guard.throwIfNotAnObject(spec.definition, 'You should provide definition for Activity');

            spec.objectType = 'Activity';

            var obj = new object(spec);
            obj.id = spec.id;
            obj.definition = spec.definition;

            return obj;
        };

        return activity;
    }
);
define('xApi/models/result',['../guard'],
    function (guard) {

        var result = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create Result';
            }

            var obj = {};

            if (typeof spec.score != typeof undefined) {
                obj.score = spec.score;
            }

            if (typeof spec.duration != typeof undefined) {
                guard.throwIfNotISODuration(spec.duration, 'You should provide duration in correct ISO duration format.');
                obj.duration = spec.duration;
            }

            if (typeof spec.response != typeof undefined) {
                obj.response = spec.response;
            }

            return obj;
        };

        return result;
    }
);
define('xApi/models/score',['../guard'],
    function (guard) {

        var score = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create Statement';
            }

            guard.throwIfNotNumber(spec, 'You should provide score');

            var obj = {};
            
            if (spec >= 0 && spec <= 1) {
                obj.scaled = spec;
            } else {
                obj.raw = spec;
            }
            
            return obj;
        };

        return score;
    }
);
define('xApi/models/context',['guard'], function (guard) {
    var context = function(spec) {
        if (typeof spec == typeof undefined) {
            throw 'You should provide a specification to create Context';
        }

        guard.throwIfNotString(spec.registration, 'You should specify registration field for the context');
        var obj = {};

        obj.registration = spec.registration;
        if (_.isObject(spec.contextActivities)) {
            obj.contextActivities = spec.contextActivities;
        }

        if (_.isObject(spec.extensions)) {
            obj.extensions = spec.extensions;
        }


        return obj;
    };

    return context;
});
define('xApi/models/contextActivities',[], function() {
    var contextActivities = function(spec) {
        if (typeof spec == typeof undefined) {
            throw 'You should provide a specification to create contextActivities';
        }

        var obj = {};

        if (typeof spec.parent != typeof undefined) {
            obj.parent = spec.parent;
        }
        
        if (typeof spec.grouping != typeof undefined) {
            obj.grouping = spec.grouping;
        }

        return obj;
    };

    return contextActivities;
});
define('xApi/models/languageMap',['../guard'],
    function (guard) {
        "use strict";

        var languageMap = function(text) {
            guard.throwIfNotString(text, 'You should provide valid text according to \'RFC 5646 Language Tag\'.');

            return {
                "en-US": text
            };
        };

        return languageMap;

    }
);
define('xApi/models/interactionDefinition',['./activityDefinition'], function (activityDefinition) {
    "use strict";

    var interactionDefinition = function (spec) {

        spec.type = "http://adlnet.gov/expapi/activities/cmi.interaction";

        return new activityDefinition(spec);
    };

    return interactionDefinition;

});
define('xApi/utils/dateTimeConverter',['../guard'], function(guard) {
    var timeToISODurationString = function (timeInMilliseconds) {
        guard.throwIfNotNumber(timeInMilliseconds, 'You should provide only number');
        timeInMilliseconds /= 1000;

        var hours = parseInt(timeInMilliseconds / 3600, 10);
        timeInMilliseconds -= hours * 3600;

        var minutes = parseInt(timeInMilliseconds / 60, 10);
        timeInMilliseconds -= minutes * 60;

        var seconds = parseInt(timeInMilliseconds, 10);

        return 'PT' + hours + 'H' + minutes + 'M' + seconds + 'S';
    };

    return {
         timeToISODurationString: timeToISODurationString
    };
});
define('xApi/statementQueue',[], function() {

    var statements = ko.observableArray([]);

    var queue = {
        statements: ko.computed({
            read: statements
        }),
        enqueue: enqueue,
        dequeue: dequeue
    };

    return queue;

    function enqueue(statement) {
        statements.push(statement);
    }

    function dequeue() {
        return statements.shift();
    }

});
define('repositories/sectionRepository',['guard', 'context'], function (guard, context) {

    return {
        get: get,
        getCollection: getCollection,
        getSectionByQuestionId: getSectionByQuestionId
    };

    function get(sectionId) {
        guard.throwIfNotString(sectionId, 'section id is not a string');

        var section = _.find(context.course.sections, function (item) {
            return item.id == sectionId;
        });

        if (!section) {
            return null;
        }

        return section;
    }

    function getCollection() {
        return context.course.sections;
    }

    function getSectionByQuestionId(questionId) {
        guard.throwIfNotString(questionId, 'Question id is not a string');

        var foundSection;
        _.each(context.course.sections, function (section) {
            _.each(section.questions, function (item) {
                if (item.id === questionId) {
                    foundSection = section;
                }
            });
        });

        return foundSection;
    }
});
define('xApi/statementSender',['./requestManager', './configuration/xApiSettings', './errorsHandler'],
    function(requestManager, xApiSettings, errorsHandler) {

        return {
            sendLrsStatement: sendLrsStatement,
            sendNpsStatement: sendNpsStatement
        };

        function sendLrsStatement(statement) {
            return sendStatement(statement, xApiSettings.xApi.lrs).fail(function(error) {
                errorsHandler.handleError(error);
            });
        }

        function sendNpsStatement(statement) {
            return sendStatement(statement, xApiSettings.xApi.nps);
        }

        function sendStatement(statement, settings) {
            var url = getUrl(settings.uri);
            
            var credentials = settings.authenticationRequired ?
                settings.credentials : xApiSettings.anonymousCredentials;

            return requestManager.sendStatement(statement, url, credentials.username, credentials.password);
        }

        function getUrl(settingsUrl) {
            var url = settingsUrl;

            if (url.indexOf("/statements") !== -1)
                return url;

            if (url.slice(-1) !== "/")
                url += "/";

            url += "statements";
            return url;
        }
    });
define('xApi/models/actor',['../guard', './object'], function (guard, object) {
  var actor = function (spec) {
    if (typeof spec == typeof undefined) {
      throw 'You should provide a specification to create an Actor'
    }

    spec.objectType = 'Agent'
    var obj = new object(spec)

    if (spec.account) {
      guard.throwIfNotString(
        spec.account.homePage,
        'You should provide homePage for Actor account'
      )
      guard.throwIfNotString(
        spec.account.name,
        'You should provide name for Actor account'
      )
      obj.account = {
        homePage: spec.account.homePage,
        name: spec.account.name,
      }
      obj.name = spec.name || spec.account.name
    } else {
      guard.throwIfNotMbox(
        spec.mbox,
        'You should provide mbox identity for Actor'
      )
      obj.mbox = spec.mbox
      obj.name = spec.name
    }

    return obj
  }

  return actor
})
;
define('xApi/activityProvider',['./models/statement', './models/activity', './models/activityDefinition', 'eventManager', './errorsHandler', './configuration/xApiSettings',
    './constants', './models/result', './models/score', './models/context', './models/contextActivities', './models/languageMap',
    './models/interactionDefinition', './utils/dateTimeConverter', './statementQueue', 'constants', 'guard', 'repositories/sectionRepository', 'progressContext', 'context',
    './statementSender', './models/actor'],
    function (statementModel, activityModel, activityDefinitionModel, eventManager, errorsHandler, xApiSettings, constants,
        resultModel, scoreModel, contextModel, contextActivitiesModel, languageMapModel, interactionDefinitionModel, dateTimeConverter, statementQueue,
        globalConstants, guard, sectionRepository, progressContext, courseContext, statementSender, actorModel) {

        "use strict";

        var subscriptions = [],
            activityProvider = {
                actor: null,
                activityName: null,
                activityUrl: null,

                initActivity: initActivity,
                initActor: initActor,
                subscribeToxApiEvents: subscribeToxApiEvents,
                turnOffxApiSubscriptions: unsubscribeFromxApiEvents,
                subscribeToNpsEvents: subscribeToNpsEvents,
                turnOffNpsSubscriptions: unsubscribeFromNpsEvents,
                rootCourseUrl: null,
                turnOffSubscriptions: turnOffSubscriptions,
                courseId: null
            },
            sessionId = function () {
                return progressContext.get().attemptId;
            };

        return activityProvider;

        function initActivity(courseId, activityName, activityUrl) {
            return Q.fcall(function () {
                if (_.isUndefined(xApiSettings.scoresDistribution.positiveVerb)) {
                    throw errorsHandler.errors.notEnoughDataInSettings;
                }

                activityProvider.activityName = activityName;
                activityProvider.activityUrl = activityUrl;
                activityProvider.rootCourseUrl = activityUrl !== undefined ? activityUrl.split("?")[0].split("#")[0] : '';
                activityProvider.courseId = courseId;
            });
        }

        function initActor(name, email, account) {
            return Q.fcall(function () {
                activityProvider.actor = createActor(name, email, account);
            });
        }

        function subscribeToxApiEvents() {
            unsubscribeFromxApiEvents();

            subscribeToEvent(eventManager.events.courseStarted, enqueueCourseStarted);
            subscribeToEvent(eventManager.events.courseFinished, enqueueCourseFinished);
            subscribeToEvent(eventManager.events.learningContentExperienced, enqueueLearningContentExperienced);
            subscribeToEvent(eventManager.events.answersSubmitted, enqueueQuestionAnsweredStatement);
        }

        function unsubscribeFromxApiEvents() {
            unsubscribeFromEvent(eventManager.events.courseStarted, enqueueCourseStarted);
            unsubscribeFromEvent(eventManager.events.courseFinished, enqueueCourseFinished);
            unsubscribeFromEvent(eventManager.events.learningContentExperienced, enqueueLearningContentExperienced);
            unsubscribeFromEvent(eventManager.events.answersSubmitted, enqueueQuestionAnsweredStatement);
        }

        function subscribeToNpsEvents() {
            unsubscribeFromNpsEvents();
            
            subscribeToEvent(eventManager.events.courseEvaluated, handleCourseEvaluated);
        }

        function unsubscribeFromNpsEvents() {
            unsubscribeFromEvent(eventManager.events.courseEvaluated, handleCourseEvaluated);
        }

        function turnOffSubscriptions() {
            _.each(subscriptions, function (subscription) {
                if (!_.isNullOrUndefined(subscription && subscription.off)) {
                    subscription.off();
                }
            });
        }

        // #region Event handlers

        function handleCourseEvaluated(data) {
            guard.throwIfNotAnObject(data, 'Data is not an object');
            guard.throwIfNotString(data.response, 'Data response is not a string');
            guard.throwIfNotNumber(data.score, 'Data score is not a number');

            var statement = createStatement(constants.verbs.evaluated,
                                            new resultModel({ score: new scoreModel(data.score), response: data.response }),
                                            createActivity(null, activityProvider.activityName, constants.activityTypes.course),
                                            createContextModel());

            return statementSender.sendNpsStatement(statement);
        }

        function enqueueCourseStarted() {
            pushStatementIfSupported(createStatement(constants.verbs.started));
        }

        function enqueueCourseFinished(course) {
            guard.throwIfNotAnObject(course, 'Course is not an object');

            if (_.isArray(course.sections)) {
                _.each(course.sections, function (section) {
                    if (_.isArray(section.questions)) {
                        _.each(section.questions, function (question) {
                            if (question.affectProgress && !question.isAnswered) {
                                enqueueQuestionAnsweredStatement(question);
                            }
                        });
                    }

                    var sectionUrl = activityProvider.rootCourseUrl + '#sections?section_id=' + section.id;
                    var score = section.affectProgress ? new scoreModel(section.score() / 100) : undefined;
                    var statement = createStatement(constants.verbs.mastered, new resultModel({ score: score }), createActivity(sectionUrl, section.title));
                    pushStatementIfSupported(statement);
                });
            }

            var result = new resultModel({
                score: new scoreModel(course.result())
            });

            var resultVerb = course.isCompleted() ? xApiSettings.scoresDistribution.positiveVerb : constants.verbs.failed;
            pushStatementIfSupported(createStatement(resultVerb, result));
            pushStatementIfSupported(createStatement(constants.verbs.stopped));

            var dfd = Q.defer();

            statementQueue.statements.subscribe(function (newValue) {
                if (newValue.length === 0) {
                    dfd.resolve();
                }
            });

            // (^\ x_x /^)
            statementQueue.enqueue(undefined);

            return dfd.promise;
        }

        function enqueueCourseProgressedStatement(course) {
            guard.throwIfNotAnObject(course, 'Course is not an object');

            var result = new resultModel({
                score: new scoreModel(course.result())
            });

            pushStatementIfSupported(createStatement(constants.verbs.progressed, result, createActivity(null, activityProvider.activityName, constants.activityTypes.course)));
        }

        function enqueueSectionProgressedStatement(section) {
            guard.throwIfNotAnObject(section, 'Section is not an object');

            var sectionUrl = activityProvider.rootCourseUrl + '#sections?section_id=' + section.id;
            var score = section.affectProgress ? new scoreModel(section.score() / 100) : undefined;
            var statement = createStatement(constants.verbs.progressed, new resultModel({ score: score }), createActivity(sectionUrl, section.title, constants.activityTypes.objective));
            pushStatementIfSupported(statement);
        }

        function enqueueLearningContentExperienced(question, spentTime) {
            pushStatementIfSupported(getLearningContentExperiencedStatement(question, spentTime));
        }

        function enqueueQuestionAnsweredStatement(question, sendParentProgress) {
            try {
                guard.throwIfNotAnObject(question, 'Question is not an object');

                var section = sectionRepository.get(question.sectionId);
                guard.throwIfNotAnObject(section, 'Section is not found');

                var parts = null;
                var contextObj = {};
                contextObj.extensions = {};
                switch (question.type) {
                    case globalConstants.questionTypes.multipleSelect:
                    case globalConstants.questionTypes.singleSelectText:
                        parts = getSelectTextQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.fillInTheBlank:
                        parts = getFillInQuestionActivityAndResult(question, contextObj);
                        break;
                    case globalConstants.questionTypes.singleSelectImage:
                        parts = getSingleSelectImageQuestionAcitivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.statement:
                        parts = getStatementQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.dragAndDrop:
                        parts = getDragAndDropTextQuestionActivityAndResult(question, contextObj);
                        break;
                    case globalConstants.questionTypes.textMatching:
                        parts = getMatchingQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.hotspot:
                        parts = getHotSpotQuestionActivityAndResult(question, contextObj);
                        break;
                    case globalConstants.questionTypes.openQuestion:
                        parts = getOpenQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.scenario:
                        parts = getScenarioQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.rankingText:
                        parts = getRankingTextQuestionActivityAndResult(question);
                        break;
                    case globalConstants.questionTypes.informationContent:
                        parts = getInformationContentActivityAndResult(question);
                        break;
                }

                var parentUrl = activityProvider.rootCourseUrl + '#sections?section_id=' + section.id;

                contextObj.contextActivities = new contextActivitiesModel({
                    parent: [createActivity(parentUrl, section.title)]
                });

                contextObj.extensions[constants.extenstionKeys.surveyMode] = question.hasOwnProperty('isSurvey') && question.isSurvey;
                contextObj.extensions[constants.extenstionKeys.questionType] = question.type;

                var context = createContextModel(contextObj);

                if (parts) {
                    var verb = question.type === globalConstants.questionTypes.informationContent ?
                        constants.verbs.experienced : constants.verbs.answered;

                    var statement = createStatement(verb, parts.result, parts.object, context);
                    if (statement) {
                        pushStatementIfSupported(statement);
                        if (sendParentProgress) {
                            enqueueSectionProgressedStatement(section);
                            enqueueCourseProgressedStatement(courseContext.course);
                        }
                    }
                }

            } catch (e) {
                console.error(e);
            }
        }

        // #endregion

        // #region Statement creation

        function getSelectTextQuestionActivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: getItemsIds(question.answers, function (item) {
                        return item.isChecked;
                    }).join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.choice,
                        correctResponsesPattern: !!question.isSurvey ? [] : [
                            getItemsIds(question.answers, function (item) {
                                return item.isCorrect;
                            }).join("[,]")
                        ],
                        choices: _.map(question.answers, function (item) {
                            return {
                                id: item.id,
                                description: new languageMapModel(item.text)
                            };
                        })
                    })
                })
            };
        }

        function getStatementQuestionActivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.chain(question.statements).filter(function (statement) {
                        return !_.isNullOrUndefined(statement.userAnswer);
                    }).map(function (statement) {
                        return statement.id + '[.]' + statement.userAnswer;
                    }).value().join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.choice,
                        correctResponsesPattern: !!question.isSurvey ? [] : [
                            _.map(question.statements, function (item) {
                                return item.id + '[.]' + item.isCorrect;
                            }).join("[,]")
                        ],
                        choices: _.map(question.statements, function (item) {
                            return {
                                id: item.id,
                                description: new languageMapModel(item.text)
                            };
                        })
                    })
                })
            };
        }

        function getSingleSelectImageQuestionAcitivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: question.checkedAnswerId || ''
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.choice,
                        correctResponsesPattern: [[question.correctAnswerId].join("[,]")],
                        choices: _.map(question.answers, function (item) {
                            return {
                                id: item.id,
                                description: new languageMapModel(item.image)
                            };
                        })
                    })
                })
            };

        }

        function getFillInQuestionActivityAndResult(question, contextObj) {
            contextObj.extensions[constants.extenstionKeys.content] = question.content;
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.map(question.answerGroups, function (item) {
                        return item.answeredText + "[.]" + item.id;
                    }).join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.fillIn,
                        correctResponsesPattern: [
                            _.flatten(_.map(question.answerGroups, function (item) {
                                return item.getCorrectText() + "[.]" + item.id;
                            })).join("[,]")
                        ]
                    })
                })
            };
        }

        function getHotSpotQuestionActivityAndResult(question, contextObj) {
            contextObj.extensions[constants.extenstionKeys.imageUrl] = question.background;
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.map(question.placedMarks, function (mark) {
                        return '(' + mark.x + ',' + mark.y + ')';
                    }).join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.other,
                        correctResponsesPattern: [_.map(question.spots, function (spot) {
                            var polygonCoordinates = _.map(spot, function (spotCoordinates) {
                                return '(' + spotCoordinates.x + ',' + spotCoordinates.y + ')';
                            });
                            return polygonCoordinates.join("[.]");
                        }).join("[,]")]
                    })
                })
            }
        }

        function getDragAndDropTextQuestionActivityAndResult(question, contextObj) {
            var answerTexts = _.map(question.answers, function (item) {
                return item.text;
            }).join("[,]");
            contextObj.extensions[constants.extenstionKeys.answerTexts] = answerTexts;
            contextObj.extensions[constants.extenstionKeys.imageUrl] = question.background;
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.map(question.answers, function (item) {
                        return '(' + item.currentPosition.x + ',' + item.currentPosition.y + ')';
                    }).join("[,]")
                }),

                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.other,
                        correctResponsesPattern: [_.map(question.answers, function (item) {
                            return '(' + item.correctPosition.x + ',' + item.correctPosition.y + ')';
                        }).join("[,]")]
                    })
                })
            }
        }

        function getMatchingQuestionActivityAndResult(question) {
            var targets = _.uniq(_.pluck(question.answers, 'value'));

            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.map(question.answers, function (answer) {
                        return answer.shortId + "[.]" + (answer.attemptedValue ? _.indexOf(targets, answer.attemptedValue) : "");
                    }).join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.matching,
                        correctResponsesPattern: [_.map(question.answers, function (answer) {
                            return answer.shortId + "[.]" + _.indexOf(targets, answer.value);
                        }).join("[,]")],
                        source: _.map(question.answers, function (answer) {
                            return { id: answer.shortId.toString(), description: new languageMapModel(answer.key) }
                        }),
                        target: _.map(targets, function (value, index) {
                            return { id: index.toString(), description: new languageMapModel(value) }
                        })
                    })
                })
            };

        }

        function getOpenQuestionActivityAndResult(question) {
            return {
                result: new resultModel({
                    response: question.answeredText
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.other
                    })
                })
            };
        }

        function getScenarioQuestionActivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.other
                    })
                })
            };
        }

        function getRankingTextQuestionActivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100),
                    response: _.map(question.rankingItems, function (item) {
                        return item.text.toLowerCase();
                    }).join("[,]")
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.sequencing,
                        correctResponsesPattern: [_.map(question.correctOrder, function (item) {
                            return item.text.toLowerCase();
                        }).join("[,]")],
                        choices: _.map(question.rankingItems, function (item) {
                            return {
                                id: item.text,
                                description: new languageMapModel(item.text)
                            };
                        })
                    })
                })
            };
        }

        function getInformationContentActivityAndResult(question) {
            return {
                result: new resultModel({
                    score: new scoreModel(question.score() / 100)
                }),
                object: new activityModel({
                    id: activityProvider.rootCourseUrl + '#section/' + question.sectionId + '/question/' + question.id,
                    definition: new interactionDefinitionModel({
                        name: new languageMapModel(question.title),
                        interactionType: constants.interactionTypes.other
                    })
                })
            };
        }

        function getLearningContentExperiencedStatement(question, spentTime) {
            guard.throwIfNotAnObject(question, 'Question is not an object');
            guard.throwIfNotNumber(spentTime, 'SpentTime is not a number');

            var section = sectionRepository.get(question.sectionId);
            guard.throwIfNotAnObject(section, 'Section is not found');

            var result = new resultModel({
                duration: dateTimeConverter.timeToISODurationString(spentTime)
            });

            var learningContentUrl = activityProvider.rootCourseUrl + '#section/' + section.id + '/question/' + question.id + '?learningContents';
            var parentUrl = activityProvider.rootCourseUrl + '#section/' + section.id + '/question/' + question.id;
            var groupingUrl = activityProvider.rootCourseUrl + '#sections?section_id=' + section.id;
            var object = createActivity(learningContentUrl, question.title);

            var context = createContextModel({
                contextActivities: new contextActivitiesModel({
                    parent: [createActivity(parentUrl, question.title)],
                    grouping: [createActivity(groupingUrl, section.title)]
                })
            });

            return createStatement(constants.verbs.experienced, result, object, context);
        }

        function createContextModel(contextSpec) {
            contextSpec = contextSpec || {};
            var contextExtensions = contextSpec.extensions || {};
            contextExtensions[constants.extenstionKeys.courseId] = activityProvider.courseId;
            contextSpec.extensions = contextExtensions;
            contextSpec.registration = sessionId();

            return new contextModel(contextSpec);
        }

        function createStatement(verb, result, activity, context) {
            var activityData = activity || createActivity(null, activityProvider.activityName);
            context = context || createContextModel();

            return statementModel({
                actor: activityProvider.actor,
                verb: verb,
                object: activityData,
                result: result,
                context: context
            });
        }

        // #endregion

        // #region Utilities

        function createActor(name, email, account) {
            var actor = {};

            try {
                actor = actorModel({
                    name: name,
                    mbox: 'mailto:' + email,
                    account: account
                });
            } catch (e) {
                errorsHandler.handleError(errorsHandler.errors.actorDataIsIncorrect);
            }

            return actor;
        }

        function createActivity(id, name, type) {
            return activityModel({
                id: id || activityProvider.activityUrl,
                definition: new activityDefinitionModel({
                    name: new languageMapModel(name),
                    type: type
                })
            });
        }


        function pushStatementIfSupported(statement) {
            if (_.contains(xApiSettings.xApi.allowedVerbs, statement.verb.display[xApiSettings.defaultLanguage])) {
                statementQueue.enqueue(statement);
            }
        }

        function subscribeToEvent(eventName, eventHandler) {
            subscriptions.push(eventManager.subscribeForEvent(eventName).then(eventHandler));
        }

        function unsubscribeFromEvent(eventName, eventHandler) {
            eventManager.unsubscribeForEvent(eventName, eventHandler);
        }

        function getItemsIds(items, filter) {
            return _.chain(items)
               .filter(function (item) {
                   return filter(item);
               })
               .map(function (item) {
                   return item.id;
               }).value();
        }

        // #endregion

    }
);
define('xApi/statementQueueHandler',['durandal/app', './statementSender', './statementQueue'], function (app, statementSender, statementQueue) {

    return {
        handle: handle
    };

    function handle() {
        var statement = statementQueue.dequeue();

        if (statement) {
            return statementSender.sendLrsStatement(statement).then(handle);
        } else {
            var subscription = statementQueue.statements.subscribe(function () {
                handle();
                subscription.dispose();
            });
        }

    }

});
define('xApi/xApiInitializer',['durandal/app', 'plugins/router', './routingManager', './requestManager', './activityProvider', './configuration/xApiSettings', './constants', './statementQueueHandler',
        './errorsHandler', 'context', 'progressContext', 'userContext', 'eventManager', 'helpers/loginHelper', 'constants'
    ],
    function(app, router, routingManager, requestManager, activityProvider, xApiSettingsConfig, constants, statementQueueHandler, errorsHandler, context, progressContext,
        userContext, eventManager, loginHelper, appConstants) {
        "use strict";

        var self = {
                isUserAuthenticated: false,
                npsSettings: null,
                xApiSettings: null,
            },
            xApiInitializer = {
                initialize: initialize,
                deactivate: deactivate,
                isLrsReportingInitialized: false,
                isNpsReportingInitialized: false,
                isInitialized: false,
                activate: activate
            };

        return xApiInitializer;

        function initialize(xApiReportingSettings, npsReportingSettings) {
            return Q.fcall(function() {
                self.xApiSettings = xApiReportingSettings;
                self.npsSettings = npsReportingSettings;

                if (!self.xApiSettings.enabled && !self.npsSettings.enabled)
                    return;

                routingManager.mapRoutes();
                self.isCourseStarted = checkIsCourseStarted();
                return initializeTracking()
                    .then(function() {
                        activateReporting();
                        if (loginHelper.getLoginMethod() !== appConstants.loginTypes.byEmailAndName && !self.isCourseStarted) {
                            return eventManager.courseStarted();
                        }
                    })
                    .fail(onInitializationFailed);
            });
        }

        function activate(user) {
            return initializeActor(user)
                .then(function() {
                    activateReporting();
                })
                .fail(onInitializationFailed);
        }

        function deactivate() {
            deactivateLrsReporting();
            deactivateNpsReporting();
            xApiInitializer.isInitialized = false;
        }

        function initializeTracking() {
            return initializeActorFromContext().then(function() {
                initializeActivity();

                return requestManager.init()
                    .then(function() {
                        xApiInitializer.isInitialized = true;
                    });
            });
        }

        function activateReporting() {
            if (!self.isUserAuthenticated)
                return;

            if (self.xApiSettings.enabled) activateXapiReporting();
            if (self.npsSettings.enabled) activateNpsReporting();
        }

        function activateXapiReporting() {
            xApiSettingsConfig.initxApi(self.xApiSettings);
            activityProvider.subscribeToxApiEvents();

            statementQueueHandler.handle();
            xApiInitializer.isLrsReportingInitialized = true;
        }

        function activateNpsReporting() {
            xApiSettingsConfig.initNps(self.npsSettings);
            activityProvider.subscribeToNpsEvents();

            xApiInitializer.isNpsReportingInitialized = true;
        }

        function deactivateLrsReporting() {
            activityProvider.turnOffxApiSubscriptions();
            xApiInitializer.isLrsReportingInitialized = false;
            app.trigger('user:authentication-skipped');
        }

        function deactivateNpsReporting() {
            activityProvider.turnOffNpsSubscriptions();
            xApiInitializer.isNpsReportingInitialized = false;
        }

        function initializeActorFromContext() {
            return Q.fcall(function() {
                var user = userContext.getCurrentUser(),
                    progress = progressContext.get();

                if (user && user.username && (constants.patterns.email.test(user.email) || user.account)) {
                    return initializeActor(user);
                }

                if (_.isObject(progress)) {
                    if (_.isObject(progress.user)) {
                        return initializeActor(progress.user);
                    }
                    if (progress.user === 0) {
                        deactivateLrsReporting();
                        return;
                    }
                }
            });
        }

        function initializeActor(user) {
            return Q.fcall(function() {
                var actor = {};

                actor.username = user.username;
                actor.email = user.email || (user.account ? user.account.name : '');
                if (user.account)
                    actor.account = user.account;

                self.isUserAuthenticated = true;
                app.trigger('user:authenticated', user);

                return activityProvider.initActor(actor.username, actor.email, actor.account);
            });
        }

        function initializeActivity() {
            return Q.fcall(function() {
                var id = context.course.id;
                var title = context.course.title;

                var url = "";
                if (window != window.top && ('referrer' in document)) {
                    url = document.referrer;
                } else {
                    url = window.location.toString();
                }

                url = url + '?course_id=' + context.course.id;

                activityProvider.initActivity(id, title, url);
            });
        }

        function onInitializationFailed(reason) {
            deactivate();
            if (self.xApiSettings.enabled)
                errorsHandler.handleError(reason);
        }

        function checkIsCourseStarted() {
            var progress = progressContext.get();
            return _.isObject(progress) && _.isObject(progress.user);
        }
    }
);
define('account/routing/guardRoute',['plugins/router', 'underscore', 'templateSettings', 'userContext', '../limitAccess/accessLimiter',
        'modules/progress/progressStorage/auth', 'progressContext'],
 function (router, _, templateSettings, userContext, accessLimiter, auth, progressContext) {
    'use strict';

    var defaultGuard = null,
        loginGuardSkipped = false;


    function guardViewUrl(route, guardView) {
        return route.queryString ? guardView + '?' + route.queryString : guardView;
    }

    return {
        createGuard: function () {
            defaultGuard = router.guardRoute;
            var loginGuardView = 'login',
                registerGuardView = 'register',
                signinGuardView = 'signin',
                noAccessGuardView = 'noaccess';

            router.guardRoute = function (model, route) {
                var user = userContext.getCurrentUser(),
                    accessLimitationEnabled = accessLimiter.accessLimitationEnabled(),
                    xapiEnabled = templateSettings.xApi.enabled,
                    npsEnabled = templateSettings.nps.enabled,
                    allowCrossDeviceSaving = templateSettings.allowCrossDeviceSaving,
                    currentRoute = route.config.route;
                    
                if (currentRoute === loginGuardView) {
                    var checkUserPermission = user && accessLimiter.userHasAccess(user);
                    if(allowCrossDeviceSaving){
                        return checkUserPermission && auth.authenticated ? '/': true;
                    }
                    return checkUserPermission  ? '/' : true;
                }

                if((currentRoute === registerGuardView || currentRoute === signinGuardView)
                    && userContext.user.email){
                        return accessLimiter.userHasAccess({email:userContext.user.email});
                    }

                if (!userContext.user.email && !loginGuardSkipped && (xapiEnabled || npsEnabled || allowCrossDeviceSaving || accessLimitationEnabled)) {
                    return guardViewUrl(route, loginGuardView);
                }

                if (currentRoute === noAccessGuardView) {
                    return accessLimitationEnabled && (!user || !accessLimiter.userHasAccess(user)) ? true : '/';
                }

                if (accessLimitationEnabled && (!user || !accessLimiter.userHasAccess(user))) {
                    return guardViewUrl(route, noAccessGuardView);
                }

                if(userContext.user.email && allowCrossDeviceSaving && !auth.authenticated){
                    return guardViewUrl(route, loginGuardView);
                }

                if (_.isFunction(defaultGuard)) {
                    return defaultGuard(model, route);
                }
                return true;
            };
        },
        skipLoginGuard: function () {
            loginGuardSkipped = true;
        }
    };
});
define('modules/progress/localStorage/index',['context', '../constants'], function (context, constants) {
    'use strict';
    
    function LocalStorageProgressProvider(courseId, templateId) {
        this.courseId = courseId;
        this.templateId = templateId;

        this.progressKey = constants.progressKey +this.courseId+this.templateId;
        this.resultKey = constants.resultKey +this.courseId+this.templateId;
        
        this.getProgress = function () {
            var progress = {};
            try {
                progress = JSON.parse(localStorage.getItem(this.progressKey));
            } catch (e) {
                console.log('Unable to restore progress from localStorage');
            }
            return progress;
        };
        
        this.saveProgress = function(progress){
            try {
                localStorage.setItem(this.progressKey, JSON.stringify(progress));
            } catch (e) {
                return false;
            }
            return true;
        };
        
        this.saveResults = function () {
            var result = {
                score: context.course.score(),
                status: context.course.getStatus()
            };

            try {
                localStorage.setItem(this.resultKey, JSON.stringify(result));
            } catch (e) {
                alert(TranslationPlugin.getTextByKey('[not enough memory to save progress]'));
            }
            return true;
        };
        
        this.removeProgress = function () {
            try {
                localStorage.removeItem(this.progressKey);
            } catch (e) {
                console.log('Unable to remove progress from localStorage');
            }
            return true;
        };
    }
    
    return LocalStorageProgressProvider;
});
define('modules/progress/progressStorage/index',[
  'context',
  '../constants',
  'requester/httpWrapper',
  './urlProvider',
  './auth',
  'helpers/requestResender',
  'helpers/idToUuid'
], function (context, constants, httpWrapper, urlProvider, auth, requestResender, idToUuid) {
  'use strict';

  function ProgressStorageProvider(courseId, templateId) {
    var _progress = null;
    var self = this;
    this.courseId = courseId;
    this.templateId = templateId;

    this.progressKey = constants.progressKey + this.courseId + this.templateId;
    this.resultKey = constants.resultKey + this.courseId + this.templateId;

    this.getCourseStatus = function () {
      if (context.course.getStatus() === 'completed') {
        return 'PASSED';
      } else if (context.course.getStatus() === 'failed') {
        return 'FAILED';
      }
      return 'IN_PROGRESS';
    };

    this.getProgress = function () {
      return _progress;
    };

    this.setProgress = function (progress) {
      _progress = progress;
    };

    this.getProgressFromServer = function () {
      if (!auth.authenticated) {
        return false;
      }
      return httpWrapper
        .get({
          url: urlProvider.learnServiceUrl + constants.learnerCoursesApi + idToUuid(this.courseId) + '/attempts/last',
          query: {
            templateId: this.templateId
          },
          headers: { Authorization: 'Bearer ' + auth.getToken() },
          withCredentials: false
        })
        .then(function (response) {
          _progress = JSON.parse(response.jsonProgress);
          return response.jsonProgress;
        })
        .fail(function (fail) {
          console.error('Get progress from server fails with error: ', fail);
          return fail;
        });
    };

    this.startNewAttempt = function () {
      return httpWrapper
        .post({
          url: urlProvider.learnServiceUrl + constants.learnerCoursesApi + idToUuid(self.courseId) + '/attempts',
          data: {
            templateId: self.templateId,
            jsonProgress: '{}',
            score: 0,
            status: 'IN_PROGRESS',
            courseUrl: urlProvider.courseLink
          },
          headers: {
            'Content-Type': constants.contentTypeJson,
            Authorization: 'Bearer ' + auth.getToken()
          },
          withCredentials: false
        })
        .then(function (newProgress) {
          return newProgress.jsonProgress;
        });
    };

    this.saveProgress = function (progress) {
      var requestOptions = {
        url:
          urlProvider.learnServiceUrl +
          constants.learnerCoursesApi +
          idToUuid(this.courseId) +
          '/attempts/last' +
          '?templateId=' +
          this.templateId,

        method: 'PUT',
        data: JSON.stringify({
          jsonProgress: progress ? JSON.stringify(progress) : '{}',
          score: context.course.score(),
          status: this.getCourseStatus()
        }),
        headers: {
          'Content-Type': constants.contentTypeJson,
          Authorization: 'Bearer ' + auth.getToken()
        },
        cache: false
      };

      return requestResender
        .send(requestOptions)
        .then(function (response) {
          return response;
        })
        .fail(function (failedResponse) {
          if (failedResponse.status === 401) {
            return auth.signout();
          }

          console.error('Save progress failed with error: ', failedResponse);
          return false;
        });
    };

    this.saveResults = function (getScore, getStatus, errorMessage) {
      //TODO: now we use only localStorage for save results
      var result = {
        score: context.course.score(),
        status: context.course.getStatus()
      };

      try {
        localStorage.setItem(this.resultKey, JSON.stringify(result));
      } catch (e) {
        console.error(TranslationPlugin.getTextByKey('[not enough memory to save progress]'));
        return false;
      }
      return true;
    };

    this.removeProgress = function (callback) {
      return httpWrapper.post({
        url: urlProvider.learnServiceUrl + constants.learnerCoursesApi + idToUuid(this.courseId) + '/attempts',
        data: {
          templateId: this.templateId,
          jsonProgress: '{}',
          score: 0,
          status: 'IN_PROGRESS',
          courseUrl: urlProvider.courseLink
        },
        headers: {
          'Content-Type': constants.contentTypeJson,
          Authorization: 'Bearer ' + auth.getToken()
        },
        withCredentials: false
      });
    };
  }

  return ProgressStorageProvider;
});

define('modules/progress/index',[
  "q",
  "underscore",
  "context",
  "userContext",
  "templateSettings",
  "./localStorage/index",
  "./progressStorage/index",
  "./progressStorage/auth",
  "account/limitAccess/accessLimiter",
  "helpers/loginHelper"
], function(
  Q,
  _,
  context,
  userContext,
  templateSettings,
  LocalStorageProvider,
  ProgressStorageProvider,
  auth,
  accessLimiter,
  loginHelper
) {
  "use strict";
  var _psProvider = null,
    _lsProvider = null;

  return {
    initialize: initialize,
    initProgressStorage: initProgressStorage,
    clearLocalStorage: clearLocalStorage
  };

  function initialize() {
    var defer = Q.defer();
    _psProvider = new ProgressStorageProvider(
      context.course.id,
      context.course.templateId
    );
    _lsProvider = new LocalStorageProvider(
      context.course.id,
      context.course.templateId
    );

    if (templateSettings.allowCrossDeviceSaving) {
      var token = loginHelper.token;

      if (!_.isNull(token)) {
        auth.setToken(token);
      }
      if (auth.authenticated) {
        initProgressStorage(resolve);
      } else {
        resolve(_lsProvider);
      }
    } else {
      resolve(_lsProvider);
    }

    return defer.promise;

    function resolve(provider) {
      defer.resolve(provider);
    }
  }

  function initProgressStorage(callback) {
    auth
      .identify()
      .then(function(user) {
        userContext.user.email = user.email;
        if (!accessLimiter.userHasAccess({ email: user.email })) {
          auth.signout();
          callback(_lsProvider);
          return false;
        }
        userContext.user.email = user.email;
        userContext.user.username = user.name;
        userContext.user.keepMeLoggedIn = !auth.shortTermAccess;
        return _psProvider.getProgressFromServer().then(function(progress) {
          var lsprogress = _lsProvider.getProgress();
          if (
            lsprogress &&
            lsprogress.user &&
            user.email === lsprogress.user.email
          ) {
            lsprogress.user.username = userContext.user.username;
            if (progress == null) {
              _psProvider.setProgress(lsprogress);
              _psProvider.saveProgress(lsprogress);
            }
            _lsProvider.removeProgress();
          }
          callback(_psProvider);
        });
      })
      .fail(function(failed) {
        if (failed.status === 404) {
          _psProvider.startNewAttempt().then(function(newProgress) {
            callback(_psProvider);
          });
        } else {
          callback(_lsProvider);
        }
      });
  }

  function clearLocalStorage() {
    _lsProvider.removeProgress();
  }
});

define('account/account',[
  "plugins/router",
  "context",
  "./header/index",
  "./socialNetworks/index",
  "./helpers/validatedValue",
  "xApi/constants",
  "userContext",
  "templateSettings",
  "modules/progress/progressStorage/urlProvider",
  "modules/progress/progressStorage/auth",
  "./limitAccess/accessLimiter",
  "xApi/xApiInitializer",
  "./routing/guardRoute",
  "modules/progress/index",
  "progressContext"
], function(
  router,
  context,
  Header,
  SocialNetworks,
  validatedValue,
  constants,
  userContext,
  templateSettings,
  urlProvider,
  auth,
  accessLimiter,
  xApiInitializer,
  guardRoute,
  progressProvider,
  progressContext
) {
  "use strict";

  var _private = {
    crossDeviceEnabled: templateSettings.allowCrossDeviceSaving,
    isActivatedPreviously: false
  };

  var viewmodel = {
    activate: activate,
    submit: checkUser,
    skip: skip,
    header: null,
    socialNetworks: null,
    email: userContext.user.email || "",
    socialLoginEnabled: false,
    skipAllowed: false,
    canActivate: canActivate,
    requestProcessing: ko.observable(false)
  };

  viewmodel.socialLoginEnabled =
    templateSettings.allowCrossDeviceSaving &&
    templateSettings.allowLoginViaSocialMedia;
  viewmodel.skipAllowed = !templateSettings.xApi.required;

  viewmodel.email = validatedValue(function(value) {
    return !!value() && constants.patterns.email.test(value());
  });

  return viewmodel;

  function canActivate() {
    var defer = Q.defer();
    if (userContext.user.email && !_private.isActivatedPreviously) {
      _private.isActivatedPreviously = true;
      viewmodel.email(userContext.user.email);
      submit(true, function(canActivate) {
        defer.resolve(canActivate);
      });
    } else {
      _private.isActivatedPreviously = true;
      defer.resolve(true);
    }
    return defer.promise;
  }

  function activate() {
    viewmodel.requestProcessing(false);
    viewmodel.header = new Header(context.course.title);
    viewmodel.socialNetworks = viewmodel.socialLoginEnabled
      ? new SocialNetworks(context.course.title, urlProvider.authServiceUrl)
      : null;
  }

  function checkUser() {
    submit(false);
  }

  function submit(replaceWindow, callback) {
    if (viewmodel.requestProcessing()) {
      return;
    }
    callback = callback || function() {};
    if (!viewmodel.email.isValid()) {
      viewmodel.email.markAsModified();
      callback(true);
      return;
    }
    userContext.user.email = viewmodel.email();
    if (!accessLimiter.userHasAccess({ email: viewmodel.email() })) {
      _navigate("noaccess", replaceWindow);
      callback(false);
      return;
    }
    if (_private.crossDeviceEnabled) {
      return userExists(replaceWindow, callback);
    } else {
      _navigate("register", replaceWindow);
      callback(false);
    }
  }

  function skip() {
    if (viewmodel.requestProcessing()) {
      return;
    }

    xApiInitializer.deactivate();
    guardRoute.skipLoginGuard();
    if (!accessLimiter.userHasAccess({ email: viewmodel.email() })) {
      router.navigate("noaccess");
      return;
    }
    templateSettings.allowCrossDeviceSaving = false;
    progressProvider.initialize().then(function(provider) {
      progressContext.use(provider);
      progressContext.restoreProgress();
    });
  }

  function userExists(replaceWindow, callback) {
    viewmodel.requestProcessing(true);
    return auth
      .exists(viewmodel.email())
      .then(function(exists) {
        callback(false);
        viewmodel.requestProcessing(false);
        if (!exists) {
          _navigate("register", replaceWindow);
        } else {
          _navigate("signin", replaceWindow);
        }
      })
      .fail(function(reason) {
        callback(false);
        viewmodel.requestProcessing(false);
        if (reason.status === 409 || reason.status === 401) {
          _navigate("signin", replaceWindow);
        }
      });
  }

  function _navigate(route, replaceWindow) {
    if (replaceWindow) {
      window.location.hash = "#" + route;
      return;
    }
    router.navigate(route);
  }
});

define('routing/routes',['underscore'], function (_) {
    'use strict';

    return {
        routes: [{
            route: ['', 'introduction'],
            moduleId: 'viewmodels/introduction',
            title: TranslationPlugin.getTextByKey('[introduction page title]')
        }, {
            route: 'sections',
            moduleId: 'viewmodels/sections',
            title: TranslationPlugin.getTextByKey('[sections page title]'),
            displayTreeOfContent: true
        }, {
            route: 'section/:sectionId/question/:questionId',
            moduleId: 'viewmodels/questions/content',
            title: TranslationPlugin.getTextByKey('[question page title]'),
            displayTreeOfContent: true,
            showExitButton: true,
            rootLinkEnabled: true
        }, {
            route: 'finish',
            moduleId: 'resultPage/resultPage',
            title: TranslationPlugin.getTextByKey('[results page title]'),
            showExitButtonInReviewMode: true
        }, {
            route: '404(/:url)',
            moduleId: 'viewmodels/404',
            title: TranslationPlugin.getTextByKey('[not found page title]')
        }],
        add: add
    };

    function add(routes) {
        var self = this;
        if (_.isObject(routes) && _isRouteCorrect(routes)) {
            this.routes.push(routes);
        } else if (_.isArray(routes)) {
            _.each(routes, function (route) {
                if (_isRouteCorrect(route)) {
                    self.routes.push(route);
                } else {
                    throw 'Route not correct';
                }
            });
        }
    };

    function _isRouteCorrect(route) {
        return _.isObject(route) && route.hasOwnProperty('route') &&
            route.hasOwnProperty('moduleId') && route.hasOwnProperty('title');
    }
});
define('account/routing/routes',[],function () {
    'use strict';

    return [
        {
            route: 'login',
            moduleId: 'account/account',
            title: TranslationPlugin.getTextByKey('[login page title]')
        }, {
            route: 'noaccess',
            moduleId: 'account/limitAccess/index',
            title: TranslationPlugin.getTextByKey('[limited access page title]')
        }, {
            route: 'signin',
            moduleId: 'account/signin/index',
            title: TranslationPlugin.getTextByKey('[sign in page title]')
        }, {
            route: 'register',
            moduleId: 'account/register/index',
            title: TranslationPlugin.getTextByKey('[register page title]')
        }
    ];
});
define('account/index',['plugins/router', 'routing/routes', 'templateSettings', 'publishSettings',
        './limitAccess/accessLimiter', './routing/guardRoute', './routing/routes', 'modules/progress/progressStorage/auth', 'modules/publishModeProvider'
    ],
    function (router, mainRoutes, templateSettings, publishSettings, accessLimiter, guardRoute, routes, auth, publishModeProvider) {
        'use strict';

        return {
            enable: enable
        };

        function enable() {
            var xAPI = templateSettings.xApi.enabled;
            var nps = templateSettings.nps.enabled;
            var crossDeviceSaving = templateSettings.allowCrossDeviceSaving;
            
            if (publishModeProvider.isScormEnabled) {
                return;
            }

            if (publishModeProvider.isPreview || publishModeProvider.isReview) {
               return;
            }
            
            accessLimiter.initialize(publishSettings.accessLimitation);

            if (xAPI || nps || crossDeviceSaving || accessLimiter.accessLimitationEnabled()) {
                guardRoute.createGuard();
                mainRoutes.add(routes);
            }
        }

    });
define('account/limitAccess/index',['plugins/router', 'context', 'userContext', './accessLimiter'], function (router, context, userContext, accessLimiter) {

    "use strict";

    return {
        courseTitle: context.course.title,
        canActivate: function () {
            var user = userContext.getCurrentUser();
            if (!accessLimiter.accessLimitationEnabled() || (user && accessLimiter.userHasAccess(user))) {
                return { redirect: '/' };
            }

            return true;
        },
        activate: function () {
        },
        backToSignin: function () {
            userContext.clear();
            router.navigate('login');
        }
    };
});
define('account/register/index',['knockout', 'plugins/router', 'context', 'userContext', '../header/index', '../helpers/validatedValue', 'modules/progress/progressStorage/auth',
        'templateSettings', 'xApi/xApiInitializer', 'eventManager', 'modules/progress/index', 'progressContext'
    ],
    function (ko, router, context, userContext, Header, validatedValue,
            auth, templateSettings, xApiInitializer, eventManager, progressProvider, progressContext) {
        'use strict';

        var whitespaceRegex = /\s/;

        var viewmodel = {
            activate: activate,
            header: null,
            name: userContext.user.username || '',
            password: null,
            rememberMe: ko.observable(false),
            isPasswordVisible: ko.observable(false),
            togglePasswordVisibility: togglePasswordVisibility,
            submit: submit,
            toggleRememberMe: toggleRememberMe,
            back: back,
            crossDeviceEnabled: false,
            requestProcessing: ko.observable(false)
        };

        viewmodel.name = validatedValue(function (value) {
            return !!value();
        });

        viewmodel.password = validatedValue(function (value) {
            return value().length >= 7 && !whitespaceRegex.test(value());
        });
        viewmodel.password.noSpaces = ko.observable(false);
        viewmodel.password.moreThanSevenSymbols = ko.observable(false);
        viewmodel.password.onKeyUp = function () {
            if (viewmodel.password().length >= 7) {
                viewmodel.password.moreThanSevenSymbols(true);
            } else {
                viewmodel.password.moreThanSevenSymbols(false);
            }
            if (!whitespaceRegex.test(viewmodel.password())) {
                viewmodel.password.noSpaces(true);
            } else {
                viewmodel.password.noSpaces(false);
            }
        }


        return viewmodel;

        function activate() {
            viewmodel.header = new Header(context.course.title);
            viewmodel.crossDeviceEnabled = templateSettings.allowCrossDeviceSaving;
            if(userContext.user.username){
                viewmodel.name(userContext.user.username);
            }
        }

        function submit() {
            if(viewmodel.requestProcessing()){
                return;
            }
            if (!viewmodel.name.isValid()) {
                viewmodel.name.markAsModified();
                return;
            }
            userContext.user.username = viewmodel.name();
            if (viewmodel.crossDeviceEnabled && !viewmodel.password.isValid()) {
                viewmodel.password.markAsModified();
                return;
            } else if (viewmodel.crossDeviceEnabled && viewmodel.password.isValid()) {
                register();
            } else {
                activateXapi(function () {
                    eventManager.courseStarted();
                    router.navigate('');
                });
            }
        }

        function toggleRememberMe() {
            if(viewmodel.requestProcessing()){
                return;
            }
            viewmodel.rememberMe(userContext.user.keepMeLoggedIn = !viewmodel.rememberMe());
        }

        function togglePasswordVisibility() {
            viewmodel.isPasswordVisible(!viewmodel.isPasswordVisible());
            viewmodel.password.hasFocus(true);
        }

        function back() {
            router.navigate('login');
        }

        function register() {
            viewmodel.requestProcessing(true);
            auth.register(userContext.user.email,
                    viewmodel.password(),
                    userContext.user.username,
                    userContext.user.keepMeLoggedIn)
                .then(function (response) {
                    return progressProvider.initProgressStorage(function(provider){
                        progressProvider.clearLocalStorage();
                        progressContext.use(provider);
                        return activateXapi(function () {
                            viewmodel.requestProcessing(false);
                            progressContext.restoreProgress();
                            if (context.isInReviewAttemptMode()) {
                                xApiInitializer.deactivate();
                            }
                            eventManager.courseStarted();
                        });
                    });
                }).fail(function (reason) {
                    viewmodel.requestProcessing(false);
                    if (reason.status == 409) {
                        router.navigate('signin');
                    }
                });
        }

        function activateXapi(callback) {
            if (xApiInitializer.isInitialized) {
                return xApiInitializer.activate(userContext.user).then(callback);
            }
            callback();
        }
    });
define('account/signin/index',['knockout', 'plugins/router', 'context', 'userContext', '../header/index', '../helpers/validatedValue', 'templateSettings',
        'modules/progress/progressStorage/auth', 'xApi/xApiInitializer', 'modules/progress/index', 'progressContext', 'eventManager'
],
    function (ko, router, context, userContext, Header, validatedValue, templateSettings, auth,
        xApiInitializer, progressProvider, progressContext, eventManager) {
        'use strict';

        var whitespaceRegex = /\s/;

        var viewmodel = {
            activate: activate,
            header: null,
            password: null,
            email: null,
            back: back,
            rememberMe: ko.observable(false),
            toggleRememberMe: toggleRememberMe,
            submit: submit,
            sendSecretLink: sendSecretLink,
            isSecretLinkSent: ko.observable(false),
            isPasswordVisible: ko.observable(false),
            togglePasswordVisibility: togglePasswordVisibility,
            emailPasswordCombination: ko.observable(false),
            requestProcessing: ko.observable(false),
            isRestorePasswordEmailSent: ko.observable(false),
            forgotPassword: forgotPassword
        };

        viewmodel.password = validatedValue(function (value) {
            return value().length >= 7 && !whitespaceRegex.test(value());
        });

        return viewmodel;

        function activate() {
            viewmodel.header = new Header(context.course.title);
            viewmodel.email = userContext.user.email;
        }

        function back() {
            router.navigate('login');
        }

        function toggleRememberMe() {
            if (viewmodel.requestProcessing()) {
                return;
            }
            viewmodel.rememberMe(userContext.user.keepMeLoggedIn = !viewmodel.rememberMe());
        }

        function submit() {
            if (viewmodel.requestProcessing()) {
                return;
            }
            if (!viewmodel.password.isValid()) {
                viewmodel.password.markAsModified();
                return;
            }
            viewmodel.requestProcessing(true);
            auth.signin(userContext.user.email,
                    viewmodel.password(),
                    userContext.user.keepMeLoggedIn)
                .then(function (response) {
                    return progressProvider.initProgressStorage(function (provider) {
                        progressProvider.clearLocalStorage();
                        progressContext.use(provider);
                        return activateXapi(function () {
                            viewmodel.requestProcessing(false);
                            progressContext.restoreProgress();
                            if (context.isInReviewAttemptMode()) {
                                xApiInitializer.deactivate();
                            }
                            eventManager.courseStarted();
                        });
                    });
                }).fail(function (reason) {
                    viewmodel.requestProcessing(false);
                    if (reason.status == 401) {
                        viewmodel.emailPasswordCombination(true);
                    }
                })
        }

        function togglePasswordVisibility() {
            viewmodel.isPasswordVisible(!viewmodel.isPasswordVisible());
            viewmodel.password.hasFocus(true);
        }

        function activateXapi(callback) {
            if (xApiInitializer.isInitialized) {
                return xApiInitializer.activate(userContext.user).then(callback);
            }
            callback();
        }

        function sendSecretLink() {
            if (viewmodel.requestProcessing() || viewmodel.isSecretLinkSent()) {
                return;
            }
            auth.sendSecretLink(context.course.id).then(function () {
                viewmodel.isRestorePasswordEmailSent(false);
                viewmodel.isSecretLinkSent(true);
                toggleValueAfterSomeTime(viewmodel.isSecretLinkSent, 5000);
            });
        }

        function forgotPassword() {
            if (viewmodel.requestProcessing() || viewmodel.isRestorePasswordEmailSent()) {
                return;
            }
            auth.forgotpassword(userContext.user.email).then(function () {
                viewmodel.isSecretLinkSent(false);
                viewmodel.isRestorePasswordEmailSent(true);
                toggleValueAfterSomeTime(viewmodel.isRestorePasswordEmailSent, 5000);
            });
        }

        function toggleValueAfterSomeTime(value, time) {
            _.delay(function () {
                value(!value());
            }, time);
        }
    });
define('dialogs/continueLater',[
  "modules/progress/progressStorage/auth",
  "context",
  "userContext",
  "templateSettings",
  "includedModules/modulesInitializer",
  "modules/publishModeProvider"
], function(
  auth,
  context,
  userContext,
  templateSettings,
  modulesInitializer,
  publishModeProvider
) {
  var viewModel = {
    activate: activate,
    progressStorageActivated: false,
    email: "",
    sendSecretLink: sendSecretLink,
    isSecretLinkSent: ko.observable(false),
    keepMeLoggedIn: ko.observable(false),
    toggleKeepMeLoggedIn: toggleKeepMeLoggedIn,
    isSecondRemarkShouldBeShown: true
  };

  return viewModel;

  function sendSecretLink() {
    auth.sendSecretLink(context.course.id).then(function() {
      viewModel.isSecretLinkSent(true);
    });
  }

  function toggleKeepMeLoggedIn() {
    viewModel.keepMeLoggedIn(
      (userContext.user.keepMeLoggedIn = !viewModel.keepMeLoggedIn())
    );
    auth.shortTermAccess = !userContext.user.keepMeLoggedIn;
  }

  function activate(data) {
    if (data) {
      viewModel.close = data.close;
      viewModel.exit = function() {
        if (!userContext.user.keepMeLoggedIn) {
          auth.signout().then(function() {
            data.exit();
          });
        } else {
          data.exit();
        }
      };
      viewModel.progressStorageActivated =
        templateSettings.allowCrossDeviceSaving &&
        !publishModeProvider.isScormEnabled;
      viewModel.isSecondRemarkShouldBeShown = !(
        templateSettings.allowCrossDeviceSaving ||
        publishModeProvider.isScormEnabled
      );
      viewModel.email = userContext.user.email;
      viewModel.keepMeLoggedIn(userContext.user.keepMeLoggedIn);
    }
  }
});

define('dialogs/dialog',["plugins/router"], function(router) {
    function Dialog() {
        var that = this;
        this.view = null;
        this.isVisible = ko.observable(false);
        this.callbacks = {};
        this.show = function(callbacks) {
            if (router.isNavigationLocked()) {
                return;
            }

            if (callbacks) that.callbacks = callbacks;

            that.isVisible(true);
        };

        this.hide = function() {
            that.isVisible(false);
            if (_.isFunction(that.callbacks.closed)) {
                that.callbacks.closed();
            }

            that.view &&
                that.view.removeEventListener("keydown", handleKeyDown);
        };

        this.compositionComplete = function(child) {
            that.view = child;
            that.view.addEventListener("keydown", handleKeyDown);

            var focusableState = getFocusableState(that.view);
            if (!focusableState.currentFocusable) {
                focusableState.firstFocusable &&
                    focusableState.firstFocusable.focus();
            }
        };

        function handleKeyDown(e) {
            var KEY_TAB = 9;

            var focusableState = getFocusableState(that.view);

            switch (e.keyCode) {
                case KEY_TAB:
                    if (focusableState.entries.length === 1) {
                        e.preventDefault();
                        break;
                    }
                    if (e.shiftKey) {
                        if (
                            document.activeElement ===
                            focusableState.firstFocusable
                        ) {
                            e.preventDefault();
                            focusableState.lastFocusable.focus();
                        }
                    } else {
                        if (
                            document.activeElement ===
                            focusableState.lastFocusable
                        ) {
                            e.preventDefault();
                            focusableState.firstFocusable.focus();
                        }
                    }
                    break;
                default:
                    break;
            }
        }

        function getFocusableState(view) {
            var result = {};
            result.entries = [].slice.call(
                view.querySelectorAll(
                    "a[href], area[href], input:not([disabled]), select:not([disabled]), textarea:not([disabled]), button:not([disabled]), [contenteditable]"
                )
            );

            result.firstFocusable = result.entries[0];
            result.lastFocusable = result.entries[result.entries.length - 1];

            result.currentFocusable = null;

            for (var i = 0; i < result.entries.length; i++) {
                if (result.entries[i] === document.activeElement) {
                    result.currentFocusable = result.entries[i];
                    break;
                }
            }

            return result;
        }
    }

    return Dialog;
});

define('dialogs/nps/figures/evaluation',['templateSettings'], function(templateSettings) {
    return {
        textColor: templateSettings.getColorValue('@text-color'),
        correctColor: templateSettings.getColorValue('@correct-color'),
        contentBodyColor: templateSettings.getColorValue('@content-body-color'),
        buttonTextColor: templateSettings.getColorValue('@button-text-color')
    };
});
define('dialogs/nps/figures/feedback',['templateSettings'], function(templateSettings) {
    var viewModel = {
        score: 0,
        feedbackColor: '',
        textColor: templateSettings.getColorValue('@text-color'),
        activate: activate
    };

    function activate(score) {
        var colorName = score > 6 ? '@correct-color' : '@secondary-color';
        viewModel.feedbackColor = templateSettings.getColorValue(colorName);
        viewModel.score = score;
    }

    return viewModel;
});
define('dialogs/nps/figures/rocket',['templateSettings'], function(templateSettings) {
    return {
        textColor: templateSettings.getColorValue('@text-color'),
        mainColor: templateSettings.getColorValue('@main-color'),
        contentBodyColor: templateSettings.getColorValue('@content-body-color')
    };
});
define('repositories/courseRepository',['context'], function (context) {

    return {
        get: get
    };

    function get() {
        return _.isObject(context.course) ? context.course : null;
    }

});
define('dialogs/nps/nps',['knockout', 'underscore', 'repositories/courseRepository', 'templateSettings', 'browserSupport'],
    function(ko, _, courseRepository, templateSettings, browserSupport) {
        var steps = {
            evaluation: 'evaluation',
            feedback: 'feedback',
            end: 'end'
        };

        var course = courseRepository.get();

        var viewModel = {
            activate: activate,
            submit: submit,
            goToFeedbackStep: goToFeedbackStep,
            score: ko.observable(0),
            feedback: ko.observable(''),
            isFeedbackEditing: ko.observable(false),
            currentStep: ko.observable(),
            callbacks: {},
            compositionComplete: compositionComplete,
            isCompositionComplete: ko.observable(false),
            evaluationStepTitle: '',
            isCourseEmbedded: window.self !== window.top,
            isReporting: ko.observable(false)
        };

        viewModel.isEvaluationStep = ko.computed(function() {
            return viewModel.currentStep() === steps.evaluation;
        });

        viewModel.isFeedbackStep = ko.computed(function() {
            return viewModel.currentStep() === steps.feedback;
        });

        viewModel.isLastStep = ko.computed(function() {
            return viewModel.currentStep() === steps.end;
        });

        return viewModel;

        function activate(data) {
            if (!data.close)
                throw 'Nps dialog activation data close() method is not specified';

            viewModel.close = data.close;
            if (data.callbacks)
                viewModel.callbacks = data.callbacks;

            viewModel.score(0);
            viewModel.feedback('');
            viewModel.currentStep(steps.evaluation);
            viewModel.isReporting(false);
            viewModel.isFeedbackEditing(false);

            var evaluationStepTitleKey = templateSettings.xApi.enabled ? '[progress has been reported]' : '[you have finished your course]';
            viewModel.evaluationStepTitle = TranslationPlugin.getTextByKey(evaluationStepTitleKey);
        }

        function compositionComplete() {
            viewModel.isCompositionComplete(true);
        }

        function goToFeedbackStep() {
            viewModel.currentStep(steps.feedback);
            if (!browserSupport.isMobileDevice) {
                viewModel.isFeedbackEditing(true);
            }
        }

        function submit() {
            viewModel.isReporting(true);

            course.evaluate({
                score: viewModel.score() / 10,
                response: viewModel.feedback().trim()
            }, function() {
                viewModel.isReporting(false);
                viewModel.currentStep(steps.end);
                if (_.isFunction(viewModel.callbacks.finalized)) {
                    viewModel.callbacks.finalized();
                }
            });
        }
    });
    define('dialogs/resultsResend/resultsResend',['knockout', 'localizationManager', 'underscore', 'repositories/courseRepository'],
    function(ko, localizationManager, _, courseRepository) {
        var course = courseRepository.get();

        var viewModel = {
            next: {},
            resend: {},
            retryCount: 0,
            resultsSendErrorTitleText: '',
            endpointNameText: '',

            isCompositionComplete: ko.observable(false),
            isResendingResults: ko.observable(false),
            resendingFailed: ko.observable(false),
            retriesExceeded: ko.observable(false),

            activate: activate,
            compositionComplete: compositionComplete,
            resendResults: resendResults,
            skipResultsSending: skipResultsSending,
            hideDialog: hideDialog
        };
       
        return viewModel;

        function activate(data) {
            throwIfNotDefined(data.callbacks, 'callbacks');
            throwIfNotDefined(data.callbacks.resend, 'callbacks.resend');
            throwIfNotDefined(data.callbacks.next, 'callbacks.next');
            throwIfNotDefined(data.close, 'close');

            throwIfNotDefined(data.resultsSendErrorTitleKey, 'resultsSendErrorTitleKey');
            throwIfNotDefined(data.endpointNameKey, 'endpointNameKey');

            viewModel.resultsSendErrorTitleText = localizeText(data.resultsSendErrorTitleKey);
            viewModel.endpointNameText = localizeText(data.endpointNameKey);

            viewModel.resend = data.callbacks.resend;
            viewModel.next = data.callbacks.next;
            viewModel.close = data.close;

            viewModel.retryCount = data.retryCount || 10;

        }

        function compositionComplete() {
            viewModel.isCompositionComplete(true);
        }
        
        function resendResults() {
            viewModel.isResendingResults(true);

            viewModel.resend(course)
                .then(function() {
                    hideDialog();
                    viewModel.next();
                })
                .catch(function() {
                    viewModel.retryCount ? viewModel.retryCount-- : viewModel.retriesExceeded(true);
                    
                    viewModel.isResendingResults(false);
                    animateNates();
                });
        }

        function skipResultsSending() {
            hideDialog();
            viewModel.next();
        }

        function hideDialog() {
            viewModel.isResendingResults(false);
            viewModel.close();
        }

        function localizeText(key) {
            return localizationManager.getLocalizedText(key);
        }

        function throwIfNotDefined(parameter, parameterName) {
            if(_.isNullOrUndefined(parameter)) {
                throw 'Can\'t activate dialog. Parameter isn\'t defined. Parameter name: ' + parameterName;
            }
        }

        function animateNates() {
            viewModel.resendingFailed(true);
            setTimeout(function() {
                viewModel.resendingFailed(false);
            }, 400);
        }
    });
define('dialogs/startNewAttempt/startNewAttempt',['knockout', 'progressContext'],
    function(ko, progressContext) {
        var viewModel = {
            activate: activate,
            callbacks: {},
            compositionComplete: compositionComplete,
            isCompositionComplete: ko.observable(false),
            isStartingNewAttempt: ko.observable(false),
            startNewAttempt: startNewAttempt,
            hide: hide
        };
       
        return viewModel;

        function activate(data) {
            if (!data.close)
                throw 'Start new attempt dialog activation data close() method is not specified';

            viewModel.close = data.close;
            if (data.callbacks)
                viewModel.callbacks = data.callbacks;
        }

        function compositionComplete() {
            viewModel.isCompositionComplete(true);
        }

        function hide() {
            viewModel.isCompositionComplete(false);
            viewModel.close && viewModel.close();
        }
        
        function startNewAttempt() {
            viewModel.isStartingNewAttempt(true);
            progressContext.remove(function() {
                window.location = window.location.href.split("#")[0];
                viewModel.isStartingNewAttempt(false);
            });
        }
    });
define('errorsHandler',['durandal/app', 'plugins/router', 'constants'],
    function (app, router, constants) {
        "use strict";

        var lastError;
        var started = false;

        function handleError(error) {
            if (error && error.navigateUrl) {
                setTimeout(function() {
                    router.navigate(error.navigateUrl, { replace: true, trigger: true });
                }, 100);
            }
        }

        function startHandle() {
            started = true;
            handleError(lastError);
        };

        app.on(constants.events.onError).then(function (error) {
            if (started) {
                handleError(error);
            } else {
                lastError = error;
            }
        });

        return {
            startHandle: startHandle
        }
    }
);
define('helpers/appOperations',['durandal/app', 'constants'],
    function(app, constants) {

        "use strict";

        var
            appOperations = {
                close: close
            };

        return appOperations;

        function close(spec) {
            spec = spec || { shouldCloseWindow: true };
            if (spec.shouldCloseWindow && window.opener && window !== window.opener) {
                window.close();
            }
            app.trigger(constants.events.appClosed);
        }
    }
);
define('modules/lti',['eventManager', 'modules/progress/progressStorage/auth', 'helpers/requestResender', 
    'errorTracking/errorTracker'],
    function (eventManager, auth, requestResender, errorTracker) {
        'use strict';
        
        var constants = {
            resultCallbackUrlParameterName: 'ltiResultCallbackUrl',
            errorMessage: 'Something went wrong and your final score has not been reported ' + 
                '({reason}). Please contact the author of the course.'
        };

        return {
            initialize: initialize
        };

        function initialize() {
            var resultCallbackUrl = auth.getValueFromUrl(constants.resultCallbackUrlParameterName);
            if (!resultCallbackUrl) {
                return;
            }

            eventManager.subscribeForEvent(eventManager.events.courseFinished)
                .then(onCourseFinished.bind(this, resultCallbackUrl));
        }

        function onCourseFinished(resultCallbackUrl, course) {
            var requestOptions = {
                url: resultCallbackUrl,
                method: 'POST',
                dataType: 'json',
                xhrFields: {
                    withCredentials: true
                },
                data: {
                    score: course.score() / 100
                }
            };

            var token = auth.getToken();
            if (token) {
                requestOptions.headers = {
                    Authorization: 'Bearer ' + token
                };
            }

            return requestResender.send(requestOptions, onError);
        }

        function onError(response) {
            var errorMessage = 'LTI reporting failed. Code: ' + response.status + ', Reason: ' + response.responseText;

            errorTracker.trackError(new Error(errorMessage));
            
            return constants.errorMessage.replace('{reason}', errorMessage);
        }
    }
);
define('modules/progress/progressStorage/certificateProvider',['./urlProvider', 'requester/httpWrapper', './auth', 'templateSettings', 'helpers/idToUuid'], function (
  urlProvider,
  httpWrapper,
  auth,
  templateSettings,
  idToUuid
) {
  'use strict';

  function CertificateProvider() {
    this.getCertificateUrl = function (courseId, templateId, logoUrl) {
      var headers = auth.headers;
      headers['Authorization'] = 'Bearer ' + auth.getToken();
      headers['X-UI-Culture'] = getCulture();

      var urlCheckPattern = /(http|https):\/\/(\w+:{0,1}\w*)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%!\-\/]))?/;

      var newLogoUrl = function () {
        if (logoUrl && !urlCheckPattern.test(logoUrl)) {
          if (logoUrl.match(/\/\//g)) {
            return 'https:' + logoUrl;
          }
          return 'https://' + logoUrl;
        } else {
          return logoUrl;
        }
      };

      return httpWrapper
        .get({
          url:
            urlProvider.learnServiceUrl +
            '/api/learner/me/courses/' +
            idToUuid(courseId) +
            '/attempts/last/certificate?templateId=' +
            templateId +
            '&logoUrl=' +
            newLogoUrl(),
          headers: headers
        })
        .then(function (response) {
          return response.url;
        });
    };
  }

  //todo: remove this temp fix
  function getCulture() {
    return templateSettings.languages.selected === 'ua' ? 'uk' : templateSettings.languages.selected;
  }

  return new CertificateProvider();
});

define('modules/questionsNavigation',['repositories/sectionRepository', 'plugins/router', 'underscore', 'modules/publishModeProvider'],
    function (sectionRepository, router, _, publishModeProvider) {
        function getNavigationContext(sectionId, questionId) {
            var section = sectionRepository.get(sectionId);

            if (section && section.questions) {
                var currentItemIndex = _getItemIndexById(section.questions, questionId);
                if (currentItemIndex > -1) {
                    return {
                        previousQuestionUrl: _getQuestionUrl(section, section.questions[currentItemIndex - 1]),
                        currentQuestionUrl: _getQuestionUrl(section, section.questions[currentItemIndex]),
                        nextQuestionUrl: _getQuestionUrl(section, section.questions[currentItemIndex + 1]),                        
                        questionsCount: section.questions.length,
                        currentQuestionIndex: currentItemIndex + 1,
                        nextSectionUrl: getNextSectionUrl(section)
                    };
                }
            }
        }

        function redirectToQuestion() {
            if (!publishModeProvider.isPreview) {
                return;
            }

            var questionId = router.getQueryStringValue('questionId');
            if (_.isNullOrUndefined(questionId)) {
                return;
            }

            var section = sectionRepository.getSectionByQuestionId(questionId);
            if (_.isNullOrUndefined(section)) {
                return;
            }

            window.location.hash = '#section/' + section.id + '/question/' + questionId;
        }

        function _getItemIndexById(collection, itemId) {
            for (var i = 0, count = collection.length; i < count; i++) {
                if (collection[i].id === itemId) {
                    return i;
                }
            }
            return -1;
        }

        function _getQuestionUrl(section, question) {
            if (section && question) {
                return '#section/' + section.id + '/question/' + question.id;
            }

            return undefined;
        }

        function getNextSectionUrl(section) {
            if (!section) {
                return;
            }

            var nextSection = getNextSection(section);
            if (nextSection) {
                return  _getQuestionUrl(nextSection, nextSection.questions[0]);
            }
        }

        function getNextSection(currentSection) {
            var sections = sectionRepository.getCollection();

            for (var i = 0; i < sections.length; i++) {
                if (sections[i].id === currentSection.id) {
                    return sections[i + 1];
                }
            }
        }

        return {
            getNavigationContext: getNavigationContext,
            redirectToQuestion: redirectToQuestion
        };
    });
define('modules/webhooks',['underscore', 'eventManager', 'userContext'], function(_, eventManager, userContext){
    'use strict';

    function WebhooksModule () {
        this.initialized = false;
        this.url = '';
    }

    WebhooksModule.prototype.initialize = function (webhooksSettings) {
        if (this.initialized || !webhooksSettings.url) return;

        this.url = webhooksSettings.url;
        this.initialized = true;
    }

    WebhooksModule.prototype.sendResults = function (course) {
        var defer = Q.defer(),
            user = userContext.getCurrentUser();

        if(!_.isNullOrUndefined(user)) {
            var data = {
                courseId: course.id,
                learnerId: user.email,
                score: course.result() * 100,
                finishedOn: (new Date()).toISOString(),
                status: course.getStatus() === 'completed' ? 'passed' : 'failed'
            };

            fetch(this.url, {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json'
                },
                body: JSON.stringify(data)
            }).then(function(response) {
                if(response.status !== 200) {
                    defer.reject();
                }

                defer.resolve();
            });
        } else {
            defer.resolve();
        }

        return defer.promise;
    }

    return new WebhooksModule();
});

define('viewmodels/questions/multipleSelect/multipleSelect',[],function () {
	"use strict";

	function MultipleSelect() {
		var self = this;

		this.question = null;

		this.questionInstructions = [];
	    this.isAnswered = ko.observable(false);
	    this.answers = null;
		this.isSurveyModeEnabled = false;
	};

	MultipleSelect.prototype.checkItem = function(item) {
		if (this.isAnswered()) {
			return;
		}

		item.isChecked(!item.isChecked());
	};

	MultipleSelect.prototype.initialize = function(question, isPreview) {
		var self = this;

		return Q.fcall(function () {
			self.question = question;

			self.isSurveyModeEnabled = !!question.isSurvey;
			self.questionInstructions = question.questionInstructions;
			self.isAnswered(question.isAnswered);
			self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);

			self.answers = _.map(question.answers, function (answer) {
				return {
					id: answer.id,
					text: answer.text,
					isChecked: ko.observable(answer.isChecked)
				};
			});
		});
	};
	
	MultipleSelect.prototype.submit = function() {
		var self = this;

		return Q.fcall(function () {
			self.question.submitAnswer(
				_.chain(self.answers)
				.filter(function (item) {
					return item.isChecked();
				})
				.map(function (item) {
					return item.id;
				}).value());

			self.isAnswered(true);
		});
	};

	MultipleSelect.prototype.tryAnswerAgain = function() {
		var self = this;

		return Q.fcall(function() {
			_.each(self.answers, function (answer) {
				answer.isChecked(false);
			});

			self.isAnswered(false);
		});
	};

	return MultipleSelect;
});
define('viewmodels/questions/singleSelectText/singleSelectText',[],function () {
    "use strict";

    function SingleSelectText() {
        this.question = null;

        this.questionInstructions = [];
        this.isAnswered = ko.observable(false);
        this.answers = null;
        this.isSurveyModeEnabled = false;
    };   

    SingleSelectText.prototype.checkItem = function (item) {
            if (this.isAnswered()) {
                return;
            }

            _.each(this.answers, function(answer) {
                answer.isChecked(false);
            });

            item.isChecked(true);
        };
        
    SingleSelectText.prototype.initialize = function (question, isPreview) {
        var self = this;

        return Q.fcall(function () {
            self.question = question;

            self.isSurveyModeEnabled = !!question.isSurvey;
            self.questionInstructions = question.questionInstructions;
            self.isAnswered(question.isAnswered);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);

            self.answers = _.map(question.answers, function (answer) {
                return {
                    id: answer.id,
                    text: answer.text,
                    isChecked: ko.observable(answer.isChecked)
                };
            });
        });
    };

    SingleSelectText.prototype.submit = function() {
        var self = this;
        
        return Q.fcall(function () {
            self.question.submitAnswer(
                _.chain(self.answers)
                .filter(function (item) {
                    return item.isChecked();
                })
                .map(function (item) {
                    return item.id;
                }).value());

            self.isAnswered(true);
        });
    };

    SingleSelectText.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function () {
            _.each(self.answers, function (answer) {
                answer.isChecked(false);
            });

            self.isAnswered(false);
        });
    };

    return SingleSelectText;
});
define('viewmodels/questions/fillInTheBlank/fillInTheBlank',['knockout', 'requester/resourceLoader'], function (ko, resourceLoader) {
  'use strict';

  function FillInTheBlank() {
    this.question = null;

    this.content = null;
    this.loadContent = loadContent;

    this.isAnswered = ko.observable(false);
    this.inputValues = ko.observableArray([]);
  }

  FillInTheBlank.prototype.initialize = function (question, isPreview) {
    var self = this;

    return question.load().then(function () {
      self.question = question;

      self.isAnswered(question.isAnswered);
      self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
      self.inputValues(
        _.map(question.answerGroups, function (answerGroup) {
          return {
            id: answerGroup.id,
            value: answerGroup.answeredText,
            answers: answerGroup.answers
          };
        })
      );

      if (self.question.hasContent) {
        return self.loadContent();
      }
    });
  };

  FillInTheBlank.prototype.submit = function () {
    var self = this;

    return Q.fcall(function () {
      self.question.submitAnswer(self.inputValues());

      self.isAnswered(true);
    });
  };

  FillInTheBlank.prototype.tryAnswerAgain = function () {
    var self = this;

    return Q.fcall(function () {
      _.each(self.inputValues(), function (blankValue) {
        blankValue.value = '';
      });
      self.isAnswered(false);
    });
  };

  return FillInTheBlank;

  function loadContent() {
    var self = this;
    return Q.fcall(function () {
      var contentUrl = 'content/' + self.question.sectionId + '/' + self.question.id + '/content.html';
      return resourceLoader
        .getLocalResource({ url: contentUrl })
        .then(function (response) {
          self.content = response;
          self.question.content = response;
        })
        .fail(function () {
          self.content = '';
        });
    });
  }
});

define('viewmodels/questions/dragAndDrop/dragAndDrop',[],function () {
    "use strict";

    function DragAndDrop() {
        this.question = null;
        this.questionInstructions = [];

        this.dropspots = [];
        this.texts = [];

        this.isPreview = ko.observable(true);

        this.isAnswered = ko.observable(false);
    };

    DragAndDrop.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            var answer = [];

            _.each(self.dropspots, function (dropspot) {
                var text = dropspot.text();
                if (text) {
                    answer.push({
                        text: text.text,
                        x: dropspot.x,
                        y: dropspot.y
                    });
                }
            });

            self.question.submitAnswer(answer);
            self.isAnswered(true);
        });
    };

    DragAndDrop.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function () {
            self.isAnswered(false);
            _.each(self.dropspots, function (dropspot) {
                dropspot.text(undefined);
            });
            _.each(self.texts, function (answer) {
                answer.dropSpot = {};
                answer.placed(false);
            });
        });
    };

    DragAndDrop.prototype.initialize = function(question, isPreview) {
        var self = this;
        
        return Q.fcall(function () {
            self.question = question;
            self.questionInstructions = question.questionInstructions;
            self.isPreview(_.isUndefined(isPreview) ? false : isPreview);

            self.isAnswered(question.isAnswered);
            self.dropspots = _.map(question.answers, function (answer) {
                return {
                    x: answer.correctPosition.x,
                    y: answer.correctPosition.y,
                    text: ko.observable(''),
                    
                };
            });

            self.texts = _.map(_.shuffle(question.answers), function (answer) {
                return {
                    id: answer.id,
                    text: answer.text,
                    placed: ko.observable(false)
                };
            });

            _.each(question.answers, function (answer) {
                var selectedDropspot = _.find(self.dropspots, function (dropspot) {
                    return dropspot.x === answer.currentPosition.x
                        && dropspot.y === answer.currentPosition.y;
                });

                if (selectedDropspot) {
                    var selectedText = _.find(self.texts, function (item) {
                        return item.id === answer.id;
                    });
                    selectedText.placed(true);
                    selectedDropspot.text(selectedText);
                }
            });
        });
    };

    return DragAndDrop;
});
define('viewmodels/questions/singleSelectImage/singleSelectImage',['browserSupport'], function (browserSupport) {
    "use strict";

    function SingleSelectImage() {
        var self = this;

        this.canHover = !browserSupport.isMobileDevice;
        this.question = null;

        this.questionInstructions = [];
        this.isAnswered = ko.observable(false);
        this.answers = null;
        this.checkedAnswerId = ko.observable(null);
    };

    SingleSelectImage.prototype.checkItem = function(item) {
            if (this.isAnswered()) {
                return;
            }
            
            this.checkedAnswerId(item.id);
        };

    SingleSelectImage.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function () {
            self.question = question;
            self.checkedAnswerId(question.checkedAnswerId);

            self.questionInstructions = question.questionInstructions;
            self.isAnswered(question.isAnswered);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
            self.answers = _.map(question.answers, function (answer) {
                return {
                    id: answer.id,
                    image: answer.image,
                    isChecked: ko.observable(answer.isChecked)
                };
            });
        });
    };

    SingleSelectImage.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            self.question.submitAnswer(self.checkedAnswerId());

            self.isAnswered(true);
        });
    };

    SingleSelectImage.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function () {
            self.checkedAnswerId(null);

            self.isAnswered(false);
        });
    };

    return SingleSelectImage;
});
define('viewmodels/questions/textMatching/textMatchingSource',[], function () {
    return function (id, key) {

        this.id = id;
        this.key = key;
        this.value = ko.observable();

        this.acceptValue = function (value) {
            this.value(value);
        };
        this.rejectValue = function () {
            this.value(null);
        };
    };
});
define('viewmodels/questions/textMatching/textMatchingTarget',[],function () {

    return function (value) {
        this.value = ko.observable(value);

        this.acceptValue = function (value) {
            this.value(value);
        };
        this.rejectValue = function () {
            this.value(null);
        };
    };

});
define('viewmodels/questions/textMatching/textMatching',['viewmodels/questions/textMatching/textMatchingSource', 'viewmodels/questions/textMatching/textMatchingTarget'], function (Source, Target) {

    function TextMatching() {
        this.question = null;
        this.questionInstructions = [];
        this.isAnswered = ko.observable(false);

        this.values = [];

        this.sources = ko.observableArray([]);
        this.targets = ko.observableArray([]);
    };

    TextMatching.prototype.acceptValue = function (value) {
        this.targets.push(value);
    };


    TextMatching.prototype.rejectValue = function (value) {
        this.targets.remove(value);
    };

    TextMatching.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function () {
            self.question = question;

            self.questionInstructions = question.questionInstructions;
            self.isAnswered(question.isAnswered);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);

            self.values = _.chain(question.answers)
                .map(function (answer) {
                    return answer.value;
                })
                .shuffle()
                .value();

            var targets = [];

            _.each(self.values, function (value) {
                targets.push(new Target(value));
            });

            var sources = [];
            _.each(question.answers, function (pair) {
                var source = new Source(pair.id, pair.key);
                var target = _.find(targets, function (target) {
                    return target.value() == pair.attemptedValue;
                });

                if (target) {
                    source.acceptValue(target.value());
                    target.rejectValue();
                } else {
                    source.value(null);
                }

                sources.push(source);
            });

            self.targets(targets);
            self.sources(_.shuffle(sources));
        });
    };

    TextMatching.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            self.question.submitAnswer(_.map(self.sources(), function (source) {
                var value = source.value() ? source.value() : null;
                return { id: source.id, value: value };
            }));
            self.isAnswered(true);
        });
    };

    TextMatching.prototype.tryAnswerAgain = function() {
        var self = this;

        return Q.fcall(function () {
            _.each(self.sources(), function (pair, index) {
                pair.rejectValue();
                self.targets()[index].acceptValue(self.values[index]);
            });

            self.isAnswered(false);
        });
    };

    return TextMatching;
});
define('viewmodels/questions/statement/statement',[],function() {
    "use strict";

    function Statement() {
        this.question = null;
        this.questionInstructions = [];
        this.statements = null;
        this.isAnswered = ko.observable(false);
        this.isSurveyModeEnabled = false;
    };

    Statement.prototype.markStatementAsTrue = function(statement) {
        if (this.isAnswered()) {
            return;
        }

        statement.userAnswer(true);
    };

    Statement.prototype.markStatementAsFalse = function(statement) {
        if (this.isAnswered()) {
            return;
        }

        statement.userAnswer(false);
    };

    Statement.prototype.initialize = function(question, isPreview) {
        this.question = question;
        this.isSurveyModeEnabled = !!question.isSurvey;
        this.questionInstructions = question.questionInstructions;
        this.isAnswered(question.isAnswered);
        this.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
        this.statements = _.map(question.statements, function(statement) {
            var result = {
                id: statement.id,
                text: statement.text,
                userAnswer: ko.observable(statement.userAnswer)
            };

            result.isTrueChecked = ko.computed(function() {
                return result.userAnswer() == true;
            });

            result.isFalseChecked = ko.computed(function() {
                return result.userAnswer() == false;
            });

            return result;
        });
    };

    Statement.prototype.submit = function() {
        var self = this;

        return Q.fcall(function() {
            var userAnswers = _.chain(self.statements)
                .filter(function(statement) {
                    return !_.isNullOrUndefined(statement.userAnswer());
                })
                .map(function(statement) {
                    return { id: statement.id, answer: statement.userAnswer() };
                })
                .value();

            self.question.submitAnswer(userAnswers);
            self.isAnswered(true);
        });
    };

    Statement.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function() {
            self.isAnswered(false);
            _.each(self.statements, function(statement) {
                statement.userAnswer(null);
            });
        });
    };

    return Statement;
});
define('viewmodels/questions/hotspot/hotspot',[],function () {
    "use strict";

    function HotSpot (){ }

    HotSpot.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function (){
            self.questionInstructions = question.questionInstructions;
            self.background = question.background;
            self.isAnswered = ko.observable(question.isAnswered);
            self.isMultiple = question.isMultiple;
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
            self.question = question;

                self.submit = function () {         
                return Q.fcall(function () {
                        question.submitAnswer(self.marks());             
                        self.isAnswered(true);
                });
                };
        
                self.marks = ko.observableArray(question.placedMarks ? _.map(question.placedMarks, function (mark) { return { x: mark.x, y: mark.y }; }) : []);

            self.addMark = function (mark) {
                if (!self.isMultiple) {
                    self.marks.removeAll();
                }
                self.marks.push(mark);
            };

            self.removeMark = function (mark) {
                self.marks.remove(mark);
            };

            self.tryAnswerAgain = function () {
                return Q.fcall(function () {
                    self.isAnswered(false);
                    self.marks.removeAll();
                });
            };
        });
    };
    
    return HotSpot;
});
define('viewmodels/questions/openQuestion/openQuestion',[], function() {
    "use strict";

    function OpenQuestion(){
        this.question = null;
        this.questionInstructions = [];
        this.answeredText = ko.observable('');
        this.isAnswered = ko.observable(false);
        
        this.feedbackView = 'questions/openQuestion/feedback.html';  
    };

    OpenQuestion.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function() {
            self.question = question;
            self.questionInstructions = question.questionInstructions;
            self.answeredText(question.answeredText);
            self.isAnswered(question.isAnswered);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
        });
    };

    OpenQuestion.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            self.question.submitAnswer(self.answeredText().trim());
            self.isAnswered(true);
        });
    };

    OpenQuestion.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function () {
            self.answeredText('');
            self.isAnswered(false);
        });
    };

    return OpenQuestion;
});
define('viewmodels/questions/scenarioQuestion/components/branchtrackProvider',['knockout'], function(ko) {
    'use strict';
    var instance = null;

    return {
        isFinished: isFinished,
        destroy: destroy,
        reset: reset,
        init: init,
        score: getScore
    };

    function isFinished() {
        if (!_.isNullOrUndefined(instance)) {
            return instance.isFinished;
        }

        return false;
    }

    function destroy() {
        if (!_.isNullOrUndefined(instance)) {
            instance.destroy();
        }
    }

    function reset() {
        if (!_.isNullOrUndefined(instance)) {
            instance.reset();
        }
    }

    function getScore() {
        if (!_.isNullOrUndefined(instance)) {
            return instance.score;
        }

        return 0;
    }

    function init(instanceId) {
        if (!_.isEmptyOrWhitespace(instanceId)) {
            instance = Branchtrack.create(instanceId);
        }
    }
});
define('viewmodels/questions/scenarioQuestion/scenarioQuestion',['knockout', 'viewmodels/questions/scenarioQuestion/components/branchtrackProvider'], function (ko, branchtrackProvider) {
    "use strict";

    function ScenarioQuestion(){
        this.question = null;
        this.questionInstructions = [];
        this.embedCode = ko.observable(null);
        this.isAnswered = ko.observable(false);
        this.isCorrect = ko.observable(false);
        
        this.feedbackView = 'questions/scenarioQuestion/feedback.html';
        this.customSubmitViewModel = 'viewmodels/questions/scenarioQuestion/submitQuestion';
    };

    ScenarioQuestion.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function () {
            self.question = question;
            self.questionInstructions = question.questionInstructions;
            self.embedCode(question.embedCode);
            self.isAnswered(question.isAnswered);
            self.isCorrect(question.isCorrectAnswered);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
            branchtrackProvider.init(question.projectId, question.masteryScore);
        });
    }

    ScenarioQuestion.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            self.question.submitAnswer(branchtrackProvider.score());
            self.isAnswered(true);
        });
    }

    ScenarioQuestion.prototype.tryAnswerAgain = function() {
        var self = this;
        
        return Q.fcall(function () {
            self.isAnswered(false);
            self.embedCode.valueHasMutated();
            branchtrackProvider.reset();
        });
    }

    ScenarioQuestion.prototype.deactivate = function() {
        branchtrackProvider.destroy();
    }

    return ScenarioQuestion;
});
define('viewmodels/questions/rankingText/rankingText',[],function () {
    "use strict";

    function RankingText() {
        this.question = null;
        this.questionInstructions = [];

        this.isAnswered = ko.observable(false);
        this.rankingItems = ko.observableArray([]);
    };

    RankingText.prototype.initialize = function(question, isPreview) {
        var self = this;

        return Q.fcall(function () {
            self.question = question;

            self.questionInstructions = question.questionInstructions;
            self.isAnswered(question.isAnswered);

            self.rankingItems(question.rankingItems);
            self.isPreview = ko.observable(_.isUndefined(isPreview) ? false : isPreview);
        });
    };

    RankingText.prototype.submit = function() {
        var self = this;

        return Q.fcall(function () {
            self.question.submitAnswer(self.rankingItems());

            self.isAnswered(true);
        });
    };

    RankingText.prototype.tryAnswerAgain = function() {
        var self = this;

        return Q.fcall(function () {
            self.isAnswered(false);
        });
    };

    return RankingText;    
});
define('viewmodels/questions/questionsViewModelFactory',['constants', 'viewmodels/questions/multipleSelect/multipleSelect', 'viewmodels/questions/singleSelectText/singleSelectText', 'viewmodels/questions/fillInTheBlank/fillInTheBlank',
    'viewmodels/questions/dragAndDrop/dragAndDrop', 'viewmodels/questions/singleSelectImage/singleSelectImage', 'viewmodels/questions/textMatching/textMatching',
    'viewmodels/questions/statement/statement', 'viewmodels/questions/hotspot/hotspot', 'viewmodels/questions/openQuestion/openQuestion',
    'viewmodels/questions/scenarioQuestion/scenarioQuestion', 'viewmodels/questions/rankingText/rankingText'],
    function (constants, MultipleSelect, SingleSelectText, FillInTheBlank, DragAndDrop, SingleSelectImage,
        TextMatching, Statement, Hotspot, OpenQuestion, ScenarioQuestion, RankingText) {
        'use strict';

        return {
            getViewModel: getViewModel
        };

        function getViewModel(type) {
            switch (type) {
                case constants.questionTypes.multipleSelect:
                    return new MultipleSelect();
                case constants.questionTypes.dragAndDrop:
                    return new DragAndDrop();
                case constants.questionTypes.singleSelectText:
                    return new SingleSelectText();
                case constants.questionTypes.fillInTheBlank:
                    return new FillInTheBlank();
                case constants.questionTypes.singleSelectImage:
                    return new SingleSelectImage();
                case constants.questionTypes.textMatching:
                    return new TextMatching();
                case constants.questionTypes.statement:
                    return new Statement();
                case constants.questionTypes.hotspot:
                    return new Hotspot();
                case constants.questionTypes.openQuestion:
                    return new OpenQuestion();
                case constants.questionTypes.scenario:
                    return new ScenarioQuestion();
                case constants.questionTypes.rankingText:
                    return new RankingText();
                default:
                    return new MultipleSelect();
            }
        }

    });
define('questionContent/questionContent',['knockout', 'plugins/router', 'constants', 'modules/questionsNavigation', 'viewmodels/questions/questionsViewModelFactory', 'templateSettings', 'localizationManager', 'context'],
    function (ko, router, constants, navigationModule, questionViewModelFactory, templateSettings, localizationManager, context) {
        "use strict";

        function QuestionContent() {
            var self = this;
            
            this.sectionId = null;
            this.question = null;
            
            this.title = '';
            this.index = 0;
            this.isAnswered = ko.observable(false);
            this.isCorrect = ko.observable(false);
            this.isExpanded = ko.observable(true);
            this.isPreview = false;
            this.isSurvey = false;
            this.copyright = templateSettings.copyright;

            this.learningContents = [];
            this.questionInstructions = [];
            this.correctFeedback = [];
            this.incorrectFeedback = [];
            this.feedbackResultText = ko.observable('');
            this.feedbackView = '';
            this.submitViewModel = '';

            this.navigationContext = null;

            this.activeQuestionViewModel = null;
            
            this.isNavigationLocked = router.isNavigationLocked;

            this.isCorrectAnswered = ko.computed(function() {
                return self.isAnswered() && self.isCorrect();
            });

            this.isWrongAnswered = ko.computed(function () {
                return self.isAnswered() && !self.isCorrect();
            });

            // this logic moved here from html view to fix blinking bug in IE11
            this.updateFeedbackResultText = function() {
                var key = '[your answer was stored]';
                if (!self.isSurvey) {
                    key = self.isCorrect() ? '[correct answer]': '[incorrect answer]';
                }

                this.feedbackResultText(localizationManager.getLocalizedText(key));
            }

            this.hideTryAgain = false;

            this.isGoToNextSectionVisible = ko.observable(false);
            this.isGoToResultsVisible = ko.observable(false);
        };

        QuestionContent.prototype.submit = function() {
            var self = this;
            return self.activeQuestionViewModel.submit().then(function () {
                self.isCorrect(self.question.isCorrectAnswered);
                self.updateFeedbackResultText();
                self.isAnswered(self.question.isAnswered);

                updateNavigationState(self);
            });
        }

        QuestionContent.prototype.tryAnswerAgain = function () {	
            var self = this;

            return self.activeQuestionViewModel.tryAnswerAgain().then(function () {
                self.isAnswered(false);
            });
        }

        QuestionContent.prototype.navigateNext = function () {
            if (router.isNavigationLocked()) {
                return;
            }

            var nextUrl = !_.isNullOrUndefined(this.navigationContext.nextQuestionUrl) ? this.navigationContext.nextQuestionUrl : 'sections';
            router.navigate(nextUrl);
        }

        QuestionContent.prototype.goToNextSection = function () {
            if (router.isNavigationLocked()) {
                return;
            }
            if (!this.navigationContext.nextSectionUrl) {
                return;
            }
    
            router.navigate(this.navigationContext.nextSectionUrl);
        }

        QuestionContent.prototype.goToResults = function () {
            if (router.isNavigationLocked()) {
                return;
            }
    
            router.navigate('#finish');
        }
        
        QuestionContent.prototype.toggleExpand = function () {
            return this.isExpanded(!this.isExpanded());
        }

        QuestionContent.prototype.activate = function (sectionId, question, index, isPreview) {
            if (!sectionId || !question) {
                return;
            }


            this.isGoToResultsVisible(false);
            this.isGoToNextSectionVisible(false);

            this.sectionId = sectionId;
            this.question = question;
            this.isPreview = _.isUndefined(isPreview) ? false : isPreview;
            this.index = _.isUndefined(index) ? 0 : index < 10 ? '0' + index : index;
            this.navigationContext = navigationModule.getNavigationContext(this.sectionId, this.question.id);
            this.title = this.question.title;
            this.isAnswered(this.question.isAnswered);
            this.isCorrect(this.question.isCorrectAnswered);
            this.isSurvey = !!question.isSurvey;

            this.learningContents = this.question.learningContents;
            this.questionInstructions = this.question.questionInstructions;
            this.correctFeedback = this.question.correctFeedback;
            this.incorrectFeedback = this.question.incorrectFeedback;
            this.updateFeedbackResultText();

            this.activeQuestionViewModel = questionViewModelFactory.getViewModel(this.question.type);
            this.feedbackView = this.activeQuestionViewModel.feedbackView || 'questions/feedback.html';
            this.submitViewModel = this.activeQuestionViewModel.customSubmitViewModel || '';

            this.hideTryAgain = templateSettings.hideTryAgain || context.isInReviewAttemptMode();

            updateNavigationState(this);

            if(isPreview){
                var self = this;

                return this.question.load().then(function(){
                    return self.activeQuestionViewModel.initialize(self.question, isPreview);
                });
            } else {
                return this.activeQuestionViewModel.initialize(this.question, isPreview);
            }
        }

        QuestionContent.prototype.deactivate = function () {
            if (_.isFunction(this.activeQuestionViewModel.deactivate)) {
                this.activeQuestionViewModel.deactivate();
            }
        }

        function updateNavigationState(viewModel) {
            if (context.course.score() === 100) {
                viewModel.isGoToResultsVisible(!templateSettings.hideFinishActionButtons);
                viewModel.isGoToNextSectionVisible(viewModel.navigationContext.questionsCount === viewModel.navigationContext.currentQuestionIndex && viewModel.navigationContext.nextSectionUrl);
            } else {
                if (viewModel.navigationContext.questionsCount === viewModel.navigationContext.currentQuestionIndex) {
                    viewModel.isGoToResultsVisible(!viewModel.navigationContext.nextSectionUrl && !templateSettings.hideFinishActionButtons);
                    viewModel.isGoToNextSectionVisible(!!viewModel.navigationContext.nextSectionUrl);
                } else {
                    viewModel.isGoToResultsVisible(false);
                    viewModel.isGoToNextSectionVisible(false);
                }
            }
        }

        return QuestionContent;
    }
);
define('repositories/questionRepository',['context', 'guard'], function (context, guard) {

    return {
        get: get
    };

    function get(sectionId, questionId) {
        guard.throwIfNotString(sectionId, 'section id is not a string');
        guard.throwIfNotString(questionId, 'Question id is not a string');

        var section = _.find(context.course.sections, function (item) {
            return item.id == sectionId;
        });

        if (!section) {
            return null;
        }
        var question = _.find(section.questions, function (item) {
            return item.id == questionId;
        });
        
        if (!question) {
            return null;
        }

        return question;
    }

});
define('resultPage/resultPage',[
  "repositories/courseRepository",
  "templateSettings",
  "plugins/router",
  "progressContext",
  "userContext",
  "xApi/xApiInitializer",
  "helpers/appOperations",
  "constants",
  "modules/progress/progressStorage/auth",
  "modules/publishModeProvider",
  "dialogs/dialog",
  "modules/progress/progressStorage/certificateProvider",
  "helpers/fileDownloader",
  "localizationManager",
  "modules/webhooks"
], function(
  courseRepository,
  templateSettings,
  router,
  progressContext,
  userContext,
  xApiInitializer,
  appOperations,
  constants,
  auth,
  publishModeProvider,
  Dialog,
  certificateProvider,
  fileDownloader,
  localizationManager,
  webhooks
) {
  "use strict";

  var course = courseRepository.get();

  var progressStatuses = constants.progressContext.statuses;

  var statuses = {
    readyToFinish: "readyToFinish",
    preparingCertificate: "preparingCertificate",
    sendingRequests: "sendingRequests",
    finished: "finished"
  };

  var viewModel = {
    score: course.score,
    title: course.title,
    sections: [],
    status: ko.observable(statuses.readyToFinish),
    statuses: statuses,
    activate: activate,
    close: close,
    finish: finish,
    isDownloadingCertificate: ko.observable(false),
    downloadCertificate: downloadCertificate,
    npsDialog: new Dialog(),
    newAttemptDialog: new Dialog(),
    resendResultsDialog: new Dialog(),

    //properties
    isInReviewAttemptMode: false,
    isCompleted: false,
    crossDeviceEnabled: false,
    allowContentPagesScoring: false,
    xAPIEnabled: false,
    scormEnabled: false,
    canDownloadCertificate: false,
    certificateDownloaded: false,
    stayLoggedIn: ko.observable(false),

    //methods
    toggleStayLoggedIn: toggleStayLoggedIn
  };

  return viewModel;

  function activate() {
    viewModel.isInReviewAttemptMode = course.isFinished;
    viewModel.npsDialog.isVisible(false);
    viewModel.newAttemptDialog.isVisible(false);
    viewModel.resendResultsDialog.isVisible(false);
    viewModel.crossDeviceEnabled = templateSettings.allowCrossDeviceSaving;
    viewModel.canDownloadCertificate =
      templateSettings.allowCrossDeviceSaving &&
      templateSettings.allowCertificateDownload &&
      course.isCompleted();

    viewModel.allowContentPagesScoring =
      templateSettings.allowContentPagesScoring;

    viewModel.xAPIEnabled = xApiInitializer.isLrsReportingInitialized;
    viewModel.scormEnabled = publishModeProvider.isScormEnabled;

    viewModel.stayLoggedIn(userContext.user.keepMeLoggedIn);
    viewModel.sections = _.chain(course.sections)
      .filter(function(section) {
        return section.affectProgress || section.hasSurveyQuestions;
      })
      .map(mapSection)
      .value();

    viewModel.isCompleted = course.isCompleted();
  }

  function close() {
    router.navigate("#sections");
  }

  function finish() {
    if (
      router.isNavigationLocked() ||
      viewModel.status() !== statuses.readyToFinish
    ) {
      return;
    }

    if (viewModel.canDownloadCertificate && !viewModel.certificateDownloaded) {
      viewModel.status(statuses.preparingCertificate);
      doFinishCourse();
      downloadCertificate();
      return;
    }

    doFinishCourse();
  }

  function doFinishCourse() {
    if (
      templateSettings.xApi.enabled &&
      xApiInitializer.isLrsReportingInitialized
    ) {
      viewModel.status(statuses.sendingRequests);
    }

    if (course.getStatus() === constants.course.statuses.inProgress) {
      var finishHandler =
        viewModel.crossDeviceEnabled || viewModel.scormEnabled
          ? progressContext.finish
          : progressContext.remove;

      course.setFinishedStatus();
      return finishHandler(function() {
        progressContext.status(progressStatuses.ignored);
        course.finish(sendWebhooks);
      });
    }

    sendWebhooks();
  }

  function downloadCertificate() {
    viewModel.isDownloadingCertificate(true);
    return certificateProvider
      .getCertificateUrl(
        course.id,
        course.templateId,
        templateSettings.logo.url
      )
      .then(function(url) {
        /* Fix for IE11 and Edge (files can`t saved without extension) */
        var filename =
          localizationManager.getLocalizedText("[certificate file name]") +
          ".pdf";
        return fileDownloader.downloadFile(url, filename);
      })
      .always(function() {
        viewModel.isDownloadingCertificate(false);
        viewModel.certificateDownloaded = true;
      });
  }

  function sendWebhooks() {
    if (webhooks.initialized) {
      return webhooks
        .sendResults(course)
        .then(function() {
          onCourseFinished();
        })
        .catch(function() {
          viewModel.status(statuses.readyToFinish);
          viewModel.resendResultsDialog.resultsSendErrorTitleKey =
            constants.dialogs.resendResults.webhooks.resultsSendErrorTitleKey;
          viewModel.resendResultsDialog.endpointNameKey =
            constants.dialogs.resendResults.webhooks.endpointNameKey;
          viewModel.resendResultsDialog.show({
            resend: webhooks.sendResults.bind(webhooks),
            next: onCourseFinished
          });
        });
    }

    onCourseFinished();
  }

  function onCourseFinished() {
    viewModel.status(statuses.finished);

    if (
      templateSettings.nps.enabled &&
      xApiInitializer.isNpsReportingInitialized
    ) {
      return viewModel.npsDialog.show({
        closed: function() {
          finalize({ close: true });
        },
        finalized: function() {
          finalize({ close: false });
        }
      });
    }

    finalize({ close: true });
  }

  function finalize(params) {
    params = params || {};
    if (auth.authenticated && !viewModel.stayLoggedIn()) auth.signout();

    course.finalize(function() {
      if (params.close) {
        appOperations.close({
          shouldCloseWindow: !viewModel.canDownloadCertificate
        });
      }
    });
  }

  function toggleStayLoggedIn() {
    viewModel.stayLoggedIn(
      (userContext.user.keepMeLoggedIn = !viewModel.stayLoggedIn())
    );
    auth.shortTermAccess = !userContext.user.keepMeLoggedIn;
  }

  function mapSection(entity) {
    var section = {};

    section.id = entity.id;
    section.title = entity.title;
    section.score = entity.score();

    section.readedContents = _.filter(entity.questions, function(question) {
      return !isQuestion(question) && question.isAnswered;
    }).length;

    section.questions = _.filter(entity.questions, function(question) {
      return isQuestion(question);
    });

    section.amountOfQuestions = _.filter(section.questions, function(question) {
      return !question.isSurvey;
    }).length;

    section.correctQuestions = _.filter(section.questions, function(question) {
      return (
        question.isAnswered && question.isCorrectAnswered && !question.isSurvey
      );
    }).length;

    section.amountOfContents =
      entity.questions.length - section.questions.length;
    section.affectProgress = entity.affectProgress;
    section.title = entity.title;

    section.isCorrect = entity.isCompleted();

    return section;
  }

  function isQuestion(question) {
    return question.type !== constants.questionTypes.informationContent;
  }
});

define('treeOfContent/treeNodes/TreeNode',[], function() {
    return function(id, title) {
        this.id = id;
        this.title = title;
    };
});
define('treeOfContent/treeNodes/QuestionTreeNode',['treeOfContent/treeNodes/TreeNode', 'plugins/router'], function (TreeNode, router) {

    return function (id, title, isInformationContent, url) {
        TreeNode.call(this, id, title);
        var self = this;
        this.url = url;
        this.isActive = ko.observable(false);
        this.isInformationContent = isInformationContent;
        this.canNavigate = ko.computed(function () {
            return !self.isActive() && !router.isNavigationLocked();
        });
    };
});
define('treeOfContent/treeNodes/SectionTreeNode',['treeOfContent/treeNodes/TreeNode', 'treeOfContent/treeNodes/QuestionTreeNode', 'repositories/sectionRepository'],
    function (TreeNode, QuestionTreeNode, sectionRepository) {

        return function (sectionId, title) {
            TreeNode.call(this, sectionId, title);
            this.children = [];

            this.isExpanded = ko.observable(false);
            this.toggleIsExpanded = toggleIsExpanded;
            this.activate = activate;
            this.activateQuestion = activateQuestion;
            this.hasActiveQuestion = ko.observable(false);
        };

        function activate() {
            var sectionId = this.id;
            this.children = _.chain(sectionRepository.get(sectionId).questions)
                .map(function (question) {
                    return new QuestionTreeNode(question.id, question.title, question.isInformationContent,
                        '#section/' + sectionId + '/question/' + question.id);
                })
                .value();

            this.hasChildren = this.children.length > 0;
        }

        function activateQuestion(url) {
            var self = this;
            self.hasActiveQuestion(false);
            this.children.forEach(function (question) {
                var isQuestionToActivate = question.url === url;
                question.isActive(isQuestionToActivate);
                if (isQuestionToActivate) {
                    self.isExpanded(true);
                    self.hasActiveQuestion(true);
                }
            });
            
            return self.hasActiveQuestion();
        }

        function toggleIsExpanded() {
            this.isExpanded(!this.isExpanded());
        }
    }
);
define('treeOfContent/treeOfContent',['repositories/courseRepository', 'treeOfContent/treeNodes/SectionTreeNode', 'plugins/router', 'treeOfContent/utils/screenResolutionChecker'],
function (courseRepository, SectionTreeNode, router, screenResolutionChecker) {
    var viewModel = {
        children: [],
        hasChildren: false,
        isExpanded: ko.observable(true),
        isQuestionActivated: false,
        state: {
            isExpanded: false,
            isExpandedForQuestionView: true
        },
        toggleIsExpanded: toggleIsExpanded,
        activate: activate,
        deactivate: deactivate,
        initializeRoute: initializeRoute,
        activateQuestion: activateQuestion
    };

    return viewModel;

    function activate() {
        viewModel.children = _.chain(courseRepository.get().sections)
                .map(function (section) {
                    var treeNode = new SectionTreeNode(section.id, section.title);
                    treeNode.activate();
                    return treeNode;
                })
                .value();

        viewModel.hasChildren = viewModel.children.length > 0;
        viewModel.initializeRoute();
        router.on('router:navigation:complete', viewModel.initializeRoute);
    }

    function deactivate() {
        router.off('router:navigation:complete', viewModel.initializeRoute);
        viewModel.children = [];
        viewModel.hasChildren = false;
    }

    function toggleIsExpanded() {
        viewModel.isExpanded(!viewModel.isExpanded());
        if (viewModel.isQuestionActivated) {
            viewModel.state.isExpandedForQuestionView = viewModel.isExpanded();
        } else {
            viewModel.state.isExpanded = viewModel.isExpanded();
        }
    }

    function initializeRoute() {
        var activeInstruction = router.activeInstruction();
        if (!activeInstruction || !activeInstruction.fragment) {
            return;
        }

        var isQuestionActivatedForPreviousRoute = viewModel.isQuestionActivated;
        viewModel.isQuestionActivated = viewModel.activateQuestion('#' + activeInstruction.fragment);
        if (viewModel.isQuestionActivated) {
            if (!isQuestionActivatedForPreviousRoute) {
                viewModel.children.forEach(function (section) {
                    section.isExpanded(section.hasActiveQuestion());
                });
            }
        } else {
            viewModel.children.forEach(function (section, index) {
                section.isExpanded(index === 0);
            });
        }

        var isExpanded = false;
        if (!screenResolutionChecker.isLowResolution()) {
            isExpanded = viewModel.isQuestionActivated ? viewModel.state.isExpandedForQuestionView : viewModel.state.isExpanded;
        }

        viewModel.isExpanded(isExpanded);
    }

    function activateQuestion(url) {
        var isQuestionActivated = false;
        viewModel.children.forEach(function (section) {
            if (section.activateQuestion(url)) {
                isQuestionActivated = true;
            }
        });

        return isQuestionActivated;
    }
});
define('viewmodels/404',[], function () {
    var viewModel = {
        homeUrl: '#sections'
    };

    return viewModel;
});
define('viewmodels/exitCourse',[
    'durandal/app', 'helpers/appOperations', 'repositories/courseRepository', 'progressContext',
    'plugins/router', 'templateSettings', 'constants',
    'xApi/xApiInitializer', 'includedModules/modulesInitializer', 'context', 'modules/publishModeProvider','dialogs/dialog'
], function(app, appOperations, courseRepository, progressContext, router, templateSettings,
    constants, xApiInitializer, modulesInitializer, context, publishModeProvider, Dialog) {
    "use strict";

    var progressStatuses = constants.progressContext.statuses;
    var course = courseRepository.get();

    var statuses = {
        readyToFinish: 'readyToFinish',
        sendingRequests: 'sendingRequests',
        finished: 'finished'
    };

    var viewModel = {
        type: ko.observable(),
        isNavigationLocked: router.isNavigationLocked,

        status: ko.observable(statuses.readyToFinish),
        statuses: statuses,

        score: null,
        isCompleted: false,
        isInReviewAttemptMode: false,
        xAPIEnabled: false,
        scormEnabled: false,
        hideFinishActionButtons: false,
        hideContinueLaterButton: false,
        activate: activate
    };

    viewModel.continueLaterPopup = new Dialog();

    viewModel.continueLaterPopup.actions = {
        close: viewModel.continueLaterPopup.hide,
        exit: exit
    };

    viewModel.isProgressSaved = ko.computed(function() {
        return progressContext.status() === progressStatuses.saved;
    });

    viewModel.isProgressNotSaved = ko.computed(function() {
        return progressContext.status() === progressStatuses.error;
    });

    viewModel.finishAction = function() {
        router.navigate('#finish');
    };

    viewModel.takeABreakAction = function() {
        viewModel.continueLaterPopup.show();
    }

    return viewModel;

    function activate(type) {
        viewModel.score = context.course.score;
        viewModel.isCompleted = context.course.isCompleted();
        viewModel.isInReviewAttemptMode = context.isInReviewAttemptMode();
        viewModel.type(type);
        viewModel.xAPIEnabled = xApiInitializer.isLrsReportingInitialized;
        viewModel.scormEnabled = publishModeProvider.isScormEnabled;
        viewModel.hideFinishActionButtons = templateSettings.hideFinishActionButtons;
        viewModel.hideContinueLaterButton = templateSettings.hideContinueLaterButton || false;
    }

    function exit() {
        if (progressContext.status() === progressStatuses.error) {
            var isCourseClosingConfirmed = confirm(TranslationPlugin.getTextByKey('[progress is not saved confirmation]'));
            if (!isCourseClosingConfirmed) {
                return;
            }
            progressContext.status(progressStatuses.ignored);
        }

        course.finalize(function() {
            appOperations.close();
        });
    }

});
define('viewmodels/introduction',['durandal/app', 'context', 'plugins/router', 'requester/resourceLoader', 'templateSettings'], function (
  app,
  context,
  router,
  resourceLoader,
  templateSettings
) {
  var viewModel = {
    courseTitle: context.course.title,
    introduction: [],
    copyright: templateSettings.copyright,
    authorContactEmail: context.course.authorContactEmail,
    authorPersonalPhone: context.course.authorPersonalPhone,
    authorShortBio: context.course.authorShortBio,
    createdBy: context.course.createdBy,
    authorAvatarUrl: context.course.authorAvatarUrl,
    allowAuthorsBio: templateSettings.allowAuthorsBio,
    isNavigationLocked: router.isNavigationLocked()
  };

  viewModel.canActivate = function () {
    if (!context.course.hasIntroductionContent && !viewModel.allowAuthorsBio) {
      return { redirect: '#sections' };
    }
    return true;
  };

  var dfd = Q.defer();

  viewModel.activate = function () {
    var promises = [];

    context.course.hasIntroductionContent &&
      context.course.introductions.forEach(function (introductionItem, index) {
        viewModel.introduction[index] = introductionItem;

        if (introductionItem.children && introductionItem.children.length) {
          introductionItem.children.forEach(function (child, childIndex) {
            return promises.push(
              resourceLoader
                .getLocalResource({
                  url: 'content/introduction/' + child.id + '.html'
                })
                .then(function (response) {
                  viewModel.introduction[index].children[childIndex].content = response;
                })
                .fail(function () {
                  viewModel.introduction = [];
                })
            );
          });
        }

        return promises.push(
          resourceLoader
            .getLocalResource({
              url: 'content/introduction/' + introductionItem.id + '.html'
            })
            .then(function (response) {
              viewModel.introduction[index].content = response;
            })
            .fail(function () {
              viewModel.introduction = [];
            })
        );
      });

    Q.allSettled(promises)
      .then(function () {
        dfd.resolve();
      })
      ['catch'](function (reason) {
        dfd.reject(reason);
      });

    return dfd.promise;
  };

  viewModel.startCourse = function () {
    if (router.isNavigationLocked()) {
      return;
    }
    router.navigate('sections');
  };

  return viewModel;
});

define('viewmodels/questions/content',['plugins/router', 'constants', 'repositories/questionRepository', 'repositories/sectionRepository', 'templateSettings', 'modules/questionsNavigation'],
    function(router, constants, questionRepository, sectionRepository, templateSettings, navigationModule) {
        "use strict";

        var viewModel = {
            section: null,
            question: null,

            startTime: null,

            masteryScore: 0,

            navigationContext: null,
            backToSections: backToSections,
            isNextQuestionAvailable: isNextQuestionAvailable,
            nextQuestionUrl: nextQuestionUrl,
            goNext: goNext,
            isPreviousQuestionAvailable: isPreviousQuestionAvailable,
            previousQuestionUrl: previousQuestionUrl,
            goPrevious: goPrevious,

            voiceOver: null,

            activeViewModel: null,

            activate: activate,
            isNavigationLocked: router.isNavigationLocked,
            deactivate: deactivate
        };

        return viewModel;

        function backToSections() {
            if (router.isNavigationLocked()) {
                return;
            }
            router.navigate('sections');
        }

        function isNextQuestionAvailable() {
            return !_.isNullOrUndefined(viewModel.navigationContext.nextQuestionUrl) && !router.isNavigationLocked();
        }

        function nextQuestionUrl() {
            return router.isNavigationLocked() ? undefined : viewModel.navigationContext.nextQuestionUrl;
        }

        function goNext() {
            if (router.isNavigationLocked() || !isNextQuestionAvailable()) {
                return;
            }
            router.navigate(viewModel.navigationContext.nextQuestionUrl);
        }

        function isPreviousQuestionAvailable() {
            return !_.isNullOrUndefined(viewModel.navigationContext.previousQuestionUrl) && !router.isNavigationLocked();
        }

        function previousQuestionUrl() {
            return router.isNavigationLocked() ? undefined : viewModel.navigationContext.previousQuestionUrl;
        }

        function goPrevious() {
            if (router.isNavigationLocked() || !isPreviousQuestionAvailable()) {
                return;
            }
            router.navigate(viewModel.navigationContext.previousQuestionUrl);
        }

        function getActiveContentViewModel(question) {
            switch (question.type) {
                case constants.questionTypes.informationContent:
                    return 'viewmodels/questions/informationContent';
                default:
                    return 'viewmodels/questions/questionContent';
            }
        }

        function activate(sectionId, questionId) {

            return Q.fcall(function() {
                viewModel.section = sectionRepository.get(sectionId);
                if (viewModel.section === null) {
                    router.navigate('404', { replace: true, trigger: true });
                    return;
                }

                viewModel.masteryScore = templateSettings.masteryScore.score;
                viewModel.isSectionMastered = ko.computed(function() {
                    return viewModel.section.score() >= templateSettings.masteryScore.score;
                });

                viewModel.question = questionRepository.get(sectionId, questionId);
                if (viewModel.question === null) {
                    router.navigate('404', { replace: true, trigger: true });
                    return;
                }


                viewModel.voiceOver = viewModel.question.voiceOver;

                viewModel.startTime = new Date();
                viewModel.navigationContext = navigationModule.getNavigationContext(viewModel.section.id, viewModel.question.id);


                return viewModel.question.load().then(function() {
                    router.activeInstruction().config.title = TranslationPlugin.getTextByKey(
                        viewModel.question.isInformationContent 
                        ? '[content page title]' 
                        : '[question page title]'
                    );

                    viewModel.activeViewModel = getActiveContentViewModel(viewModel.question);
                });
            });
        }

        function deactivate() {
            if (viewModel.question && viewModel.question.learningContents.length > 0) {
                viewModel.question.learningContentExperienced(new Date() - viewModel.startTime);
            }
        }
    }
);
define('viewmodels/questions/expandableBlock',['knockout'], function (ko) {
    "use strict";

    function ExpandableBlock() {
        this.content = undefined;
        this.children = undefined;
        this.isExpanded = ko.observable();
    };

    ExpandableBlock.prototype.activate = function(data) {
        this.content = data.content;
        this.children = data.children;
        this.isExpanded(false);
    }

    ExpandableBlock.prototype.toggleExpand = function() {
        return this.isExpanded(!this.isExpanded());
    }

    return ExpandableBlock;
});
define('viewmodels/questions/informationContent',['modules/questionsNavigation', 'plugins/router', 'templateSettings', 'context'], function (navigationModule, router, templateSettings, context) {
    "use strict";

    var viewModel = {
        title: null,
        learningContents: null,
        navigateNext: navigateNext,
        goToNextSection: goToNextSection,
        goToResults: goToResults,
        copyright: templateSettings.copyright,

        activate: activate,
        isGoToNextSectionVisible: ko.observable(false),
        isGoToResultsVisible: ko.observable(false),
        isNavigationLocked: router.isNavigationLocked
    };

    return viewModel;

    function navigateNext() {
        if (router.isNavigationLocked()) {
            return;
        }

        var nextUrl = !_.isNullOrUndefined(viewModel.navigationContext.nextQuestionUrl) ? viewModel.navigationContext.nextQuestionUrl : 'sections';
        router.navigate(nextUrl);
    }

    function goToNextSection() {
        if (router.isNavigationLocked()) {
            return;
        }
        if (!viewModel.navigationContext.nextSectionUrl) {
            return;
        }

        router.navigate(viewModel.navigationContext.nextSectionUrl);
    }

    function goToResults() {
        if (router.isNavigationLocked()) {
            return;
        }
        router.navigate('#finish');
    }

    function activate(sectionId, question) {
        return Q.fcall(function () {
            question.submitAnswer();

            viewModel.navigationContext = navigationModule.getNavigationContext(sectionId, question.id);
            viewModel.id = question.id;
            viewModel.title = question.title;
            viewModel.learningContents = question.learningContents;

            if (context.course.score() === 100) {
                viewModel.isGoToResultsVisible(!templateSettings.hideFinishActionButtons);
                viewModel.isGoToNextSectionVisible(viewModel.navigationContext.questionsCount === viewModel.navigationContext.currentQuestionIndex && viewModel.navigationContext.nextSectionUrl);
            } else {
                if (viewModel.navigationContext.questionsCount === viewModel.navigationContext.currentQuestionIndex) {
                    viewModel.isGoToResultsVisible(!viewModel.navigationContext.nextSectionUrl && !templateSettings.hideFinishActionButtons);
                    viewModel.isGoToNextSectionVisible(!!viewModel.navigationContext.nextSectionUrl);
                } else {
                    viewModel.isGoToResultsVisible(false);
                    viewModel.isGoToNextSectionVisible(false);
                }
            }
        });
    }
});
define('viewmodels/questions/questionContent',['questionContent/questionContent'],
	function (QuestionContent) {
	    "use strict";

	    return new QuestionContent();
	});
define('viewmodels/questions/scenarioQuestion/submitQuestion',['knockout', 'viewmodels/questions/scenarioQuestion/components/branchtrackProvider'], function (ko, branchtrackProvider) {
    'use strict';

    var submitCallback = function() {};

    var viewmodel = {
        submit: submit,
        canBeSubmitted: ko.observable(false),
        activate: activate
    };

    return viewmodel;

    function activate(activationData) {
        if (_.isFunction(activationData.submit)) {
            submitCallback = activationData.submit;
            viewmodel.canBeSubmitted(branchtrackProvider.isFinished());
        }
    }

    function submit() {
        viewmodel.canBeSubmitted(branchtrackProvider.isFinished());
        if (viewmodel.canBeSubmitted()) {
            submitCallback();
        }
    }
});
define('viewmodels/sections',['underscore', 'context', 'repositories/courseRepository', 'plugins/router', 'templateSettings'],
    function (_, context, repository, router, templateSettings) {

        var
            sections = [],
            masteryScore = 0,
            courseTitle = context.course.title,
            sectionsLayout = null,
            sectionThumbnail = { width: 284, height: 170 },
            
            canActivate = function () {
                if (templateSettings.sectionsPage && !templateSettings.sectionsPage.enabled && !router.isNavigationLocked()) {
                    var section = _.find(context.course.sections, function(section) {
                        return section.questions.length > 0;
                    });

                    if(section) {
                        return { redirect: '#section/' + section.id + '/question/' + section.questions[0].id };
                    } else {
                        return { redirect: '#finish' };
                    }
                }

                return true;
            },

            activate = function () {
                var course = repository.get();
                if (course == null) {
                    router.navigate('404');
                    return;
                }
                this.sectionsLayout = templateSettings.sectionsLayout;
                this.masteryScore = templateSettings.masteryScore.score;

                if (sectionsLayout === "List") {
                    sectionThumbnail.width = 100;
                    sectionThumbnail.height = 70;
                }

                this.sections = _.map(course.sections, function (item) {
                    return {
                        id: item.id,
                        title: item.title,
                        imageUrl: item.imageUrl,
                        imageWidth: sectionThumbnail.width,
                        imageHeight: sectionThumbnail.height,
                        score: item.score(),
                        isMastered: item.score() >= templateSettings.masteryScore.score,
                        scoreTooltipText: getScoreTooltipText(templateSettings.masteryScore.score, item.score()),
                        questions: item.questions,
                        affectProgress: item.affectProgress,
                        goToFirstQuestion: function () {
                            if (router.isNavigationLocked()) {
                                return;
                            }
                            router.navigate('#/section/' + item.id + '/question/' + item.questions[0].id);
                        }
                    };
                });
            }

        return {
            activate: activate,
            canActivate: canActivate,
            isNavigationLocked: router.isNavigationLocked,
            caption: 'Sections and questions',
            courseTitle: courseTitle,
            sectionsLayout: sectionsLayout,
            isInReviewAttemptMode: context.isInReviewAttemptMode(),
            getSectionButtonAccessibilityLabel: getSectionButtonAccessibilityLabel,

            masteryScore: masteryScore,
            sections: sections
        };

        function getSectionButtonAccessibilityLabel(buttonTranslationKey, section) {
            return TranslationPlugin.getTextByKey(buttonTranslationKey) + ' ' + section.title + ' ' + section.score + '% ' + section.scoreTooltipText;
        }
        
        function getScoreTooltipText(masteryScore, score) {
            var scoreToComplete = masteryScore - score;
            return scoreToComplete > 0 ? scoreToComplete + '% ' + TranslationPlugin.getTextByKey('[to complete]') : TranslationPlugin.getTextByKey('[completed]');
        }

    });
define('viewmodels/shell',[
  "knockout",
  "underscore",
  "durandal/app",
  "modules/progress/progressStorage/auth",
  "durandal/composition",
  "plugins/router",
  "routing/routes",
  "context",
  "includedModules/modulesInitializer",
  "templateSettings",
  "progressContext",
  "constants",
  "userContext",
  "errorsHandler",
  "modules/progress/index",
  "account/index",
  "xApi/xApiInitializer",
  "modules/publishModeProvider",
  "modules/questionsNavigation",
  "modules/lti",
  "modules/webhooks",
  "helpers/loginHelper",
  "modules/progress/progressStorage/urlProvider"
], function(
  ko,
  _,
  app,
  auth,
  composition,
  router,
  routes,
  context,
  modulesInitializer,
  templateSettings,
  progressContext,
  constants,
  userContext,
  errorsHandler,
  progressProvider,
  account,
  xApiInitializer,
  publishModeProvider,
  questionsNavigation,
  lti,
  webhooks,
  loginHelper,
  urlProvider
) {
  "use strict";
  var viewmodel = {
    router: null,
    cssName: null,
    isInReviewMode: false,
    title: "",
    createdOn: null,
    logo: {
      url: "",
      maxWidth: "",
      maxHeight: ""
    },
    pdfExportEnabled: ko.observable(false),
    isClosed: ko.observable(false),
    isNavigatingToAnotherView: ko.observable(false),
    inIframe: window.self !== window.top,

    viewSettings: ko.computed(viewSettings),
    activate: activate
  };

  viewmodel.router = router;
  viewmodel.cssName = ko.computed(function() {
    var activeItem = router.activeItem();
    if (_.isObject(activeItem)) {
      var moduleId = activeItem.__moduleId__;
      moduleId = moduleId.slice(moduleId.lastIndexOf("/") + 1);
      return moduleId;
    }
    return "";
  });

  viewmodel.isInReviewMode = router.getQueryStringValue("reviewApiUrl");

  router.on("router:route:activating").then(function(newView) {
    if (
      !publishModeProvider.isScormEnabled &&
      !viewmodel.pdfExportEnabled() === templateSettings.pdfExport.enabled
    ) {
      var isNotAccountModule =
        newView &&
        newView.__moduleId__.slice(0, newView.__moduleId__.indexOf("/")) !==
          "account";
      viewmodel.pdfExportEnabled(isNotAccountModule);
    }

    var currentView = router.activeItem();
    if (
      newView &&
      currentView &&
      newView.__moduleId__ === currentView.__moduleId__
    ) {
      return;
    }
    viewmodel.isNavigatingToAnotherView(true);
  });

  app.on(constants.events.appClosed).then(function() {
    viewmodel.isClosed(true);
  });

  return viewmodel;

  //public methods
  function activate() {
    return context
      .initialize()
      .then(initializeUrlProvider)
      .then(loginHelper.initialize)
      .then(auth.authorize)
      .catch(function(e) {
        console.error(e);
      })
      .then(userContext.initialize)
      .then(lti.initialize)
      .then(account.enable)
      .then(initializeWebhooks)
      .then(initializeProgressProvider)
      .then(initxApi)
      .then(initApp)
      .then(initRouter)
      .catch(function(e) {
        console.error(e);
      });

    function initializeUrlProvider() {
      return urlProvider.initialize();
    }

    function initializeWebhooks() {
      if (templateSettings.xApi.enabled) {
        webhooks.initialize(templateSettings.webhooks);
      }
    }

    function initxApi() {
      return xApiInitializer.initialize(
        templateSettings.xApi,
        templateSettings.nps
      );
    }

    function initializeProgressProvider() {
      if (
        !publishModeProvider.isScormEnabled &&
        !publishModeProvider.isPreview &&
        !publishModeProvider.isReview
      ) {
        return progressProvider.initialize().then(function(provider) {
          progressContext.use(provider);
        });
      }
    }

    function initApp() {
      return Q.fcall(function() {
        viewmodel.logo.url = templateSettings.logo.url;
        viewmodel.logo.maxWidth = templateSettings.logo.maxWidth || "300px";
        viewmodel.logo.maxHeight = templateSettings.logo.maxHeight || "100px";
        viewmodel.title = app.title = context.course.title;
        viewmodel.createdOn = context.course.createdOn;
        progressContext.restoreProgress();
        if (context.isInReviewAttemptMode()) {
          xApiInitializer.deactivate();
        }
        app.trigger(constants.events.appInitialized);
      });
    }

    function initRouter() {
      questionsNavigation.redirectToQuestion();
      return router
        .map(routes.routes)
        .buildNavigationModel()
        .mapUnknownRoutes("viewmodels/404", "404")
        .activate()
        .then(function() {
          errorsHandler.startHandle();
        });
    }
  }

  function viewSettings() {
    var settings = {
      rootLinkEnabled: false,
      exitButtonVisible: false,
      treeOfContentVisible: false,
      isInReviewAttemptMode: false
    };

    var activeInstruction = router.activeInstruction();
    if (_.isObject(activeInstruction)) {
      settings.rootLinkEnabled =
        !!activeInstruction.config.rootLinkEnabled &&
        !router.isNavigationLocked();
      settings.exitButtonVisible =
        showExitButton(activeInstruction.config) &&
        !templateSettings.hideFinishActionButtons;
      settings.treeOfContentVisible =
        templateSettings.treeOfContent.enabled &&
        !!activeInstruction.config.displayTreeOfContent;
    }
    settings.isInReviewAttemptMode = context.isInReviewAttemptMode();

    return settings;
  }

  function showExitButton(activeInstructionConfig) {
    return (
      !!activeInstructionConfig.showExitButton ||
      (!!activeInstructionConfig.showExitButtonInReviewMode &&
        context.isInReviewAttemptMode())
    );
  }
});

define('xApi/models/verb',['../guard'],
    function (guard) {

        var verb = function (spec) {

            if (typeof spec == typeof undefined) {
                throw 'You should provide a specification to create Verb';
            }

            guard.throwIfNotVerbId(spec.id, 'You should provide valid id for Verb');
            guard.throwIfNotLanguageMap(spec.display, 'You should provide valid language map as display for Verb');

            return {
                id: spec.id,
                display: spec.display
            };
        };

        return verb;
    }
);
define('xApi/viewmodels/xAPIError',['knockout', 'durandal/app', 'plugins/router', '../xApiInitializer', '../configuration/xApiSettings'],
    function (ko, app, router, xApiInitializer, xApiSettings) {

    var
        navigateBackUrl = '',

        allowToContinue = ko.observable(false),

        restartCourse = function () {
            var rootUrl = location.toString().replace(location.hash, '');
            router.navigate(rootUrl, { replace: true, trigger: true });
        },
        
        continueLearning = function () {
            if (!allowToContinue()) {
                return;
            }

            xApiInitializer.deactivate();
            router.navigate(navigateBackUrl);
            app.trigger('user:authentication-skipped');
        },

        activate = function (backUrl) {
            navigateBackUrl = backUrl;
            allowToContinue(!xApiSettings.xApi.required);
        };

        return {
            allowToContinue: allowToContinue,
            restartCourse: restartCourse,
            continueLearning: continueLearning,
        
            activate: activate
        };
});
/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The entrance transition module.
 * @module entrance
 * @requires system
 * @requires composition
 * @requires jquery
 */
define('transitions/entrance',['durandal/system', 'durandal/composition', 'jquery'], function(system, composition, $) {
    var fadeOutDuration = 50;
    var endValues = {
        marginRight: 0,
        marginLeft: 0,
        opacity: 1
    };
    var clearValues = {
        marginLeft: '',
        marginRight: '',
        opacity: '',
        display: ''
    };

    /**
     * @class EntranceModule
     * @constructor
     */
    var entrance = function(context) {
        return system.defer(function (dfd) {
            function endTransition() {
                if (context.bindingContext.$data && context.bindingContext.$data.isNavigatingToAnotherView && ko.isObservable(context.bindingContext.$data.isNavigatingToAnotherView)) {
                    context.bindingContext.$data.isNavigatingToAnotherView(false);
                }
                dfd.resolve();
            }

            function scrollIfNeeded() {
                if (!context.keepScrollPosition) {
                    $(document).scrollTop(0);
                }
            }

            if (!context.child) {
                $(context.activeView).fadeOut(fadeOutDuration, endTransition);
            } else {
                var duration = context.duration || 300;
                var fadeOnly = !!context.fadeOnly;

                function startTransition() {
                    scrollIfNeeded();
                    context.triggerAttach();

                    var startValues = {
                        marginLeft: fadeOnly ? '0' : '20px',
                        marginRight: fadeOnly ? '0' : '-20px',
                        opacity: 0,
                        display: 'block'
                    };

                    var $child = $(context.child);

                    $child.css(startValues);
                    $child.animate(endValues, {
                        duration: duration,
                        easing: 'swing',
                        always: function () {
                            $child.css(clearValues);
                            endTransition();
                        }
                    });
                }
                
                if (context.activeView) {
                    $(context.activeView).fadeOut({ duration: fadeOutDuration, always: startTransition });
                } else {
                    startTransition();
                }
                
            }
        }).promise();
    };

    return entrance;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * The dialog module enables the display of message boxes, custom modal dialogs and other overlays or slide-out UI abstractions. Dialogs are constructed by the composition system which interacts with a user defined dialog context. The dialog module enforced the activator lifecycle.
 * @module dialog
 * @requires system
 * @requires app
 * @requires composition
 * @requires activator
 * @requires viewEngine
 * @requires jquery
 * @requires knockout
 */
define('plugins/dialog',['durandal/system', 'durandal/app', 'durandal/composition', 'durandal/activator', 'durandal/viewEngine', 'jquery', 'knockout'], function (system, app, composition, activator, viewEngine, $, ko) {
    var contexts = {},
        dialogCount = 0,
        dialog;

    /**
     * Models a message box's message, title and options.
     * @class MessageBox
     */
    var MessageBox = function(message, title, options) {
        this.message = message;
        this.title = title || MessageBox.defaultTitle;
        this.options = options || MessageBox.defaultOptions;
    };

    /**
     * Selects an option and closes the message box, returning the selected option through the dialog system's promise.
     * @method selectOption
     * @param {string} dialogResult The result to select.
     */
    MessageBox.prototype.selectOption = function (dialogResult) {
        dialog.close(this, dialogResult);
    };

    /**
     * Provides the view to the composition system.
     * @method getView
     * @return {DOMElement} The view of the message box.
     */
    MessageBox.prototype.getView = function(){
        return viewEngine.processMarkup(MessageBox.defaultViewMarkup);
    };

    /**
     * Configures a custom view to use when displaying message boxes.
     * @method setViewUrl
     * @param {string} viewUrl The view url relative to the base url which the view locator will use to find the message box's view.
     * @static
     */
    MessageBox.setViewUrl = function(viewUrl){
        delete MessageBox.prototype.getView;
        MessageBox.prototype.viewUrl = viewUrl;
    };

    /**
     * The title to be used for the message box if one is not provided.
     * @property {string} defaultTitle
     * @default Application
     * @static
     */
    MessageBox.defaultTitle = app.title || 'Application';

    /**
     * The options to display in the message box of none are specified.
     * @property {string[]} defaultOptions
     * @default ['Ok']
     * @static
     */
    MessageBox.defaultOptions = ['Ok'];

    /**
     * The markup for the message box's view.
     * @property {string} defaultViewMarkup
     * @static
     */
    MessageBox.defaultViewMarkup = [
        '<div data-view="plugins/messageBox" class="messageBox">',
            '<div class="modal-header">',
                '<h3 data-bind="text: title"></h3>',
            '</div>',
            '<div class="modal-body">',
                '<p class="message" data-bind="text: message"></p>',
            '</div>',
            '<div class="modal-footer" data-bind="foreach: options">',
                '<button class="btn" data-bind="click: function () { $parent.selectOption($data); }, text: $data, css: { \'btn-primary\': $index() == 0, autofocus: $index() == 0 }"></button>',
            '</div>',
        '</div>'
    ].join('\n');

    function ensureDialogInstance(objOrModuleId) {
        return system.defer(function(dfd) {
            if (system.isString(objOrModuleId)) {
                system.acquire(objOrModuleId).then(function (module) {
                    dfd.resolve(system.resolveObject(module));
                }).fail(function(err){
                    system.error('Failed to load dialog module (' + objOrModuleId + '). Details: ' + err.message);
                });
            } else {
                dfd.resolve(objOrModuleId);
            }
        }).promise();
    }

    /**
     * @class DialogModule
     * @static
     */
    dialog = {
        /**
         * The constructor function used to create message boxes.
         * @property {MessageBox} MessageBox
         */
        MessageBox:MessageBox,
        /**
         * The css zIndex that the last dialog was displayed at.
         * @property {number} currentZIndex
         */
        currentZIndex: 1050,
        /**
         * Gets the next css zIndex at which a dialog should be displayed.
         * @method getNextZIndex
         * @return {number} The next usable zIndex.
         */
        getNextZIndex: function () {
            return ++this.currentZIndex;
        },
        /**
         * Determines whether or not there are any dialogs open.
         * @method isOpen
         * @return {boolean} True if a dialog is open. false otherwise.
         */
        isOpen: function() {
            return dialogCount > 0;
        },
        /**
         * Gets the dialog context by name or returns the default context if no name is specified.
         * @method getContext
         * @param {string} [name] The name of the context to retrieve.
         * @return {DialogContext} True context.
         */
        getContext: function(name) {
            return contexts[name || 'default'];
        },
        /**
         * Adds (or replaces) a dialog context.
         * @method addContext
         * @param {string} name The name of the context to add.
         * @param {DialogContext} dialogContext The context to add.
         */
        addContext: function(name, dialogContext) {
            dialogContext.name = name;
            contexts[name] = dialogContext;

            var helperName = 'show' + name.substr(0, 1).toUpperCase() + name.substr(1);
            this[helperName] = function (obj, activationData) {
                return this.show(obj, activationData, name);
            };
        },
        createCompositionSettings: function(obj, dialogContext) {
            var settings = {
                model:obj,
                activate:false,
                transition: false
            };

            if (dialogContext.attached) {
                settings.attached = dialogContext.attached;
            }

            if (dialogContext.compositionComplete) {
                settings.compositionComplete = dialogContext.compositionComplete;
            }

            return settings;
        },
        /**
         * Gets the dialog model that is associated with the specified object.
         * @method getDialog
         * @param {object} obj The object for whom to retrieve the dialog.
         * @return {Dialog} The dialog model.
         */
        getDialog:function(obj){
            if(obj){
                return obj.__dialog__;
            }

            return undefined;
        },
        /**
         * Closes the dialog associated with the specified object.
         * @method close
         * @param {object} obj The object whose dialog should be closed.
         * @param {object} results* The results to return back to the dialog caller after closing.
         */
        close:function(obj){
            var theDialog = this.getDialog(obj);
            if(theDialog){
                var rest = Array.prototype.slice.call(arguments, 1);
                theDialog.close.apply(theDialog, rest);
            }
        },
        /**
         * Shows a dialog.
         * @method show
         * @param {object|string} obj The object (or moduleId) to display as a dialog.
         * @param {object} [activationData] The data that should be passed to the object upon activation.
         * @param {string} [context] The name of the dialog context to use. Uses the default context if none is specified.
         * @return {Promise} A promise that resolves when the dialog is closed and returns any data passed at the time of closing.
         */
        show: function(obj, activationData, context) {
            var that = this;
            var dialogContext = contexts[context || 'default'];

            return system.defer(function(dfd) {
                ensureDialogInstance(obj).then(function(instance) {
                    var dialogActivator = activator.create();

                    dialogActivator.activateItem(instance, activationData).then(function (success) {
                        if (success) {
                            var theDialog = instance.__dialog__ = {
                                owner: instance,
                                context: dialogContext,
                                activator: dialogActivator,
                                close: function () {
                                    var args = arguments;
                                    dialogActivator.deactivateItem(instance, true).then(function (closeSuccess) {
                                        if (closeSuccess) {
                                            dialogCount--;
                                            dialogContext.removeHost(theDialog);
                                            delete instance.__dialog__;

                                            if (args.length === 0) {
                                                dfd.resolve();
                                            } else if (args.length === 1) {
                                                dfd.resolve(args[0]);
                                            } else {
                                                dfd.resolve.apply(dfd, args);
                                            }
                                        }
                                    });
                                }
                            };

                            theDialog.settings = that.createCompositionSettings(instance, dialogContext);
                            dialogContext.addHost(theDialog);

                            dialogCount++;
                            composition.compose(theDialog.host, theDialog.settings);
                        } else {
                            dfd.resolve(false);
                        }
                    });
                });
            }).promise();
        },
        /**
         * Shows a message box.
         * @method showMessage
         * @param {string} message The message to display in the dialog.
         * @param {string} [title] The title message.
         * @param {string[]} [options] The options to provide to the user.
         * @return {Promise} A promise that resolves when the message box is closed and returns the selected option.
         */
        showMessage:function(message, title, options){
            if(system.isString(this.MessageBox)){
                return dialog.show(this.MessageBox, [
                    message,
                    title || MessageBox.defaultTitle,
                    options || MessageBox.defaultOptions
                ]);
            }

            return dialog.show(new this.MessageBox(message, title, options));
        },
        /**
         * Installs this module into Durandal; called by the framework. Adds `app.showDialog` and `app.showMessage` convenience methods.
         * @method install
         * @param {object} [config] Add a `messageBox` property to supply a custom message box constructor. Add a `messageBoxView` property to supply custom view markup for the built-in message box.
         */
        install:function(config){
            app.showDialog = function(obj, activationData, context) {
                return dialog.show(obj, activationData, context);
            };

            app.showMessage = function(message, title, options) {
                return dialog.showMessage(message, title, options);
            };

            if(config.messageBox){
                dialog.MessageBox = config.messageBox;
            }

            if(config.messageBoxView){
                dialog.MessageBox.prototype.getView = function(){
                    return config.messageBoxView;
                };
            }
        }
    };

    /**
     * @class DialogContext
     */
    dialog.addContext('default', {
        blockoutOpacity: .2,
        removeDelay: 200,
        /**
         * In this function, you are expected to add a DOM element to the tree which will serve as the "host" for the modal's composed view. You must add a property called host to the modalWindow object which references the dom element. It is this host which is passed to the composition module.
         * @method addHost
         * @param {Dialog} theDialog The dialog model.
         */
        addHost: function(theDialog) {
            var body = $('body');
            var blockout = $('<div class="modalBlockout"></div>')
                .css({ 'z-index': dialog.getNextZIndex(), 'opacity': this.blockoutOpacity })
                .appendTo(body);

            var host = $('<div class="modalHost"></div>')
                .css({ 'z-index': dialog.getNextZIndex() })
                .appendTo(body);

            theDialog.host = host.get(0);
            theDialog.blockout = blockout.get(0);

            if (!dialog.isOpen()) {
                theDialog.oldBodyMarginRight = body.css("margin-right");
                theDialog.oldInlineMarginRight = body.get(0).style.marginRight;

                var html = $("html");
                var oldBodyOuterWidth = body.outerWidth(true);
                var oldScrollTop = html.scrollTop();
                $("html").css("overflow-y", "hidden");
                var newBodyOuterWidth = $("body").outerWidth(true);
                body.css("margin-right", (newBodyOuterWidth - oldBodyOuterWidth + parseInt(theDialog.oldBodyMarginRight, 10)) + "px");
                html.scrollTop(oldScrollTop); // necessary for Firefox
            }
        },
        /**
         * This function is expected to remove any DOM machinery associated with the specified dialog and do any other necessary cleanup.
         * @method removeHost
         * @param {Dialog} theDialog The dialog model.
         */
        removeHost: function(theDialog) {
            $(theDialog.host).css('opacity', 0);
            $(theDialog.blockout).css('opacity', 0);

            setTimeout(function() {
                ko.removeNode(theDialog.host);
                ko.removeNode(theDialog.blockout);
            }, this.removeDelay);

            if (!dialog.isOpen()) {
                var html = $("html");
                var oldScrollTop = html.scrollTop(); // necessary for Firefox.
                html.css("overflow-y", "").scrollTop(oldScrollTop);

                if(theDialog.oldInlineMarginRight) {
                    $("body").css("margin-right", theDialog.oldBodyMarginRight);
                } else {
                    $("body").css("margin-right", '');
                }
            }
        },
        attached: function (view) {
            //To prevent flickering in IE8, we set visibility to hidden first, and later restore it
            $(view).css("visibility", "hidden");
        },
        /**
         * This function is called after the modal is fully composed into the DOM, allowing your implementation to do any final modifications, such as positioning or animation. You can obtain the original dialog object by using `getDialog` on context.model.
         * @method compositionComplete
         * @param {DOMElement} child The dialog view.
         * @param {DOMElement} parent The parent view.
         * @param {object} context The composition context.
         */
        compositionComplete: function (child, parent, context) {
            var theDialog = dialog.getDialog(context.model);
            var $child = $(child);
            var loadables = $child.find("img").filter(function () {
                //Remove images with known width and height
                var $this = $(this);
                return !(this.style.width && this.style.height) && !($this.attr("width") && $this.attr("height"));
            });

            $child.data("predefinedWidth", $child.get(0).style.width);

            var setDialogPosition = function () {
                //Setting a short timeout is need in IE8, otherwise we could do this straight away
                setTimeout(function () {
                    //We will clear and then set width for dialogs without width set 
                    if (!$child.data("predefinedWidth")) {
                        $child.css({ width: '' }); //Reset width
                    }
                    var width = $child.outerWidth(false);
                    var height = $child.outerHeight(false);
                    var windowHeight = $(window).height();
                    var constrainedHeight = Math.min(height, windowHeight);

                    $child.css({
                        'margin-top': (-constrainedHeight / 2).toString() + 'px',
                        'margin-left': (-width / 2).toString() + 'px'
                    });

                    if (!$child.data("predefinedWidth")) {
                        //Ensure the correct width after margin-left has been set
                        $child.outerWidth(width);
                    }

                    if (height > windowHeight) {
                        $child.css("overflow-y", "auto");
                    } else {
                        $child.css("overflow-y", "");
                    }

                    $(theDialog.host).css('opacity', 1);
                    $child.css("visibility", "visible");

                    $child.find('.autofocus').first().focus();
                }, 1);
            };

            setDialogPosition();
            loadables.load(setDialogPosition);

            if ($child.hasClass('autoclose')) {
                $(theDialog.blockout).click(function () {
                    theDialog.close();
                });
            }
        }
    });

    return dialog;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Enables common http request scenarios.
 * @module http
 * @requires jquery
 * @requires knockout
 */
define('plugins/http',['jquery', 'knockout'], function($, ko) {
    /**
     * @class HTTPModule
     * @static
     */
    return {
        /**
         * The name of the callback parameter to inject into jsonp requests by default.
         * @property {string} callbackParam
         * @default callback
         */
        callbackParam:'callback',
        /**
         * Makes an HTTP GET request.
         * @method get
         * @param {string} url The url to send the get request to.
         * @param {object} [query] An optional key/value object to transform into query string parameters.
         * @return {Promise} A promise of the get response data.
         */
        get:function(url, query) {
            return $.ajax(url, { data: query });
        },
        /**
         * Makes an JSONP request.
         * @method jsonp
         * @param {string} url The url to send the get request to.
         * @param {object} [query] An optional key/value object to transform into query string parameters.
         * @param {string} [callbackParam] The name of the callback parameter the api expects (overrides the default callbackParam).
         * @return {Promise} A promise of the response data.
         */
        jsonp: function (url, query, callbackParam) {
            if (url.indexOf('=?') == -1) {
                callbackParam = callbackParam || this.callbackParam;

                if (url.indexOf('?') == -1) {
                    url += '?';
                } else {
                    url += '&';
                }

                url += callbackParam + '=?';
            }

            return $.ajax({
                url: url,
                dataType:'jsonp',
                data:query
            });
        },
        /**
         * Makes an HTTP POST request.
         * @method post
         * @param {string} url The url to send the post request to.
         * @param {object} data The data to post. It will be converted to JSON. If the data contains Knockout observables, they will be converted into normal properties before serialization.
         * @return {Promise} A promise of the response data.
         */
        post:function(url, data) {
            return $.ajax({
                url: url,
                data: ko.toJSON(data),
                type: 'POST',
                contentType: 'application/json',
                dataType: 'json'
            });
        }
    };
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Enables automatic observability of plain javascript object for ES5 compatible browsers. Also, converts promise properties into observables that are updated when the promise resolves.
 * @module observable
 * @requires system
 * @requires binder
 * @requires knockout
 */
define('plugins/observable',['durandal/system', 'durandal/binder', 'knockout'], function(system, binder, ko) {
    var observableModule,
        toString = Object.prototype.toString,
        nonObservableTypes = ['[object Function]', '[object String]', '[object Boolean]', '[object Number]', '[object Date]', '[object RegExp]'],
        observableArrayMethods = ['remove', 'removeAll', 'destroy', 'destroyAll', 'replace'],
        arrayMethods = ['pop', 'reverse', 'sort', 'shift', 'splice'],
        additiveArrayFunctions = ['push', 'unshift'],
        arrayProto = Array.prototype,
        observableArrayFunctions = ko.observableArray.fn,
        logConversion = false;

    /**
     * You can call observable(obj, propertyName) to get the observable function for the specified property on the object.
     * @class ObservableModule
     */

    function shouldIgnorePropertyName(propertyName){
        var first = propertyName[0];
        return first === '_' || first === '$';
    }

    function isNode(obj) {
        return !!(obj && obj.nodeType !== undefined && system.isNumber(obj.nodeType));
    }

    function canConvertType(value) {
        if (!value || isNode(value) || value.ko === ko || value.jquery) {
            return false;
        }

        var type = toString.call(value);

        return nonObservableTypes.indexOf(type) == -1 && !(value === true || value === false);
    }

    function makeObservableArray(original, observable) {
        var lookup = original.__observable__, notify = true;

        if(lookup && lookup.__full__){
            return;
        }

        lookup = lookup || (original.__observable__ = {});
        lookup.__full__ = true;

        observableArrayMethods.forEach(function(methodName) {
            original[methodName] = function() {
                notify = false;
                var methodCallResult = observableArrayFunctions[methodName].apply(observable, arguments);
                notify = true;
                return methodCallResult;
            };
        });

        arrayMethods.forEach(function(methodName) {
            original[methodName] = function() {
                if(notify){
                    observable.valueWillMutate();
                }

                var methodCallResult = arrayProto[methodName].apply(original, arguments);

                if(notify){
                    observable.valueHasMutated();
                }

                return methodCallResult;
            };
        });

        additiveArrayFunctions.forEach(function(methodName){
            original[methodName] = function() {
                for (var i = 0, len = arguments.length; i < len; i++) {
                    convertObject(arguments[i]);
                }

                if(notify){
                    observable.valueWillMutate();
                }

                var methodCallResult = arrayProto[methodName].apply(original, arguments);

                if(notify){
                    observable.valueHasMutated();
                }

                return methodCallResult;
            };
        });

        original['splice'] = function() {
            for (var i = 2, len = arguments.length; i < len; i++) {
                convertObject(arguments[i]);
            }

            if(notify){
                observable.valueWillMutate();
            }

            var methodCallResult = arrayProto['splice'].apply(original, arguments);

            if(notify){
                observable.valueHasMutated();
            }

            return methodCallResult;
        };

        for (var i = 0, len = original.length; i < len; i++) {
            convertObject(original[i]);
        }
    }

    /**
     * Converts an entire object into an observable object by re-writing its attributes using ES5 getters and setters. Attributes beginning with '_' or '$' are ignored.
     * @method convertObject
     * @param {object} obj The target object to convert.
     */
    function convertObject(obj){
        var lookup, value;

        if(!canConvertType(obj)){
            return;
        }

        lookup = obj.__observable__;

        if(lookup && lookup.__full__){
            return;
        }

        lookup = lookup || (obj.__observable__ = {});
        lookup.__full__ = true;

        if (system.isArray(obj)) {
            var observable = ko.observableArray(obj);
            makeObservableArray(obj, observable);
        } else {
            for (var propertyName in obj) {
                if(shouldIgnorePropertyName(propertyName)){
                    continue;
                }

                if(!lookup[propertyName]){
                    value = obj[propertyName];

                    if(!system.isFunction(value)){
                        convertProperty(obj, propertyName, value);
                    }
                }
            }
        }

        if(logConversion) {
            system.log('Converted', obj);
        }
    }

    function innerSetter(observable, newValue, isArray) {
        var val;
        observable(newValue);
        val = observable.peek();

        //if this was originally an observableArray, then always check to see if we need to add/replace the array methods (if newValue was an entirely new array)
        if (isArray) {
            if (!val) {
                //don't allow null, force to an empty array
                val = [];
                observable(val);
                makeObservableArray(val, observable);
            }
            else if (!val.destroyAll) {
                makeObservableArray(val, observable);
            }
        } else {
            convertObject(val);
        }
    }

    /**
     * Converts a normal property into an observable property using ES5 getters and setters.
     * @method convertProperty
     * @param {object} obj The target object on which the property to convert lives.
     * @param {string} propertyName The name of the property to convert.
     * @param {object} [original] The original value of the property. If not specified, it will be retrieved from the object.
     * @return {KnockoutObservable} The underlying observable.
     */
    function convertProperty(obj, propertyName, original){
        var observable,
            isArray,
            lookup = obj.__observable__ || (obj.__observable__ = {});

        if(original === undefined){
            original = obj[propertyName];
        }

        if (system.isArray(original)) {
            observable = ko.observableArray(original);
            makeObservableArray(original, observable);
            isArray = true;
        } else if (typeof original == "function") {
            if(ko.isObservable(original)){
                observable = original;
            }else{
                return null;
            }
        } else if(system.isPromise(original)) {
            observable = ko.observable();

            original.then(function (result) {
                if(system.isArray(result)) {
                    var oa = ko.observableArray(result);
                    makeObservableArray(result, oa);
                    result = oa;
                }

                observable(result);
            });
        } else {
            observable = ko.observable(original);
            convertObject(original);
        }

        Object.defineProperty(obj, propertyName, {
            configurable: true,
            enumerable: true,
            get: observable,
            set: ko.isWriteableObservable(observable) ? (function (newValue) {
                if (newValue && system.isPromise(newValue)) {
                    newValue.then(function (result) {
                        innerSetter(observable, result, system.isArray(result));
                    });
                } else {
                    innerSetter(observable, newValue, isArray);
                }
            }) : undefined
        });

        lookup[propertyName] = observable;
        return observable;
    }

    /**
     * Defines a computed property using ES5 getters and setters.
     * @method defineProperty
     * @param {object} obj The target object on which to create the property.
     * @param {string} propertyName The name of the property to define.
     * @param {function|object} evaluatorOrOptions The Knockout computed function or computed options object.
     * @return {KnockoutObservable} The underlying computed observable.
     */
    function defineProperty(obj, propertyName, evaluatorOrOptions) {
        var computedOptions = { owner: obj, deferEvaluation: true },
            computed;

        if (typeof evaluatorOrOptions === 'function') {
            computedOptions.read = evaluatorOrOptions;
        } else {
            if ('value' in evaluatorOrOptions) {
                system.error('For defineProperty, you must not specify a "value" for the property. You must provide a "get" function.');
            }

            if (typeof evaluatorOrOptions.get !== 'function') {
                system.error('For defineProperty, the third parameter must be either an evaluator function, or an options object containing a function called "get".');
            }

            computedOptions.read = evaluatorOrOptions.get;
            computedOptions.write = evaluatorOrOptions.set;
        }

        computed = ko.computed(computedOptions);
        obj[propertyName] = computed;

        return convertProperty(obj, propertyName, computed);
    }

    observableModule = function(obj, propertyName){
        var lookup, observable, value;

        if (!obj) {
            return null;
        }

        lookup = obj.__observable__;
        if(lookup){
            observable = lookup[propertyName];
            if(observable){
                return observable;
            }
        }

        value = obj[propertyName];

        if(ko.isObservable(value)){
            return value;
        }

        return convertProperty(obj, propertyName, value);
    };

    observableModule.defineProperty = defineProperty;
    observableModule.convertProperty = convertProperty;
    observableModule.convertObject = convertObject;

    /**
     * Installs the plugin into the view model binder's `beforeBind` hook so that objects are automatically converted before being bound.
     * @method install
     */
    observableModule.install = function(options) {
        var original = binder.binding;

        binder.binding = function(obj, view, instruction) {
            if(instruction.applyBindings && !instruction.skipConversion){
                convertObject(obj);
            }

            original(obj, view);
        };

        logConversion = options.logConversion;
    };

    return observableModule;
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Serializes and deserializes data to/from JSON.
 * @module serializer
 * @requires system
 */
define('plugins/serializer',['durandal/system'], function(system) {
    /**
     * @class SerializerModule
     * @static
     */
    return {
        /**
         * The name of the attribute that the serializer should use to identify an object's type.
         * @property {string} typeAttribute
         * @default type
         */
        typeAttribute: 'type',
        /**
         * The amount of space to use for indentation when writing out JSON.
         * @property {string|number} space
         * @default undefined
         */
        space:undefined,
        /**
         * The default replacer function used during serialization. By default properties starting with '_' or '$' are removed from the serialized object.
         * @method replacer
         * @param {string} key The object key to check.
         * @param {object} value The object value to check.
         * @return {object} The value to serialize.
         */
        replacer: function(key, value) {
            if(key){
                var first = key[0];
                if(first === '_' || first === '$'){
                    return undefined;
                }
            }

            return value;
        },
        /**
         * Serializes the object.
         * @method serialize
         * @param {object} object The object to serialize.
         * @param {object} [settings] Settings can specify a replacer or space to override the serializer defaults.
         * @return {string} The JSON string.
         */
        serialize: function(object, settings) {
            settings = (settings === undefined) ? {} : settings;

            if(system.isString(settings) || system.isNumber(settings)) {
                settings = { space: settings };
            }

            return JSON.stringify(object, settings.replacer || this.replacer, settings.space || this.space);
        },
        /**
         * Gets the type id for an object instance, using the configured `typeAttribute`.
         * @method getTypeId
         * @param {object} object The object to serialize.
         * @return {string} The type.
         */
        getTypeId: function(object) {
            if (object) {
                return object[this.typeAttribute];
            }

            return undefined;
        },
        /**
         * Maps type ids to object constructor functions. Keys are type ids and values are functions.
         * @property {object} typeMap.
         */
        typeMap: {},
        /**
         * Adds a type id/constructor function mampping to the `typeMap`.
         * @method registerType
         * @param {string} typeId The type id.
         * @param {function} constructor The constructor.
         */
        registerType: function() {
            var first = arguments[0];

            if (arguments.length == 1) {
                var id = first[this.typeAttribute] || system.getModuleId(first);
                this.typeMap[id] = first;
            } else {
                this.typeMap[first] = arguments[1];
            }
        },
        /**
         * The default reviver function used during deserialization. By default is detects type properties on objects and uses them to re-construct the correct object using the provided constructor mapping.
         * @method reviver
         * @param {string} key The attribute key.
         * @param {object} value The object value associated with the key.
         * @param {function} getTypeId A custom function used to get the type id from a value.
         * @param {object} getConstructor A custom function used to get the constructor function associated with a type id.
         * @return {object} The value.
         */
        reviver: function(key, value, getTypeId, getConstructor) {
            var typeId = getTypeId(value);
            if (typeId) {
                var ctor = getConstructor(typeId);
                if (ctor) {
                    if (ctor.fromJSON) {
                        return ctor.fromJSON(value);
                    }

                    return new ctor(value);
                }
            }

            return value;
        },
        /**
         * Deserialize the JSON.
         * @method deserialize
         * @param {string} text The JSON string.
         * @param {object} [settings] Settings can specify a reviver, getTypeId function or getConstructor function.
         * @return {object} The deserialized object.
         */
        deserialize: function(text, settings) {
            var that = this;
            settings = settings || {};

            var getTypeId = settings.getTypeId || function(object) { return that.getTypeId(object); };
            var getConstructor = settings.getConstructor || function(id) { return that.typeMap[id]; };
            var reviver = settings.reviver || function(key, value) { return that.reviver(key, value, getTypeId, getConstructor); };

            return JSON.parse(text, reviver);
        }
    };
});

/**
 * Durandal 2.0.1 Copyright (c) 2012 Blue Spire Consulting, Inc. All Rights Reserved.
 * Available via the MIT license.
 * see: http://durandaljs.com or https://github.com/BlueSpire/Durandal for details.
 */
/**
 * Layers the widget sugar on top of the composition system.
 * @module widget
 * @requires system
 * @requires composition
 * @requires jquery
 * @requires knockout
 */
define('plugins/widget',['durandal/system', 'durandal/composition', 'jquery', 'knockout'], function(system, composition, $, ko) {
    var kindModuleMaps = {},
        kindViewMaps = {},
        bindableSettings = ['model', 'view', 'kind'],
        widgetDataKey = 'durandal-widget-data';

    function extractParts(element, settings){
        var data = ko.utils.domData.get(element, widgetDataKey);

        if(!data){
            data = {
                parts:composition.cloneNodes(ko.virtualElements.childNodes(element))
            };

            ko.virtualElements.emptyNode(element);
            ko.utils.domData.set(element, widgetDataKey, data);
        }

        settings.parts = data.parts;
    }

    /**
     * @class WidgetModule
     * @static
     */
    var widget = {
        getSettings: function(valueAccessor) {
            var settings = ko.utils.unwrapObservable(valueAccessor()) || {};

            if (system.isString(settings)) {
                return { kind: settings };
            }

            for (var attrName in settings) {
                if (ko.utils.arrayIndexOf(bindableSettings, attrName) != -1) {
                    settings[attrName] = ko.utils.unwrapObservable(settings[attrName]);
                } else {
                    settings[attrName] = settings[attrName];
                }
            }

            return settings;
        },
        /**
         * Creates a ko binding handler for the specified kind.
         * @method registerKind
         * @param {string} kind The kind to create a custom binding handler for.
         */
        registerKind: function(kind) {
            ko.bindingHandlers[kind] = {
                init: function() {
                    return { controlsDescendantBindings: true };
                },
                update: function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
                    var settings = widget.getSettings(valueAccessor);
                    settings.kind = kind;
                    extractParts(element, settings);
                    widget.create(element, settings, bindingContext, true);
                }
            };

            ko.virtualElements.allowedBindings[kind] = true;
            composition.composeBindings.push(kind + ':');
        },
        /**
         * Maps views and module to the kind identifier if a non-standard pattern is desired.
         * @method mapKind
         * @param {string} kind The kind name.
         * @param {string} [viewId] The unconventional view id to map the kind to.
         * @param {string} [moduleId] The unconventional module id to map the kind to.
         */
        mapKind: function(kind, viewId, moduleId) {
            if (viewId) {
                kindViewMaps[kind] = viewId;
            }

            if (moduleId) {
                kindModuleMaps[kind] = moduleId;
            }
        },
        /**
         * Maps a kind name to it's module id. First it looks up a custom mapped kind, then falls back to `convertKindToModulePath`.
         * @method mapKindToModuleId
         * @param {string} kind The kind name.
         * @return {string} The module id.
         */
        mapKindToModuleId: function(kind) {
            return kindModuleMaps[kind] || widget.convertKindToModulePath(kind);
        },
        /**
         * Converts a kind name to it's module path. Used to conventionally map kinds who aren't explicitly mapped through `mapKind`.
         * @method convertKindToModulePath
         * @param {string} kind The kind name.
         * @return {string} The module path.
         */
        convertKindToModulePath: function(kind) {
            return 'widgets/' + kind + '/viewmodel';
        },
        /**
         * Maps a kind name to it's view id. First it looks up a custom mapped kind, then falls back to `convertKindToViewPath`.
         * @method mapKindToViewId
         * @param {string} kind The kind name.
         * @return {string} The view id.
         */
        mapKindToViewId: function(kind) {
            return kindViewMaps[kind] || widget.convertKindToViewPath(kind);
        },
        /**
         * Converts a kind name to it's view id. Used to conventionally map kinds who aren't explicitly mapped through `mapKind`.
         * @method convertKindToViewPath
         * @param {string} kind The kind name.
         * @return {string} The view id.
         */
        convertKindToViewPath: function(kind) {
            return 'widgets/' + kind + '/view';
        },
        createCompositionSettings: function(element, settings) {
            if (!settings.model) {
                settings.model = this.mapKindToModuleId(settings.kind);
            }

            if (!settings.view) {
                settings.view = this.mapKindToViewId(settings.kind);
            }

            settings.preserveContext = true;
            settings.activate = true;
            settings.activationData = settings;
            settings.mode = 'templated';

            return settings;
        },
        /**
         * Creates a widget.
         * @method create
         * @param {DOMElement} element The DOMElement or knockout virtual element that serves as the target element for the widget.
         * @param {object} settings The widget settings.
         * @param {object} [bindingContext] The current binding context.
         */
        create: function(element, settings, bindingContext, fromBinding) {
            if(!fromBinding){
                settings = widget.getSettings(function() { return settings; }, element);
            }

            var compositionSettings = widget.createCompositionSettings(element, settings);

            composition.compose(element, compositionSettings, bindingContext);
        },
        /**
         * Installs the widget module by adding the widget binding handler and optionally registering kinds.
         * @method install
         * @param {object} config The module config. Add a `kinds` array with the names of widgets to automatically register. You can also specify a `bindingName` if you wish to use another name for the widget binding, such as "control" for example.
         */
        install:function(config){
            config.bindingName = config.bindingName || 'widget';

            if(config.kinds){
                var toRegister = config.kinds;

                for(var i = 0; i < toRegister.length; i++){
                    widget.registerKind(toRegister[i]);
                }
            }

            ko.bindingHandlers[config.bindingName] = {
                init: function() {
                    return { controlsDescendantBindings: true };
                },
                update: function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) {
                    var settings = widget.getSettings(valueAccessor);
                    extractParts(element, settings);
                    widget.create(element, settings, bindingContext, true);
                }
            };

            composition.composeBindings.push(config.bindingName + ':');
            ko.virtualElements.allowedBindings[config.bindingName] = true;
        }
    };

    return widget;
});

/**
 * @license RequireJS text 2.0.12 Copyright (c) 2010-2014, The Dojo Foundation All Rights Reserved.
 * Available via the MIT or new BSD license.
 * see: http://github.com/requirejs/text for details
 */
/*jslint regexp: true */
/*global require, XMLHttpRequest, ActiveXObject,
  define, window, process, Packages,
  java, location, Components, FileUtils */

define('text',['module'], function (module) {
    'use strict';

    var text, fs, Cc, Ci, xpcIsWindows,
        progIds = ['Msxml2.XMLHTTP', 'Microsoft.XMLHTTP', 'Msxml2.XMLHTTP.4.0'],
        xmlRegExp = /^\s*<\?xml(\s)+version=[\'\"](\d)*.(\d)*[\'\"](\s)*\?>/im,
        bodyRegExp = /<body[^>]*>\s*([\s\S]+)\s*<\/body>/im,
        hasLocation = typeof location !== 'undefined' && location.href,
        defaultProtocol = hasLocation && location.protocol && location.protocol.replace(/\:/, ''),
        defaultHostName = hasLocation && location.hostname,
        defaultPort = hasLocation && (location.port || undefined),
        buildMap = {},
        masterConfig = (module.config && module.config()) || {};

    text = {
        version: '2.0.12',

        strip: function (content) {
            //Strips <?xml ...?> declarations so that external SVG and XML
            //documents can be added to a document without worry. Also, if the string
            //is an HTML document, only the part inside the body tag is returned.
            if (content) {
                content = content.replace(xmlRegExp, "");
                var matches = content.match(bodyRegExp);
                if (matches) {
                    content = matches[1];
                }
            } else {
                content = "";
            }
            return content;
        },

        jsEscape: function (content) {
            return content.replace(/(['\\])/g, '\\$1')
                .replace(/[\f]/g, "\\f")
                .replace(/[\b]/g, "\\b")
                .replace(/[\n]/g, "\\n")
                .replace(/[\t]/g, "\\t")
                .replace(/[\r]/g, "\\r")
                .replace(/[\u2028]/g, "\\u2028")
                .replace(/[\u2029]/g, "\\u2029");
        },

        createXhr: masterConfig.createXhr || function () {
            //Would love to dump the ActiveX crap in here. Need IE 6 to die first.
            var xhr, i, progId;
            if (typeof XMLHttpRequest !== "undefined") {
                return new XMLHttpRequest();
            } else if (typeof ActiveXObject !== "undefined") {
                for (i = 0; i < 3; i += 1) {
                    progId = progIds[i];
                    try {
                        xhr = new ActiveXObject(progId);
                    } catch (e) {}

                    if (xhr) {
                        progIds = [progId];  // so faster next time
                        break;
                    }
                }
            }

            return xhr;
        },

        /**
         * Parses a resource name into its component parts. Resource names
         * look like: module/name.ext!strip, where the !strip part is
         * optional.
         * @param {String} name the resource name
         * @returns {Object} with properties "moduleName", "ext" and "strip"
         * where strip is a boolean.
         */
        parseName: function (name) {
            var modName, ext, temp,
                strip = false,
                index = name.indexOf("."),
                isRelative = name.indexOf('./') === 0 ||
                             name.indexOf('../') === 0;

            if (index !== -1 && (!isRelative || index > 1)) {
                modName = name.substring(0, index);
                ext = name.substring(index + 1, name.length);
            } else {
                modName = name;
            }

            temp = ext || modName;
            index = temp.indexOf("!");
            if (index !== -1) {
                //Pull off the strip arg.
                strip = temp.substring(index + 1) === "strip";
                temp = temp.substring(0, index);
                if (ext) {
                    ext = temp;
                } else {
                    modName = temp;
                }
            }

            return {
                moduleName: modName,
                ext: ext,
                strip: strip
            };
        },

        xdRegExp: /^((\w+)\:)?\/\/([^\/\\]+)/,

        /**
         * Is an URL on another domain. Only works for browser use, returns
         * false in non-browser environments. Only used to know if an
         * optimized .js version of a text resource should be loaded
         * instead.
         * @param {String} url
         * @returns Boolean
         */
        useXhr: function (url, protocol, hostname, port) {
            var uProtocol, uHostName, uPort,
                match = text.xdRegExp.exec(url);
            if (!match) {
                return true;
            }
            uProtocol = match[2];
            uHostName = match[3];

            uHostName = uHostName.split(':');
            uPort = uHostName[1];
            uHostName = uHostName[0];

            return (!uProtocol || uProtocol === protocol) &&
                   (!uHostName || uHostName.toLowerCase() === hostname.toLowerCase()) &&
                   ((!uPort && !uHostName) || uPort === port);
        },

        finishLoad: function (name, strip, content, onLoad) {
            content = strip ? text.strip(content) : content;
            if (masterConfig.isBuild) {
                buildMap[name] = content;
            }
            onLoad(content);
        },

        load: function (name, req, onLoad, config) {
            //Name has format: some.module.filext!strip
            //The strip part is optional.
            //if strip is present, then that means only get the string contents
            //inside a body tag in an HTML string. For XML/SVG content it means
            //removing the <?xml ...?> declarations so the content can be inserted
            //into the current doc without problems.

            // Do not bother with the work if a build and text will
            // not be inlined.
            if (config && config.isBuild && !config.inlineText) {
                onLoad();
                return;
            }

            masterConfig.isBuild = config && config.isBuild;

            var parsed = text.parseName(name),
                nonStripName = parsed.moduleName +
                    (parsed.ext ? '.' + parsed.ext : ''),
                url = req.toUrl(nonStripName),
                useXhr = (masterConfig.useXhr) ||
                         text.useXhr;

            // Do not load if it is an empty: url
            if (url.indexOf('empty:') === 0) {
                onLoad();
                return;
            }

            //Load the text. Use XHR if possible and in a browser.
            if (!hasLocation || useXhr(url, defaultProtocol, defaultHostName, defaultPort)) {
                text.get(url, function (content) {
                    text.finishLoad(name, parsed.strip, content, onLoad);
                }, function (err) {
                    if (onLoad.error) {
                        onLoad.error(err);
                    }
                });
            } else {
                //Need to fetch the resource across domains. Assume
                //the resource has been optimized into a JS module. Fetch
                //by the module name + extension, but do not include the
                //!strip part to avoid file system issues.
                req([nonStripName], function (content) {
                    text.finishLoad(parsed.moduleName + '.' + parsed.ext,
                                    parsed.strip, content, onLoad);
                });
            }
        },

        write: function (pluginName, moduleName, write, config) {
            if (buildMap.hasOwnProperty(moduleName)) {
                var content = text.jsEscape(buildMap[moduleName]);
                write.asModule(pluginName + "!" + moduleName,
                               "define(function () { return '" +
                                   content +
                               "';});\n");
            }
        },

        writeFile: function (pluginName, moduleName, req, write, config) {
            var parsed = text.parseName(moduleName),
                extPart = parsed.ext ? '.' + parsed.ext : '',
                nonStripName = parsed.moduleName + extPart,
                //Use a '.js' file name so that it indicates it is a
                //script that can be loaded across domains.
                fileName = req.toUrl(parsed.moduleName + extPart) + '.js';

            //Leverage own load() method to load plugin value, but only
            //write out values that do not have the strip argument,
            //to avoid any potential issues with ! in file names.
            text.load(nonStripName, req, function (value) {
                //Use own write() method to construct full module value.
                //But need to create shell that translates writeFile's
                //write() to the right interface.
                var textWrite = function (contents) {
                    return write(fileName, contents);
                };
                textWrite.asModule = function (moduleName, contents) {
                    return write.asModule(moduleName, fileName, contents);
                };

                text.write(pluginName, nonStripName, textWrite, config);
            }, config);
        }
    };

    if (masterConfig.env === 'node' || (!masterConfig.env &&
            typeof process !== "undefined" &&
            process.versions &&
            !!process.versions.node &&
            !process.versions['node-webkit'])) {
        //Using special require.nodeRequire, something added by r.js.
        fs = require.nodeRequire('fs');

        text.get = function (url, callback, errback) {
            try {
                var file = fs.readFileSync(url, 'utf8');
                //Remove BOM (Byte Mark Order) from utf8 files if it is there.
                if (file.indexOf('\uFEFF') === 0) {
                    file = file.substring(1);
                }
                callback(file);
            } catch (e) {
                if (errback) {
                    errback(e);
                }
            }
        };
    } else if (masterConfig.env === 'xhr' || (!masterConfig.env &&
            text.createXhr())) {
        text.get = function (url, callback, errback, headers) {
            var xhr = text.createXhr(), header;
            xhr.open('GET', url, true);

            //Allow plugins direct access to xhr headers
            if (headers) {
                for (header in headers) {
                    if (headers.hasOwnProperty(header)) {
                        xhr.setRequestHeader(header.toLowerCase(), headers[header]);
                    }
                }
            }

            //Allow overrides specified in config
            if (masterConfig.onXhr) {
                masterConfig.onXhr(xhr, url);
            }

            xhr.onreadystatechange = function (evt) {
                var status, err;
                //Do not explicitly handle errors, those should be
                //visible via console output in the browser.
                if (xhr.readyState === 4) {
                    status = xhr.status || 0;
                    if (status > 399 && status < 600) {
                        //An http 4xx or 5xx error. Signal an error.
                        err = new Error(url + ' HTTP status: ' + status);
                        err.xhr = xhr;
                        if (errback) {
                            errback(err);
                        }
                    } else {
                        callback(xhr.responseText);
                    }

                    if (masterConfig.onXhrComplete) {
                        masterConfig.onXhrComplete(xhr, url);
                    }
                }
            };
            xhr.send(null);
        };
    } else if (masterConfig.env === 'rhino' || (!masterConfig.env &&
            typeof Packages !== 'undefined' && typeof java !== 'undefined')) {
        //Why Java, why is this so awkward?
        text.get = function (url, callback) {
            var stringBuffer, line,
                encoding = "utf-8",
                file = new java.io.File(url),
                lineSeparator = java.lang.System.getProperty("line.separator"),
                input = new java.io.BufferedReader(new java.io.InputStreamReader(new java.io.FileInputStream(file), encoding)),
                content = '';
            try {
                stringBuffer = new java.lang.StringBuffer();
                line = input.readLine();

                // Byte Order Mark (BOM) - The Unicode Standard, version 3.0, page 324
                // http://www.unicode.org/faq/utf_bom.html

                // Note that when we use utf-8, the BOM should appear as "EF BB BF", but it doesn't due to this bug in the JDK:
                // http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4508058
                if (line && line.length() && line.charAt(0) === 0xfeff) {
                    // Eat the BOM, since we've already found the encoding on this file,
                    // and we plan to concatenating this buffer with others; the BOM should
                    // only appear at the top of a file.
                    line = line.substring(1);
                }

                if (line !== null) {
                    stringBuffer.append(line);
                }

                while ((line = input.readLine()) !== null) {
                    stringBuffer.append(lineSeparator);
                    stringBuffer.append(line);
                }
                //Make sure we return a JavaScript string and not a Java string.
                content = String(stringBuffer.toString()); //String
            } finally {
                input.close();
            }
            callback(content);
        };
    } else if (masterConfig.env === 'xpconnect' || (!masterConfig.env &&
            typeof Components !== 'undefined' && Components.classes &&
            Components.interfaces)) {
        //Avert your gaze!
        Cc = Components.classes;
        Ci = Components.interfaces;
        Components.utils['import']('resource://gre/modules/FileUtils.jsm');
        xpcIsWindows = ('@mozilla.org/windows-registry-key;1' in Cc);

        text.get = function (url, callback) {
            var inStream, convertStream, fileObj,
                readData = {};

            if (xpcIsWindows) {
                url = url.replace(/\//g, '\\');
            }

            fileObj = new FileUtils.File(url);

            //XPCOM, you so crazy
            try {
                inStream = Cc['@mozilla.org/network/file-input-stream;1']
                           .createInstance(Ci.nsIFileInputStream);
                inStream.init(fileObj, 1, 0, false);

                convertStream = Cc['@mozilla.org/intl/converter-input-stream;1']
                                .createInstance(Ci.nsIConverterInputStream);
                convertStream.init(inStream, "utf-8", inStream.available(),
                Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);

                convertStream.readString(inStream.available(), readData);
                convertStream.close();
                inStream.close();
                callback(readData.value);
            } catch (e) {
                throw new Error((fileObj && fileObj.path || '') + ': ' + e);
            }
        };
    }
    return text;
});


define('text!account/account.html',[],function () { return '<section id="account" class="account unselectable" data-bind="autofocus">\n    <!-- ko compose: header --><!-- /ko -->\n    <article class="content">\n        <h2 data-translate-text="[login header]"></h2>\n        <!-- ko if: socialNetworks -->\n        <!-- ko compose: socialNetworks --><!-- /ko -->\n        <div class="or-separator">\n            <span class="or-separator-text" data-translate-text="[or]"></span>\n        </div>\n        <!-- /ko -->\n        <div class="input-container" data-bind="css: { error: !email.isValid() && email.isModified() }">\n            <input class="input-field autofocus" type="email" ondrop="return false;"\n                data-bind="value: email, valueUpdate: \'input\',  css: { empty: !email().length }, event: { blur: email.trim }, hasFocus: email.hasFocus, submitOnEnter: submit" />\n            <label class="input-label" data-translate-text="[login email field]"></label>\n            <span class="input-icon icon-mail"></span>\n            <p class="error-message" data-translate-text="[tracking and tracing email is not valid]"></p>\n        </div>\n        <button class="button primary small autofocus" data-bind="click: submit, css: { important: !requestProcessing() }">\n            <!-- ko ifnot: requestProcessing -->\n            <span data-translate-text="[OK, email are correct]"></span>\n            <!-- /ko -->\n            <!-- ko if: requestProcessing -->\n            <div class="pulsar-loader align-center">\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n            </div>\n            <!-- /ko -->\n        </button>\n        <!-- ko if: skipAllowed -->\n        <button class="button link" data-bind="click: skip" data-translate-text="[login skip]"></button>\n        <!-- /ko -->\n        <hr />\n        <p class="account-hint" data-translate-text="[login explanation]"></p>\n    </article>\n</section>\n';});


define('text!account/header/index.html',[],function () { return '<header data-bind="focusHandler: {}">\n    <div class="page-header">\n        <h1 class="course-title" data-bind="css: { small: isCourseTitleTooLong }">\n            <span class="course-caption" data-bind="text: courseTitle, reviewable: { type: \'course\', property: \'title\' }"></span>\n        </h1>\n    </div>\n</header>\n';});


define('text!account/limitAccess/index.html',[],function () { return '<section id="noaccess" class="unselectable">\n    <header class="page-header">\n        <h1 class="course-title" data-bind="css: {small: courseTitle.length>100}">\n            <span class="course-caption" data-bind="text: courseTitle, reviewable: {type: \'course\', property: \'title\'}">"course_name"</span>\n        </h1>\n    </header>\n    <section class="content noaccess">\n        <div class="noaccess-icon">\n            <span class="icon-attention"></span>\n        </div>\n        <div class="noaccess-notification" data-translate-text="[the owner of the course has limited the access]"></div>\n        <button class="button primary small autofocus back-to-signin" data-bind="click: backToSignin">\n            <span class="icon-home"></span>\n            <span data-translate-text="[back to sign in]" data-translate-title="[back to sign in]"></span>\n        </button>\n    </section>\n</section>\n';});


define('text!account/register/index.html',[],function () { return '<section id="register" class="account unselectable" data-bind="autofocus">\n    <!-- ko compose: header --><!-- /ko -->\n    <article class="content second-step">\n        <div class="back-button-container" data-bind="click: back">\n            <span class="back-button-arrow icon-left_arrow"></span><span class="back-button-text" data-translate-text="[back]"></span>\n        </div>\n        <!-- ko if: crossDeviceEnabled -->\n        <h2 data-translate-text="[set your name and password]"></h2>\n        <!-- /ko -->\n        <!-- ko ifnot: crossDeviceEnabled -->\n        <h2 data-translate-text="[set your name]"></h2>\n        <!-- /ko -->\n        <div class="input-container" data-bind="css: { error: !name.isValid() && name.isModified() }">\n            <input class="input-field" type="text" ondrop="return false;"\n                data-bind="value: name, valueUpdate: \'afterkeydown\',  css: { empty: !name().length, autofocus: !name().length }, event: { blur: name.trim }, hasFocus: name.hasFocus, submitOnEnter: submit" />\n            <label class="input-label" data-translate-text="[login name field]"></label>\n            <span class="input-icon icon-name"></span>\n            <p class="error-message" data-translate-text="[tracking and tracing name is not valid]"></p>\n        </div>\n        <!-- ko if: crossDeviceEnabled -->\n        <div class="input-container" data-bind="css: { error: !password.isValid() && password.isModified() }">\n            <input class="input-field" type="password" ondrop="return false;"\n                data-bind="value: password, valueUpdate: \'afterkeydown\',  css: { empty: !password().length, autofocus: name().length }, event: { keyup: password.onKeyUp, blur: password.trim }, hasFocus: password.hasFocus, attr: { type: isPasswordVisible() ? \'text\' : \'password\'}, submitOnEnter: submit" />\n            <label class="input-label" data-translate-text="[password]"></label>\n            <span class="input-icon icon-password2"></span>\n            <span class="input-icon align-right pointer password" data-bind="click: togglePasswordVisibility, css: { \'icon-eye_lock\': isPasswordVisible(), \'icon-eye\': !isPasswordVisible() }"></span>\n            <div class="validation-container">\n                <span class="validator" data-bind="css: { correct: password.moreThanSevenSymbols }">\n                    <span class="validator-icon">\n                        <span class="icon-check_mark"></span>\n                    </span>\n                    <span class="validator-text" data-translate-text="[at least 7 characters]"></span>\n                </span>\n                <span class="validator" data-bind="css: { correct: password.noSpaces }">\n                    <span class="validator-icon">\n                        <span class="icon-check_mark"></span>\n                    </span>\n                    <span class="validator-text" data-translate-text="[no spaces]"></span>\n                </span>\n            </div>\n        </div>\n        <button class="button check small" data-bind="click: toggleRememberMe, css: { active: rememberMe }">\n            <span data-translate-text="[remember me on this device]"></span>\n        </button>\n        <!-- /ko -->\n        <button class="button primary small" data-bind="click: submit">\n            <!-- ko ifnot: requestProcessing -->\n            <span data-translate-text="[OK, go to the course]"></span>\n            <!-- /ko -->\n            <!-- ko if: requestProcessing -->\n            <div class="pulsar-loader align-center">\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n            </div>\n            <!-- /ko -->\n        </button>\n    </article>\n</section>';});


define('text!account/signin/index.html',[],function () { return '<section id="signin" class="account unselectable" data-bind="autofocus">\n    <!-- ko compose: header --><!-- /ko -->\n    <article class="content second-step">\n        <div class="back-button-container" data-bind="click: back">\n            <span class="back-button-arrow icon-left_arrow"></span><span class="back-button-text" data-translate-text="[back]"></span>\n        </div>\n        <h2 data-translate-text="[type your password]"></h2>\n        <h3>(<span data-bind="text: email"></span>)</h3>\n        <div class="input-container" data-bind="css: { error: !password.isValid() && password.isModified() || emailPasswordCombination() }">\n            <input class="input-field autofocus" type="password" ondrop="return false;"\n                data-bind="value: password, valueUpdate: \'input\',  css: { empty: !password().length }, event: { blur: password.trim }, hasFocus: password.hasFocus, attr: { type: isPasswordVisible() ? \'text\' : \'password\'}, submitOnEnter: submit" />\n            <label class="input-label" data-translate-text="[password]"></label>\n            <span class="input-icon icon-password2"></span>\n            <span class="input-icon align-right pointer password" data-bind="click: togglePasswordVisibility, css: { \'icon-eye_lock\': isPasswordVisible(), \'icon-eye\': !isPasswordVisible() }"></span>\n            <p class="error-message" data-bind="visible: !emailPasswordCombination()" data-translate-text="[login email / password combination is not valid]" data-translate-title="[login email / password combination is not valid]"></p>\n            <p class="error-message" data-bind="visible: emailPasswordCombination()" data-translate-text="[login email / password combination is not valid]" data-translate-title="[login email / password combination is not valid]"></p>\n            <p class="forgot-password" data-bind="click: forgotPassword" data-translate-text="[forgot password]"></p>\n        </div>\n        <button class="button check small" data-bind="click: toggleRememberMe, css: { active: rememberMe }">\n            <span data-translate-text="[remember me on this device]"></span>\n        </button>\n        <button class="button primary small autofocus" data-bind="click: submit">\n            <!-- ko ifnot: requestProcessing -->\n            <span data-translate-text="[OK, go to the course]"></span>\n            <!-- /ko -->\n            <!-- ko if: requestProcessing -->\n            <div class="pulsar-loader align-center">\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n                <div class="pulsar-loader-item"></div>\n            </div>\n            <!-- /ko -->\n        </button>\n        <div class="account-hint">\n        </div>\n        <!-- ko if: isRestorePasswordEmailSent() -->\n        <div class="secret-link-confirm-container show">\n            <span class="secret-link-confirm-wrapper">\n                <span class="icon-envelope"></span>\n                <!-- ko if: isRestorePasswordEmailSent -->\n                <span class="secret-link-confirm-text" data-translate-text="[forgot password email sent]" data-translate-title="[forgot password email sent]"></span>\n                <!-- /ko -->\n            </span>\n        </div>\n        <!-- /ko -->\n    </article>\n</section>';});


define('text!account/socialNetworks/index.html',[],function () { return '<div class="social-container">\n    <h4 data-translate-text="[authorize with]"></h4>\n    <div class="link-wrapper">\n        <a href="#" data-bind="attr: { href: facebookAuthLink }" class="social-button facebook"><span class="icon-facebook"></span></a>\n        <a href="#" data-bind="attr: { href: linkedinAuthLink }" class="social-button linkedin"><span class="icon-linkedin"></span></a>\n        <a href="#" data-bind="attr: { href: googleAuthLink }" class="social-button google"><span class="icon-google"></span></a>\n    </div>\n</div>';});


define('text!dialogs/continueLater.html',[],function () { return '<section class="popup-layout" data-bind="scrollbar: { checkDOMChanges: true }, keyDown: { 27: { handler: close } }">\n    <div class="popup">\n        <span class="close-btn material-icons" data-bind="click: close">clear</span>\n        <!-- ko ifnot: progressStorageActivated -->\n        <div class="popup-wrapper">\n            <article>\n                <header>\n                    <h2>\n                        <span data-translate-text="[note that]"></span>\n                    </h2>\n                </header>\n                <section>\n                    <ul>\n                        <!-- ko if: exit -->\n                        <li data-translate-text="[finish course take brake first remark]"></li>\n                        <!-- ko if: isSecondRemarkShouldBeShown -->\n                        <li data-translate-text="[finish course take brake second remark]"></li>\n                        <!-- /ko -->\n                        <!-- /ko -->\n                    </ul>\n                </section>\n                <footer data-bind="autofocus">\n                    <span class="button-wrapper">\n                        <button class="button block default" data-bind="click: close" data-translate-text="[cancel, continue course]" data-translate-title="[cancel, continue course]"></button>\n                    </span>\n                    <!-- ko if: exit -->\n                    <span class="button-wrapper">\n                        <button class="button block primary autofocus" data-bind="click: exit" data-translate-text="[OK, close course]" data-translate-title="[OK, close course]"></button>\n                    </span>\n                    <!-- /ko -->\n                </footer>\n            </article>\n        </div>\n        <!-- /ko -->\n        <!-- ko if: progressStorageActivated -->\n        <div class="popup-wrapper">\n            <article class="without-note">\n                <div class="icon-container pt70">\n                    <span class="icon-wrapper">\n                        <span class="icon-all_devices"></span>\n                    </span>\n                </div>\n                <header>\n                    <h3>\n                        <span data-translate-text="[continue anytime]"></span>\n                    </h3>\n                    <p>\n                        <span>\n                            <span data-translate-text="[get secret link to]"></span>&nbsp;<span class="bold" data-bind="text: email"></span><span data-translate-text="[you can sign in to the course]"></span>\n                        </span>\n                    </p>\n                </header>\n                <section>\n                    <div class="popup-button-wrapper">\n                        <button class="button primary small autofocus" data-bind="click: sendSecretLink, css: { hide: isSecretLinkSent }">\n                            <span data-translate-text="[send me the link]"></span>\n                        </button>\n                        <div class="secret-link-confirm-container" data-bind="css: { show: isSecretLinkSent }">\n                            <span class="secret-link-confirm-wrapper">\n                                <span class="icon-envelope"></span>\n                                <span class="secret-link-confirm-text" data-translate-text="[the letter has been successfully sent]" data-translate-title="[the letter has been successfully sent]"></span>\n                            </span>\n                        </div>\n                    </div>\n                    <div class="popup-button-wrapper">\n                        <button class="button small default" data-bind="click: exit, changeClass: { delay: 500, fromClass: \'default\', toClass: \'primary\', value: isSecretLinkSent }">\n                            <!-- ko if: keepMeLoggedIn -->\n                            <span data-translate-text="[close course]"></span>\n                            <!-- /ko -->\n                            <!-- ko ifnot: keepMeLoggedIn -->\n                            <span data-translate-text="[close course and logout]"></span>\n                            <!-- /ko -->\n                        </button>\n                    </div>\n                    <div class="popup-button-wrapper">\n                        <button class="button check small" data-bind="click: toggleKeepMeLoggedIn, css: { active: keepMeLoggedIn }">\n                            <span data-translate-text="[remember me on this device]"></span>\n                        </button>\n                    </div>\n                </section>\n            </article>\n        </div>\n        <!-- ko ifnot: progressStorageActivated -->\n        <div class="popup-wrapper">\n            <article class="without-note">\n                <header>\n                    <div class="icon-container">\n                        <span class="icon-no_connection"></span>\n                    </div>\n                    <h3>\n                        <span data-translate-text="[continue anytime in browser]"></span>\n                    </h3>\n                    <p>\n                        <span data-translate-text="[note about continue course]"></span>\n                    </p>\n                </header>\n                <footer class="background-transparent">\n                    <span class="button-wrapper">\n                        <button class="button block primary small" data-bind="click: exit" data-translate-text="[OK, close course]" data-translate-title="[OK, close course]"></button>\n                    </span>\n                </footer>\n            </article>\n        </div>\n        <!-- /ko -->\n        <!-- /ko -->\n    </div>\n</section>';});


define('text!dialogs/nps/figures/evaluation.html',[],function () { return '<?xml version="1.0" encoding="UTF-8"?>\n<svg width="300px" height="120px" viewBox="0 0 300 120" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">\n    <!-- Generator: Sketch 44.1 (41455) - http://www.bohemiancoding.com/sketch -->\n    <title>Group</title>\n    <desc>Created with Sketch.</desc>\n    <defs>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-1">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-2">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n        <circle id="path-3" cx="151" cy="50" r="32"></circle>\n        <filter x="-13.3%" y="-10.2%" width="126.6%" height="126.6%" filterUnits="objectBoundingBox" id="filter-4">\n            <feOffset dx="0" dy="2" in="SourceAlpha" result="shadowOffsetOuter1"></feOffset>\n            <feGaussianBlur stdDeviation="2.5" in="shadowOffsetOuter1" result="shadowBlurOuter1"></feGaussianBlur>\n            <feColorMatrix values="0 0 0 0 0   0 0 0 0 0   0 0 0 0 0  0 0 0 0.1 0" type="matrix" in="shadowBlurOuter1"></feColorMatrix>\n        </filter>\n    </defs>\n    <g id="Page-1" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd">\n        <g id="nps-xapi-enabled" transform="translate(-577.000000, -236.000000)">\n            <g id="Group" transform="translate(577.000000, 236.000000)">\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M85.9375,24.4478528 L24.2138056,24.4478528 C24.2138056,24.4478528 13.3928571,25.5180635 13.3928571,33.404908 C13.3928571,41.2917525 24.2138056,42.3619632 24.2138056,42.3619632 L60.9375,42.4259422 C60.9375,42.4259422 69.53125,43.9614373 69.53125,50.8711656 C69.53125,57.780894 60.9375,59.8282209 60.9375,59.8282209 L41.9642857,59.8282209 C41.9642857,59.8282209 26.5625,60.6902176 26.5625,69.2331288 C26.5625,77.77604 41.9642857,79.0858896 41.9642857,79.0858896 L10.9375,79.0858896 C10.9375,79.0858896 5.84683802e-16,80.8133216 0,87.9789658 C0,95.14461 10.9375,97 10.9375,97 L250.782522,97 C250.782522,97 264.285714,95.5924628 264.285714,87.9789658 C264.285714,80.3654689 250.782522,79.0858896 250.782522,79.0858896 L227.331152,79.0858896 C227.331152,79.0858896 213.637423,78.3376034 213.392857,70.5766871 C213.148291,62.8157708 227.331152,61.363716 227.331152,61.363716 L254.545455,60.5 C254.545455,60.5 268.75,62.4194542 268.75,51.6389132 C268.75,40.8583723 255.681818,42.3619632 255.681818,42.3619632 L288.28125,42.4259422 C288.28125,42.4259422 300,41.7802657 300,33.404908 C300,25.0295503 288.28125,24 288.28125,24 L85.9375,24.4478528 Z" id="Path-8" opacity="0.05" filter="url(#filter-1)"></path>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M216.405846,39.484982 C216.088518,39.8101473 215.943659,40.2786626 216.018645,40.7379796 L216.37108,42.8989795 L214.525398,41.8786865 C214.329411,41.7702184 214.11411,41.7162233 213.898923,41.7162233 C213.683735,41.7162233 213.468548,41.7704573 213.272447,41.8786865 L211.426766,42.8989795 L211.7792,40.7379796 C211.854187,40.2787821 211.709327,39.8102668 211.391999,39.484982 L209.898979,37.9547217 L211.962462,37.6393519 C212.400904,37.5724553 212.780038,37.282769 212.976139,36.8650235 L213.899036,34.8989795 L214.82182,36.8650235 C215.018034,37.282769 215.396942,37.5723358 215.835497,37.6393519 L217.898979,37.9547217 L216.405846,39.484982 Z" id="Shape-Copy-5" fill-rule="nonzero" opacity="0.300000012" transform="translate(213.898979, 38.898979) scale(-1, 1) rotate(-15.000000) translate(-213.898979, -38.898979) "></path>\n                <ellipse data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" id="Oval-6" opacity="0.100000001" filter="url(#filter-2)" cx="151" cy="117" rx="30" ry="3"></ellipse>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M252,68 L252,65.9970301 C252,65.4530363 251.552285,65 251,65 C250.443865,65 250,65.4463856 250,65.9970301 L250,68 L247.99703,68 C247.453036,68 247,68.4477153 247,69 C247,69.5561352 247.446386,70 247.99703,70 L250,70 L250,72.0029699 C250,72.5469637 250.447715,73 251,73 C251.556135,73 252,72.5536144 252,72.0029699 L252,70 L254.00297,70 C254.546964,70 255,69.5522847 255,69 C255,68.4438648 254.553614,68 254.00297,68 L252,68 Z" id="Combined-Shape" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M93,107 L93,104.99703 C93,104.453036 92.5522847,104 92,104 C91.4438648,104 91,104.446386 91,104.99703 L91,107 L88.9970301,107 C88.4530363,107 88,107.447715 88,108 C88,108.556135 88.4463856,109 88.9970301,109 L91,109 L91,111.00297 C91,111.546964 91.4477153,112 92,112 C92.5561352,112 93,111.553614 93,111.00297 L93,109 L95.0029699,109 C95.5469637,109 96,108.552285 96,108 C96,107.443865 95.5536144,107 95.0029699,107 L93,107 Z" id="Combined-Shape-Copy-3" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M61,6 L61,3.99703014 C61,3.45303631 60.5522847,3 60,3 C59.4438648,3 59,3.4463856 59,3.99703014 L59,6 L56.9970301,6 C56.4530363,6 56,6.44771525 56,7 C56,7.55613518 56.4463856,8 56.9970301,8 L59,8 L59,10.0029699 C59,10.5469637 59.4477153,11 60,11 C60.5561352,11 61,10.5536144 61,10.0029699 L61,8 L63.0029699,8 C63.5469637,8 64,7.55228475 64,7 C64,6.44386482 63.5536144,6 63.0029699,6 L61,6 Z" id="Combined-Shape-Copy" opacity="0.300000012"></path>\n                <circle data-bind="lessColor: [{ color: textColor, attr: \'stroke\' }]" id="Oval-7" stroke-width="2" opacity="0.300000012" cx="31" cy="47" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: textColor, attr: \'stroke\' }]" id="Oval-7-Copy" stroke-width="2" opacity="0.300000012" cx="234" cy="4" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: textColor, attr: \'stroke\' }]" id="Oval-7-Copy-2" stroke-width="2" opacity="0.300000012" cx="199" cy="107" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: textColor, attr: \'stroke\' }]" id="Oval-7-Copy-3" stroke-width="2" opacity="0.300000012" cx="78" cy="73" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: correctColor, attr: \'fill\' }]" id="Oval-Copy-2" cx="125" cy="28" r="12"></circle>\n                <circle data-bind="lessColor: [{ color: correctColor, attr: \'fill\' }]" id="Oval-Copy-3" cx="174" cy="73" r="12"></circle>\n                <g id="Oval-Copy">\n                    <use fill="black" fill-opacity="1" filter="url(#filter-4)" xlink:href="#path-3"></use>\n                    <use data-bind="lessColor: [{ color: correctColor, attr: \'fill\' }]" fill-rule="evenodd" xlink:href="#path-3"></use>\n                </g>\n                <polyline data-bind="lessColor: [{ color: buttonTextColor, attr: \'stroke\' }]" id="Path-2-Copy" stroke-width="2" points="140 51.8381204 147.706943 59 163 43"></polyline>\n            </g>\n        </g>\n    </g>\n</svg>';});


define('text!dialogs/nps/figures/feedback.html',[],function () { return '<?xml version="1.0" encoding="UTF-8"?>\n<svg width="300px" height="112px" viewBox="0 0 300 112" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">\n    <!-- Generator: Sketch 44.1 (41455) - http://www.bohemiancoding.com/sketch -->\n    <title>Group</title>\n    <desc>Created with Sketch.</desc>\n    <defs>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-1">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-2">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n    </defs>\n    <g id="Page-1" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd">\n        <g id="8-10-rate-feedback" transform="translate(-570.000000, -275.000000)">\n            <g id="Group" transform="translate(570.000000, 275.000000)">\n                <path data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\', func: \'fadein\', dimension: \'10%\' }]" d="M85.9375,24.4478528 L24.2138056,24.4478528 C24.2138056,24.4478528 13.3928571,25.5180635 13.3928571,33.404908 C13.3928571,41.2917525 24.2138056,42.3619632 24.2138056,42.3619632 L60.9375,42.4259422 C60.9375,42.4259422 69.53125,43.9614373 69.53125,50.8711656 C69.53125,57.780894 60.9375,59.8282209 60.9375,59.8282209 L41.9642857,59.8282209 C41.9642857,59.8282209 26.5625,60.6902176 26.5625,69.2331288 C26.5625,77.77604 41.9642857,79.0858896 41.9642857,79.0858896 L10.9375,79.0858896 C10.9375,79.0858896 5.84683802e-16,80.8133216 0,87.9789658 C0,95.14461 10.9375,97 10.9375,97 L250.782522,97 C250.782522,97 264.285714,95.5924628 264.285714,87.9789658 C264.285714,80.3654689 250.782522,79.0858896 250.782522,79.0858896 L227.331152,79.0858896 C227.331152,79.0858896 213.637423,78.3376034 213.392857,70.5766871 C213.148291,62.8157708 227.331152,61.363716 227.331152,61.363716 L254.545455,60.5 C254.545455,60.5 268.75,62.4194542 268.75,51.6389132 C268.75,40.8583723 255.681818,42.3619632 255.681818,42.3619632 L288.28125,42.4259422 C288.28125,42.4259422 300,41.7802657 300,33.404908 C300,25.0295503 288.28125,24 288.28125,24 L85.9375,24.4478528 Z" id="Path-8" opacity="0.1" filter="url(#filter-1)"></path>\n                <path data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\', func: \'fadein\', dimension: \'10%\' }]" d="M216.405846,39.484982 C216.088518,39.8101473 215.943659,40.2786626 216.018645,40.7379796 L216.37108,42.8989795 L214.525398,41.8786865 C214.329411,41.7702184 214.11411,41.7162233 213.898923,41.7162233 C213.683735,41.7162233 213.468548,41.7704573 213.272447,41.8786865 L211.426766,42.8989795 L211.7792,40.7379796 C211.854187,40.2787821 211.709327,39.8102668 211.391999,39.484982 L209.898979,37.9547217 L211.962462,37.6393519 C212.400904,37.5724553 212.780038,37.282769 212.976139,36.8650235 L213.899036,34.8989795 L214.82182,36.8650235 C215.018034,37.282769 215.396942,37.5723358 215.835497,37.6393519 L217.898979,37.9547217 L216.405846,39.484982 Z" id="Shape-Copy-5" fill-rule="nonzero" opacity="0.300000012" transform="translate(213.898979, 38.898979) scale(-1, 1) rotate(-15.000000) translate(-213.898979, -38.898979) "></path>\n                <ellipse data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" id="Oval-6" opacity="0.1" filter="url(#filter-2)" cx="150" cy="105" rx="30" ry="4"></ellipse>\n                <path data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\', func: \'fadein\', dimension: \'10%\' }]" d="M252,68 L252,65.9970301 C252,65.4530363 251.552285,65 251,65 C250.443865,65 250,65.4463856 250,65.9970301 L250,68 L247.99703,68 C247.453036,68 247,68.4477153 247,69 C247,69.5561352 247.446386,70 247.99703,70 L250,70 L250,72.0029699 C250,72.5469637 250.447715,73 251,73 C251.556135,73 252,72.5536144 252,72.0029699 L252,70 L254.00297,70 C254.546964,70 255,69.5522847 255,69 C255,68.4438648 254.553614,68 254.00297,68 L252,68 Z" id="Combined-Shape" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\', func: \'fadein\', dimension: \'10%\' }]" d="M93,107 L93,104.99703 C93,104.453036 92.5522847,104 92,104 C91.4438648,104 91,104.446386 91,104.99703 L91,107 L88.9970301,107 C88.4530363,107 88,107.447715 88,108 C88,108.556135 88.4463856,109 88.9970301,109 L91,109 L91,111.00297 C91,111.546964 91.4477153,112 92,112 C92.5561352,112 93,111.553614 93,111.00297 L93,109 L95.0029699,109 C95.5469637,109 96,108.552285 96,108 C96,107.443865 95.5536144,107 95.0029699,107 L93,107 Z" id="Combined-Shape-Copy-3" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\', func: \'fadein\', dimension: \'10%\' }]" d="M61,6 L61,3.99703014 C61,3.45303631 60.5522847,3 60,3 C59.4438648,3 59,3.4463856 59,3.99703014 L59,6 L56.9970301,6 C56.4530363,6 56,6.44771525 56,7 C56,7.55613518 56.4463856,8 56.9970301,8 L59,8 L59,10.0029699 C59,10.5469637 59.4477153,11 60,11 C60.5561352,11 61,10.5536144 61,10.0029699 L61,8 L63.0029699,8 C63.5469637,8 64,7.55228475 64,7 C64,6.44386482 63.5536144,6 63.0029699,6 L61,6 Z" id="Combined-Shape-Copy" opacity="0.300000012"></path>\n                <circle data-bind="lessColor: [{ color: feedbackColor, attr: \'stroke\', func: \'fade\', dimension: \'40%\'}]" id="Oval-7" stroke-width="2" opacity="0.300000012" cx="31" cy="49" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: feedbackColor, attr: \'stroke\', func: \'fade\', dimension: \'40%\'}]" id="Oval-7-Copy" stroke-width="2" opacity="0.300000012" cx="234" cy="4" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: feedbackColor, attr: \'stroke\', func: \'fade\', dimension: \'40%\'}]" id="Oval-7-Copy-2" stroke-width="2" opacity="0.300000012" cx="199" cy="107" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: feedbackColor, attr: \'stroke\', func: \'fade\', dimension: \'40%\'}]" id="Oval-7-Copy-3" stroke-width="2" opacity="0.300000012" cx="78" cy="73" r="3"></circle>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'stroke\' }]" d="M147.5,79.5 L147.5,100 C147.5,101.380712 148.619288,102.5 150,102.5 C151.380712,102.5 152.5,101.380712 152.5,100 L152.5,79.5 L147.5,79.5 Z" id="Rectangle-2-Copy" opacity="0.3"></path>\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M148,79 L152,79 L152,100 L152,100 C152,101.104569 151.104569,102 150,102 L150,102 L150,102 C148.895431,102 148,101.104569 148,100 L148,79 Z" id="Rectangle-2-Copy-2" opacity="0.2"></path>\n                <rect id="Rectangle-26" data-bind="lessColor: [{ color: feedbackColor, attr: \'fill\' }]" x="120" y="20" width="60" height="60" rx="5"></rect>\n                <rect id="Rectangle-26-Copy" data-bind="lessColor: [{ color: feedbackColor, attr: \'stroke\', func: \'shade\', dimension: \'10%\'}]"  stroke-width="2" x="121" y="21" width="58" height="58" rx="5"></rect>\n                <text y="62" transform="translate(100)" id="8" font-family="RobotoSlab-Regular, Roboto Slab" font-size="32" font-weight="normal" fill="#FFFFFF">\n                     <tspan x="50" text-anchor="middle" data-bind="text: score"></tspan>\n                </text>\n            </g>\n        </g>\n    </g>\n</svg>';});


define('text!dialogs/nps/figures/rocket.html',[],function () { return '<?xml version="1.0" encoding="UTF-8"?>\n<svg width="300px" height="141px" viewBox="0 0 300 141" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">\n    <!-- Generator: Sketch 44.1 (41455) - http://www.bohemiancoding.com/sketch -->\n    <title>Group 2</title>\n    <desc>Created with Sketch.</desc>\n    <defs>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-1">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n        <filter x="0.0%" y="0.0%" width="100.0%" height="100.0%" filterUnits="objectBoundingBox" id="filter-2">\n            <feGaussianBlur stdDeviation="0" in="SourceGraphic"></feGaussianBlur>\n        </filter>\n    </defs>\n    <g id="Page-1" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd">\n        <g id="thank-you-screen" transform="translate(-570.000000, -213.000000)">\n            <g id="Group-2" transform="translate(570.000000, 213.000000)">\n                <path data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" d="M85.9375,47.4478528 L24.2138056,47.4478528 C24.2138056,47.4478528 13.3928571,48.5180635 13.3928571,56.404908 C13.3928571,64.2917525 24.2138056,65.3619632 24.2138056,65.3619632 L60.9375,65.4259422 C60.9375,65.4259422 69.53125,66.9614373 69.53125,73.8711656 C69.53125,80.780894 60.9375,82.8282209 60.9375,82.8282209 L41.9642857,82.8282209 C41.9642857,82.8282209 26.5625,83.6902176 26.5625,92.2331288 C26.5625,100.77604 41.9642857,102.08589 41.9642857,102.08589 L10.9375,102.08589 C10.9375,102.08589 5.84683802e-16,103.813322 0,110.978966 C0,118.14461 10.9375,120 10.9375,120 L250.782522,120 C250.782522,120 264.285714,118.592463 264.285714,110.978966 C264.285714,103.365469 250.782522,102.08589 250.782522,102.08589 L227.331152,102.08589 C227.331152,102.08589 213.637423,101.337603 213.392857,93.5766871 C213.148291,85.8157708 227.331152,84.363716 227.331152,84.363716 L254.545455,83.5 C254.545455,83.5 268.75,85.4194542 268.75,74.6389132 C268.75,63.8583723 255.681818,65.3619632 255.681818,65.3619632 L288.28125,65.4259422 C288.28125,65.4259422 300,64.7802657 300,56.404908 C300,48.0295503 288.28125,47 288.28125,47 L85.9375,47.4478528 Z" id="Path-8" opacity="0.05" filter="url(#filter-1)"></path>\n                <g id="startup" transform="translate(151.861436, 54.861436) scale(-1, 1) rotate(-330.000000) translate(-151.861436, -54.861436) translate(112.361436, 15.361436)">\n                    <g id="Layer_1" transform="translate(-0.000000, 0.000000)">\n                        <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M55.932,51.6565578 L51.6567151,51.4921056 L51.8211673,55.7673904 C51.9856195,62.8382052 57.9052689,68.7578546 64.9760837,68.9223068 L69.2513685,69.086759 L69.0869163,64.8114741 C68.9224641,57.7406594 63.0026574,51.82101 55.932,51.6565578 Z" id="Shape" fill-opacity="0.5" fill-rule="nonzero"></path>\n                        <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M78.9529462,47.8744721 C78.1308426,46.3945598 76.9796773,45.0790996 75.8286693,43.9279343 C72.8688446,40.9681096 69.2512112,38.9948406 65.3046733,38.0082849 C57.9051116,36.1994681 49.6832888,38.1727371 43.9279343,43.9279343 C38.1725797,49.6831315 36.1994681,57.9049542 38.0082849,65.3046733 C38.9948406,69.086759 40.9681096,72.8686873 43.9279343,75.828512 C45.0789422,76.9795199 46.5588546,78.1306853 47.8744721,78.9527888 C39.323745,79.2816932 30.4442709,76.1574163 23.8668127,69.5799582 C18.6048147,64.3179602 15.64499,57.9049542 14.6584343,51.1630438 C13.3429741,41.4613088 16.467251,31.4306693 23.8668127,24.0311076 C31.2663745,16.6315458 41.4614661,13.5072689 50.998749,14.8227291 C57.7406594,15.8092849 64.3181175,18.7692669 69.4156633,24.0311076 C76.1575737,30.4442709 79.2818506,39.3239024 78.9529462,47.8744721 Z" id="Shape" fill-rule="nonzero" opacity="0.600000024"></path>\n                        <path data-bind="lessColor: [{ color: contentBodyColor, attr: \'fill\' }, { color: mainColor, attr: \'stroke\' }]" d="M64.8114741,48.0389243 C60.2072849,19.9204323 36.0351733,-0.798498008 6.10770717,0.0236055777 L0.188057769,0.188057769 L0.0236055777,6.10770717 C-0.798498008,36.0351733 19.9204323,60.2072849 48.0389243,64.8114741 L64.8114741,48.0389243 Z" id="Shape" stroke-width="1.73333333" fill-rule="nonzero"></path>\n                        <g id="Group" transform="translate(-0.000000, 0.000000)"></g>\n                        <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M48.0389243,64.8114741 C49.6832888,65.1403785 51.4921056,65.3048307 53.1364701,65.4692829 L65.3046733,53.3010797 C65.1402211,51.4922629 64.9757689,49.8478984 64.6468645,48.2035339 L48.0389243,64.8114741 Z" id="Shape" fill-rule="nonzero"></path>\n                        <circle data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" id="Oval" fill-rule="nonzero" cx="27.7496176" cy="27.8199622" r="11.7197802" opacity="0.5"></circle>\n                        <circle data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" id="Oval" fill-rule="nonzero" cx="27.4635159" cy="27.5338606" r="8.8794741" opacity="0.2"></circle>\n                        <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M21.8937012,23.2091614 C25.0179781,20.0848845 29.9510717,20.0848845 33.0753486,23.2091614 C36.0351733,26.1689861 36.1996255,30.9376275 33.404253,34.0619044 C33.5687052,33.8974522 33.7331574,33.8974522 33.7331574,33.733 C37.1863386,30.2798187 37.1863386,24.6890737 33.7331574,21.2358924 C30.2799761,17.7827112 24.6892311,17.7827112 21.2360498,21.2358924 C17.7828685,24.6890737 17.7828685,30.2798187 21.2360498,33.733 C21.400502,33.8974522 21.5649542,33.8974522 21.5649542,34.0619044 C18.9338765,30.9376275 18.9338765,26.1689861 21.8937012,23.2091614 Z" id="Shape" fill="#ACDEDB" fill-rule="nonzero"></path>\n                    </g>\n                </g>\n                <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M216.405846,62.484982 C216.088518,62.8101473 215.943659,63.2786626 216.018645,63.7379796 L216.37108,65.8989795 L214.525398,64.8786865 C214.329411,64.7702184 214.11411,64.7162233 213.898923,64.7162233 C213.683735,64.7162233 213.468548,64.7704573 213.272447,64.8786865 L211.426766,65.8989795 L211.7792,63.7379796 C211.854187,63.2787821 211.709327,62.8102668 211.391999,62.484982 L209.898979,60.9547217 L211.962462,60.6393519 C212.400904,60.5724553 212.780038,60.282769 212.976139,59.8650235 L213.899036,57.8989795 L214.82182,59.8650235 C215.018034,60.282769 215.396942,60.5723358 215.835497,60.6393519 L217.898979,60.9547217 L216.405846,62.484982 Z" id="Shape-Copy-5" fill-rule="nonzero" opacity="0.300000012" transform="translate(213.898979, 61.898979) scale(-1, 1) rotate(-15.000000) translate(-213.898979, -61.898979) "></path>\n                <ellipse data-bind="lessColor: [{ color: textColor, attr: \'fill\' }]" id="Oval-6" opacity="0.100000001" filter="url(#filter-2)" cx="148" cy="137" rx="25" ry="4"></ellipse>\n                <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M252,91 L252,88.9970301 C252,88.4530363 251.552285,88 251,88 C250.443865,88 250,88.4463856 250,88.9970301 L250,91 L247.99703,91 C247.453036,91 247,91.4477153 247,92 C247,92.5561352 247.446386,93 247.99703,93 L250,93 L250,95.0029699 C250,95.5469637 250.447715,96 251,96 C251.556135,96 252,95.5536144 252,95.0029699 L252,93 L254.00297,93 C254.546964,93 255,92.5522847 255,92 C255,91.4438648 254.553614,91 254.00297,91 L252,91 Z" id="Combined-Shape" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M93,130 L93,127.99703 C93,127.453036 92.5522847,127 92,127 C91.4438648,127 91,127.446386 91,127.99703 L91,130 L88.9970301,130 C88.4530363,130 88,130.447715 88,131 C88,131.556135 88.4463856,132 88.9970301,132 L91,132 L91,134.00297 C91,134.546964 91.4477153,135 92,135 C92.5561352,135 93,134.553614 93,134.00297 L93,132 L95.0029699,132 C95.5469637,132 96,131.552285 96,131 C96,130.443865 95.5536144,130 95.0029699,130 L93,130 Z" id="Combined-Shape-Copy-3" opacity="0.300000012"></path>\n                <path data-bind="lessColor: [{ color: mainColor, attr: \'fill\' }]" d="M61,29 L61,26.9970301 C61,26.4530363 60.5522847,26 60,26 C59.4438648,26 59,26.4463856 59,26.9970301 L59,29 L56.9970301,29 C56.4530363,29 56,29.4477153 56,30 C56,30.5561352 56.4463856,31 56.9970301,31 L59,31 L59,33.0029699 C59,33.5469637 59.4477153,34 60,34 C60.5561352,34 61,33.5536144 61,33.0029699 L61,31 L63.0029699,31 C63.5469637,31 64,30.5522847 64,30 C64,29.4438648 63.5536144,29 63.0029699,29 L61,29 Z" id="Combined-Shape-Copy" opacity="0.300000012"></path>\n                <circle data-bind="lessColor: [{ color: mainColor, attr: \'stroke\' }]" id="Oval-7" stroke-width="2" opacity="0.300000012" cx="31" cy="72" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: mainColor, attr: \'stroke\' }]" id="Oval-7-Copy" stroke-width="2" opacity="0.300000012" cx="234" cy="27" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: mainColor, attr: \'stroke\' }]" id="Oval-7-Copy-2" stroke-width="2" opacity="0.300000012" cx="199" cy="130" r="3"></circle>\n                <circle data-bind="lessColor: [{ color: mainColor, attr: \'stroke\' }]" id="Oval-7-Copy-3" stroke-width="2" opacity="0.300000012" cx="78" cy="96" r="3"></circle>\n            </g>\n        </g>\n    </g>\n</svg>';});


define('text!dialogs/nps/nps.html',[],function () { return '<section class="fullscreen-popup-container" data-bind="css: { attached: isCompositionComplete }">\n    <div class="fp-content-row">\n        <div class="fp-content-cell">\n            <div class="fp-content">\n                <div class="fp-content-step evaluation" data-bind="css: { show: isEvaluationStep && isCompositionComplete, hide: !isEvaluationStep() }">\n                    <figure class="fp-image">\n                        <!-- ko compose: { model: \'dialogs/nps/figures/evaluation\' } -->\n                        <!-- /ko -->\n                    </figure>\n                    <h1 class="fp-title" data-bind="text: evaluationStepTitle"></h1>\n                    <h2 class="fp-subtitle" data-translate-text="[rate your overall experience]"></h2>\n                    <div class="fp-rating-container">\n                        <div class="rating-control" data-bind="if: isEvaluationStep, rating: { value: score, max: 10, onValueUpdated: goToFeedbackStep }"></div>\n                        <div class="fp-rating-tips">\n                            <span class="fp-rating-tip left" data-translate-text="[poor]"></span>\n                            <span class="fp-rating-tip right" data-translate-text="[excellent]"></span>\n                        </div>\n                    </div>\n                    <hr />\n                </div>\n                <div class="fp-content-step feedback" data-bind="css: { show: isFeedbackStep, hide: !isFeedbackStep() }">\n                    <h1 class="fp-title" data-translate-text="[you rated this course with]"></h1>\n                    <figure class="fp-image">\n                        <!-- ko compose: { model: \'dialogs/nps/figures/feedback\', activationData: score() } -->\n                        <!-- /ko -->\n                    </figure>\n                    <h2 class="fp-subtitle" data-translate-text="[course score reason]"></h2>\n                        <div class="fp-field">\n                            <div class="fp-field-input" data-bind="hasFocus: isFeedbackEditing, editableText: { text: feedback, onEnterKeyPress: submit }, scrollbar: { checkDOMChanges: true }, attr: { contentEditable: isFeedbackStep }"></div>\n                        </div>\n                        <div class="btn-submit-wrapper">\n                            <button href="#" class="button small primary" data-bind="click: submit, disable: isReporting() || !isFeedbackStep(), css: { progress: isReporting }" data-translate-title="[submit]">\n                                <!-- ko if: isReporting -->\n                                    <span class="button-caption-loader"></span>\n                                <!-- /ko -->\n                                <!-- ko ifnot: isReporting -->\n                                    <span class="button-caption" data-translate-text="[submit]"></span>\n                                <!-- /ko -->\n                            </button>\n                        </div>\n                        <hr />\n                </div>\n                <div class="fp-content-step last-step delay" data-bind="css: { show: isLastStep, hide: !isEvaluationStep() && !isLastStep() }">\n                    <figure class="fp-image">\n                        <!-- ko compose: { model: \'dialogs/nps/figures/rocket\' } -->\n                        <!-- /ko -->\n                    </figure>\n                    <h1 class="fp-title" data-translate-text="[thank you for your feedback]"></h1>\n                    <!-- ko ifnot: isCourseEmbedded -->\n                    <h2 class="fp-subtitle" data-translate-text="[you can close the page now]"></h2>\n                    <!-- /ko -->\n                    <hr />\n                </div>\n            </div>\n        </div>\n    </div>\n</section>';});


define('text!dialogs/resultsResend/resultsResend.html',[],function () { return '<section class="fullscreen-popup-container resend-results" data-bind="css: { attached: isCompositionComplete }, keyDown: { 27: { handler: hideDialog } }">\n    <div class="fp-content-row">\n        <div class="fp-content-cell">\n            <div class="fp-content">\n                <div class="fp-content-step send-error" data-bind="css: { show: isCompositionComplete }">\n                    <h3 data-bind="text: resultsSendErrorTitleText"></h3>\n                    <div class="note" data-bind="css: { shake: resendingFailed }">\n                        <p>\n                            <span class="material-icons incorrect-color">info</span>\n                            <span data-translate-text="[try to submit again key]"></span>\n                        </p>\n                        <p>\n                            <span class="material-icons incorrect-color">info</span>\n                            <span data-translate-text="[contanct with author key]"></span>\n                        </p>\n                        <p>\n                            <span class="material-icons incorrect-color">info</span>\n                            <span data-translate-text="[general-message-key]"></span>\n                            <span class="error-message" data-translate-text="[error-message-key]"></span>\n                        </p>\n                    </div>\n                    <div class="result-buttons" data-bind="autofocus">\n                        <button class="button block cancel default" data-bind="click: hideDialog, css: { disabled: isResendingResults }, disable: isResendingResults" data-translate-text="[back]" data-translate-title="[back]"></button>\n                        <button href="#" class="button block primary autofocus submit" data-bind="click: resendResults, disable: isResendingResults, css: { progress: isResendingResults }"\n                            data-translate-title="[resend-results-key]">\n                            <!-- ko if: isResendingResults -->\n                            <span class="button-caption-loader"></span>\n                            <span>&nbsp;</span>\n                            <!-- /ko -->\n                            <!-- ko ifnot: isResendingResults -->\n                            <span class="button-caption" data-translate-text="[resend-results-key]"></span>\n                            <!-- /ko -->\n                        </button>\n                        <!-- ko if: retriesExceeded -->\n                        <div class="link-container">\n                            <a class="button link" data-bind="click: skipResultsSending, localize: { text: { key: \'[close without reporting results]\', replace: { endpointName: endpointNameText } } }"></a>\n                        </div>\n                        <!-- /ko -->\n                    </div>\n                </div>\n            </div>\n            <div class="overlay-close-button" data-bind="click: hideDialog">\n                <span class="overlay-close-button-icon material-icons">clear</span>\n                <div class="overlay-close-button-title">esc</div>\n            </div>\n        </div>\n    </div>\n</section>';});


define('text!dialogs/startNewAttempt/startNewAttempt.html',[],function () { return '<section class="fullscreen-popup-container start-new-attempt" data-bind="css: { attached: isCompositionComplete }, keyDown: { 27: { handler: hide } }">\n    <div class="fp-content-row">\n        <div class="fp-content-cell">\n            <div class="fp-content">\n                <div class="fp-content-step evaluation" data-bind="css: { show: isCompositionComplete }">\n                    <h3 data-translate-text="[start new attempt dialog title]"></h3>\n                    <div class="note">\n                        <p><span class="material-icons text-main-color">info</span> <span data-translate-text="[start new attempt dialog text]"></span></p>\n                    </div>\n                    <div class="result-buttons" data-bind="autofocus">\n                        <button class="button block cancel default" data-bind="click: hide, css: { disabled: isStartingNewAttempt }, disable: isStartingNewAttempt" data-translate-text="[back]" data-translate-title="[back]"></button>\n                        <button href="#" class="button block primary autofocus submit" data-bind="click: startNewAttempt, disable: isStartingNewAttempt, css: { progress: isStartingNewAttempt }"\n                            data-translate-title="[start new attempt]">\n                            <!-- ko if: isStartingNewAttempt -->\n                            <span class="button-caption-loader"></span>\n                            <span>&nbsp;</span>\n                            <!-- /ko -->\n                            <!-- ko ifnot: isStartingNewAttempt -->\n                            <span class="button-caption" data-translate-text="[start new attempt]"></span>\n                            <!-- /ko -->\n                        </button>\n                    </div>\n                </div>\n            </div>\n            <div class="overlay-close-button" data-bind="click: hide">\n                <span class="overlay-close-button-icon material-icons">clear</span>\n                <div class="overlay-close-button-title">esc</div>\n            </div>\n        </div>\n    </div>\n</section>';});


define('text!header/header.html',[],function () { return '<section class="header clearfix" data-bind="css: \'module-\' + cssName(), style: { height: (height() ? height() + \'px\' : \'\') }">\n    <div class="header-overlay" data-bind="css: { fullscreen: !backgroundProps.isBodyEnabled }, style: { backgroundImage: \'url(\'+ backgroundProps.image.url +\')\', backgroundPosition: backgroundProps.image.backgroundPosition, backgroundSize: backgroundProps.image.backgroundSize, backgroundRepeat: backgroundProps.image.backgroundRepeat, backgroundColor: backgroundProps.color }"></div>\n    <!-- ko if: backgroundProps.brightness > 0 -->\n    <div class="header-brightness-holder" data-bind="css: { fullscreen: !backgroundProps.isBodyEnabled }, style: {backgroundColor: \'white\', opacity: backgroundProps.brightness }"></div>\n    <!-- /ko -->\n    <!-- ko ifnot: backgroundProps.brightness > 0 -->\n    <div class="header-brightness-holder" data-bind="css: { fullscreen: !backgroundProps.isBodyEnabled }, style: {backgroundColor: \'black\', opacity: -backgroundProps.brightness }"></div>\n    <!-- /ko -->\n    <section class="header-items-holder-wrapper">\n        <div class="header-items-holder" data-bind="css: { \'review-attempt\' : viewSettings().isInReviewAttemptMode }">\n            <span class="logo-img">\n                <!-- ko if: viewSettings().rootLinkEnabled  -->\n                <!-- ko if: logo.url -->\n                <a href="#sections">\n                    <img alt="" src="" data-bind="attr: { src: logo.url }, style: { \'max-width\': logo.maxWidth, \'max-height\': logo.maxHeight }">\n                </a>\n                <!-- /ko -->\n                <!-- ko ifnot: logo.url -->\n                <a href="#sections">\n                    <span class="default-logo"></span>\n                </a>\n                <!-- /ko -->\n                <!-- /ko -->\n                <!-- ko ifnot: viewSettings().rootLinkEnabled -->\n                <!-- ko if: logo.url -->\n                <img alt="" src="" data-bind="attr: { src: logo.url }, style: { \'max-width\': logo.maxWidth, \'max-height\': logo.maxHeight }">\n                <!-- /ko -->\n                <!-- ko ifnot: logo.url -->\n                <span class="default-logo"></span>\n                <!-- /ko -->\n                <!-- /ko -->\n            </span>\n\n            <!-- ko if: pdfExportEnabled -->\n            <div class="download-course-container">\n                <div class="download-course-wrapper default" data-bind="downloadAsPdf: { title: title, serviceUrl: pdfServiceUrl, version: version }">\n                    <div class="download-as-pdf-button" data-translate-title="[download course in pdf format]"></div>\n                    <div class="download-as-pdf-status-wrapper">\n                        <div class="download-as-pdf-status">\n                            <span class="default" data-translate-text="[download as pdf]"></span>\n                            <span class="proggress" data-translate-text="[download as pdf preparing]"></span>\n                            <span class="error" data-translate-text="[download as pdf error]"></span>\n                            <span class="server-error"></span>\n                        </div>\n                    </div>\n                    <span class="disabled-tooltip-text">This option doesn\'t work in preview. Publish the course to use the PDF download option.</span>\n                </div>\n            </div>\n            <!-- /ko -->\n            <div class="summary-container">\n                <!-- ko if: viewSettings().exitButtonVisible -->\n                <!-- ko compose: \'viewmodels/exitCourse\' -->\n                <!-- /ko -->\n                <!-- /ko -->\n            </div>\n        </div>\n    </section>\n</section>';});


define('text!questionContent/questionContent.html',[],function () { return '<div class="question-header" data-bind="css: {\'survey\': isSurvey, \'correct\': isCorrectAnswered() && !isSurvey, \'incorrect\': !isCorrectAnswered() && !isSurvey}, attr: {\'data-count\': index}">\n    <span class="question-header-icon" data-bind="css: {\'icon-survey\': isSurvey, \'icon-correct_check\': isCorrectAnswered() && !isSurvey, \'icon-incorrect_cross\': !isCorrectAnswered() && !isSurvey}"></span>\n    <div class="question-title">\n        <a class="question-title-link" data-bind="text: title, attr: {href: navigationContext.currentQuestionUrl}"></a>\n    </div>\n</div>\n<div class="question-content">\n    <!-- ko compose: activeQuestionViewModel --><!-- /ko -->\n</div>';});


define('text!resultPage/resultPage.html',[],function () { return '<section data-bind="blockBodyScroll: npsDialog.isVisible() || newAttemptDialog.isVisible() || resendResultsDialog.isVisible()">\r\n    <div class="result-header" data-bind="headerHeightDepender">\r\n        <div class="course-title">\r\n            <span class="course-caption" data-bind=\'text: title\'>""</span>\r\n        </div>\r\n        <div class="overall-progress-wrapper">\r\n            <div class="overall-progress">\r\n                <div class="overall-progress-progressbar-wrapper">\r\n                    <div class="overall-progress-progressbar" data-bind="css: { success: isCompleted, fail: !isCompleted }">\r\n                        <canvas class="overall-progress-progressbar-canvas canvas-styles result" width="159" height="159" data-bind="circleProgress: { progress: score(), lineWidth:2,radius:75, isMastered: isCompleted }"></canvas>\r\n                        <div class="overall-progress-progressbar-text">\r\n                            <div class="score" data-bind="text: score()">0</div>\r\n                            <!-- ko if: isCompleted -->\r\n                            <div class="status correct-color" data-translate-text="[success]"></div>\r\n                            <!-- /ko -->\r\n                            <!-- ko ifnot: isCompleted -->\r\n                            <div class="status incorrect-color" data-translate-text="[failed]"></div>\r\n                            <!-- /ko -->\r\n                        </div>\r\n                    </div>\r\n                </div>\r\n            </div>\r\n        </div>\r\n        <div class="result-buttons">\r\n            <span class="button-wrapper" data-bind="autofocus">\r\n                <span class="button-wrapper">\r\n                    <!-- ko ifnot: isInReviewAttemptMode -->\r\n                    <button class="button block default cancel" data-bind="click: close" data-translate-text="[continue course]" data-translate-title="[continue course]"></button>\r\n                    <!-- /ko -->\r\n                    <!-- ko if: isInReviewAttemptMode -->\r\n                    <button class="button block cancel" data-bind="click: close, css: { primary: isCompleted, default: !isCompleted, autofocus: isCompleted }" data-translate-text="[review attempt]" data-translate-title="[review attempt]"></button>\r\n                    <!-- /ko -->\r\n                </span>                \r\n                <!-- ko ifnot: isInReviewAttemptMode -->\r\n                    <!-- ko if: ((!xAPIEnabled && !crossDeviceEnabled) || scormEnabled) || ((xAPIEnabled || crossDeviceEnabled) && !scormEnabled) -->\r\n                        <!-- ko if: canDownloadCertificate -->\r\n                        <button class="button block primary autofocus submit" data-bind="click: finish" data-translate-text="[submit results and download certificate]" data-translate-title="[submit results and download certificate]"></button>\r\n                        <!-- /ko -->\r\n                        <!-- ko ifnot: canDownloadCertificate -->\r\n                        <button class="button block primary autofocus submit" data-bind="click: finish" data-translate-text="[finish course]" data-translate-title="[finish course]"></button>\r\n                        <!-- /ko -->\r\n                    <!-- /ko -->\r\n                <!-- /ko -->\r\n                <!-- ko if: isInReviewAttemptMode -->\r\n                <button class="button block submit" data-bind="click: newAttemptDialog.show, css: { primary: !isCompleted, default: isCompleted, autofocus: !isCompleted }" data-translate-text="[start new attempt]" data-translate-title="[start new attempt]"></button>\r\n                <!-- /ko -->\r\n            </span>\r\n        </div>\r\n        <!-- ko if: crossDeviceEnabled && !scormEnabled && !isInReviewAttemptMode -->\r\n        <button class="button check small mb10" type="submit" data-bind="click: toggleStayLoggedIn, css: { active: stayLoggedIn }">\r\n            <span data-translate-text="[keep me logged-in for the next courses]" data-translate-title="[keep me logged-in for the next courses]"></span>\r\n        </button>\r\n        <!-- /ko -->\r\n    </div>\r\n    <!-- ko if: sections.length -->\r\n    <div class="results-summary-heading">\r\n        <div class="section-wrapper">\r\n            <h3 class="results-summary-caption" data-translate-text="[detailed summary]"></h3>\r\n            <!-- ko if: isInReviewAttemptMode -->\r\n            <!-- ko if: canDownloadCertificate -->\r\n            <button class="button default download-certificate" data-bind="click: downloadCertificate, css: { disabled: isDownloadingCertificate }, disabled: isDownloadingCertificate">\r\n                <!-- ko ifnot: isDownloadingCertificate -->\r\n                <span data-translate-text="[download certificate]"></span>\r\n                <!-- /ko -->\r\n                <!-- ko if: isDownloadingCertificate -->\r\n                    <div class="pulsar-loader align-center secondary">\r\n                        <div class="pulsar-loader-item"></div>\r\n                        <div class="pulsar-loader-item"></div>\r\n                        <div class="pulsar-loader-item"></div>\r\n                    </div>\r\n                <!-- /ko -->\r\n            </button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n    <div class="result-page-content" data-bind="foreach: sections">\r\n        <div class="section noselect">\r\n            <div class="section-header" data-bind="css: {correct: isCorrect, wrong: !isCorrect}">\r\n                <div class="section-header-wrapper">\r\n                    <div class="section-header-label" data-translate-text="[section]" data-bind="attr: {\'data-index\': \' \' + ($parent.sections.indexOf($data) + 1)}"></div>\r\n                    <div class="section-header-title" data-bind="text: title"></div>\r\n                    <!-- ko if: amountOfQuestions || ($parent.allowContentPagesScoring && amountOfContents) -->\r\n                    <div class="section-header-analytic">\r\n                        <span class="section-header-analytic-icon" data-bind="css: {\'icon-correct_check\': isCorrect, \'icon-incorrect_cross\': !isCorrect}"></span>\r\n                        <!-- ko if: amountOfQuestions -->\r\n                        <div class="section-header-correct" data-bind="attr: {\'data-count\': correctQuestions + \'/\' + amountOfQuestions}" data-translate-text="[questions correct]"></div>\r\n                        <!-- /ko -->\r\n                        <!-- ko if: $parent.allowContentPagesScoring && amountOfContents -->\r\n                        <div class="section-header-content-readed" data-bind="attr: {\'data-count\': readedContents + \'/\' + amountOfContents}" data-translate-text="[contents readed]"></div>\r\n                        <!-- /ko -->\r\n                    </div>\r\n                    <!-- /ko -->\r\n                </div>\r\n            </div>\r\n            <!-- ko if: questions.length -->\r\n            <div class="section-content">\r\n                <!-- ko foreach: questions -->\r\n                <!-- ko compose: { model: \'questionContent/questionContent\', activationData: [ $parent.id, $data, ($parent.questions.indexOf($data) + 1), true ] } --><!-- /ko -->\r\n                <!-- /ko -->\r\n            </div>\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <div class="result-page-go-top noselect icon-up_arrow" data-bind="scrollTop: 1000"></div>\r\n    <div class="finish-course-layout" data-bind="visible: status() == statuses.sendingRequests || status() == statuses.preparingCertificate">\r\n        <div class="finish-course-layout-block"></div>\r\n        <div class="finish-course-layout-loader">\r\n            <!-- ko if: status() == statuses.sendingRequests -->\r\n            <span class="finish-course-layout-loader-text" data-translate-text="[tracking and tracing reporting progress]"></span>\r\n            <!-- /ko -->\r\n             <!-- ko if: status() == statuses.preparingCertificate -->\r\n             <span class="finish-course-layout-loader-text" data-translate-text="[preparing course certificate]"></span>\r\n             <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- ko if: npsDialog.isVisible -->\r\n    <!-- ko compose: { view: \'dialogs/nps/nps\', model: \'dialogs/nps/nps\', compositionComplete: npsDialog.compositionComplete, activationData: { close: npsDialog.hide, callbacks: npsDialog.callbacks } } --><!-- /ko -->\r\n    <!-- /ko -->\r\n    <!-- ko if: newAttemptDialog.isVisible -->\r\n    <!-- ko compose: { view: \'dialogs/startNewAttempt/startNewAttempt\', model: \'dialogs/startNewAttempt/startNewAttempt\', compositionComplete: newAttemptDialog.compositionComplete, activationData: { close: newAttemptDialog.hide, callbacks: newAttemptDialog.callbacks } } --><!-- /ko -->\r\n    <!-- /ko -->\r\n    <!-- ko if: resendResultsDialog.isVisible -->\r\n    <!-- ko compose: { view: \'dialogs/resultsResend/resultsResend\', model: \'dialogs/resultsResend/resultsResend\', compositionComplete: resendResultsDialog.compositionComplete, activationData: { close: resendResultsDialog.hide, callbacks: resendResultsDialog.callbacks, resultsSendErrorTitleKey: resendResultsDialog.resultsSendErrorTitleKey, endpointNameKey: resendResultsDialog.endpointNameKey } } --><!-- /ko -->\r\n    <!-- /ko -->\r\n</section>';});


define('text!treeOfContent/treeOfContent.html',[],function () { return '<section class="tree-of-content-panel unselectable" data-bind="css: { expanded: isExpanded }, tabFocus">\n    <div tabindex="0" class="tree-of-content-panel-tail" data-bind="click: toggleIsExpanded.bind($data), localize: { title: isExpanded() ? \'[collapse]\' : \'[expand]\'}, submitOnEnter">\n        <i class="tree-of-content-panel-tail-icon material-icons" data-bind="text: isExpanded() ? \'&#xE314;\' : \'menu\'"></i>\n    </div>\n    <div class="tree-of-content-container" data-bind="css: { expanded: isExpanded }, treeOfContentAnimations: { expanded: isExpanded }, attr: { \'aria-hidden\': !isExpanded() }">\n        <div class="tree-of-content" data-bind="scrollbar: { checkDOMChanges: true }">\n            <header class="tree-of-content-header">\n                <button class="tree-of-content-header-btn menu" data-bind="click: toggleIsExpanded.bind($data)">\n                    <i class="tree-of-content-header-btn-icon menu material-icons">menu</i>\n                </button>\n                <div class="tree-of-content-header-title" data-translate-text="[sections]"></div>\n                <button class="tree-of-content-header-btn close" data-bind="click: toggleIsExpanded.bind($data)">\n                    <i class="tree-of-content-header-btn-icon close material-icons">close</i>\n                </button>\n            </header>\n            <ul class="tree-of-content-items" data-bind="foreach: children">\n                <li class="tree-of-content-section-item" data-bind="css:{ expanded : isExpanded, active: hasActiveQuestion }">\n                    <div class="tree-of-content-section-title-wrapper" data-bind="click: toggleIsExpanded.bind($data), keyDown: { 13: { handler: toggleIsExpanded.bind($data) } }, attr: { tabindex: $parent.isExpanded() ? \'0\' : \'-1\' }">\n                        <i class="tree-of-content-section-title-icon material-icons" data-bind="text: isExpanded() ? \'&#xE313;\' : \'&#xE315;\', localize: { title: isExpanded() ? \'[collapse]\' : \'[expand]\'}"></i>\n                        <div class="tree-of-content-section-title" data-bind="limitedLengthText: { text: title, maxLength: 70 }, attr: { title: title }"></div>\n                    </div>\n                    <!-- ko if: isExpanded -->\n                    <div class="tree-of-content-questions-wrapper" data-bind="visible: isExpanded">\n                        <!-- ko ifnot: hasChildren -->\n                        <div class="tree-of-content-questions-empty" data-translate-text="[no questions]"></div>\n                        <!-- /ko -->\n                        <!-- ko if: hasChildren -->\n                        <ul class="tree-of-content-questions" data-bind="foreach: children">\n                            <li class="tree-of-content-question-item" data-bind="css: { active: isActive }">\n                                <!-- ko if: canNavigate -->\n                                <a class="tree-of-content-question-title-wrapper navigatable" href="#" data-bind="attr: { href: url, tabindex: $parent.isExpanded() && $root.isExpanded() ? \'0\' : \'-1\' }">\n                                    <i class="tree-of-content-question-title-icon material-icons" data-bind="css: { question: !isInformationContent }, text: isInformationContent ? \'info_outline\' : \'&#xE8FD;\'"></i>\n                                    <span class="tree-of-content-question-title" data-bind="text: title"></span>\n                                </a>\n                                <!-- /ko -->\n                                <!-- ko ifnot: canNavigate -->\n                                <div class="tree-of-content-question-title-wrapper">\n                                    <i class="tree-of-content-question-title-icon material-icons" data-bind="css: { question: !isInformationContent }, text: isInformationContent ? \'info_outline\' : \'&#xE8FD;\'"></i>\n                                    <span class="tree-of-content-question-title" data-bind="text: title"></span>\n                                </div>\n                                <!-- /ko -->\n                            </li>\n                        </ul>\n                        <!-- /ko -->\n                        <!-- /ko -->\n                    </div>\n                    <!-- /ko -->\n                </li>\n            </ul>\n        </div>\n    </div>\n</section>';});


define('text!views/404.html',[],function () { return '<section id="errorPage404" class="error error-404">\r\n    <div class="error-content">\r\n        <div class="text-404">\r\n            <div class="title-404">404</div>\r\n            <div class="page-404-hint" data-translate-text="[page not found message]"></div>\r\n            <a href="#" data-bind="attr: { href: homeUrl }" class="button primary medium">\r\n                <span data-translate-text="[home]"></span>\r\n            </a>\r\n        </div>\r\n    </div>\r\n</section>';});


define('text!views/exitCourse.html',[],function () { return '<section class="summary-wrapper" data-bind="blockBodyScroll: continueLaterPopup.isVisible()">\r\n    <!-- ko if: type() === \'extended\' -->\r\n    <section class="summary-extended-wrapper">\r\n        <div class="summary-extended-background"></div>\r\n        <div class="summary-extended">\r\n            <div class="overall-progress-wrapper">\r\n                <div class="overall-progress">\r\n                    <div>\r\n                        <div class="overall-progress-label-wrapper">\r\n                            <span class="overall-progress-label" data-translate-text="[Overall progress]"></span>\r\n                        </div>\r\n                        <div class="overall-progress-progressbar-wrapper">\r\n                            <div class="overall-progress-progressbar" data-bind="css: { success: isInReviewAttemptMode && isCompleted, fail: !isCompleted && isInReviewAttemptMode }">\r\n                                <canvas class="overall-progress-progressbar-canvas canvas-styles" width="80" height="80" data-bind="circleProgress: { progress: score(), lineWidth:2,radius:39, isMastered: isCompleted }"></canvas>\r\n                                <div class="overall-progress-progressbar-text" data-bind="text: score()"></div>\r\n                            </div>\r\n                        </div>\r\n                    </div>\r\n                </div>\r\n            </div>\r\n            <!-- ko ifnot: hideFinishActionButtons -->\r\n            <div class="exit-course-wrapper">\r\n                <div class="exit-course-container">\r\n                    <!-- ko ifnot: isInReviewAttemptMode -->\r\n                    <!-- ko if: !xAPIEnabled && !scormEnabled -->\r\n                    <button class="button small primary exit-course-button" data-translate-text="[finish course when tracking disabled]" data-bind="click: finishAction, css: { disabled: isNavigationLocked }"></button>\r\n                    <!-- /ko -->\r\n                    <!-- ko if: xAPIEnabled || scormEnabled -->\r\n                    <button class="button small primary exit-course-button" data-translate-text="[finish course]" data-bind="click: finishAction, css: { disabled: isNavigationLocked }"></button>\r\n                    <!-- /ko -->\r\n                    <!-- /ko -->\r\n                    <!-- ko if: isInReviewAttemptMode -->\r\n                    <button class="button small primary exit-course-button finished" data-translate-text="[go to results]" data-bind="click: finishAction, css: { disabled: isNavigationLocked }"></button>\r\n                    <!-- /ko -->\r\n                    <!-- ko ifnot: hideContinueLaterButton -->\r\n                    <div class="or-separator">\r\n                        <span class="or-separator-text" data-translate-text="[or]"></span>\r\n                    </div>\r\n                    <!-- /ko -->\r\n                    <div class="exit-course-link-wrapper">\r\n                        <!-- ko ifnot: isInReviewAttemptMode -->\r\n                        <!-- ko ifnot: hideContinueLaterButton -->\r\n                        <a class="exit-course-link" href="#" data-translate-title="[take a break and continue later]" data-translate-text="[take a break and continue later]" data-bind="click: takeABreakAction"></a>\r\n                        <!-- /ko -->\r\n                        <!-- /ko -->\r\n                        <!-- ko if: isInReviewAttemptMode -->\r\n                        <a class="exit-course-link" href="#" data-translate-title="[close course]" data-translate-text="[close course]" data-bind="click: takeABreakAction"></a>\r\n                        <!-- /ko -->\r\n                    </div>\r\n                </div>\r\n            </div>\r\n            <!-- /ko -->\r\n        </div>\r\n    </section>\r\n    <!-- /ko -->\r\n\r\n    <!-- ko ifnot: hideContinueLaterButton -->\r\n\r\n    <!-- ko ifnot: type() ===\'extended\' -->\r\n    <section class="summary-short">\r\n        <button class="exit-course-button" data-bind="click: takeABreakAction, css: { disabled: isNavigationLocked }">\r\n            <span class="material-icons">exit_to_app</span>\r\n            <!-- ko ifnot: isInReviewAttemptMode -->\r\n            <span data-translate-text="[take a brake]"></span>\r\n            <!-- /ko -->\r\n            <!-- ko if: isInReviewAttemptMode -->\r\n            <span data-translate-text="[close course]"></span>\r\n            <!-- /ko -->\r\n        </button>\r\n    </section>\r\n\r\n    <!-- /ko -->  \r\n    \r\n    <!-- /ko -->    \r\n    \r\n\r\n    <!-- ko if: continueLaterPopup.isVisible -->\r\n        <!-- ko compose: { view: \'dialogs/continueLater\', model: \'dialogs/continueLater\', compositionComplete: continueLaterPopup.compositionComplete, activationData: [ continueLaterPopup.actions, continueLaterPopup.isVisible ] } --><!-- /ko -->\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/introduction.html',[],function () { return '<section id="introduction" class="unselectable" data-bind="autofocus">\r\n    <div class="">\r\n        <header class="page-header">\r\n            <h1 class="course-title" data-bind="css: {small: courseTitle.length>100}">\r\n                <span class="course-caption" data-bind="text: courseTitle, reviewable: {type: \'course\', property: \'title\'}">"course_name"</span>\r\n            </h1>\r\n            <div class="course-action-block intro">\r\n                <button class="button small primary autofocus" href="#" data-bind="click: startCourse, css: { disabled: isNavigationLocked }" data-translate-text="[start course]" data-translate-title="[start course]">Start course</button>\r\n            </div>\r\n        </header>\r\n        <div class="course-intro-text styled-content content">\r\n            <!-- ko if: allowAuthorsBio -->\r\n                <div class="author-information-wrapper">\r\n                    <h3 class="cource-intro-title" data-bind="text: \'About the author\'"></h3>\r\n                    <div class="author-information-content">\r\n                        <div class="author-information-image">\r\n                            <!-- ko ifnot: authorAvatarUrl -->\r\n                                <span class="material-icons">person</span>\r\n                            <!-- /ko -->\r\n                            <!-- ko if: authorAvatarUrl -->\r\n                                <img data-bind="attr: {src: (authorAvatarUrl + \'?width=356&height=356&scaleBySmallerSide=true\')}" alt="">\r\n                            <!-- /ko -->\r\n                        </div>\r\n                        <div class="author-information-text">\r\n                            <h3 class="author-information-name" data-bind="text: createdBy"></h3>\r\n                            <!-- ko if: authorContactEmail || authorPersonalPhone -->\r\n                                <div class="author-information-contacts">\r\n                                    <!-- ko if: authorContactEmail -->\r\n                                        <div class="author-information-email">\r\n                                            <i class="material-icons">drafts</i>\r\n                                            <span data-bind="text: authorContactEmail"></span>\r\n                                        </div>\r\n                                    <!-- /ko -->\r\n                                    <!-- ko if: authorPersonalPhone -->\r\n                                        <div class="author-information-phone">\r\n                                            <i class="material-icons">call</i>\r\n                                            <span data-bind="text: authorPersonalPhone"></span>\r\n                                        </div>\r\n                                    <!-- /ko -->\r\n                                </div>\r\n                            <!-- /ko -->\r\n                            <!-- ko if: authorShortBio -->\r\n                                <div class="author-information-bio" data-bind="text: authorShortBio"></div>\r\n                            <!-- /ko -->\r\n                        </div>\r\n                    </div>\r\n                </div>\r\n            <!-- /ko -->\r\n            <!-- ko if: introduction -->\r\n                <div data-bind="elementsWrap: {}, reviewable: {type: \'course\', property: \'introduction\'}">\r\n                    <!-- ko foreach: introduction -->\r\n                        <!-- ko ifnot: children.length -->\r\n                            <div class="introduction-item styled-content" data-bind="contentBlock: content"></div>\r\n                        <!-- /ko -->\r\n                        <!-- ko if: children.length -->\r\n                            <div class="introduction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></div>\r\n                        <!-- /ko -->\r\n                    <!-- /ko -->\r\n                </div>\r\n                \r\n            <!-- /ko -->\r\n            <!-- ko if: copyright --> \r\n            <div class="copyright-text" data-bind="text: copyright"></div>\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n</section>\r\n';});


define('text!views/questions/content.html',[],function () { return '<section id="content" class="question unselectable" data-bind="keyDown: { 37: { handler: goPrevious }, 39: { handler: goNext } }">\r\n    <div>\r\n        <header class="page-header">\r\n            <div class="navigation-home-link-wrapper" data-bind="click: backToSections">\r\n                <div tabindex="0" class="navigation-home-link material-icons" data-bind="submitOnEnter: backToSections, css: { disabled: isNavigationLocked }">home</div>\r\n            </div>\r\n            <div class="section-progress-wrapper">\r\n                <div class="section-image-wrapper" data-bind="style: { \'background-image\': \'url(\' + section.imageUrl + \')\' }, css: { \'affect-progress\': section.affectProgress }">\r\n                    <!-- ko if: section.affectProgress -->\r\n                    <div class="sections-canvas-wrapper">\r\n                        <canvas tabindex="0" class="canvas-styles" width="80" height="80" data-bind="circleProgress: { progress: section.score(), isMastered: isSectionMastered(), lineWidth: 2 },toolTip:{progress: section.score(), masteryScore:masteryScore}"></canvas>\r\n                    </div>\r\n                    <span class="progress-text" data-bind="text: section.score()"></span>\r\n\r\n                    <!-- /ko -->\r\n                </div>\r\n            </div>\r\n            <div class="section-info">\r\n                <span data-bind="text: section.title, reviewable: {type: \'section\', property: \'title\', id: section.id, title: section.title}">section title</span>\r\n            </div>\r\n            <div class="questions-navigation-container">\r\n                <div class="question-navigation-bar">\r\n                    <a class="previous" data-bind="attr: { href: previousQuestionUrl(), title: \'previous\'},css:{disabled : !isPreviousQuestionAvailable()}" data-translate-title="[previous question]">&nbsp;</a>\r\n                    <span class="navigation-progress"><span data-bind="text: navigationContext.currentQuestionIndex + \' \'"></span><span>/</span><span data-bind="text: \' \' + navigationContext.questionsCount"></span></span>\r\n                    <a class="next" data-bind="attr: { href: nextQuestionUrl(), title: \'next\' }, css:{disabled: !isNextQuestionAvailable() }" data-translate-title="[next question]">&nbsp;</a>\r\n                </div>\r\n            </div>\r\n        </header>\r\n        <!-- ko if: voiceOver -->\r\n        <div class="voiceOver" data-bind="mediaPlayer: { embedCode: voiceOver}, reviewable: {type: question.isInformationContent ? \'informationContent\' : \'question\', property: \'voiceOver\', id: question.id, title: question.title}"></div>\r\n        <!-- /ko -->\r\n        <!-- ko compose: { model: activeViewModel, activationData: [ section.id, question ] } --><!-- /ko -->\r\n    </div>\r\n    <div class="question-navigation-bar bottom" style="float: right">\r\n        <a class="previous" data-bind="attr: { href: previousQuestionUrl(), title: \'previous\'},css:{disabled : !isPreviousQuestionAvailable()}" data-translate-title="[previous question]">&nbsp;</a>\r\n        <span class="navigation-progress"><span data-bind="text: navigationContext.currentQuestionIndex + \' \'"></span><span>/</span><span data-bind="text: \' \' + navigationContext.questionsCount"></span></span>\r\n        <a class="next" data-bind="attr: { href: nextQuestionUrl(), title: \'next\' }, css:{disabled: !isNextQuestionAvailable() }" data-translate-title="[next question]">&nbsp;</a>\r\n    </div>\r\n</section>';});


define('text!views/questions/dragAndDrop/dragAndDrop.html',[],function () { return '<section class="overlay-wrapper">\n    <!-- ko if: questionInstructions.length > 0 -->\n    <section id="question-instructions" class="question-instructions" data-animate="">\n        <ul class="question-instruction-items unstyled">\n            <!-- ko foreach: questionInstructions-->\n            <!-- ko ifnot: children.length -->\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\n            <!-- /ko -->\n            <!-- ko if: children.length -->\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n            <!-- /ko -->\n            <!-- /ko -->\n        </ul>\n    </section>\n    <!-- /ko -->\n    <section class="drag-and-drop-text">\n        <div>            \n            <!-- ko if: !$root.isPreview() -->\n            <div class="drag-and-drop-text-draggable-container" data-bind="draggableTextContainer: {}, scope: question.id">\n                <div class="drag-and-drop-text-draggable-container-message" data-translate-text="[drag and drop question all texts are placed]">\n                    All texts are placed\n                </div>\n                <!-- ko foreach: texts -->\n                <div class="drag-and-drop-text-draggable" data-bind="draggableText: {}, scope: $parent.question.id">\n                    <div data-bind="text: _.isObject(text) ? text.text : text, attr: {title: _.isObject(text) ? text.text : text}">\n                    </div>\n                </div>\n                <!-- /ko -->\n            </div>            \n            <!-- /ko -->\n            <!-- ko if: $root.isPreview -->\n            <div class="drag-and-drop-text-draggable-container" data-bind="visible: false">\n                <!-- ko foreach: texts -->\n                <div class="drag-and-drop-text-draggable">\n                    <div data-bind="text: _.isObject(text) ? text.text : text, attr: {title: _.isObject(text) ? text.text : text}">\n                    </div>\n                </div>\n                <!-- /ko -->\n            </div>            \n            <!-- /ko -->\n            <div class="drag-and-drop-text-background-wrapper">\n                <div data-bind="background: question.background" class="drag-and-drop-text-background">\n                    <!-- ko foreach: dropspots -->\n                    <div class="drag-and-drop-text-dropspot" data-bind="dropspot: $data, scope: $parent.question.id">\n                        <!-- ko ifnot: $root.isPreview() || $root.isAnswered() -->\n                        <div class="drag-and-drop-text-dropspot-tooltip-image">\n                        </div>\n                        <!-- /ko -->\n                    </div>\n                    <!-- /ko -->\n                </div>\n            </div>\n\n        </div>\n    </section>\n    <!-- ko if: isAnswered() && !$root.isPreview() -->\n    <div class="question-overlay"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/expandableBlock.html',[],function () { return '<section class="expandable-block" data-bind="css: { collapsed: !isExpanded() }, toggleVisible: { isExpanded: isExpanded() }">\n    <div class="expandable-block-content" data-bind="click: toggleExpand, submitOnEnter">\n        <div class="expandable-block-content-title" data-bind="html: content"></div>\n        <button class="expand-button" data-bind="css: { expand: !isExpanded() }">\n            <span class="material-icons expand-button-icon" data-bind="css: { visible: isExpanded() }">remove</span>\n            <span class="material-icons expand-button-icon" data-bind="css: { visible: !isExpanded() }">add</span>\n        </button>\n    </div>\n    <!-- ko if: children.length > 0 -->\n    <ul class="expandable-block-children-items unstyled" data-animate>\n        <!-- ko foreach: children -->\n        <li class="expandable-block-children styled-content" data-bind="contentBlock: content"></li>\n        <!-- /ko -->\n    </ul>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/feedback.html',[],function () { return '<section id="feedback" data-bind="scrollIntoView: isAnswered">\r\n    <!-- ko if: isAnswered() -->\r\n    <!-- ko if: isCorrectAnswered() || (isWrongAnswered() && isSurvey) -->\r\n    <div class="result-block-wrapper" data-bind="autofocus">\r\n        <div class="question-result-block" data-bind="text: feedbackResultText, title: feedbackResultText"></div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: correctFeedback">\r\n            <!--ko foreach: correctFeedback-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n    <!-- ko if: isWrongAnswered() && !isSurvey -->\r\n    <div class="result-block-wrapper" data-bind="autofocus">\r\n        <div class="question-result-block incorrect" data-bind="text: feedbackResultText, title: feedbackResultText">\r\n        </div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: incorrectFeedback">\r\n            <!--ko foreach: incorrectFeedback-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            \r\n            <!-- ko if: hideTryAgain -->\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n\r\n            <!-- ko ifnot: hideTryAgain -->\r\n            <button class="button medium primary autofocus" data-bind="click: tryAnswerAgain" data-translate-text="[try again]" data-translate-title="[try again]">Try again</button>\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium default" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium default" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium default" data-bind="click: goToResults, css: { disabled: isNavigationLocked }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n            \r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/questions/fillInTheBlank/fillInTheBlank.html',[],function () { return '<section class="overlay-wrapper">\n    <!-- ko if: content -->\n    <div class="question-content-wrapper blank-content fill-in-the-blank">\n        <div class="question-content blank-answers styled-content" data-bind="elementsWrap: {}, fillInTheBlankAnswers: { inputValues:inputValues, content:content, isAnswered: isAnswered() || isPreview() }"></div>\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isAnswered() && !isPreview() -->\n    <div class="question-overlay"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/hotspot/hotspot.html',[],function () { return '<section class="overlay-wrapper">\n    <!-- ko if: questionInstructions.length > 0 -->\n    <section id="question-instructions" class="question-instructions" data-animate="">\n        <ul class="question-instruction-items unstyled">\n            <!-- ko foreach: questionInstructions-->\n            <!-- ko ifnot: children.length -->\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\n            <!-- /ko -->\n            <!-- ko if: children.length -->\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n            <!-- /ko -->\n            <!-- /ko -->\n        </ul>\n    </section>\n    <!-- /ko -->\n    <!-- ko ifnot: isPreview -->\n    <div class="hotspot">\n        <div class="question-image-wrapper">\n            <div class="hotspot-background" data-bind="area: { click: addMark }, background: background">\n                <ul data-bind="foreach: marks">\n                    <li class="pulsemark" data-bind="style: { left: x + \'px\', top: y + \'px\' }, click: $parent.removeMark, clickBubble: !$parent.isMultiple, css: { \'multiple\': $parent.isMultiple }">\n                        <div class="pin"></div>\n                        <div class="mark">\n                            <span class="impulse"></span>\n                        </div>\n                    </li>\n                </ul>\n            </div>\n        </div>\n        <!-- ko if: isAnswered() -->\n        <div class="question-overlay"></div>\n        <!-- /ko -->\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isPreview -->\n    <div class="hotspot">\n        <div class="question-image-wrapper">\n            <div class="hotspot-background" data-bind="background: background">\n                <ul data-bind="foreach: marks">\n                    <li class="pulsemark" data-bind="style: { left: x + \'px\', top: y + \'px\' }, css: { \'multiple\': $parent.isMultiple }">\n                        <div class="pin"></div>\n                        <div class="mark">\n                            <span class="impulse"></span>\n                        </div>\n                    </li>\n                </ul>\n            </div>\n        </div>\n    </div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/informationContent.html',[],function () { return '<section id="informationContent" class="content information-content" data-bind="reviewable: {type: \'informationContent\', id: id, title: title}, autofocus">\n    <div class="learning-content-section">\n        <div tabindex="-1" class="information-content-title autofocus">\n            <span data-bind="text: title"></span>\n        </div>\n        <!-- ko if: learningContents.length > 0 -->\n        <section id="learning-contents" class="learning-contents">\n            <ul class="learning-content-items unstyled">\n                <!-- ko foreach: learningContents-->\n                <!-- ko ifnot: children.length -->\n                <li class="learning-content-item styled-content" data-bind="contentBlock: content"></li>\n                <!-- /ko -->\n                <!-- ko if: children.length -->\n                <li class="learning-content-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n                <!-- /ko -->\n                <!-- /ko -->\n            </ul>\n        </section>\n        <!-- /ko -->\n    </div>\n    <div class="result-block">\n        <div class="result-btn-holder">\n            <!-- ko if: navigationContext.nextQuestionUrl -->\n            <button class="button medium primary" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\n            <!-- /ko -->\n            <!-- ko if: isGoToNextSectionVisible() -->\n            <button class="button medium primary" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\n            <!-- /ko -->\n            <!-- ko if: isGoToResultsVisible -->\n            <button class="button medium default" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\n            <!-- /ko -->\n        </div>\n    </div>\n    <!-- ko if: copyright --> \n    <div class="copyright-text" data-bind="text: copyright"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/multipleSelect/multipleSelect.html',[],function () { return '<section class="overlay-wrapper">\r\n    <!-- ko if: questionInstructions.length > 0 -->\r\n    <section id="question-instructions" class="question-instructions" data-animate="">\r\n        <ul class="question-instruction-items unstyled">\r\n            <!-- ko foreach: questionInstructions-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </ul>\r\n    </section>\r\n    <!-- /ko -->\r\n    <!-- ko if: isSurveyModeEnabled -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode answered caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isPreview -->\r\n    <ul class="question-answers-list" data-bind="foreach: answers">\r\n        <li class="checkbox" role="checkbox" data-bind="click: $parent.checkItem.bind($parent), keyDown: { 32: { handler: $parent.checkItem.bind($parent) }, 13: { handler: $parent.checkItem.bind($parent) } },\r\n            css: { checked: isChecked(), enabled: !$parent.isAnswered()},\r\n            attr: { tabindex: $parent.isAnswered() ? \'-1\' : \'0\', \'aria-checked\': isChecked() ? \'true\' : \'false\' }">\r\n            <span class="checkbox-icon-wrapper" >\r\n                <button tabindex="-1" class="checkbox-icon" data-bind="css: { checked: isChecked() }" ></button>\r\n            </span>\r\n            <label class="checkbox-text" data-bind="text: text"></label>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isPreview -->\r\n    <ul class="question-answers-list" data-bind="foreach: answers">\r\n        <li class="checkbox" data-bind="css: { checked: isChecked() }">\r\n            <span class="checkbox-icon-wrapper">\r\n                <button tabindex="-1" class="checkbox-icon" data-bind="css: { checked: isChecked() }"></button>\r\n            </span>\r\n            <label class="checkbox-text" data-bind="text: text"></label>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="question-overlay"></div>\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/questions/openQuestion/feedback.html',[],function () { return '<section id="feedback" data-bind="scrollIntoView: isAnswered">\r\n    <!-- ko if: isCorrectAnswered -->\r\n    <div class="result-block-wrapper" data-bind="autofocus">\r\n        <div class="question-result-block" data-translate-text="[your answer was stored]" data-translate-title="[your answer was stored]">Your answer has been stored</div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: correctFeedback">\r\n            <!--ko foreach: correctFeedback-->\r\n             <!-- ko ifnot: children.length -->\r\n             <li class="styled-content" data-bind="contentBlock: content"></li>\r\n             <!-- /ko -->\r\n             <!-- ko if: children.length -->\r\n             <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n             <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n    <!-- ko if: isWrongAnswered -->\r\n    <div class="result-block-wrapper" data-bind="autofocus">\r\n        <div class="question-result-block incorrect" data-translate-text="[your answer is empty and was not stored]" data-translate-title="[your answer is empty and was not stored]">Your answer is empty and was not stored</div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: incorrectFeedback">\r\n            <!--ko foreach: incorrectFeedback-->\r\n             <!-- ko ifnot: children.length -->\r\n             <li class="styled-content" data-bind="contentBlock: content"></li>\r\n             <!-- /ko -->\r\n             <!-- ko if: children.length -->\r\n             <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n             <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            <!-- ko if: hideTryAgain -->\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n\r\n            <!-- ko ifnot: hideTryAgain -->\r\n            <button class="button medium primary autofocus" data-bind="click: tryAnswerAgain" data-translate-text="[try again]" data-translate-title="[try again]">Try again</button>\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium default" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium default" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium default" data-bind="click: goToResults, css: { disabled: isNavigationLocked }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n</section>\r\n';});


define('text!views/questions/openQuestion/openQuestion.html',[],function () { return '<section class="overlay-wrapper">\r\n    <!-- ko if: questionInstructions.length > 0 -->\r\n    <section id="question-instructions" class="question-instructions" data-animate="">\r\n        <ul class="question-instruction-items unstyled">\r\n            <!-- ko foreach: questionInstructions-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </ul>\r\n    </section>\r\n    <!-- /ko -->\r\n    <div class="edit-field-wrapper">\r\n        <textarea class="open-question-edit-field edit-field selectable" data-bind="textArea: answeredText, attr:{ disabled: isAnswered }, keyDown: {37: {preventDefault: false}, 39: {preventDefault: false}}" data-translate-placeholder="[type your answer here]"></textarea>\r\n    </div>\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="question-overlay"></div>\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/questions/questionContent.html',[],function () { return '<section id="questionContent" data-bind="autofocus">\r\n    <!-- ko if: learningContents.length > 0 -->\r\n    <section class="content learning-content" data-bind="css:{collapsed:!isExpanded()},toggleVisible:{isExpanded:isExpanded()}, reviewable: {type: \'question\', property: \'learningContent\', id: question.id, title: question.title}">\r\n        <div tabindex="0" class="toggle-expand-container" data-bind="click: toggleExpand, submitOnEnter, css: { \'autofocus\': !isAnswered() }">\r\n            <div class="learning-content-title" >\r\n                <span class="material-icons"> school</span>\r\n                <span data-translate-text="[learning content]">Learning content</span>\r\n            </div>\r\n            <a class="expand-button" data-bind="css:{expand:!isExpanded()}"></a>\r\n        </div>\r\n            <section id="learning-contents" class="learning-contents" data-animate="">\r\n                <ul class="learning-content-items unstyled">\r\n                    <!-- ko foreach: learningContents-->\r\n                    <!-- ko ifnot: children.length -->\r\n                    <li class="learning-content-item styled-content" data-bind="contentBlock: content"></li>\r\n                    <!-- /ko -->\r\n                    <!-- ko if: children.length -->\r\n                    <li class="learning-content-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n                    <!-- /ko -->\r\n                    <!-- /ko -->\r\n                </ul>\r\n            </section>\r\n    </section>\r\n    <!-- /ko -->\r\n    <section class="content learning-content" data-bind="reviewable: {type: \'question\', id: question.id, title: question.title}">\r\n        <div class="learning-content-title question-content-title">\r\n            <span class="material-icons"> help_outline</span>\r\n            <span data-translate-text="[question]">question</span>\r\n        </div>\r\n        <div tabindex="-1" class="question-title selectable" data-bind="text: title, css: { \'autofocus\': !isAnswered() }">\r\n            question title\r\n        </div>\r\n        <!-- ko compose: activeQuestionViewModel --><!-- /ko -->\r\n        <!-- ko compose: { view: feedbackView } --><!-- /ko -->\r\n        <!-- ko ifnot: isAnswered -->\r\n        <!-- ko ifnot: submitViewModel -->\r\n        <div class="btn-submit-wrapper" data-bind="autofocus">\r\n            <button href="#" class="button medium primary autofocus" data-bind="click: submit" data-translate-text="[submit]" data-translate-title="[submit]">Submit</button>\r\n        </div>\r\n        <!-- /ko -->\r\n        <!-- ko if: submitViewModel -->\r\n        <!-- ko compose: { model: submitViewModel, activationData: { submit: submit.bind($root) } } --><!-- /ko-->\r\n        <!-- /ko -->\r\n        <!-- /ko -->\r\n        <!-- ko if: copyright --> \r\n        <div class="copyright-text" data-bind="text: copyright"></div>\r\n        <!-- /ko -->\r\n    </section>\r\n</section>';});


define('text!views/questions/rankingText/rankingText.html',[],function () { return '<section class="overlay-wrapper">\n    <!-- ko if: questionInstructions.length > 0 -->\n    <section id="question-instructions" class="question-instructions" data-animate="">\n        <ul class="question-instruction-items unstyled">\n            <!-- ko foreach: questionInstructions-->\n            <!-- ko ifnot: children.length -->\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\n            <!-- /ko -->\n            <!-- ko if: children.length -->\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n            <!-- /ko -->\n            <!-- /ko -->\n        </ul>\n    </section>\n    <!-- /ko -->\n    <!-- ko ifnot: isPreview -->\n    <div data-bind="dragula: { data: rankingItems }" class="gu-unselectable">\n        <div class="ranking-item gu-draggable" data-bind="preventDefaultTouchEvent: true">\n            <span class="ranking-item-icon"></span>\n            <span class="ranking-item-text" data-bind="text: text"></span>\n        </div>\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isPreview -->\n    <div data-bind="foreach: { data: rankingItems }" class="gu-unselectable">\n        <div class="ranking-item gu-draggable">\n            <span class="ranking-item-icon"></span>\n            <span class="ranking-item-text" data-bind="text: text"></span>\n        </div>\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isAnswered() || isPreview() -->\n    <div class="question-overlay"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/scenarioQuestion/feedback.html',[],function () { return '<section id="feedback">\r\n    <!-- ko if: isCorrectAnswered -->\r\n    <div class="result-block-wrapper" data-bind="scrollIntoView: isAnswered, autofocus">\r\n        <div class="question-result-block" data-translate-text="[scenario has been passed successfully]" data-translate-title="[scenario has been passed successfully]">Scenario has been passed successfully</div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: correctFeedback">\r\n            <!--ko foreach: correctFeedback-->\r\n             <!-- ko ifnot: children.length -->\r\n             <li class="styled-content" data-bind="contentBlock: content"></li>\r\n             <!-- /ko -->\r\n             <!-- ko if: children.length -->\r\n             <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n             <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n    <!-- ko if: isWrongAnswered -->\r\n    <div class="result-block-wrapper" data-bind="scrollIntoView: isAnswered, autofocus">\r\n        <div class="question-result-block incorrect" data-translate-text="[scenario has not been passed successfully]" data-translate-title="[scenario has not been passed successfully]">Scenario has NOT been passed successfully</div>\r\n        <ul class="question-feedback-holder unstyled" data-bind="visible: incorrectFeedback">\r\n            <!--ko foreach: incorrectFeedback-->\r\n             <!-- ko ifnot: children.length -->\r\n             <li class="styled-content" data-bind="contentBlock: content"></li>\r\n             <!-- /ko -->\r\n             <!-- ko if: children.length -->\r\n             <li class="expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n             <!-- /ko -->\r\n            <!--/ko-->\r\n        </ul>\r\n        <div class="result-btn-holder">\r\n            <!-- ko if: hideTryAgain -->\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium primary autofocus" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium primary autofocus" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium autofocus" data-bind="click: goToResults, css: { disabled: isNavigationLocked, default: navigationContext.nextQuestionUrl || isGoToNextSectionVisible(), primary: !navigationContext.nextQuestionUrl && !isGoToNextSectionVisible() }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n\r\n            <!-- ko ifnot: hideTryAgain -->\r\n            <button class="button medium primary autofocus" data-bind="click: tryAnswerAgain" data-translate-text="[try again]" data-translate-title="[try again]">Try again</button>\r\n            <!-- ko if: navigationContext.nextQuestionUrl -->\r\n            <button class="button medium default" data-bind="click: navigateNext, css: { disabled: isNavigationLocked }" data-translate-text="[next]" data-translate-title="[next]">Next</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToNextSectionVisible() -->\r\n            <button class="button medium default" data-bind="click: goToNextSection, css: { disabled: isNavigationLocked }" data-translate-text="[next section]" data-translate-title="[next section]">Next section</button>\r\n            <!-- /ko -->\r\n            <!-- ko if: isGoToResultsVisible -->\r\n            <button class="button medium default" data-bind="click: goToResults, css: { disabled: isNavigationLocked }" data-translate-text="[go to results]" data-translate-title="[next section]">Go to results</button>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </div>\r\n    </div>\r\n    <!-- /ko -->\r\n</section>\r\n';});


define('text!views/questions/scenarioQuestion/scenarioQuestion.html',[],function () { return '<section class="overlay-wrapper">\n    <!-- ko if: questionInstructions.length > 0 -->\n    <section id="question-instructions" class="question-instructions" data-animate="">\n        <ul class="question-instruction-items unstyled">\n            <!-- ko foreach: questionInstructions-->\n            <!-- ko ifnot: children.length -->\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\n            <!-- /ko -->\n            <!-- ko if: children.length -->\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n            <!-- /ko -->\n            <!-- /ko -->\n        </ul>\n    </section>\n    <!-- /ko -->\n    <!-- ko ifnot: isPreview() -->\n    <div class="scenario-question-container">\n        <!-- ko if: !isAnswered() -->\n        <div class="scenario-question" data-bind="html: embedCode"></div>\n        <!-- /ko -->\n        <!-- ko if: isAnswered() -->\n        <div class="scenario-question-overlay">\n            <div class="scenario-question-overlay-container">\n                <!-- ko if: isAnswered -->\n                <span class="scenario-question-overlay-text" data-translate-text="[scenario has been submitted]" data-translate-title="[scenario has been submitted]"></span>\n                <!-- /ko --> \n            </div>\n        </div>\n        <!-- /ko -->\n    </div>\n    <!-- /ko -->\n    <!-- ko if: !isAnswered() && isPreview() -->\n    <div class="question-result-block incorrect" data-translate-text="[scenario has not been finished yet]" data-translate-title="[scenario has not been finished yet]">Scenario has NOT been finished yet</div>       \n    <!-- /ko -->\n    <!-- ko if: isAnswered() && isPreview() -->\n    <!-- ko if: isCorrect -->\n    <div class="question-result-block" data-translate-text="[scenario has been passed successfully]" data-translate-title="[scenario has been passed successfully]">Scenario has been passed successfully</div>      \n    <!-- /ko -->\n    <!-- ko ifnot: isCorrect -->\n    <div class="question-result-block incorrect" data-translate-text="[scenario has not been passed successfully]" data-translate-title="[scenario has not been passed successfully]">Scenario has NOT been passed successfully</div>      \n    <!-- /ko -->\n    <!-- /ko -->\n\n    <!-- ko if: isAnswered() || isPreview() -->\n    <div class="question-overlay"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/questions/scenarioQuestion/submitQuestion.html',[],function () { return '<div class="btn-submit-wrapper submit-question-tooltip-container" data-bind="css: { show: !canBeSubmitted() }">\n    <button href="#" class="button medium primary" data-bind="click: submit, customFocus: {}" data-translate-text="[submit]" data-translate-title="[submit]">Submit</button>\n    <div class="submit-question-tooltip" data-bind="visible: !canBeSubmitted()" data-translate-text="[scenario has not been finished yet]" data-translate-title="[scenario has not been finished yet]">\n    </div>\n</div>';});


define('text!views/questions/singleSelectImage/singleSelectImage.html',[],function () { return '<section class="overlay-wrapper single-select-image-wrapper">\r\n    <!-- ko if: questionInstructions.length > 0 -->\r\n    <section id="question-instructions" class="question-instructions" data-animate="">\r\n        <ul class="question-instruction-items unstyled">\r\n            <!-- ko foreach: questionInstructions-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </ul>\r\n    </section>\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isPreview -->\r\n    <ul class="single-select-image" data-bind="foreach: answers, css:{canHover:canHover}">\r\n        <li class="select-image" data-bind="css: { checked: id == $parent.checkedAnswerId(), enabled: !$parent.isAnswered() }">\r\n            <div class="select-image-wrapper">\r\n                <div tabindex="0" class="preview-thumbnail" data-bind="click: $parent.checkItem.bind($parent), keyDown: { 32: { handler: $parent.checkItem.bind($parent) }, 13: { handler: $parent.checkItem.bind($parent) } }, attr: { tabindex: $parent.isAnswered() ? \'-1\' : \'0\' }">   \r\n                    <img class="thumbnail" data-bind="attr:{\'data-image\': image, src: image + \'?height=200&amp;width=250&amp;scaleBySmallerSide=true\' }">\r\n                </div>\r\n                <div class="preview-button-holder">\r\n                    <button class="preview-button-wrapper material-icons" data-bind="previewImage: $data, enable: !$parent.isAnswered(), attr: { id: id }">zoom_in</button>\r\n                </div>\r\n            </div>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isPreview -->\r\n    <ul class="single-select-image" data-bind="foreach: answers">\r\n        <li class="select-image" data-bind="css: { checked: id == $parent.checkedAnswerId() }">\r\n            <div class="select-image-wrapper">\r\n                <div class="preview-thumbnail">   \r\n                    <img class="thumbnail" data-bind="attr:{\'data-image\': image, src: image + \'?height=200&amp;width=250&amp;scaleBySmallerSide=true\' }">\r\n                </div>\r\n            </div>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="question-overlay"></div>\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/questions/singleSelectText/singleSelectText.html',[],function () { return '<section class="overlay-wrapper">\r\n    <!-- ko if: questionInstructions.length > 0 -->\r\n    <section id="question-instructions" class="question-instructions" data-animate="">\r\n        <ul class="question-instruction-items unstyled">\r\n            <!-- ko foreach: questionInstructions-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </ul>\r\n    </section>\r\n    <!-- /ko -->\r\n\r\n    <!-- ko if: isSurveyModeEnabled -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode answered caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isPreview -->\r\n    <ul class="question-answers-list" data-bind="foreach: answers">\r\n        <li class="radio" role="radio" data-bind="click: $parent.checkItem.bind($parent), keyDown: { 32: { handler: $parent.checkItem.bind($parent) }, 13: { handler: $parent.checkItem.bind($parent) } },\r\n            css: { checked: isChecked(), enabled: !$parent.isAnswered() },\r\n            attr: { tabindex: $parent.isAnswered() ? \'-1\' : \'0\', \'aria-checked\': isChecked() ? \'true\' : \'false\' }">\r\n            <span class="radio-icon-wrapper">\r\n                <button tabindex="-1" class="radio-icon"></button>\r\n            </span>\r\n            <label class="radio-text" data-bind="text: text"></label>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isPreview -->\r\n    <ul class="question-answers-list" data-bind="foreach: answers">\r\n        <li class="radio" data-bind="css: { checked: isChecked() }">\r\n            <span class="radio-icon-wrapper">\r\n                <button tabindex="-1" class="radio-icon"></button>\r\n            </span>\r\n            <label class="radio-text" data-bind="text: text"></label>\r\n        </li>\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="question-overlay"></div>\r\n    <!-- /ko -->\r\n</section>';});


define('text!views/questions/statement/statement.html',[],function () { return '\r\n<div class="statement-content">\r\n    <!-- ko if: questionInstructions.length > 0 -->\r\n    <section id="question-instructions" class="question-instructions" data-animate="">\r\n        <ul class="question-instruction-items unstyled">\r\n            <!-- ko foreach: questionInstructions-->\r\n            <!-- ko ifnot: children.length -->\r\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\r\n            <!-- /ko -->\r\n            <!-- ko if: children.length -->\r\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\r\n            <!-- /ko -->\r\n            <!-- /ko -->\r\n        </ul>\r\n    </section>\r\n    <!-- /ko -->\r\n    <!-- ko if: isSurveyModeEnabled -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode answered caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- ko ifnot: isAnswered() || isPreview() -->\r\n    <div class="text-matching-caption" data-translate-text="[survey mode caption]"></div>\r\n    <!-- /ko -->\r\n    <!-- /ko -->\r\n    <!-- ko if: statements.length -->\r\n    <ul class="statement">\r\n        <li class="statement-headers-group" aria-hidden="true">\r\n            <div class="statement-option" data-translate-text="[statement question true text]" data-translate-title="[statement question true tooltip]">\r\n                True\r\n            </div>\r\n            <div class="statement-option" data-translate-text="[statement question false text]" data-translate-title="[statement question false tooltip]">\r\n                False\r\n            </div>\r\n        </li>\r\n        <!-- ko ifnot: isPreview() -->\r\n        <!-- ko foreach: statements -->\r\n        <li class="statement-item" data-bind="css:{ enabled: !$parent.isAnswered(), checked: isTrueChecked() || isFalseChecked() }, attr: { \'aria-labelledby\': id, \'aria-setsize\': $parent.statements.length, \'aria-posinset\': $index() + 1 }" role="radiogroup" >\r\n            <span class="statement-option" >\r\n                <span class="radio-icon-wrapper">\r\n                    <button class="radio-icon" data-bind="click:$parent.markStatementAsTrue.bind($parent), css: { checked: isTrueChecked() }, attr:{ \'aria-checked\': isTrueChecked() ? \'true\' : \'false\', disabled: $parent.isAnswered() }" role="radio" aria-setsize="2" aria-posinset="1">\r\n                        <span class="radio-label" data-translate-text="[statement question true text]"></span>\r\n                    </button>\r\n                </span>\r\n            </span>\r\n            <span class="statement-option-separator-wrapper">\r\n                <span class="statement-option-separator"></span>\r\n            </span>\r\n            <span class="statement-option" >\r\n                <span class="radio-icon-wrapper">\r\n                    <button class="radio-icon checked" data-bind="click:$parent.markStatementAsFalse.bind($parent), css: { checked: isFalseChecked() }, attr:{ \'aria-checked\': isFalseChecked() ? \'true\' : \'false\', disabled: $parent.isAnswered() }" role="radio" aria-setsize="2" aria-posinset="2">\r\n                        <span class="radio-label" data-translate-text="[statement question false text]"></span>\r\n                    </button>\r\n                </span>\r\n            </span>\r\n            <span class="statement-label" data-bind="text: text, attr: { id: id }"></span>\r\n        </li>\r\n        <!-- /ko -->\r\n        <!-- /ko -->\r\n        <!-- ko if: isPreview() -->\r\n        <!-- ko foreach: statements -->\r\n        <li class="statement-item" data-bind="css:{ checked: isTrueChecked() || isFalseChecked() }, attr: { \'aria-labelledby\': id, \'aria-setsize\': $parent.statements.length, \'aria-posinset\': $index() + 1 }" role="radiogroup">\r\n            <span class="statement-option">\r\n                <span class="radio-icon-wrapper">\r\n                    <button class="radio-icon" data-bind="css: { checked: isTrueChecked() }, attr: { \'aria-checked\': isTrueChecked() ? \'true\' : \'false\' }" disabled="disabled" role="radio" aria-setsize="2" aria-posinset="1">\r\n                        <span class="radio-label" data-translate-text="[statement question true text]"></span>\r\n                    </button>\r\n                </span>\r\n            </span>\r\n            <span class="statement-option-separator-wrapper">\r\n                <span class="statement-option-separator"></span>\r\n            </span>\r\n            <span class="statement-option">\r\n                <span class="radio-icon-wrapper">\r\n                    <button class="radio-icon checked" data-bind="css: { checked: isFalseChecked() }, attr: { \'aria-checked\': isFalseChecked() ? \'true\' : \'false\' }" disabled="disabled" role="radio" aria-setsize="2" aria-posinset="2">\r\n                        <span class="radio-label" data-translate-text="[statement question false text]"></span>\r\n                    </button>\r\n                </span>\r\n            </span>\r\n            <span class="statement-label" data-bind="text: text, attr: { id: id }"></span>\r\n        </li>\r\n        <!-- /ko -->\r\n        <!-- /ko -->\r\n    </ul>\r\n    <!-- /ko -->\r\n    <!-- ko if: isAnswered() || isPreview() -->\r\n    <div class="question-overlay" data-bind="visible: isAnswered() || isPreview()"></div>\r\n    <!-- /ko -->\r\n</div>';});


define('text!views/questions/textMatching/textMatching.html',[],function () { return '<section class="textmatching overlay-wrapper">\n    <!-- ko if: questionInstructions.length > 0 -->\n    <section id="question-instructions" class="question-instructions" data-animate="">\n        <ul class="question-instruction-items unstyled">\n            <!-- ko foreach: questionInstructions-->\n            <!-- ko ifnot: children.length -->\n            <li class="question-instruction-item styled-content" data-bind="contentBlock: content"></li>\n            <!-- /ko -->\n            <!-- ko if: children.length -->\n            <li class="question-instruction-item expandable-block-wrapper" data-bind="compose: { model:\'viewmodels/questions/expandableBlock\', activationData: $data }"></li>\n            <!-- /ko -->\n            <!-- /ko -->\n        </ul>\n    </section>\n    <!-- /ko -->\n    <!-- ko ifnot: isPreview-->\n    <div data-bind="overlay: isAnswered, syncHeight: {}">\n        <div class="text-matching-caption" data-translate-text="[text matching question hint]">\n            Drag items from right column to the left to match the pairs\n        </div>\n        <!-- ko foreach: sources-->\n        <div class="text-matching-row">\n            <div class="text-matching-column">\n                <div class="text-matching-source-wrapper">\n                    <div class="text-matching-source">\n                        <div class="text-matching-source-text-wrapper">\n                            <div class="text-matchig-source-text" data-bind="text: key"></div>\n                        </div>\n                    </div>\n                </div>\n            </div>\n            <div class="text-matching-column">\n                <div class="text-matching-target-placeholder-wrapper" data-bind="droppable: $data, accept: 1, scope: $parent.id">\n                    <div class="text-matching-target-placeholder">\n                        <div class="text-matching-target-placeholder-text-wrapper">\n                            <div class="text-matching-target-placeholder-text" data-translate-text="[text matching question drop here]">Drop here</div>\n                        </div>\n                    </div>\n                    <!-- ko if: value -->\n                    <!-- ko with: value -->\n                    <div class="text-matching-target" data-bind="draggable: {}, scope: $parents[1].id">\n                        <div class="text-matching-target-text-wrapper">\n                            <div class="text-matching-target-text" data-bind="text: $data"></div>\n                        </div>\n                    </div>\n                    <!-- /ko -->\n                    <!-- /ko -->\n                </div>\n            </div>\n            <div class="text-matching-column empty">&nbsp;</div>\n            <!-- ko with: $parent.targets()[$index()] -->\n            <div class="text-matching-column" data-bind="droppable: $data, accept: 1, scope: $parents[1].id">\n                <!-- ko if: value -->\n                <!-- ko with: value -->\n                <div class="text-matching-target-wrapper">\n                    <div class="text-matching-target" data-bind="draggable: {}, scope: $parents[2].id">\n                        <div class="text-matching-target-text-wrapper">\n                            <div class="text-matching-target-text" data-bind="text: $data"></div>\n                        </div>\n                    </div>\n                </div>\n                <!-- /ko -->\n                <!-- /ko -->\n            </div>\n            <!-- /ko -->\n        </div>\n        <!-- /ko -->\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isPreview-->\n    <div data-bind="syncHeight: {}">\n        <!-- ko foreach: sources-->\n        <div class="text-matching-row">\n            <div class="text-matching-column">\n                <div class="text-matching-source-wrapper">\n                    <div class="text-matching-source">\n                        <div class="text-matching-source-text-wrapper">\n                            <div class="text-matchig-source-text" data-bind="text: key"></div>\n                        </div>\n                    </div>\n                </div>\n            </div>\n            <div class="text-matching-column">\n                <div class="text-matching-target-placeholder-wrapper">\n                    <div class="text-matching-target-placeholder">\n                    </div>\n                    <!-- ko if: value -->\n                    <!-- ko with: value -->\n                    <div class="text-matching-target">\n                        <div class="text-matching-target-text-wrapper">\n                            <div class="text-matching-target-text" data-bind="text: $data"></div>\n                        </div>\n                    </div>\n                    <!-- /ko -->\n                    <!-- /ko -->\n                </div>\n            </div>\n            <div class="text-matching-column empty">&nbsp;</div>\n            <!-- ko with: $parent.targets()[$index()] -->\n            <div class="text-matching-column">\n                <!-- ko if: value -->\n                <!-- ko with: value -->\n                <div class="text-matching-target-wrapper">\n                    <div class="text-matching-target">\n                        <div class="text-matching-target-text-wrapper">\n                            <div class="text-matching-target-text" data-bind="text: $data"></div>\n                        </div>\n                    </div>\n                </div>\n                <!-- /ko -->\n                <!-- /ko -->\n            </div>\n            <!-- /ko -->\n        </div>\n        <!-- /ko -->\n    </div>\n    <!-- /ko -->\n    <!-- ko if: isAnswered() || isPreview() -->\n    <div class="question-overlay"></div>\n    <!-- /ko -->\n</section>';});


define('text!views/sections.html',[],function () { return '<section id="sectionsPage" class="sections " data-view="views/sections" data-active-view="true" style="">\r\n    <header class="page-header">\r\n        <h1 class="course-title" data-bind="css: {small: courseTitle.length>100}">\r\n            <span class="course-caption" data-bind="text: courseTitle, reviewable: {type: \'course\', property: \'title\'}">"course_name"</span>\r\n        </h1>\r\n        <div class="course-action-block">\r\n            <!-- ko compose: { model: \'viewmodels/exitCourse\', activationData: [ \'extended\'] } -->\r\n            <!-- /ko -->\r\n        </div>\r\n    </header>\r\n    <!-- ko if: $root.sectionsLayout==="List" -->\r\n        <h1 class="sections-title-container">\r\n            <span class="sections-title" data-translate-text="[sections]"></span>\r\n        </h1>\r\n    <!-- /ko -->\r\n    <ul class="sections-list autofocus" data-bind="foreach: sections">\r\n        <li tabindex="0" class="sections-list-item" data-bind="click: goToFirstQuestion, submitOnEnter, css: { disabled: $parent.isNavigationLocked }, css:$root.sectionsLayout, reviewable: {type: \'section\', property: \'title\', id: id, title: title}">\r\n            <!-- ko if: $root.sectionsLayout==="Tiles" -->\r\n            <div class="section-image-holder">\r\n                <div class="section-image" data-bind="containImage: { imageUrl: imageUrl, imageWidth: imageWidth, imageHeight: imageHeight }"></div>\r\n                <!-- ko ifnot: $root.isInReviewAttemptMode -->\r\n                <button tabindex="-1" class="button small primary" data-translate-text="[start]" \r\n                    data-bind="css: { disabled: $parent.isNavigationLocked }, attr: { \'aria-label\': $parent.getSectionButtonAccessibilityLabel(\'[start]\', $data) }"></button>\r\n                <!-- /ko -->\r\n                <!-- ko if: $root.isInReviewAttemptMode -->\r\n                <button tabindex="-1" class="button small primary" data-translate-text="[view]"\r\n                    data-bind="css: { disabled: $parent.isNavigationLocked }, attr: { \'aria-label\': $parent.getSectionButtonAccessibilityLabel(\'[view]\', $data) }"></button>\r\n                <!-- /ko -->\r\n            </div>\r\n            <!-- ko if: affectProgress -->\r\n            <div class="section-progress" aria-hidden="true">\r\n                <canvas class="canvas-styles" width="48" height="48" data-bind="circleProgress: { progress: score, radius: 22, isMastered: isMastered, lineWidth:2 },toolTip:{progress: score, masteryScore:$parent.masteryScore}"></canvas>\r\n                <span class="section-progress-text-wrapper" data-bind="text:score"></span>\r\n            </div>\r\n            <!-- /ko -->\r\n            <!-- ko ifnot: affectProgress -->\r\n            <div class="section-progress-no-questions" data-translate-text="[there are no questions]">No questions</div>\r\n            <!-- /ko -->\r\n            <div class="section-info">\r\n                <span data-bind="text: title" aria-hidden="true">section title</span>\r\n            </div>\r\n            <!--/ko-->\r\n            <!-- ko if: $root.sectionsLayout==="List" -->\r\n            <div class="section-info">\r\n                <div class="section-image-holder">\r\n                    <div class="section-image" data-bind="containImage: { imageUrl: imageUrl, imageWidth: imageWidth, imageHeight: imageHeight }"></div>\r\n                </div>\r\n                <div class="section-info-holder">\r\n                    <div class="section-info-title" aria-hidden="true" data-bind="text: title"></div>\r\n                    <div class="section-info-title clamped" aria-hidden="true" data-bind="text: title"></div>\r\n                    <!-- ko ifnot: affectProgress -->\r\n                    <span class="section-progress-no-questions" data-translate-text="[there are no questions]">No questions</span>\r\n                    <!-- /ko -->\r\n                    <!-- ko if: affectProgress -->\r\n                    <div class="section-progress" aria-hidden="true">\r\n                        <div class="progressbar-wrapper canvas-styles">\r\n                            <div class=\'progressbar\'>\r\n                                <div class="progressbar-value" data-bind="style: {width: score+\'%\'}, css:{mastered:score >= $parent.masteryScore}"></div>\r\n                                <div class="mastered-score-tooltip" data-bind="style: {left:$parent.masteryScore+\'%\'}, css:{mastered:score >= $parent.masteryScore}">\r\n                                    <span class="material-icons">done</span>\r\n                                    <span class="mastered-score-tooltip-text">\r\n                                        <span data-bind="text: scoreTooltipText"></span>\r\n                                    </span>\r\n                                </div>\r\n                            </div>\r\n                        </div>\r\n                        <span class="progress-value" data-bind="text: score + \'%\'"></span>\r\n                        <div class="progress-value-container">\r\n                            <span class="progress-value" data-bind="text: score + \'%\', css: { visible: score < $parent.masteryScore }"></span>\r\n                            <span class="progress-value-text" data-bind="css: { visible: score < $parent.masteryScore }" data-translate-text="[to complete]"></span>\r\n                            <span class="progress-value-text completed" data-bind="css: { visible: score >= $parent.masteryScore }">\r\n                                <span class="icon-correct_check"></span>\r\n                                <span data-translate-text="[completed]"></span>\r\n                            </span>\r\n                        </div>\r\n                    </div>\r\n                    <!-- /ko -->\r\n                </div>\r\n                <div class="section-start-button-holder">\r\n                    <!-- ko ifnot: $root.isInReviewAttemptMode -->\r\n                    <button tabindex="-1" class="button small primary" data-translate-text="[start]"\r\n                        data-bind="css: { disabled: $parent.isNavigationLocked }, attr: { \'aria-label\': $parent.getSectionButtonAccessibilityLabel(\'[start]\', $data) }"></button>\r\n                    <!-- /ko -->\r\n                    <!-- ko if: $root.isInReviewAttemptMode -->\r\n                    <button tabindex="-1" class="button small primary" data-translate-text="[view]"\r\n                        data-bind="css: { disabled: $parent.isNavigationLocked }, attr: { \'aria-label\': $parent.getSectionButtonAccessibilityLabel(\'[view]\', $data) }"></button>\r\n                    <!-- /ko -->                    \r\n                </div>\r\n            </div>\r\n            <!--/ko-->\r\n        </li>\r\n    </ul>\r\n</section>';});


define('text!views/shell.html',[],function () { return '<section class="shell">\n    <!-- ko if: viewSettings().treeOfContentVisible -->\n    <!-- ko compose: \'treeOfContent/treeOfContent\' -->\n    <!-- /ko -->\n    <!-- /ko -->\n    <!-- ko compose: {model: \'header/header\', activationData: [viewSettings, pdfExportEnabled]} -->\n    <!-- /ko -->\n    <section class="content-container" data-bind="css: { \'review-mode\' : isInReviewMode }, css: \'module-\' + cssName()">\n        <div id="content" class="content-wrapper">\n            <!-- ko router: { cacheViews: false, transition: \'entrance\' } -->\n            <!-- /ko -->\n        </div>\n    </section>\n    <div class="blockout" data-bind="css: { active: isClosed }">\n        <div class="blockout-text-wrapper">\n            <!-- ko if: inIframe -->\n            <span class="blockout-text" data-translate-text="[the course has been finished]"></span>\n            <!-- /ko -->\n            <!-- ko ifnot: inIframe -->\n            <span class="blockout-text" data-translate-text="[please close the browser tab now]"></span>\n            <!-- /ko -->\n        </div>\n    </div>\n    <div data-bind="widget: {kind:\'imagePreview\'}"></div>\n</section>\n';});


define('text!widgets/imagePreview/view.html',[],function () { return '<div tabindex="0" class="overlay-holder" data-bind="fade: isVisible">\n    <div class="overlay" data-bind="click:closePreview"></div>\n    <div class="center">\n        <div class="center-holder">\n            <div class="center-content">\n                <div class="image-holder">\n                    <div data-bind="imageLoader: { isLoaded: isLoaded, imageUrl: imageUrl}">\n                    </div>\n                    <!--ko ifnot:isLoaded-->\n                    <div class="loader-image"></div>\n                    <!--/ko-->\n                </div>                \n                <div class="overlay-close-button-holder" data-bind="if: isLoaded">\n                    <a tabindex="0" class="overlay-close-button material-icons" data-bind="click: closePreview, hasFocus: isVisible, submitOnEnter, keyDown: { 27: { handler: closePreview } }">clear</a>\n                    <span class="overlay-close-button-subtext">Esc</span>\n                </div>\n            </div>\n        </div>\n    </div>\n</div>';});


define('text!xApi/views/xAPIError.html',[],function () { return '<section id="xapiError" class="container xapi-error">\r\n    <div class="error-content">\r\n        <div class="xapi-error-container">\r\n            <h1 data-translate-text="[tracking and tracing error]">Something is wrong</h1>\r\n            <!-- ko if: allowToContinue() -->\r\n            <div class="xapi-error-container-description" data-translate-text="[tracking and tracing error hint]">If you continue without restarting, your learning progress will not be reported.</div>\r\n            <!-- /ko -->\r\n            <!-- ko ifnot: allowToContinue() -->\r\n            <div class="xapi-error-container-description" data-translate-text="[tracking and tracing error hint without continue]">Restart the course. Contact your course provider if the problem persists.</div>\r\n            <!-- /ko -->\r\n            <div class="xapi-error-container-buttons clearfix">\r\n                <!-- ko if: allowToContinue() -->\r\n                <button class="button default large" data-bind="click: continueLearning" data-translate-text="[tracking and tracing continue anyway]" data-translate-title="[tracking and tracing continue anyway]">Continue anyway</button>\r\n                <button class="button primary large" data-bind="click: restartCourse" data-translate-text="[tracking and tracing restart course]" data-translate-title="[tracking and tracing restart course]">Restart course</button>\r\n\r\n                <!-- /ko -->\r\n                <!-- ko ifnot: allowToContinue() -->\r\n                <button class="button primary large" data-bind="click: restartCourse" data-translate-text="[tracking and tracing restart course]" data-translate-title="[tracking and tracing restart course]">Restart course</button>\r\n                <!-- /ko -->\r\n            </div>\r\n        </div>\r\n    </div>\r\n</section>';});

//# sourceMappingURL=main.js.map